package sota.ast.nodes;

import java.util.ArrayList;
import java.util.HashSet;

import sota.Logger;

/**
 * Superklasse fr Knoten mit Verzweigung abgeleitete Klassen: ASTIfNode, ASTIterationNode, ASTSwitchNode, ASTTryNode
 * 
 * <PRE>
 * 		...
 * 		ASTNode 			entryNode
 * 		(ASTLabelNode)
 * 		ASTBranchNode		this
 *    	   &gt;  ...
 * 		ASTStatementNode 	endNode
 * 		...
 * </PRE>
 * 
 */
/**
 * @author Quark I
 *
 */
public abstract class ASTBranchNode extends ASTStatementNode {

	/** Bei der Konstruktion zu setzende Informationen und Verknpfungen * */

	// der Eingangsknoten, der vor dem Branchblock inklusive aller Labels steht,
	// oder der erste LabelNode
	protected ASTNode entryNode;

	// der Endknoten, der den Ausgang aus dem Branchblock markiert
	// kann unerreichbar sein! sonst in setOfExitNodes enthalten
	protected ASTStatementNode endNode;

	/***************************************************************************/

	// Menge aller Knoten, die den Branchblock verlassen (break, return,
	// endNode, throw)
	protected HashSet<ASTNode> setOfExitNodes;

	// Menge aller Knoten, zu denen der BranchNode-Eingang verzweigt (fr CFG)
	protected HashSet<ASTNode> setOfFollowUpNodes;

	protected HashSet<ArrayList<Integer>> setOfBIPaths;

	// zu implementieren: die entsprechenden folloupNodes einfgen
	public abstract void completeNode();

	/**
	 * Der Konstruktor.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 */
	public ASTBranchNode(int ln, String s) {
		super(ln, s);
		endNode = null;
		setOfExitNodes = new HashSet<ASTNode>();
		setOfFollowUpNodes = new HashSet<ASTNode>();
		setOfBIPaths = new HashSet<ArrayList<Integer>>();
	}

	/**
	 * Legt Eingangsknoten fest.
	 * 
	 * @param en
	 *            der Eingangsknoten
	 */
	public void setEntryNode(ASTNode en) {
		entryNode = en;
		return;
	}

	/**
	 * Legt den Endknoten fest.
	 * 
	 * @param en
	 *            der Endknoten
	 */
	public void setEndNode(ASTStatementNode en) {
		endNode = en;
		return;
	}

	/**
	 * Liefert den Eingangsknoten.
	 * 
	 * @return der Eingangsknoten
	 */
	public ASTNode getEntryNode() {
		return entryNode;
	}

	/**
	 * Liefert den Endknoten.
	 * 
	 * @return der Endknoten
	 */
	public ASTStatementNode getEndNode() {
		return endNode;
	}

	/**
	 * Fgt einen Ausgangsknoten hinzu.
	 * 
	 * @param n
	 *            der Ausgangsknoten
	 */
	public void addExitNode(ASTNode n) {
		setOfExitNodes.add(n);
	}

	/**
	 * Gibt die Menge der Ausgangsknoten zurck.
	 * 
	 * @return die Menge
	 */
	public HashSet<ASTNode> getSetOfExitNodes() {
		return setOfExitNodes;
	}

	/**
	 * Fgt einen Folgeknoten hinzu, zu dem der Kontrollfluss verzweigen kann.
	 * 
	 * @param n
	 *            der Folgeknoten.
	 */
	public void addFollowUpNode(ASTNode n) {
		setOfFollowUpNodes.add(n);
	}

	/**
	 * Gibt die Menge an Folgeknoten zurck, zu denen die Anweisung verzweigen kann.
	 * 
	 * @return die Menge
	 */
	public HashSet<ASTNode> getSetOfFollowUpNodes() {
		return setOfFollowUpNodes;
	}

	/**
	 * Ermittelt, ob ein Knoten innerhalb des Branch-Konstruktes ist.
	 * 
	 * @param i
	 *            die Knoten-ID
	 * @return true, wenn enthalten, sonst false
	 */
	public boolean isWithinBranchConstruct(int i) {

		return entryNode.getID() < i && endNode.getID() > i;
	}

	// Printfunktionen

	public void printExits() {
		Logger.log("< Exits:");
		for (ASTNode n : setOfExitNodes) {
			Logger.log(" " + n.getID());
		}
		Logger.log(" >");
	}

	public String getExitString() {
		String s = "< Exits:";
		for (ASTNode n : setOfExitNodes)
			s += " " + n.getID();
		return s + " >";
	}

}