package sota.ast.nodes;

import sota.Logger;

/**
 * AST-Knoten fr Case-Statements (inkl. Default).
 * 
 * <PRE>
 *   	..
 *   	ASTNode				switchNode.entryNode
 * 		ASTSwitchNode		switchNode
 * 		  ASTCaseNode		this
 * 		  &gt;  ..
 * 		  ASTCaseNode		switchNode.[listOfCaseNodes]
 * 	      &gt;  ..
 *      ASTStatementNode	switchNode.endNode	
 *      ..
 * </PRE>
 */
public abstract class ASTCaseNode extends ASTStatementNode {

	/** Bei der Konstruktion zu setzende Informationen und Verknpfungen * */

	protected ASTSwitchNode switchNode = null;

	protected String constant = "";

	/**********************************************************************/

	protected int nrSelects;

	protected boolean isDefault = false;

	/**
	 * Konstruktof fr Default-Knoten.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 */
	public ASTCaseNode(int ln, String s) {
		super(ln, s);
		nrSelects = 0;
		isDefault = true;
	}

	/**
	 * Konstruktor fr Case-Knoten.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param c
	 *            die Konstante
	 */
	public ASTCaseNode(int ln, String s, String c) {
		super(ln, s);
		nrSelects = 0;
		constant = c;
	}

	/**
	 * Setzt den Switchknoten.
	 * 
	 * @param switchNode
	 *            der Switchknoten
	 */
	public void setSwitchNode(ASTSwitchNode switchNode) {
		controlingBranchNode = switchNode;
		this.switchNode = switchNode;
	}

	/**
	 * Liefert den Switchknoten.
	 * 
	 * @return der Switchknoten
	 */
	public ASTSwitchNode getSwitchNode() {
		return switchNode;
	}

	/**
	 * Gibt zurck, ob dieser Knoten ein Default ist.
	 * 
	 * @return true,wenn default, sonst false
	 */
	public boolean isDefault() {
		return isDefault;
	}

	/**
	 * Liefert die Konstante eines Case zurck
	 * 
	 * @return die Konstante
	 */
	public String getConstant() {
		return constant;
	}

	/**
	 * Setzt die Konstante eines Case.
	 * 
	 * @param s
	 *            die Konstante
	 */
	public void setConstant(String s) {
		constant = s;
		return;
	}

	/**
	 * Erhht die Anzahl der Switch-Verzweigungen zu diesem Case.
	 */
	public void incNrSelects() {
		// Logger.log("inc.selection of " + id);
		nrSelects++;
	}

	/**
	 * Erhht die Anzahl der Switch-Verzweigungen zu diesem Case.
	 * 
	 * @param selection
	 *            die Anzahl
	 */
	public void incNrSelects(int selection) {
		// Logger.log("inc.selection of " + id + " by " + selection + "(new: " + nrSelects+selection + ")");
		nrSelects += selection;
	}

	/**
	 * Setzt die Anzahl der Switch-Verzweigungen zu diesem Case.
	 * 
	 * @param i
	 *            die Anzahl
	 */
	public void setNrSelects(int i) {
		nrSelects = i;
	}

	/**
	 * Gibt die Anzahl der Switch-Verzweigungen zu diesem Case zurck.
	 * 
	 * @return die Anzahl
	 */
	public int getNrSelects() {
		return nrSelects;
	}

	/**
	 * Liefert den Abschlieenden Knoten des Case-Blockes zurck.
	 * 
	 * @return
	 */
	public ASTNode getEndOfCaseBlock() {
		ASTNode n = this;
		while (!(n.getNextNode() instanceof ASTCaseNode) && (n.getNextNode() != switchNode.getEndNode()))
			n = n.getNextNode();
		return n;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#printNode()
	 */
	public void printNode() {
		Logger.log("---");
		printReach();
		if (constant == "") {
			Logger.logln("<" + id + "><Default><" + nrHits + "><S: " + nrSelects + ">---" + content);
		} else {
			Logger.logln("<" + id + "><Case: " + constant + "><" + nrHits + "><S: " + nrSelects + ">---" + content);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#getNodeString()
	 */
	public String getNodeString() {
		String s = super.getNodeString();
		if (constant == "")
			s += "<Default>";
		else
			s += "<Case: " + constant + ">";
		return s + "<" + nrHits + "><S: " + nrSelects + ">";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#resetHits()
	 */
	public void resetHits() {
		super.resetHits();
		// Logger.log("reseting nrselections of " + id);
		nrSelects = 0;
	}

}