package sota.ast.nodes;

import sota.Logger;

/**
 * AST-Knoten fr Catches und Finally in Ausnahme-Blcke.
 * 
 * <PRE>
 *   	..
 *   	ASTNode				tryNode.entryNode
 * 		ASTTryNode			tryNode
 * 		  ASTStatementNode	tryNode.entryOfTryBlock
 * 		  &gt;  ..
 * 		  ASTNode			tryNode.endOfTryBlock
 * 	    ASTCatchNode		this
 *        &gt;  ..
 * 	    ASTCatchNode		tryNode.[listOfCatchNodes]
 *        &gt;  ..
 *      ASTStatementNode	tryNode.endNode	
 *      ..
 * </PRE>
 * 
 */
public abstract class ASTCatchNode extends ASTStatementNode {

	/** Bei der Konstruktion zu setzende Informationen und Verknpfungen * */

	// Referenz auf zugehrigen ASTTryNode
	protected ASTTryNode tryNode;

	protected String parameter = "";

	/************************************************************************/

	protected boolean isFinally = false;

	/**
	 * Konstruktor fr die Erstellung eines Finally-Knotens
	 * 
	 * @param nl
	 *            die Zeilennummer
	 * @param s
	 *            der Parameter
	 */
	public ASTCatchNode(int nl, String s) {
		super(nl, s);
		isFinally = true;
	}

	/**
	 * Konstruktor fr die Erstellung eines Catch-Knotens.
	 * 
	 * @param nl
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param para
	 *            der Parameter
	 */
	public ASTCatchNode(int nl, String s, String para) {
		super(nl, s);
		parameter = para;
	}

	/**
	 * Liefert den Parameter des Catches.
	 * 
	 * @return der Parameter
	 */
	public String getParameter() {
		return parameter;
	}

	/**
	 * Liefert das Ende des CatchBlockes.
	 * 
	 * @return der Endknoten
	 */
	public ASTNode getEndOfCatchBlock() {
		ASTNode n = this;
		while (!(n.getNextNode() instanceof ASTCatchNode) && (n.getNextNode() != tryNode.getEndNode())) {
			n = n.getNextNode();
		}
		return n;
	}

	/**
	 * Setzt den Parameter des Catches.
	 * 
	 * @param s
	 *            der Parameter
	 */
	public void setParameter(String s) {
		parameter = s;
		return;
	}

	/**
	 * Setzt den TryNode fr das Catch.
	 * 
	 * @param tn
	 *            der TryNode
	 */
	public void setTryNode(ASTTryNode tn) {
		controlingBranchNode = tn;
		tryNode = tn;
	}

	/**
	 * Liefert den TryNode.
	 * 
	 * @return der TryNode
	 */
	public ASTTryNode getTryNode() {
		return tryNode;
	}

	/**
	 * Sagt aus, ob dieser Knoten ein Finally ist.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean isFinally() {
		return isFinally;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#printNode()
	 */
	public void printNode() {
		Logger.log("---");
		printReach();
		Logger.logln("<" + id + "><Catch: " + parameter + ">---" + content);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#getNodeString()
	 */
	public String getNodeString() {
		return super.getNodeString() + "<Catch: " + parameter + ">";
	}

}