package sota.ast.nodes;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Stack;
import java.util.Vector;

import sota.Logger;
import sota.ast.util.ASTManager;

/**
 * AST-Knoten fr eine Funktionsstruktur.
 * */
public abstract class ASTFunctionNode extends ASTScopeNode implements IStatementNode {

	protected Stack<ASTBranchNode> stackOfBranchNodes;

	protected String arguments;

	protected String modifiers;

	protected int nrPrimaryConditions = 0;

	protected Vector<Integer> nrAtoms = new Vector<Integer>();

	protected ASTStatementNode entryNode;

	protected ArrayList<IConditionNode> listOfConditionNodes;

	protected HashSet<ASTStatementNode> setOfExitNodes = new HashSet<ASTStatementNode>();

	private int nrHits;

	/**
	 * Konstruktor fr einen Funktionsknoten.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param content
	 *            der Inhalt
	 * @param scope
	 *            der Scope der Funktion
	 * @param modifiers
	 *            die Modifiers
	 * @param name
	 *            der Name
	 * @param arguments
	 *            die Argumente
	 * @param parent
	 *            der bergeordnete IScopeNode
	 */
	public ASTFunctionNode(int ln, String content, String scope, String modifiers, String name, String arguments,
			IScopeNode parent) {
		super(ln, scope, name, parent);
		this.content = content;
		this.arguments = arguments;
		this.modifiers = modifiers;
		setQualifiedName();

		listOfConditionNodes = new ArrayList<IConditionNode>();
		stackOfBranchNodes = new Stack<ASTBranchNode>();
		functionNode = this;

	}
	
	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTNode#createASCContent(int)
	 */
	public abstract void createASCContent(int level);


	/**
	 * Entscheidet, ob die Funktion private deklariert is.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean isPrivate() {
		return modifiers.contains(" private ");
	}

	/**
	 * Entscheidet, ob die Funktion protected deklariert is.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean isProtected() {
		return modifiers.contains(" protected ");
	}

	/**
	 * Fgt einen IConditionNode hinzu.
	 * 
	 * @param cs
	 *            der IConditionNode
	 */
	public void addConditionNode(IConditionNode cs) {
		listOfConditionNodes.add(cs);
	}

	/**
	 * Gibt das Argument zurck.
	 * 
	 * @return das Argument
	 */
	public String getArguments() {
		return arguments;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getName()
	 */
	public String getName() {
		return name + " (" + arguments + ")";
	}

	/**
	 * Liefert den Funktionsnamen.
	 * 
	 * @return der Name
	 */
	public String getFunctionName() {
		return name;
	}

	/**
	 * Setzt den Eingangsknoten der Funktion.
	 * 
	 * @param en
	 *            der Eingangsknoten
	 */
	public void setEntryNode(ASTStatementNode en) {
		entryNode = en;
		// entryNode.setControlingBranchNode(this);
	}

	/**
	 * Liefert den Eingangsknoten der Funktion.
	 * 
	 * @return der Eingangsknoten
	 */
	public ASTStatementNode getEntryNode() {
		return entryNode;
	}

	/**
	 * Fgt einen Ausgangsknoten hinzu.
	 * 
	 * @param en
	 *            der Ausgangsknoten
	 */
	public void addExitNode(ASTStatementNode en) {
		setOfExitNodes.add(en);
	}

	/**
	 * Liefert die Menge der Ausgangsknoten der Funktion.
	 * 
	 * @return die Meng
	 */
	public HashSet<ASTStatementNode> getSetOfExitNodes() {
		return setOfExitNodes;
	}
	
	/**
	 * frbt alle erreichbaren Knoten und sammelt die ReturnKnoten
	 */
	public void findReachableNodes() {
		findReachableNodes(next);
	}

	/**
	 * Luft den Kontrollfluss ab und registriert fr jeden Knoten seine Ausfhrbarkeit.
	 * 
	 * @param n
	 *            der ASTKnoten
	 */
	private void findReachableNodes(ASTNode n) {

		do {
			// falls Ende des akt. BranchNodes - als ExitNode hinzufgen + aus Stack
			// nehmen

			if (!stackOfBranchNodes.empty() && stackOfBranchNodes.peek().getEndNode() == n) {
				stackOfBranchNodes.pop().addExitNode(n);
			}

			// schon gefrbt -> Abbruch
			if (n.isReachable() == true) {
				return;
			}
			// sonst frben
			n.isReachable(true);

			// falls Funktionsende -> Abbruch

			if (n == endNode) {
				return;
			}
			// falls ein Ausgang des akt. BranchNodes - aus dem Stack nehmen +
			// Abbruch
			else if (!stackOfBranchNodes.empty() && stackOfBranchNodes.peek().getSetOfExitNodes().contains(n)) {
				stackOfBranchNodes.pop();
				return;
			}

			if (n instanceof ASTDoNode || n instanceof ASTBranchNode || n instanceof ASTJumpNode
					|| n instanceof ASTCompoundNode) {
				break;
			} else {
				n.getNextNode().isDirectReachable(true);
				n = n.getNextNode();
			}
		} while (true);

		// falls DoNode -> IterationNode auf Stack, nchsten frben
		if (n instanceof ASTDoNode) {
			stackOfBranchNodes.push(((ASTDoNode) n).getIterationNode());
			n.getNextNode().isDirectReachable(true);
			findReachableNodes(n.getNextNode());
			return;
		}
		// falls BranchNode
		else if (n instanceof ASTBranchNode) {
			// auf Stack, falls durch DoNode schon auf dem stack, merken
			if (!(n instanceof ASTIterationNode && ((ASTIterationNode) n).hasDoNode())) {
				stackOfBranchNodes.push((ASTBranchNode) n);

				// falls IfNode
			}
			if (n instanceof ASTIfNode) {
				// trueNode frben
				n.getNextNode().isDirectReachable(true);
				findReachableNodes(n.getNextNode());

				if (stackOfBranchNodes.empty() || stackOfBranchNodes.peek() != n) {
					stackOfBranchNodes.push((ASTBranchNode) n);
					// falseNode frben
				}
				findReachableNodes(((ASTIfNode) n).getElseNode());
				return;
			}
			// falls IterationNode
			else if (n instanceof ASTIterationNode) {
				Condition cond = ((ASTIterationNode) n).getCondition();
				// falls kein DoWhile, nchsten Knoten frben
				if (!((ASTIterationNode) n).hasDoNode()) {
					n.getNextNode().isDirectReachable(true);
					findReachableNodes(n.getNextNode());
				}

				// falls foreach oder keine immer wahre Bedingung
				if (cond == null || !cond.isTrueAtom()) {

					// falls DoWhile, dann EndNode direct erreichbar
					if (((ASTIterationNode) n).hasDoNode()) {
						((ASTIterationNode) n).getEndNode().isDirectReachable(true);
					}
					// Endknoten frben und merken
					findReachableNodes(((ASTIterationNode) n).getEndNode());
					((ASTIterationNode) n).addExitNode(((ASTIterationNode) n).getEndNode());
				}
				// sonst vom Stack nehmen
				else {
					if (!stackOfBranchNodes.isEmpty()) {
						stackOfBranchNodes.pop();
					}
				}
				return;
			}
			// falls SwitchNode
			else if (n instanceof ASTSwitchNode) {
				// frbe alle Cases und Default

				for (ASTCaseNode cn : ((ASTSwitchNode) n).getListOfCases()) {
					findReachableNodes(cn);
					if (stackOfBranchNodes.empty() || stackOfBranchNodes.peek() != n) {
						stackOfBranchNodes.push((ASTBranchNode) n);
					}
				}
				return;
			}
			// falls TryNode
			else if (n instanceof ASTTryNode) {
				// tryBlock frben
				n.getNextNode().isDirectReachable(true);
				findReachableNodes(n.getNextNode());
				if (stackOfBranchNodes.empty() || stackOfBranchNodes.peek() != n) {
					stackOfBranchNodes.push((ASTBranchNode) n);
				}
				// frbe alle Catches und Finally
				for (ASTCatchNode cn : ((ASTTryNode) n).getListOfCatches()) {
					findReachableNodes(cn);
					if (stackOfBranchNodes.empty() || stackOfBranchNodes.peek() != n) {
						stackOfBranchNodes.push((ASTBranchNode) n);

					}
				}
				return;
			}
		}
		// falls JumpNode
		else if (n instanceof ASTJumpNode) {
			Stack<ASTBranchNode> stack = new Stack<ASTBranchNode>();
			// falls Return
			if (((ASTJumpNode) n).isReturn()) {
				// merke fr Funktion
				setOfExitNodes.add((ASTJumpNode) n);
				// merke fr alle BranchNodes auf dem Stack
				for (ASTBranchNode bn : stackOfBranchNodes) {
					bn.addExitNode(n);
				}
				return;
			}
			// kein Return

			ASTBranchNode bn;
			int targetID = ((ASTJumpNode) n).getTargetNode().getID();
			// schaue BranchNodes auf dem Stack an

			while (!stackOfBranchNodes.isEmpty()) {
				bn = stackOfBranchNodes.peek();
				// wenn Jump aus Bereich des BranchNodes rausfhrt
				if (targetID < bn.getID() || targetID > bn.getEndNode().getID()) {
					// vermerke Jump als Exit
					bn.addExitNode(n);
					// sichere BranchNode und schau nchsten BranchNode an
					stack.push(stackOfBranchNodes.pop());
					if (!stackOfBranchNodes.empty()) {
						bn = stackOfBranchNodes.peek();
						// Abbruch, wenn kein BranchNode mehr da
					} else {
						break;
					}
				}
				// Abbruch, wenn Jump nicht aus Bereich des BranchNode rausfhrt
				else {
					break;
				}
			}

			// frbe JumpZiel, auer JumpNode ist Throw und fhrt aus Funktion
			// raus (d.h. auf endNode)!
			if (!(((ASTJumpNode) n).isThrow() && ((ASTJumpNode) n).getTargetNode() == endNode))
				findReachableNodes(((ASTJumpNode) n).getTargetNode());

			// hole gesicherte BranchNodes zurck
			while (!stack.empty()) {
				stackOfBranchNodes.push(stack.pop());
			}
			return;
		}
		// falls CompoundNode
		else if (n instanceof ASTCompoundNode) {
			ASTCompoundNode cn = (ASTCompoundNode) n;

			if (!cn.isStartNode()) {
				// falls Ende eines Iterationcompounds -> Abbruch, da Loop
				// IterationNode | LoopNode | CompoundNode
				if (cn.getStartNode().getPrevNode().getPrevNode() instanceof ASTIterationNode) {
					return;
				}
				// falls Ende eines TryBlockes
				else if (cn.getStartNode().getPrevNode() instanceof ASTTryNode) {
					ASTTryNode tn = (ASTTryNode) cn.getStartNode().getPrevNode();
					// falls kein FinallyNode, beim Endknoten weiterfrben
					if (tn.getFinallyNode() == null) {
						findReachableNodes(tn.getEndNode());
					}
					return;
				}
				// falls Ende eines CatchBlockes
				else if (cn.getStartNode().getPrevNode() instanceof ASTCatchNode) {
					ASTCatchNode ctn = (ASTCatchNode) cn.getStartNode().getPrevNode();
					// falls kein FinallyNode, Abbruch
					if (!ctn.isFinally()) {
						if (!stackOfBranchNodes.empty() && stackOfBranchNodes.peek() == ctn.getTryNode())
							stackOfBranchNodes.pop();

						return;
					} else {
						if (!ctn.getTryNode().getSetOfExitNodes().contains(ctn.getTryNode().getEndNode())) {
							if (!stackOfBranchNodes.empty() && stackOfBranchNodes.peek() == ctn.getTryNode())
								stackOfBranchNodes.pop();

							return;
						}
					}
				}
			}

		}

		n.getNextNode().isDirectReachable(true);
		findReachableNodes(n.getNextNode());
		return;
	}

	/**
	 * Vollendet Struktur einer Funktion, wenn der AST fertig konstruiert ist (erst nach Verknpfung der Sprungziele
	 * aufrufen!).
	 * 
	 * @param ast
	 * der ASTManager
	 */
	public void completeFunction(ASTManager ast) {
		completeBranchNodes();
		numerateConditions();

		next.isDirectReachable(true);
		findReachableNodes();

		HashSet<ASTStatementNode> unreachableNodes = new HashSet<ASTStatementNode>();
		for (ASTStatementNode n : setOfExitNodes) {
			if (!n.isReachable()) {
				unreachableNodes.add(n);
			}
		}
		setOfExitNodes.removeAll(unreachableNodes);
	}

	/**
	 * Nummeriert die Bedingungen und ihre Atome durch.
	 */
	public void numerateConditions() {
		int i = 0;
		for (IConditionNode n : listOfConditionNodes) {
			i = n.numerateCondition(nrPrimaryConditions);
			nrAtoms.add(nrPrimaryConditions++, i);
		}
	}

	/**
	 * Vollendet alle BranchNodes.
	 */
	public void completeBranchNodes() {
		ASTNode n = next;
		while (n != endNode) {
			if (n instanceof ASTBranchNode)
				((ASTBranchNode) n).completeNode();
			n = n.getNextNode();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IStatementNode#getNrHits()
	 */
	public int getNrHits() {
		return nrHits;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IStatementNode#setNrHits(int)
	 */
	public void setNrHits(int i) {
		nrHits = i;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IStatementNode#incNrHits()
	 */
	public int incNrHits() {
		return ++nrHits;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IStatementNode#incNrHits(int)
	 */
	public int incNrHits(int i) {
		nrHits += i;
		return nrHits;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IStatementNode#decNrHits(int)
	 */
	public int decNrHits(int i) {
		nrHits -= i;
		if (nrHits < 0)
			nrHits = 0;
		return nrHits;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IStatementNode#resetHits()
	 */
	public void resetHits() {
		nrHits = 0;
		metrics.reset();
	}
	
	/** Print-Funktionen */

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTNode#printNode()
	 */
	public void printNode() {
		int i = 0;

		Logger.log("---<" + id + "><->" + endNode.getID() + "><Function><" + fileNode.getName() + ": "
				+ fileNode.getPackageName() + ". " + scopeName + "><" + name + "><" + arguments + ">");
		printExits();
		Logger.log("<" + nrHits + "><Paths: " + i + ">");

		Logger.logln("---\n" + content);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTNode#getNodeString()
	 */
	public String getNodeString() {
		int i = 0;
		String s = super.getNodeString() + "<Function->" + endNode.getID() + "><" + fileNode.getName() + ": "
				+ fileNode.getPackageName() + ". " + scopeName + "><" + name + "><" + arguments + ">" + getExitString()
				+ "<" + nrHits + "><Paths: " + i + ">";

		return s;
	}

	/**
	 * @return
	 */
	public String getExitString() {
		String s = "< Exits:";
		for (ASTNode n : setOfExitNodes)
			s += " " + n.getID();
		return s + " >";
	}

	/**
	 * 
	 */
	public void printExits() {
		Logger.log("< Exits:");
		for (ASTNode n : setOfExitNodes) {
			Logger.log(" " + n.getID());
		}
		Logger.log(" >");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTNode#printCFG()
	 */
	public void printCFG() {
		Logger.logln("\n\n\nFunction " + scopeName + " " + name);
		Logger.logln("(" + id + "#" + nrHits + ")-P");
		Logger.logln(" |");
		ASTNode n = next;
		while (n != endNode) {
			n.printCFG();
			n = n.getNextNode();
		}
		endNode.printCFG();
	}


}
