package sota.ast.nodes;

import java.util.HashSet;
import java.util.TreeSet;

import sota.Logger;
import sota.ast.util.Metrics;

/**
 * AST-Knoten fr If- oder If-Else-Anweisung
 * 
 * <PRE>
 *   	..
 *   	ASTNode				entryNode
 * 		ASTIfNode			this
 * 		  ASTStatementNode	trueNode
 * 		  &gt;  ..
 * 		  ASTJumpNode		-&gt; endNode
 * 	    ASTNode				elseNode
 *        ASTStatementNode	falseNode
 *        &gt;  ..
 *      ASTStatementNode	endNode	
 *      ..
 * </PRE>
 */
/**
 * @author Quark I
 * 
 */
public abstract class ASTIfNode extends ASTBranchNode implements IConditionNode {

	/** Bei der Konstruktion zu setzende Informationen und Verknpfungen * */

	protected Condition cond;

	protected ASTNode elseNode;

	protected ASTStatementNode trueNode;
	protected ASTStatementNode falseNode;

	/**
	 * Von ASTBranchNode:
	 * 
	 * protected ASTNode entryNode protected ASTStatementNode endNode
	 * 
	 * 
	 ***/

	protected HashSet<String> setOfTruthVectors = new HashSet<String>();

	protected int nrConditions = 0;

	/**
	 * Konstruktor fr If-Knoten.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param c
	 *            die Bedingung
	 */
	public ASTIfNode(int ln, String s, Condition c) {
		super(ln, s);
		cond = c;
		cond.setPrimary(true);
		nrConditions = cond.getNrConditions();

		if (!cond.isOperator()) {
			cond.setAtom(true);
		}
		elseNode = null;
	}

	/**
	 * Setzt den Else-Knoten.
	 * 
	 * @param en
	 *            der Else-Knoten
	 */
	public void setElseNode(ASTNode en) {
		elseNode = en;
	}

	/**
	 * Gibt den Else-Knoten zurck.
	 * 
	 * @return der Else-Knoten
	 */
	public ASTNode getElseNode() {
		return elseNode;
	}

	/**
	 * Entscheidet, ob ein Else vorliegt.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean hasElse() {
		return elseNode != null;
	}

	/**
	 * Setzt den True-Knoten.
	 * 
	 * @param tn
	 *            der True-Knoten
	 */
	public void setTrueNode(ASTStatementNode tn) {
		trueNode = tn;
		trueNode.setControlingBranchNode(this);
	}

	/**
	 * Gibt den True-Knoten zurck.
	 * 
	 * @return der True-Knoten
	 * */
	public ASTStatementNode getTrueNode() {
		return trueNode;
	}

	/**
	 * Setzt den False-Knoten.
	 * 
	 * @param fn
	 *            der False-Knoten
	 */
	public void setFalseNode(ASTStatementNode fn) {
		falseNode = fn;
		falseNode.setControlingBranchNode(this);
	}

	/**
	 * Liefert den False-Knoten zurck.
	 * 
	 * @return der False-Knoten
	 */
	public ASTStatementNode getFalseNode() {
		return falseNode;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTBranchNode#completeNode()
	 */
	@Override
	public void completeNode() {
		addFollowUpNode(trueNode);
		addFollowUpNode(falseNode);
	}

	/** Start Of ConditionMethods * */

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#fillConditions(java.util.TreeSet)
	 */
	public void fillConditions(TreeSet<String> list) {
		setOfTruthVectors.addAll(list);
		for (String s : list) {
			cond.applyTruthVector(s);
		}
		return;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#getCondition()
	 */
	public Condition getCondition() {
		return cond;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#resetHits()
	 */
	public void resetHits() {
		super.resetHits();
		setOfTruthVectors.clear();
		if (cond != null)
			cond.resetHits();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#getTruthVectors()
	 */
	public HashSet<String> getTruthVectors() {
		return setOfTruthVectors;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#getEvaluatedTVString(java.lang.String)
	 */
	public String getEvaluatedTVString(String s) {
		if (cond == null)
			return "";
		return (cond.evaluateTruthVector(s) ? "1" : "0") + cond.getEvaluatedTVString(s);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#hasNonTrivialCondition()
	 */
	public boolean hasNonTrivialCondition() {
		if (cond == null)
			return false;
		return cond.isNonTrivial();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#numerateCondition(int)
	 */
	public int numerateCondition(int i) {
		if (cond == null)
			return i;
		return cond.numerate(i);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#computeConditionCoverage()
	 */
	public void computeConditionCoverage() {
		if (cond == null)
			return;

		cond.computeConditionCoverage(setOfTruthVectors);
	}

	
	
	/** Ausgabe und Info-Methoden * */

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#printNode()
	 */
	public void printNode() {
		Logger.log("---");
		printReach();
		Logger.log("<" + id + "><If -> " + endNode.getID() + " >");
		printExits();
		Logger.log("<nrConds: " + nrConditions + "><" + nrHits + "><w:" + trueNode.getNrHits() + ", f:"
				+ falseNode.getNrHits() + ">");
		if (cond != null) {
			cond.print();
		}
		if (setOfTruthVectors.size() > 0) {
			Logger.log(" <TruthVectors: ");
			for (String s : setOfTruthVectors) {
				Logger.log(s + "->" + cond.evaluateTruthVector(s) + " ");
			}
			Logger.log(">");
		}
		Logger.logln("---\n" + content);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#getNodeString()
	 */
	public String getNodeString() {
		String s = super.getNodeString() + "<If -> " + endNode.getID() + " >" + getExitString() + "<nrConds: "
				+ nrConditions + "><" + nrHits + "><w:" + trueNode.getNrHits() + ", f:" + falseNode.getNrHits()
				+ ">---\n";
		if (cond != null)
			s += cond.getCondString();

		if (setOfTruthVectors.size() > 0) {
			s += "---<TruthVectors: ";
			for (String s2 : setOfTruthVectors) {
				s += s2 + "->" + cond.evaluateTruthVector(s2) + " ";
			}
			s += ">";
		}
		return s;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#getMetrics()
	 */
	public Metrics getMetrics() {
		if (cond != null)
			return cond.metrics;
		return new Metrics();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#printCFG()
	 */
	public void printCFG() {
		Logger.logln("(" + id + "#" + nrHits + " -> " + elseNode.getID() + ")-P");
		Logger.logln(" |");
	}

}