package sota.ast.nodes;

import java.util.HashSet;
import java.util.TreeSet;

import sota.Logger;
import sota.ast.util.Metrics;

/**
 * AST-Knotenklasse fr alle Schleifenstrukturen.
 * 
 * <PRE>
 *   	.. 									 *   	..
 *   	ASTNode				entryNode		 *   	ASTNode				entryNode						
 *   	(ASTLabelNode)						 *		(ASTLabelNode)
 * 		ASTIterationNode	this			 *		ASTDoNode			doNode
 * 		  ASTStatementNode	loopNode		 *		  ASTStatementNode	loopNode
 * 		  &gt;  ..								 * 		  &gt; ..
 *      ASTStatementNode	endNode			 *      ASTIterationNode	this		
 *      ..									 *      ASTStatementNode	endNode
 * 
 * </PRE>
 * */
public abstract class ASTIterationNode extends ASTBranchNode implements IConditionNode {

	/** Bei der Konstruktion zu setzende Informationen und Verknpfungen * */

	protected ASTDoNode doNode;

	protected ASTStatementNode loopNode;

	protected Condition cond;

	protected String forInit;
	protected String forUpdate;

	protected HashSet<String> setOfTruthVectors = new HashSet<String>();
	protected boolean isForNode = false;
	protected boolean hasDoNode = false;
	protected int nrSingleLoops = 0;
	protected int nrMultipleLoops = 0;
	protected int nrSkips = 0;
	protected int nrConditions = 0;

	/**
	 * Konstruktor fr While-Schleifen.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param c
	 *            die Bedingung
	 */
	public ASTIterationNode(int ln, String s, Condition c) {
		super(ln, s);
		cond = c;
		cond.setPrimary(true);
		nrConditions = cond.getNrConditions();
		if (!cond.isOperator()) {
			cond.setAtom(true);
		}
		doNode = null;
		loopNode = null;
	}

	/**
	 * Der Konstruktor fr eine For-Schleife.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param _forInit
	 *            die For-Initialisierung
	 * @param _cond
	 *            die Bedingung
	 * @param _forUpdate
	 *            das For-Update
	 */
	public ASTIterationNode(int ln, String s, String _forInit, Condition _cond, String _forUpdate) {
		super(ln, s);
		cond = _cond;
		cond.setPrimary(true);
		nrConditions = cond.getNrConditions();
		if (cond == null)
			Logger.log("Error@ASTIterationNode");
		if (!cond.isOperator()) {
			cond.setAtom(true);
		}
		forInit = _forInit.trim();
		forUpdate = _forUpdate.trim();
		// Logger.log(s + "\n" + forInit + "\n" + forUpdate);
		isForNode = true;
		doNode = null;
		loopNode = null;
	}

	/**
	 * Konstruktor fr For-Each-Schleifen.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 */
	public ASTIterationNode(int ln, String s) {
		super(ln, s);
		isForNode = true;
		this.cond = null;
		doNode = null;
		loopNode = null;
		return;
	}

	/**
	 * Setzt den Do-Knoten.
	 * 
	 * @param dn
	 *            der Do-Knoten
	 */
	public void setDoNode(ASTDoNode dn) {
		doNode = dn;
		hasDoNode = true;
	}

	/**
	 * Liefert den Do-Knoten
	 * 
	 * @return der Do-Knoten
	 */
	public ASTDoNode getDoNode() {
		return doNode;
	}

	/**
	 * Setzt den Loop-Knoten.
	 * 
	 * @param ln
	 *            der Loop-Knoten
	 */
	public void setLoopNode(ASTStatementNode ln) {
		loopNode = ln;
		loopNode.setControlingBranchNode(this);
		return;
	}

	/**
	 * Liefert den Loop-Knoten.
	 * 
	 * @return der Loopknoten
	 */
	public ASTStatementNode getLoopNode() {
		return loopNode;
	}

	/**
	 * Entscheidet, ob die Schleife eine For-Schleife ist.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean isForNode() {
		return isForNode;
	}

	/**
	 * Entscheidet, ob die Funktion eine Do-While-Schleife ist.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean hasDoNode() {
		return hasDoNode;
	}

	/**
	 * Liefert die Anzahl an Schleifenauslassungen.
	 * 
	 * @return die Anzahl
	 */
	public int getNrSkips() {
		return nrSkips;
	}

	/**
	 * Liefert die Anzahl an Schleifendurchlufen insgesamt.
	 * 
	 * @return die Anzahl
	 */
	public int getNrLoops() {
		return loopNode.getNrHits();
	}

	/**
	 * Liefert die Anzahl an einmaligen Schleifendurchlufen.
	 * 
	 * @return die Anzahl
	 */
	public int getNrSingleLoops() {
		return nrSingleLoops;
	}

	/**
	 * Liefert die Anzahl an mehrmaligen Schleifendurchlufen.
	 * 
	 * @return die Anzahl
	 */
	public int getNrMultipleLoops() {
		return nrMultipleLoops;
	}

	/**
	 * Erhht die Anzahl an Schleifenauslassungen um eins.
	 */
	public void incNrSkips() {
		nrSkips++;
	}

	/**
	 * Erhht die Anzahl an Schleifenauslassungen um einen Wert.
	 * 
	 * @param i
	 *            der Wert
	 */
	public void incNrSkips(int i) {
		nrSkips += i;
	}

	/**
	 * Erhht die Anzahl an einmaligen Schleifenausfhrungen um eins.
	 */
	public void incNrSingleLoops() {
		nrSingleLoops++;
	}

	/**
	 * Erhht die Anzahl an einmaligen Schleifenausfhrungen um einen Wert.
	 * 
	 * @param i
	 *            der Wert
	 */
	public void incNrSingleLoops(int i) {
		nrSingleLoops += i;
	}

	/**
	 * Erhht die Anzahl an mehrmaligen Schleifenausfhrungen um eins.
	 */
	public void incNrMultipleLoops() {
		nrMultipleLoops++;
	}

	/**
	 * Erhht die Anzahl an mehrmaligen Schleifenausfhrungen um einen Wert.
	 * 
	 * @param i
	 *            der Wert
	 */
	public void incNrMultipleLoops(int i) {
		nrMultipleLoops += i;
	}

	/** Prints ** */

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#printNode()
	 */
	public void printNode() {
		Logger.log("---");
		printReach();
		if (hasDoNode) {
			Logger.log("<" + id + "><Iterate " + entryNode.getID() + " , Do: " + doNode.getID() + " -> "
					+ endNode.getID() + ">");
		} else {
			Logger.log("<" + id + "><Iterate " + entryNode.getID() + " -> " + endNode.getID() + ">");

		}
		printExits();
		Logger.log("<" + nrHits + "><hits: " + loopNode.getNrHits() + ", multiple: " + nrMultipleLoops + ", skips: "
				+ nrSkips + ">\n");

		cond.print();

		if (setOfTruthVectors.size() > 0) {
			Logger.log(" <TruthVectors: ");
			for (String s : setOfTruthVectors) {
				Logger.log(s + "->" + cond.evaluateTruthVector(s) + " ");
			}
			Logger.log(">");
		}
		Logger.log("---\n" + content);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#getNodeString()
	 */
	public String getNodeString() {
		String s = super.getNodeString();
		if (hasDoNode) {
			s += "<Iterate " + entryNode.getID() + " , Do: " + doNode.getID() + " -> " + endNode.getID() + ">";
		} else {
			s += "<Iterate " + entryNode.getID() + " -> " + endNode.getID() + ">";

		}

		if (isForNode) {
			s += "<ForInit: " + forInit + " , ForUpdate: " + forUpdate + ">";
		}

		s += getExitString() + "<skips: " + nrSkips + ", singleLoops: " + nrSingleLoops + ", multipleLoops: "
				+ nrMultipleLoops + ">---\n";
		if (cond != null)
			s += cond.getCondString();
		else
			s += " No Condition! ";

		if (setOfTruthVectors.size() > 0) {
			s += "---<TruthVectors: ";
			for (String s2 : setOfTruthVectors) {
				s += s2 + "->" + cond.evaluateTruthVector(s2) + " ";
			}
			s += ">";
		}

		return s;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#printCFG()
	 */
	public void printCFG() {
		if (hasDoNode)
			Logger.logln("(" + id + "#" + nrHits + " -> " + doNode.getID() + ")");
		else
			Logger.logln("(" + id + "#" + nrHits + " -> " + endNode.getID() + ")");
		Logger.logln(" |");
	}

	/* Start Of ConditionMethods */

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#fillConditions(java.util.TreeSet)
	 */
	public void fillConditions(TreeSet<String> list) {
		setOfTruthVectors.addAll(list);
		for (String s : list) {
			cond.applyTruthVector(s);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#getCondition()
	 */
	public Condition getCondition() {
		return cond;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#resetHits()
	 */
	public void resetHits() {
		super.resetHits();
		nrMultipleLoops = 0;
		nrSingleLoops = 0;
		nrSkips = 0;
		setOfTruthVectors.clear();
		if (cond != null)
			cond.resetHits();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#getTruthVectors()
	 */
	public HashSet<String> getTruthVectors() {
		return setOfTruthVectors;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#getEvaluatedTVString(java.lang.String)
	 */
	public String getEvaluatedTVString(String s) {
		if (cond == null)
			return "";
		return (cond.evaluateTruthVector(s) ? "1" : "0") + cond.getEvaluatedTVString(s);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#hasNonTrivialCondition()
	 */
	public boolean hasNonTrivialCondition() {
		if (cond == null)
			return false;
		return cond.isNonTrivial();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTBranchNode#completeNode()
	 */
	public void completeNode() {
		addFollowUpNode(loopNode);
		addFollowUpNode(endNode);
	}

	/** numeriert die Konditionen, startend mit Parameterwert */
	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#numerateCondition(int)
	 */
	public int numerateCondition(int i) {
		if (cond == null)
			return i;
		return cond.numerate(i);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#computeConditionCoverage()
	 */
	public void computeConditionCoverage() {
		if (cond == null)
			return;
		cond.computeConditionCoverage(setOfTruthVectors);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IConditionNode#getMetrics()
	 */
	public Metrics getMetrics() {
		if (cond != null)
			return cond.metrics;
		return new Metrics();
	}

	/* End Of ConditionMethods */
}