package sota.ast.nodes;

import sota.Logger;

/**
 * Basisknoten des AST, der fr eine Quellcodesegment steht. Alle anderen Knoten (auer Bedingungs- und Dateiknoten)
 * leiten sich aus diesem ab.
 * */
public class ASTNode implements Comparable<ASTNode> {
	// Eindeutige ID ber gesamte Projekt
	protected int id;

	/** Bei der Konstruktion zu setzende Informationen und Verknpfungen * */

	protected int lineNr = 0;

	protected String content;

	protected ASTFunctionNode functionNode;

	protected boolean isVirtual;

	/** Automatisch erstellte Daten ****************************************/

	protected boolean isDirectReachable;
	protected boolean isReachable;

	// enthlt augementierten Content fr diesen ASTKnoten
	protected String ascContent;
	protected ASTNode next;
	protected ASTNode prev;

	// der Verzweigungsknoten, zu dem dieser Knoten die Entscheidung im Log hlt (true-false-nodes bei if z.b.)
	// oder null
	protected ASTBranchNode controlingBranchNode = null;

	public static final int NODE = 0;
	public static final int DO = 1;
	public static final int SCOPE = 10;
	public static final int STATEMENT = 20;
	public static final int CASE = 21;
	public static final int CATCH = 22;
	public static final int JUMP = 23;
	public static final int BRANCH = 30;
	public static final int SWITCH = 31;
	public static final int TRY = 32;
	public static final int CONDITION = 40;
	public static final int IF = 41;
	public static final int ITERATION = 42;

	/**
	 * Konstruktor fr einfache ASTKnoten.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 */
	public ASTNode(int ln, String s) {
		isReachable = false;
		isDirectReachable = false;
		content = s;
		ascContent = "";
		isVirtual = false;
		lineNr = ln;
	}

	/**
	 * Konstruktor fr virtuelle ASTKnoten.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 */
	public ASTNode(int ln) {
		isReachable = false;
		isDirectReachable = false;
		content = "";
		ascContent = "";
		isVirtual = true;
		lineNr = ln;
	}

	/**
	 * Liefert den zugehrigen Funktionsknoten zurck.
	 * 
	 * @return der ASTFunctionNode oder null
	 */
	public ASTFunctionNode getFunctionNode() {
		return functionNode;
	}

	/**
	 * Setzt den zugehrigen Funktionsknoten.
	 * 
	 * @param functionNode
	 *            der ASTFunctionNode
	 */
	public void setFunctionNode(ASTFunctionNode functionNode) {
		this.functionNode = functionNode;
		switch (getType()) {
		case IF:
		case ITERATION:
			functionNode.addConditionNode((IConditionNode) this);
		}
	}

	/**
	 * Setzt den zugehrigen BranchNode, dessen Verzweigung mit diesem Knoten geloggt wird. (trueNode, falseNode,
	 * caseNode,...)
	 * 
	 * @param node
	 *            der ASTBranchNode
	 */
	public void setControlingBranchNode(ASTBranchNode node) {
		controlingBranchNode = node;
	}

	/**
	 * Liefert den BranchNode, dessen Verzweigung durch diesen Knoten geloggt wird, oder null.
	 * 
	 * @return der ASTBranchNode, oder null
	 */
	public ASTBranchNode getControlingBranchNode() {
		return controlingBranchNode;
	}

	// Getter und Setter

	/**
	 * Legt die ID fest.
	 * 
	 * @param i
	 *            die ID
	 */
	public void setID(int i) {
		id = i;
		return;
	}

	/**
	 * Gibt die ID zurck.
	 * 
	 * @return die ID
	 */
	public int getID() {
		return id;
	}

	/**
	 * Setzt die Knotenreferenz auf den Vorgnger.
	 * 
	 * @param n
	 *            der Vorgngerknoten
	 */
	public void setPrevNode(ASTNode n) {
		prev = n;
		return;
	}

	/**
	 * Gibt den Vorgngerknoten zurck.
	 * 
	 * @return der Vorgngerknoten
	 */
	public ASTNode getPrevNode() {
		return prev;
	}

	/**
	 * Setzt die Knotenreferenz auf den Nachfolger.
	 * 
	 * @param n
	 *            der Nachfolgeknoten
	 */
	public void setNextNode(ASTNode n) {
		next = n;
		return;
	}

	/**
	 * Gibt den Nachfolger zurck.
	 * 
	 * @return der Nachfolgeknoten
	 */
	public ASTNode getNextNode() {
		return next;
	}

	/**
	 * Setzt den Knoteninhalt, d.h. das durch ihn reprsentierte Quellcodesegment.
	 * 
	 * @param s
	 *            der Inhalt
	 */
	public void setContent(String s) {
		content = s;
		return;
	}

	/**
	 * Liefert das Quellcodesegment des Knotens zurck.
	 * 
	 * @return der Inhalt
	 */
	public String getContent() {
		return content;
	}

	/**
	 * Liefert die Zeilennummer innerhalb der Datei.
	 * 
	 * @return die Zeilennummer
	 */
	public int getLineNr() {
		return lineNr;
	}

	/**
	 * Gibt den Typ des Knotens zurck.
	 * 
	 * @return der Typ
	 */
	public int getType() {
		if (this instanceof ASTIfNode) {
			return IF;
		}
		if (this instanceof ASTSwitchNode) {
			return SWITCH;
		}
		if (this instanceof ASTIterationNode) {
			return ITERATION;
		}
		if (this instanceof ASTDoNode) {
			return DO;
		}
		if (this instanceof ASTJumpNode) {
			return JUMP;
		}
		if (this instanceof IStatementNode) {
			return STATEMENT;
		}
		if (this instanceof ASTTryNode) {
			return TRY;
		}
		return NODE;
	}

	// Funktionen fr boolsche Parameter

	/**
	 * Entscheidet, ob der Knoten virtuell ist.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean isVirtual() {
		return isVirtual;
	}

	/**
	 * Setzt den Knoten auf virtuell bzw. nicht-virtuell.
	 * 
	 * @param b
	 *            true, wenn virtuell, sonst false
	 */
	public void isVirtual(boolean b) {
		isVirtual = b;
	}

	/**
	 * Entscheidet, ob der Knoten im Kontrollfluss erreichbar ist.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean isReachable() {
		return isReachable;
	}

	/**
	 * Legt fest, ob dieser Knoten im Kontrollfluss erreichbar ist.
	 * 
	 * @param b
	 *            true, wenn ja, sonst false
	 */
	public void isReachable(boolean b) {
		isReachable = b;
	}

	/**
	 * Entscheidet, ob der Knoten im Kontrollfluss von seinem Vorgnger erreichbar ist.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean isDirectReachable() {
		return isDirectReachable;
	}

	/**
	 * Legt fest, ob der Knoten ber seinen Vorgnger erreichbar ist.
	 * 
	 * @param b
	 *            true, falls ja, sonst nein
	 */
	public void isDirectReachable(boolean b) {
		isDirectReachable = b;
	}

	// Funktionen fr Instrumentierung

	/**
	 * Erstellt die Instrumentierung des Knotens.
	 * 
	 * @param level
	 *            das gewhlte Instrumenationslevel
	 */
	public void createASCContent(int level) {
		ascContent += content;
	}

	/**
	 * Legt den instrumenierten Inhalt des Knotens fest.
	 * 
	 * @param s
	 *            der instrumentierte Inhalt
	 */
	public void setASCContent(String s) {
		ascContent = s;
	}

	/**
	 * Fgt an den instrumentierten Inhalt einen String an.
	 * 
	 * @param s
	 *            der anzufgende String
	 */
	public void appendToASCContent(String s) {
		ascContent += s;
	}

	/**
	 * Liefert den instrumentierten Inhalt des Knotens zurck.
	 * 
	 * @return der instrumentierte Inhalt
	 */
	public String getASCContent() {
		return ascContent;
	}

	// Konsolenausgabe-Funktionen bzw. NodeView-Ausgabe

	/**
	 * Ausgabe des Inhaltes.
	 */
	public void print() {
		Logger.logln(content);
	}

	/**
	 * Ausgabe als Kontrollflusselement.
	 */
	public void printCFG() {
		Logger.logln("(" + id + ")");
		Logger.logln(" |");
	}

	/**
	 * Ausgabe der Erreichbarkeit.
	 */
	public void printReach() {
		if (isDirectReachable) {
			Logger.log("<D");
		} else if (isReachable) {
			Logger.log("<R");
		} else {
			Logger.log("<N");
		}
		Logger.log(">");
	}

	/**
	 * Ausgabe des Knoteninhaltes sowie zustzlicher Informationen.
	 */
	public void printNode() {
		Logger.logln("---");
		printReach();
		if (isVirtual) {
			Logger.logln("<" + id + "><V>----------");
		} else {
			Logger.logln("<" + id + ">----------\n" + content);
		}
	}

	/**
	 * Gibt einen String mit der Darstellung der Erreichbarkeit zurck.
	 * 
	 * @return der Erreichbarkeitsstring
	 */
	public String getReachString() {
		String s = "";
		if (isDirectReachable) {
			s += "<D";
		} else if (isReachable) {
			s += "<R";
		} else {
			s += "<N";
		}
		s += ">";
		return s;
	}

	/**
	 * Gibt eine Stringdarstellung des Knotens und zustzlicher Strukturinformationen zurck.
	 * 
	 * @return der Knotenstring
	 */
	public String getNodeString() {
		String s = "---" + "<L:" + lineNr + ">-<ID:" + id + ">-" + getReachString();
		if (isVirtual)
			s += "<V>";
		return s + "-";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(ASTNode n) {
		/*
		 * if (this instanceof ASTScopeNode && n instanceof ASTScopeNode) { ASTScopeNode sn1 = (ASTScopeNode) this, sn2
		 * = (ASTScopeNode) n; return (sn1.getIDString()).compareTo(sn2.getIDString()); }
		 */
		return id - n.getID();
	}

}