package sota.ast.nodes;

import java.util.TreeSet;

import sota.Logger;
import sota.ast.util.Metrics;

/**
 * AST-Knoten fr Scopestatements, also ASTClassNode und ASTFunctionNode.
 * 
 * <PRE>
 * ASTScopeNode this &gt; .. ASTNode endNode
 * </PRE>
 * 
 * */
public abstract class ASTScopeNode extends ASTNode implements IScopeNode {

	/** Bei der Konstruktion zu setzende Informationen und Verknpfungen * */

	// der zugehrige Scope, ohne Package/Namespace
	protected String scopeName;

	// Name der Datei, der Klasse oder der Funktion.
	protected String name;

	// Verschachtelter Name bis zur TopLevelKlasse
	protected String qualifiedName;

	// bergeordnete ScopeNode
	protected IScopeNode parent;

	protected FileNode fileNode;

	protected TreeSet<ASTFunctionNode> setOfFunctionNodes;

	protected TreeSet<ASTClassNode> setOfClasseNodes;

	protected ASTNode endNode;

	/**************************************************/

	protected boolean isTopLevel = false;
	protected boolean isFirstLevelFunction = false;

	protected Metrics metrics;

	// Konstruktor

	/**
	 * Der Konstruktor des ScopeNodes.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param scope
	 *            der Scope des Knotens
	 * @param name
	 *            der Name
	 * @param parent
	 *            der bergeordnete ScopeNode
	 */
	public ASTScopeNode(int ln, String scope, String name, IScopeNode parent) {
		super(ln, "");
		setOfFunctionNodes = new TreeSet<ASTFunctionNode>();
		setOfClasseNodes = new TreeSet<ASTClassNode>();
		this.scopeName = scope;
		this.name = name;
		this.parent = parent;

		metrics = new Metrics();
		if (scope == "")
			isTopLevel = true;

		if (this instanceof ASTFunctionNode && parent.isTopLevel())
			isFirstLevelFunction = true;
	}

	// Implementationen der Funktionen des Interface ScopeNode

	/**
	 * Bestimmt den qualifizierten Namen des Scopes.
	 */
	protected void setQualifiedName() {
		IScopeNode p = parent;
		qualifiedName = "";
		while (!(p instanceof FileNode) && p != null) {
			qualifiedName = p.getName() + "." + qualifiedName;
			p = p.getParent();
		}
		qualifiedName += getName();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#addClassNode(sota.ast.nodes.ASTClassNode)
	 */
	public void addClassNode(ASTClassNode cn) {
		setOfClasseNodes.add(cn);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getClassNodes()
	 */
	public TreeSet<ASTClassNode> getClassNodes() {
		return setOfClasseNodes;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#addFunctionNode(sota.ast.nodes.ASTFunctionNode)
	 */
	public void addFunctionNode(ASTFunctionNode fn) {
		setOfFunctionNodes.add(fn);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getFunctionNodes()
	 */
	public TreeSet<ASTFunctionNode> getFunctionNodes() {
		return setOfFunctionNodes;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#setEndNode(sota.ast.nodes.ASTNode)
	 */
	public void setEndNode(ASTNode dn) {
		endNode = dn;
		if (this instanceof ASTFunctionNode)
			((ASTFunctionNode) this).addExitNode((ASTStatementNode) dn);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getEndNode()
	 */
	public ASTNode getEndNode() {
		return endNode;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getStartNode()
	 */
	public ASTNode getStartNode() {
		return this;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#isTopLevel()
	 */
	public boolean isTopLevel() {
		return isTopLevel;
		// return this.scope.equals("");
	}

	/**
	 * Entscheidet, ob es sich um eine Funktion der obersten Klassenebene handelt.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean isFirstLevelFunction() {
		return isFirstLevelFunction;
		// return this.scope.equals("");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getScope()
	 */
	public String getScope() {
		return scopeName;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getIDString()
	 */
	public String getIDString() {
		String idString = fileNode.getFullQualifiedName() + ":";

		if (!scopeName.equals("")) {
			idString += scopeName + "::";
		}

		idString += name;

		if (this instanceof ASTFunctionNode) {
			idString += "(" + ((ASTFunctionNode) this).arguments + ")";
		}

		return idString;
	}

	// Konsolenausgabefunktionen

	/**
	 * Ausgabe der Funktionen.
	 * 
	 * @param indent
	 *            der Einschub
	 */
	public void printFunctions(int indent) {
		for (ASTScopeNode sn : setOfClasseNodes) {
			print_indent(indent);
			Logger.logln(" Class: " + sn.getScope() + "::\t" + sn.getName());
			sn.printFunctions(indent + 1);
		}
		for (ASTScopeNode sn : setOfFunctionNodes) {
			print_indent(indent);
			Logger.logln(" Function: " + sn.getScope() + "::\t" + sn.getName() + " ("
					+ ((ASTFunctionNode) sn).getArguments() + ")");
			sn.printFunctions(indent + 1);
		}
	}

	/**
	 * Ausgabe eines Einschubs.
	 * 
	 * @param indent
	 *            die Gre des Einschubs
	 */
	public void print_indent(int indent) {
		for (int i = 0; i < indent; i++) {
			Logger.log("\t");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getParent()
	 */
	public IScopeNode getParent() {
		return parent;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getFileNode()
	 */
	public FileNode getFileNode() {
		return fileNode;
	}

	/**
	 * Setzt den zugehrigen FileNode.
	 * 
	 * @param n
	 *            der FileNode
	 */
	public void setFileNode(FileNode n) {
		fileNode = n;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getMetrics()
	 */
	public Metrics getMetrics() {
		return metrics;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#setMetrics(sota.ast.util.Metrics)
	 */
	public void setMetrics(Metrics metrics) {
		this.metrics = metrics;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getFullyQualifiedName(boolean, boolean)
	 */
	public String getFullyQualifiedName(boolean showPackages, boolean showFiles) {

		return (showPackages && fileNode.getPackageName().length() > 0 ? fileNode.getPackageName() + "." : "")
				+ (showFiles ? fileNode.getName() + ":" : "") + qualifiedName;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getFirstLineNr()
	 */
	public int getFirstLineNr() {
		return this.lineNr;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getLastLineNr()
	 */
	public int getLastLineNr() {
		ASTNode n = next;
		while (!(n instanceof ASTCompoundNode))
			n = n.getNextNode();
		return ((ASTCompoundNode) n).getEndNode().getLineNr();
	}

}