package sota.ast.nodes;

import java.util.ArrayList;

import sota.Logger;

/**
 * AST-Knoten fr Switch-Statements
 * 
 * <PRE>
 *   	..
 *   	ASTNode				entryNode
 * 		ASTSwitchNode		this
 * 		  ASTCaseNode		[listOfCaseNodes]
 * 		  &gt;  ..
 * 		  ASTCaseNode		[listOfCaseNodes]
 * 	      &gt;  ..
 *      ASTStatementNode	endNode	
 *      ..
 * </PRE>
 */
public abstract class ASTSwitchNode extends ASTBranchNode {

	/** Bei der Konstruktion zu setzende Informationen und Verknpfungen * */

	protected String expression;

	protected ArrayList<ASTCaseNode> listOfCaseNodes = new ArrayList<ASTCaseNode>();

	/**
	 * Von ASTBranchNode: *
	 * 
	 * protected ASTNode entryNode protected ASTStatementNode endNode
	 * 
	 * / /
	 ***********************************************************/

	protected boolean hasDefault;

	/**
	 * Konstruktor fr das Switch.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Auswahlausdruck
	 */
	public ASTSwitchNode(int ln, String s) {
		super(ln, s);
		hasDefault = false;
	}

	/**
	 * Fgt einen Case-Knoten hinzu.
	 * 
	 * @param cn
	 *            der Case-Knoten
	 */
	public void addCaseNode(ASTCaseNode cn) {
		cn.setSwitchNode(this);
		listOfCaseNodes.add(cn);
		if (cn.isDefault()) {
			hasDefault = true;
		}
	}

	/**
	 * Gibt die Liste der Cases zurck.
	 * 
	 * @return die Liste der Case-Knoten
	 */
	public ArrayList<ASTCaseNode> getListOfCases() {
		return listOfCaseNodes;
	}

	/**
	 * Gibt die Anzahl an Cases zurck.
	 * 
	 * @return die Anzahl an Cases
	 */
	public int getNrOfCases() {
		return listOfCaseNodes.size();
	}

	/**
	 * Liefert den Auswahlstring des Switch zurck, d.h. eine Abbildung der Case-Auswahlen auf einen 0-1-String.
	 * 
	 * @return der String
	 */
	public String getSelectionString() {
		String s = "";
		for (ASTCaseNode n : listOfCaseNodes) {
			if (n.getNrSelects() > 0) {
				s += "1";
			} else {
				s += "0";
			}
		}
		return s;
	}

	/**
	 * Liefert den Switch-Ausdruck zurck.
	 * 
	 * @return der Ausdruck
	 */
	public String getExpression() {
		return expression;
	}

	/**
	 * Entscheidet, ob das Switch ein Default besitzt oder nicht.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean hasDefault() {
		return hasDefault;
	}

	/**
	 * Gibt den Default-Knoten zurck, wenn vorhanden.
	 * 
	 * @return der Default-Knoten oder null
	 */
	public ASTNode getDefaultNode() {
		for (ASTCaseNode cn : listOfCaseNodes)
			if (cn.isDefault)
				return cn;
		return null;
	}

	/** Print ** */
	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#printNode()
	 */
	public void printNode() {
		Logger.log("---");
		printReach();
		Logger.log("<" + id + "><Switch -> " + endNode.getID() + "><Cases: " + listOfCaseNodes.size() + ">");
		printExits();
		Logger.logln("<" + nrHits + "><Selection: " + getSelectionString() + ">---\n" + content);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#getNodeString()
	 */
	public String getNodeString() {
		return super.getNodeString() + "<Switch -> " + endNode.getID() + "><Cases: " + listOfCaseNodes.size() + ">"
				+ getExitString() + "<" + nrHits + "><Selection: " + getSelectionString() + ">";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#printCFG()
	 */
	public void printCFG() {
		Logger.log("(" + id + "#" + nrHits + " ->");
		for (ASTCaseNode n : listOfCaseNodes)
			Logger.log(" " + n.getID());
		Logger.logln(")");
		Logger.logln("");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTBranchNode#completeNode()
	 */
	public void completeNode() {
		for (ASTCaseNode n : listOfCaseNodes) {
			addFollowUpNode(n);
		}
		if (!hasDefault)
			addFollowUpNode(endNode);
	}

}