package sota.ast.nodes;

import java.util.ArrayList;

import sota.Logger;

/**
 * AST-Knoten fr Ausnahme-Blcke
 * 
 * <PRE>
 *   	..
 *   	ASTNode				entryNode
 * 		ASTTryNode			this
 * 		  ASTStatementNode	entryOfTryBlock
 * 		  &gt;  ..
 * 		  ASTNode			endOfTryBlock
 * 	    ASTCatchNode		[listOfCatchNodes]
 *        &gt;  ..
 * 	    ASTCatchNode		[listOfCatchNodes]
 *        &gt;  ..
 *      ASTStatementNode	endNode	
 *      ..
 * </PRE>
 */
public abstract class ASTTryNode extends ASTBranchNode {

	/** Bei der Konstruktion zu setzende Informationen und Verknpfungen * */

	protected ASTStatementNode entryOfTryBlock = null;

	protected ASTNode endOfTryBlock = null;

	// enthlt sowohl Catch als auch Finally
	protected ArrayList<ASTCatchNode> listOfCatchNodes;

	protected ASTCatchNode finallyNode;

	/***********************************************************************/

	protected int nrHitsWithoutException = 0;

	/**
	 * Der Konstruktor fr Switch-Knoten.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 */
	public ASTTryNode(int ln, String s) {
		super(ln, s);
		listOfCatchNodes = new ArrayList<ASTCatchNode>();
		finallyNode = null;
	}

	/**
	 * Gibt das Ende des TryBlocks zurck.
	 * 
	 * @return der Endknoten
	 */
	public ASTNode getEndOfTryBlock() {
		return endOfTryBlock;
	}

	/**
	 * Setzt das Ende des TryBlocks.
	 * 
	 * @param n
	 *            das Ende des TryBlocks
	 */
	public void setEndOfTryBlock(ASTNode n) {
		endOfTryBlock = n;
	}

	/**
	 * Gibt den Eingang des TryBlocks zurck.
	 * 
	 * @return der Eingangsknoten
	 */
	public ASTStatementNode getEntryOfTryBlock() {
		return entryOfTryBlock;
	}

	/**
	 * Setzt den Eingang des TryBlocks.
	 * 
	 * @param n
	 *            der Eingang des TryBlocks
	 */
	public void setEntryOfTryBlock(ASTStatementNode n) {
		entryOfTryBlock = n;
		// entryOfTryBlock.setControlNode(this);
	}

	/**
	 * Fgt einen Catch-Node hinzu.
	 * 
	 * @param cn
	 *            der CatchNode
	 */
	public void addCatchNode(ASTCatchNode cn) {
		listOfCatchNodes.add(cn);
		cn.setTryNode(this);
		if (cn.isFinally())
			finallyNode = cn;
	}

	/**
	 * Gibt den Finally-Knoten zurck.
	 * 
	 * @return der Finally-Knoten, oder null
	 */
	public ASTCatchNode getFinallyNode() {
		return finallyNode;
	}

	/**
	 * Entscheidet, ob das Try ein Finally besitzt.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean hasFinally() {
		return finallyNode != null;
	}

	/**
	 * Gibt die Liste der CatchNodes zurck.
	 * 
	 * @return die Liste der ASTCatchNodes
	 */
	public ArrayList<ASTCatchNode> getListOfCatches() {
		return listOfCatchNodes;
	}

	/**
	 * Gibt die Anzahl an Auswertungen ohne Ausnahme zurck.
	 * 
	 * @return die Anzahl an Auswertungen ohne Ausnahme
	 */
	public int getNrHitsWithoutException() {
		return nrHitsWithoutException;
	}

	/**
	 * Erhht die Anzahl an Auswertungen ohne Ausnahme um eins.
	 */
	public void incNrHitsWithoutException() {
		nrHitsWithoutException++;
	}

	/**
	 * Setzt die Anzahl an Auswertungen ohne Ausnahme.
	 * 
	 * @param integer
	 *            die Anzahl der Auswertungen ohne Ausnahme
	 */
	public void setNrHitsWithoutException(Integer integer) {
		nrHitsWithoutException = integer;

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTBranchNode#completeNode()
	 */
	public void completeNode() {
		addFollowUpNode(next);
		for (ASTCatchNode n : listOfCatchNodes) {
			addFollowUpNode(n);
		}
		addFollowUpNode(finallyNode);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#resetHits()
	 */
	public void resetHits() {
		super.resetHits();
		nrHitsWithoutException = 0;
	}

	/** Print ** */
	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#printNode()
	 */
	public void printNode() {
		Logger.log("---");
		printReach();
		Logger.log("<" + id + "><Try -> " + endNode.getID() + "><" + listOfCatchNodes.size() + ">");
		printExits();
		Logger.logln("---\n" + content);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.ASTStatementNode#getNodeString()
	 */
	public String getNodeString() {
		return super.getNodeString() + "<Try -> " + endNode.getID() + "><" + listOfCatchNodes.size() + ">"
				+ getExitString();
	}

}