package sota.ast.nodes;

import java.util.ArrayList;
import java.util.HashSet;

import sota.Logger;
import sota.ast.util.Metrics;

/**
 * Knoten fr Bedingungen. Bedingungen sind Atome oder Operatoren, die aus Subbedingungen aufgebaut sind.
 * */
public abstract class Condition {
	// Typunterscheidung

	protected boolean isPrimary = false; // Primre Bedingung, d.h. nicht in
	// anderer Bedingung enthalten
	protected boolean isAtom = false; // ist boolean Atom
	protected boolean isTrueAtom = false; // "true"
	protected boolean isFalseAtom = false; // "false"
	protected boolean isOperator = false; // ist boolean Operator

	protected boolean hasShortCircuit = false; // Operator "||" oder "&&"
	protected boolean shortCircuitValue = true; // bei "||" true, bei "&&" false

	protected boolean isEmpty = false; // leer, d.h. wahr (nur in for-schleife)

	protected int nrConditions = 0;
	protected Metrics metrics = new Metrics();

	// Id der PrimaryCondition innerhalb einer Funktion
	protected int conditionId = -1;

	// Id des Atoms innerhalb einer komplexen Bedingung (-1, falls nicht Atom)
	protected int atomId = -1;

	protected int nrTrueEvaluations = 0;

	protected int nrFalseEvaluations = 0;

	// falls binrer Operator, linke und rechte Teilbedingung
	// falls unrer Operator, dann ist leftCondition == null
	protected Condition leftCondition;
	// falls binrer Operator, linke und rechte Teilbedingung
	// falls unrer Operator, dann ist leftCondition == null
	protected Condition rightCondition;

	protected String ascContent = null;

	protected String content = "";

	/**
	 * Konstruktor fr binren Operator.
	 * 
	 * @param lC
	 *            linke Bedingung
	 * @param op
	 *            der Operator
	 * @param rC
	 *            rechte Bedingung
	 */
	public Condition(Condition lC, String op, Condition rC) {
		isOperator = true;
		leftCondition = lC;
		content = op;
		rightCondition = rC;
		nrConditions = getListOfConditions().size();
	}

	/**
	 * Konstruktor fr unren (Nicht-)Operator
	 * 
	 * @param op
	 *            unrer Operator
	 * @param rC
	 *            rechte Bedingung
	 */
	public Condition(String op, Condition rC) {
		isOperator = true;
		leftCondition = null;
		content = op;
		rightCondition = rC;
		nrConditions = getListOfConditions().size();
	}

	/**
	 * Konstruktor fr atomare Bedingung
	 * 
	 * @param atom
	 *            das Atom
	 */
	public Condition(String atom) {
		isAtom = true;
		if (atom.equals("true")) {
			isTrueAtom = true;
		} else if (atom.equals("false")) {
			isFalseAtom = true;
		}
		content = atom;
		nrConditions = 1;
	}

	/**
	 * Konstruktor fr leere, d.h. wahre Bedingung
	 */
	public Condition() {
		isAtom = true;
		isEmpty = true;
		isTrueAtom = true;
		content = "";
		nrConditions = 1;
	}

	// Konstruktor fr CondAssignment, momentan nicht genutzt
	/*
	 * public Condition(Condition cond, Condition tC, Condition fC) { condition = cond; leftCondition = tC;
	 * leftCondition.setAtom(false); rightCondition = fC; rightCondition.setAtom(false); content = cond.getContent() +
	 * " ? " + tC.getContent() + ":" + fC.getContent(); }
	 */

	/**
	 * Liefert die ID der Bedingung
	 * 
	 * @return die ID
	 */
	public int getConditionID() {
		return conditionId;
	}

	/**
	 * Liefert die ID des Atoms
	 * 
	 * @return die ID
	 */
	public int getAtomID() {
		return atomId;
	}

	/**
	 * Liefert den Inhalt der Bedingung.
	 * 
	 * @return der Inhalt
	 */
	public String getContent() {
		return content;
	}

	/**
	 * Liefert den instrumentierten Inhalt.
	 * 
	 * @return der instrumentierte Inhalt
	 */
	public String getASCContent() {
		if (ascContent == null) {
			ascContent = content;
		}
		return ascContent;
	}

	/**
	 * Liefert die linke Bedingung.
	 * 
	 * @return die linke Bedingung oder null
	 */
	public Condition getLeftCondition() {
		return leftCondition;
	}

	/**
	 * Liefert die rechte Bedingung.
	 * 
	 * @return die rechte Bedingung oder null
	 */
	public Condition getRightCondition() {
		return rightCondition;
	}

	/**
	 * Gibt die Liste der Subbedingungen von links nach rechts zurck.
	 * 
	 * @return die Bedingungsliste
	 */
	public ArrayList<Condition> getListOfConditions() {
		ArrayList<Condition> l = new ArrayList<Condition>();

		l.add(this);

		if (leftCondition != null) {
			// if(leftCondition.isCondAssignment() && leftCondition.isAtom())
			// l.add(leftCondition);
			l.addAll(leftCondition.getListOfConditions());
		}
		if (rightCondition != null) {
			// if(rightCondition.isCondAssignment() && rightCondition.isAtom())
			// l.add(rightCondition);
			l.addAll(rightCondition.getListOfConditions());
		}
		return l;
	}

	/**
	 * Setzt die Bedingung auf Atom.
	 * 
	 * @param b
	 *            true, wenn Atom, sonst false.
	 */
	public void setAtom(boolean b) {
		isAtom = b;
	}

	/**
	 * Entscheidet, ob die Bedingung leer ist.
	 * 
	 * @return true, wenn leer, sonst false
	 */
	public boolean isEmpty() {
		return isEmpty;
	}

	/**
	 * Entscheidet, ob die Bedingung ein Atom ist.
	 * 
	 * @return true, wenn Atom, sonst false
	 */
	public boolean isAtom() {
		return isAtom;
	}

	/**
	 * Entscheidet, ob die Bedingung ein true-Atom ist.
	 * 
	 * @return true, wenn true-Atom, sonst false
	 */
	public boolean isTrueAtom() {
		return isTrueAtom;
	}

	/**
	 * Entscheidet, ob die Bedingung ein false-Atom ist.
	 * 
	 * @return true, wenn false-Atom, sonst false
	 */
	public boolean isFalseAtom() {
		return isFalseAtom;
	}

	/**
	 * Entscheidet, ob die Bedingung ein Operator ist.
	 * 
	 * @return true, wenn Operator, sonst false
	 */
	public boolean isOperator() {
		return isOperator;
	}

	// fr Bool-Test fr |, & und ^ im Parser notwendig
	/**
	 * Entscheidet, ob die Bedingung ein boolscher Wert ist.
	 * 
	 * @return true, wenn boolsch, sonst false
	 */
	public boolean isBoolean() {
		return isOperator() || isAtom();
	}

	/**
	 * Entscheidet, ob die Bedingung ein Kurzschlussauswertung hat.
	 * 
	 * @return true, wenn vorhanden, sonst false
	 */
	public boolean hasShortCircuit() {
		return hasShortCircuit;
	}

	/**
	 * Setzt die Art der Kurzschlussauswertung.
	 * 
	 * @param b
	 *            true, wenn ||, sonst false
	 */
	public void setShortCircuitValue(boolean b) {
		hasShortCircuit = true;
		shortCircuitValue = b;
	}

	/**
	 * Gibt die Art der Kurzschlussauswertung zurck.
	 * 
	 * @return true, wenn ||, sonst false
	 */
	public boolean getShortCircuitValue() {
		return shortCircuitValue;
	}

	// Nummerierungsfunktion fr die PrimaryCondition
	// welcher die Atome nummeriert
	/*
	 * Bekommt: ID der PrimaryCondition Result: Anzahl Atome
	 */
	/**
	 * Nummeriert die Bedingung und ihre Atome durch.
	 * 
	 * @param i
	 *            der Offset der primren Bedingung
	 * @return Anzahl der Atome
	 */
	public int numerate(int i) {
		conditionId = i;

		if (isAtom) {
			atomId = 0;
			return 1;
		}

		int nrAtoms = 0;
		if (leftCondition != null) {
			nrAtoms = leftCondition.numerateAtoms(i, nrAtoms);
		}
		if (rightCondition != null) {
			nrAtoms = rightCondition.numerateAtoms(i, nrAtoms);
		}
		return nrAtoms;
	}

	// rek. Nummerierungsfunktion fr die Conditions
	/*
	 * Bekommt: ID der PrimaryCondition Result: Anzahl Atome
	 */
	/**
	 * Rekursive Nummerierungsfunktio fr Bedingungen.
	 * 
	 * @param i
	 *            die ID der primren Bedingung
	 * @param a
	 *            Anzahl der Atome
	 * @return Anzahl der Atome
	 */
	public int numerateAtoms(int i, int a) {
		conditionId = i;
		if (leftCondition != null) {
			a = leftCondition.numerateAtoms(i, a);
		}
		if (rightCondition != null) {
			a = rightCondition.numerateAtoms(i, a);
		}
		if (isAtom()) {
			atomId = a++;
		}
		return a;
	}

	/** Print ** */
	/**
	 * Ausgabefunktion
	 */
	public void print() {
		Logger.logln();
		if (isOperator) {
			/*
			 * if (localID > 0) { Logger.log("(LID:" + localID + ")"); }
			 */
			Logger.log(" <" + content + ">");
			if (leftCondition != null) {
				leftCondition.print();
			}
			rightCondition.print();
		} else {
			Logger.log(" <" + content + "> (" + conditionId + ", " + atomId + ")");
		}
	}

	/**
	 * Appliziert einen Wahrheitsvektor auf die Bedingung.
	 * 
	 * @param s
	 *            der Wahrheitsvektor
	 * @return die boolsche Auswertung
	 */
	public abstract boolean applyTruthVector(String s);

	/**
	 * Wertet einen Wahrheitsvektor fr diese Bedingung aus, ohne ihn zu applizieren.
	 * 
	 * @param s
	 *            der Wahrheitsvektor
	 * @return die boolsche Auswertung
	 */
	public abstract boolean evaluateTruthVector(String s);

	/**
	 * Ermittelt die Anzahl an Wahr-Falsch-Auswertungen fr alle Atome.
	 * 
	 * @param l
	 *            die Liste der Auswertungen fr den rekursiven Aufruf
	 * @return die Liste mit den Auswertungen von wahr ([0]) und falsch ([1]) fr alle Atome von links nach rechts
	 */
	public abstract ArrayList<Integer[]> getAtomEvaluation(ArrayList<Integer[]> l);

	/**
	 * Ermittelt die Anzahl an Wahr-Falsch-Auswertungen fr die Bedingung und alle Teilbedingungen.
	 * 
	 * @param l
	 *            die Liste der Auswertungen fr den rekursiven Aufruf
	 * @return die Liste mit den Auswertungen von wahr ([0]) und falsch ([1]) fr alle Bedingungen, absteigend von links
	 *         nach rechts
	 */
	public abstract ArrayList<Integer[]> getConditionEvaluation(ArrayList<Integer[]> l);

	/**
	 * Liefert eine Stringdarstellung der Bedingung.
	 * 
	 * @return die Stringdarstellung
	 */
	public String getCondString() {
		String s = "---";
		if (isOperator) {
			s += "<" + content + "><t: " + nrTrueEvaluations + ", f: " + nrFalseEvaluations + ">---\n";
			if (leftCondition != null) {
				s += leftCondition.getCondString();
			}
			s += rightCondition.getCondString();
		} else {
			s += "<" + content + "><" + conditionId + ", " + atomId + "><t: " + nrTrueEvaluations + ", f: "
					+ nrFalseEvaluations + ">---\n";
		}
		return s;
	}

	/**
	 * Setzt alle berdeckungsinformationen zurck.
	 */
	public void resetHits() {
		nrTrueEvaluations = 0;
		nrFalseEvaluations = 0;
		if (leftCondition != null)
			leftCondition.resetHits();
		if (rightCondition != null)
			rightCondition.resetHits();
		metrics.reset();
	}

	/**
	 * Berechnet die ntigen Werte fr die Bedingungsberdeckung. Vorher wurden schon die TVs appliziert, damit Atomare
	 * und Bedingungsbelegung berechnet werden kann
	 * 
	 * @param setOfTruthVectors
	 *            die Menge der Wahrheitsvektoren der Bedingung fr die aktuellen Tests
	 */
	public void computeConditionCoverage(HashSet<String> setOfTruthVectors) {

		ArrayList<Integer[]> list = getAtomEvaluation(null);

		metrics.nrAtoms = 0;
		metrics.nrAtomEvaluations = 0;
		metrics.nrConditions = 0;
		metrics.nrConditionEvaluations = 0;
		metrics.nrMcdcPairs = 0;

		if (isNonTrivial()) {
			// Zhlt Atome und wie hufig diese wahr und falsch wurden
			// -> C2
			if (list != null && list.size() > 0) {
				for (Integer[] i : list) {
					metrics.nrAtoms++;
					if (i[0] > 0)
						metrics.nrAtomEvaluations++;
					if (i[1] > 0)
						metrics.nrAtomEvaluations++;
				}
			}

			list = getConditionEvaluation(null);

			// Zhlt Bedingungen und wie hufige diese wahr und falsch wurden
			// -> MMDC
			if (list != null && list.size() > 0) {
				for (Integer[] i : list) {
					metrics.nrConditions++;
					if (i[0] > 0)
						metrics.nrConditionEvaluations++;
					if (i[1] > 0)
						metrics.nrConditionEvaluations++;
				}
			}

			// Vergleicht alle Wahrheitsvektoren paarweise miteinander
			// und schaut, ob diese Mcdc-Paare sind
			atoms: for (int i = 0; i < metrics.nrAtoms; i++) {
				for (String s1 : setOfTruthVectors) {
					if (s1.length() <= i)
						break;

					for (String s2 : setOfTruthVectors) {
						if (s2.length() <= i)
							break;

						if (areMcdcPairs(s1, s2, i)) {

							// Logger.logln(s1 + " and " + s2 + "are mcdc-pairs for " + i + " in " + this.content);

							metrics.nrMcdcPairs++;
							// fr jedes Atom nur ein MCDC-Paar zhlen!
							continue atoms;

						}
					}
				}
			}

			int[] i = computeNrAtomCombinations();
			metrics.nrAtomCombinations = i[0] + i[1];
			metrics.nrAtomCombinationEvaluations = setOfTruthVectors.size();
			metrics.computeCoverageMetrics();
			// System.out.println(metrics.nrAtomCombinationEvaluations + " - " +
			// metrics.nrAtomCombinations);
		}

	}

	/**
	 * Testet zwei Wahrheitsvektoren darauf, ob sie fr ein Atom ein MCDC-Paar bilden.
	 * 
	 * @param s1
	 *            der erste Wahrheitsvektor
	 * @param s2
	 *            der zweite Wahrheitsvektor
	 * @param i
	 *            die ID des Atoms
	 * @return true, wenn MCDC-Paar, sonst false
	 */
	public boolean areMcdcPairs(String s1, String s2, int i) {
		if (s1.substring(0, i).equals(s2.substring(0, i))
				&& s1.substring(i + 1, s1.length()).equals(s2.substring(i + 1, s2.length()))
				&& (evaluateTruthVector(s1) != evaluateTruthVector(s2)))
			return true;
		return false;
	}

	/**
	 * Berechnet die Anzahl an mglichen Kombinationen der Atomauswertung.
	 * 
	 * @return die Anzahl an Atomkombinationen, zu wahr ([0]) bzw. falsch ([1])
	 */
	public abstract int[] computeNrAtomCombinations();

	/**
	 * Berechnet aus einem Truthvector einen Belegungstring. Dabei werden nicht ausgewertete Variablen (bei
	 * short-circuit-evaluation) durch "-" gekennzeichnet. Bsp.: ((falsch && falsch) || wahr) Truthvector: 011 (zweite
	 * Wert wurde nicht geloggt, daher per default true) Eval.TV.: 0-1
	 * 
	 * @param s
	 *            der TruthVector, der umgewandelt werden soll
	 * @return der berechnete String mit den Bedingungsbelegungen
	 */
	public abstract String getEvaluatedTVString(String s);

	/**
	 * Gibt die Anzahl MCDC-Paare zurck.
	 * 
	 * @return die Anzahl an MCDC-Paaren
	 */
	public int getNrMcdcPairs() {
		return metrics.nrMcdcPairs;
	}

	/**
	 * Gibt die Anzahl an Atomen zurck.
	 * 
	 * @return die Atomanzahl
	 */
	public int getNrAtoms() {
		return metrics.nrAtoms;
	}

	/**
	 * Gibt die Anzahl an Atomauswertungen zurck.
	 * 
	 * @return die Anzahl an Atomauswertungen
	 */
	public int getNrAtomEvaluations() {
		return metrics.nrAtomEvaluations;
	}

	/**
	 * Gibt die Anzahl an an Bedingungen zurck.
	 * 
	 * @return die Anzahl an Bedingungen
	 */
	public int getNrConditions() {
		if (metrics.nrConditions > 0 && nrConditions != metrics.nrConditions)
			System.out.println("Huh, ungleiche nrConditions?!");
		return nrConditions;
	}

	/**
	 * Gibt die Anzahl an Bedingungsauswertungen zurck.
	 * 
	 * @return die Anzahl an Bedingungsauswertungen
	 */
	public int getNrConditionEvaluations() {
		return metrics.nrConditionEvaluations;
	}

	/**
	 * Zhlt die Atome.
	 * 
	 * @return die Anzahl an Atomen
	 */
	public abstract int countNrAtoms();

	/**
	 * Gibt ein Atom zurck.
	 * 
	 * @param i
	 *            die ID des Atoms
	 * @return das Atom
	 */
	public abstract Condition getAtomCondition(int i);

	/**
	 * Gibt den Inhalt der Bedingung wieder.
	 * 
	 * @return der Inhalt
	 */
	public String getConditionContent() {
		if (isAtom)
			return content;
		if (isEmpty)
			return "true";
		if (isOperator) {
			if (leftCondition != null) {
				return " ( " + leftCondition.getConditionContent() + " " + content + " "
						+ rightCondition.getConditionContent() + " ) ";
			}
			return " ( " + content + " " + rightCondition.getConditionContent() + " ) ";
		}
		return "???";
	}

	/**
	 * Entscheidet, ob die Bedingung trival ist (true-, false-Atom, leeres Atom).
	 * 
	 * @return true, wenn trival, sonst false
	 */
	public boolean isNonTrivial() {
		return !(isEmpty || isTrueAtom || isFalseAtom);
	}

	/**
	 * Setzt, ob Bedingung primr ist.
	 * 
	 * @param b
	 *            true, wenn ja, sonst false
	 */
	public void setPrimary(boolean b) {
		isPrimary = b;
	}

	/**
	 * Entscheidet, ob Bedingung primr ist. return true, wenn ja, sonst false
	 */
	public boolean isPrimary() {
		return isPrimary;
	}

	/**
	 * Gibt die Metrik der Bedingung zurck.
	 * 
	 * @return die Metrik
	 */
	public Metrics getMetrics() {
		return metrics;
	}

	/**
	 * Ausgabefunktion
	 */
	public void printFull() {
		Logger.logln();
		Logger.logln("isAtom? : " + isAtom);
		Logger.logln("isOperator? : " + isOperator);
		Logger.logln("isEmpty? : " + isEmpty);
		Logger.logln("isTrueAtom? : " + isTrueAtom);
		Logger.logln("isFalseAtom? : " + isFalseAtom);
		Logger.logln("hasShortCircuit? : " + hasShortCircuit);
		Logger.logln("shortCircuitValue? : " + shortCircuitValue);

		Logger.logln("Id: " + conditionId + " " + atomId);
		// Logger.logln("localID: " + localID);
		Logger.logln("Content: " + content);
		Logger.logln("LeftCondition: " + leftCondition);
		Logger.logln("RightCondition: " + rightCondition);
	}
}