package sota.ast.nodes;

import java.util.TreeSet;

import sota.Logger;
import sota.ast.util.IScheme;
import sota.ast.util.Metrics;

/**
 * Knoten, der eine Quelldatei reprsentiert.
 * */
abstract public class FileNode implements IScopeNode, Comparable<FileNode> {

	protected int idOffset;
	protected int nrNodes;

	protected ASTNode startNode;
	protected ASTNode endNode;

	protected String name;
	protected String fullName;
	protected String packageName = "";
	protected TreeSet<ASTClassNode> setOfClassNodes;
	protected TreeSet<ASTFunctionNode> setOfFunctionNodes;
	protected TreeSet<ASTScopeNode> setOfTopLevelScopeNodes;

	protected boolean isInstrumented = false;

	protected Metrics metrics;
	// Ist die Quelldatei instrumentiert?
	protected boolean isDirty = false;
	// Wurde eine instrumentierte Quelldatei geparst? Alternative: Quelle nicht instrumentiert oder Backup geparst.
	protected boolean isParsedDirty = false;

	// Konstruktor

	/**
	 * Konstruktor fr den FileNode.
	 * 
	 * @param name
	 *            der Name der Datei
	 * @param idOffset
	 *            der ID-Offset fr die ASTNodes des FileNodes
	 */
	public FileNode(String name, int idOffset) {
		setOfFunctionNodes = new TreeSet<ASTFunctionNode>();
		setOfClassNodes = new TreeSet<ASTClassNode>();
		setOfTopLevelScopeNodes = new TreeSet<ASTScopeNode>();
		String s = name.substring(name.lastIndexOf('\\') + 1);
		s = s.substring(s.lastIndexOf('/') + 1);
		this.name = s;
		this.fullName = name;
		this.idOffset = idOffset;
		nrNodes = 0;
		metrics = new Metrics();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#createASCContent(int)
	 */
	public abstract void createASCContent(int level);

	/**
	 * Setzt die berdeckungsdaten fr alle Knoten der Datei zurck.
	 */
	public void resetHits() {
		for (ASTNode i = startNode; i != endNode; i = i.getNextNode()) {
			if (i instanceof IStatementNode)
				((IStatementNode) i).resetHits();
		}
	}

	/**
	 * Fgt einen Knoten an die Liste der ASTNodes an.
	 * 
	 * @param n
	 *            der ASTNode
	 */
	public void append(ASTNode n) {
		if (nrNodes == 0) {
			startNode = n;
			n.prev = null;
			endNode = n;
		} else {
			endNode.setNextNode(n);
			n.setPrevNode(endNode);
			endNode = n;
		}
		n.setID(idOffset + nrNodes++);
	}

	/**
	 * Gibt Anzahl der Knoten zurck.
	 * 
	 * @return die Anzahl der Knoten
	 */
	public int getNrNodes() {
		return nrNodes;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getStartNode()
	 */
	public ASTNode getStartNode() {
		return startNode;
	}

	/**
	 * Gibt den letzten Knoten zurck.
	 * 
	 * @return der letzte Knoten
	 */
	public ASTNode getLastNode() {
		return endNode;
	}

	// Implementationen der Funktionen des Interface ScopeNode

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#addClassNode(sota.ast.nodes.ASTClassNode)
	 */
	public void addClassNode(ASTClassNode cn) {
		// Logger.logln("FN.addClass: " + this.getIDString() + " <- "+
		// cn.getIDString());
		setOfClassNodes.add(cn);
		if (cn.isTopLevel())
			setOfTopLevelScopeNodes.add(cn);
		cn.setFileNode(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#addFunctionNode(sota.ast.nodes.ASTFunctionNode)
	 */
	public void addFunctionNode(ASTFunctionNode fn) {
		setOfFunctionNodes.add(fn);
		if (fn.isTopLevel())
			setOfTopLevelScopeNodes.add(fn);
		fn.setFileNode(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#isTopLevel()
	 */
	public boolean isTopLevel() {
		return true;
	}

	/**
	 * Entscheidet, ob die Datei instrumentiert ist oder nicht.
	 * 
	 * @return true, wenn instrumentiert, sonst false
	 */
	public boolean isInstrumented() {
		return isInstrumented;
	}

	/**
	 * Legt fest, ob die Datei instrumentiert ist oder nicht.
	 * 
	 * @param b
	 *            true, wenn instrumentiert, sonst false
	 */
	public void isInstrumented(boolean b) {
		isInstrumented = b;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getName()
	 */
	public String getName() {
		return name;
	}

	/**
	 * Gibt den vollen Namen zurck.
	 * 
	 * @return der Name
	 */
	public String getFullName() {
		return fullName;
	}

	/**
	 * Setzt das Package der Datei.
	 * 
	 * @param s
	 *            das Package
	 */
	public void setPackageName(String s) {
		this.packageName = s;
	}

	/**
	 * Liefert das Package der Datei.
	 * 
	 * @return das Package
	 */
	public String getPackageName() {
		return packageName;
	}

	/**
	 * Gibt den vollqualifizierten Namen zurck.
	 * 
	 * @return der vollqualifizierte Name
	 */
	public String getFullQualifiedName() {
		return packageName + (packageName.length() > 0 ? "." : "") + name;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getScope()
	 */
	public String getScope() {
		return "";
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getIDString()
	 */
	public String getIDString() {
		return getFullQualifiedName();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#setEndNode(sota.ast.nodes.ASTNode)
	 */
	public void setEndNode(ASTNode dn) {
		endNode = dn;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getEndNode()
	 */
	public ASTNode getEndNode() {
		return endNode;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getFunctionNodes()
	 */
	public TreeSet<ASTFunctionNode> getFunctionNodes() {
		return setOfFunctionNodes;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getClassNodes()
	 */
	public TreeSet<ASTClassNode> getClassNodes() {
		return setOfClassNodes;
	}

	/**
	 * Gibt alle TopLevel-ScopeNodes zurck.
	 * 
	 * @return die Menge der TopLevel-ScopeNodes
	 */
	public TreeSet<ASTScopeNode> getTopLevelScopeNodes() {
		return setOfTopLevelScopeNodes;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(FileNode file) {
		return name.compareTo(file.getName());
	}

	// Konsolenausgabefunktionen

	/**
	 * Ausgabefunktion
	 */
	public void print() {
		Logger.logln(" \nFile : " + name + "\n");
		Logger.logln(" \tnrClasses: " + setOfClassNodes.size());
		Logger.logln(" \tnrFunctions: " + setOfFunctionNodes.size() + "\n");
		for (ASTNode n = startNode; n != null; n = n.next) {
			n.print();
		}
	}

	/**
	 * Ausgabefunktion fr CFG
	 */
	public void printCFG() {
		Logger.logln(" \nCFG of File : " + name + "\n");
		for (ASTFunctionNode n : setOfFunctionNodes)
			n.printCFG();
	}

	/**
	 * Ausgabefunktion fr Knoten
	 */
	public void printNodes() {
		Logger.logln(" \nNodes of File : " + name + "\n");
		for (ASTNode n = startNode; n != null; n = n.getNextNode()) {
			n.printNode();
		}
	}

	/**
	 * Ausgabefunktion fr Bedingungen
	 */
	public void printConditions() {
		Logger.logln(" \nConditions of File : " + name + "\n");
		for (ASTNode n = startNode; n != null; n = n.getNextNode()) {
			if (n instanceof IConditionNode) {
				n.print();
				((IConditionNode) n).getCondition().print();
			}
		}
	}

	/**
	 * Ausgabefunktion fr Funktionen
	 * 
	 * @param indent
	 *            der Einschub
	 */
	public void printFunctions(int indent) {
		for (ASTScopeNode sn : setOfClassNodes) {
			print_indent(indent);
			Logger.logln(" Class: " + sn.getScope() + "::\t" + sn.getName());
			sn.printFunctions(indent + 1);
		}
		for (ASTScopeNode sn : setOfFunctionNodes) {
			print_indent(indent);
			Logger.logln(" Function: " + sn.getScope() + "::\t" + sn.getName() + " ("
					+ ((ASTFunctionNode) sn).getArguments() + ")");
			sn.printFunctions(indent + 1);
		}
	}

	/**
	 * Einschub der Ausgabe.
	 * 
	 * @param indent
	 *            die Gre des Einschubs
	 */
	public void print_indent(int indent) {
		for (int i = 0; i < indent; i++) {
			Logger.log("\t");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getParent()
	 */
	public IScopeNode getParent() {
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getMetrics()
	 */
	public Metrics getMetrics() {
		return metrics;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#setMetrics(sota.ast.util.Metrics)
	 */
	public void setMetrics(Metrics metrics) {
		this.metrics = metrics;
	}

	/**
	 * Erstellt den instrumentierten Inhalt fr einen ScopeNode (rekursiv).
	 * 
	 * @param node
	 *            der IScopeNode
	 * @param ischeme
	 *            das gewhlte IScheme
	 */
	public void createASCContentForScope(IScopeNode node, IScheme ischeme) {
		IScopeNode sn;
		ASTNode n = null;
		int currLevel = ischeme.getInstrumentationLevel(node);

		if (node instanceof FileNode) {
			startNode.createASCContent(currLevel);
			n = startNode.getNextNode();
		} else if (node instanceof ASTScopeNode) {
			n = ((ASTScopeNode) node).getNextNode();
		}

		node.createASCContent(currLevel);

		do {
			if (n instanceof IScopeNode) {
				sn = (IScopeNode) n;
				createASCContentForScope(sn, ischeme);
				n = sn.getEndNode();
			} else {
				n.createASCContent(currLevel);
				n = n.getNextNode();
			}
		} while (n != node.getEndNode());

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getFileNode()
	 */
	public FileNode getFileNode() {
		return this;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getFullyQualifiedName(boolean, boolean)
	 */
	public String getFullyQualifiedName(boolean showPackage, boolean showFile) {
		if (showPackage && packageName != null && packageName.length() > 0)
			return packageName + "." + name;

		return name;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getFirstLineNr()
	 */
	public int getFirstLineNr() {
		return 1;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.nodes.IScopeNode#getLastLineNr()
	 */
	public int getLastLineNr() {
		return metrics.nrLines;
	}

	/**
	 * Setzt den Quellcode auf verschmutzt, d.h. instrumentiert.
	 * 
	 * @param b
	 *            true, wenn instrumentiert, sonst false
	 */
	public void setDirty(boolean b) {
		isDirty = b;
	}

	/**
	 * Entscheidet, ob der Quellcode instrumentiert ist.
	 * 
	 * @return true, wenn instrumentiert, sonst false
	 */
	public boolean isDirty() {
		return isDirty;
	}

	/**
	 * Setzt, ob die Datei instrumentiert geparst wurde.
	 * 
	 * @param b
	 *            true, wenn instrumentiert, sonst false
	 */
	public void setParsedDirty(boolean b) {
		isParsedDirty = b;
	}

	/**
	 * Entscheidet, ob die Datei instrumentiert geparst wurde.
	 * 
	 * @return true, wenn Datei instrumentiert geparst, sonst false
	 */
	public boolean isParsedDirty() {
		return isParsedDirty;
	}

	/**
	 * Setzt den instrumentierten Inhalt aller Knoten zurck.
	 */
	public void clearASCContent() {
		for (ASTNode n = startNode; n != endNode; n = n.getNextNode())
			n.ascContent = "";
	}

}