package sota.ast.util;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.TreeSet;
import java.util.Vector;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.draw2d.graph.DirectedGraph;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.dialogs.ProgressMonitorDialog;
import org.eclipse.jface.operation.IRunnableWithProgress;
import org.eclipse.ui.IWorkbenchWindow;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import sota.Logger;
import sota.app.Application;
import sota.ast.nodes.ASTClassNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.FileNode;
import sota.gui.GuiUpdateListener;
import sota.gui.ProjectChangeListener;
import sota.gui.graph.GraphManager;
import sota.gui.provider.IProjectTreeView;
import sota.lang.IASTBuilder;
import sota.lang.LangClassProvider;
import sota.lang.LanguageSpecification;
import sota.lang.java.JavaLanguageSpecification;

/**
 * Verwaltet das Projekt und bietet Zugriff auf die wichtigsten UseCases; ist damit zentrale Klasse zur Verwaltung und
 * Manipulation des gesamten AST.
 * 
 * Biete Zugriff auf folgende UseCases:<br>
 * - Erstellung des AST ber einen ASTBuilder<br>
 * - statische Analyse des Quellcodes<br>
 * - Instrumentation des Quellcodes<br>
 * - Programmkompilation und -ausfhrung<br>
 * - Einlesen der Logdateien, inkl. Verarbeitung der Testdaten<br>
 * - Applikation der berdeckungsdaten auf den AST und Berechnung der berdeckungsmae<br>
 * - Erstellung einer Reportdatei<br>
 * <br>
 * 
 * Verwaltet:<br>
 * - die Projektkonfiguration<br>
 * - alle Daten des Abstrakten Syntaxbaum<br>
 * - die Testdaten<br>
 * - die Listener des Projekts<br>
 */
public class ASTManager {

	private boolean VERBOSE = false;

	// der aktuelle FileNode
	private FileNode currFileNode;

	// die Datenstze des AST, geglieder nach FileNodes, Klassen und Funktionen
	private List<ASTNode> listOfNodes;
	private TreeSet<FileNode> setOfFileNodes;
	private TreeSet<ASTClassNode> setOfClassNodes;
	private TreeSet<ASTClassNode> setOfTopLevelClassNodes;
	private TreeSet<ASTFunctionNode> setOfFunctionNodes;
	private TreeSet<ASTFunctionNode> setOfTopLevelFunctionNodes;

	// die Projektkonfiguration
	private ProjectConfig config;

	// die angemeldeten Listener
	private Vector<ProjectChangeListener> changeListeners = new Vector<ProjectChangeListener>();
	private Vector<GuiUpdateListener> guiUpdateListeners = new Vector<GuiUpdateListener>();

	// die ProjectTreeViews, die bei Doppelklick eine Auswahl aktualisiern (ProjectView, CoverageView, MetricsView)
	private Vector<IProjectTreeView> projectTreeViews = new Vector<IProjectTreeView>();

	// ein Flag zum Abbruch von ProgressMonitoren
	private boolean[] flag = new boolean[] { false };

	/****************************************/
	/* Konstruktoren */
	/****************************************/

	/**
	 * Der Konstruktor, der den ASTManager anhand einer Konfiguration erstellt.
	 * 
	 * @param c
	 *            die Konfiguration
	 */
	public ASTManager(ProjectConfig c) {
		setOfFileNodes = new TreeSet<FileNode>();
		setOfClassNodes = new TreeSet<ASTClassNode>();
		setOfTopLevelClassNodes = new TreeSet<ASTClassNode>();
		setOfFunctionNodes = new TreeSet<ASTFunctionNode>();
		setOfTopLevelFunctionNodes = new TreeSet<ASTFunctionNode>();
		listOfNodes = new ArrayList<ASTNode>();
		config = c;
	}

	/**
	 * Der Konstruktor, der einene ASTManager mit einer neuen Konfiguration erstellt.
	 */
	public ASTManager() {
		this(new ProjectConfig());
	}

	/**
	 * Initialisiert den ASTManager anhand einer Konfiguration neu.
	 * 
	 * @param c
	 *            die Konfiguration
	 */
	public void reinit(ProjectConfig c) {
		Map<String, LanguageSpecification> setOfLangSpec = config.setOfLangSpecs;
		setOfFileNodes = new TreeSet<FileNode>();
		setOfClassNodes = new TreeSet<ASTClassNode>();
		setOfTopLevelClassNodes = new TreeSet<ASTClassNode>();
		setOfTopLevelFunctionNodes = new TreeSet<ASTFunctionNode>();
		setOfFunctionNodes = new TreeSet<ASTFunctionNode>();
		listOfNodes = new ArrayList<ASTNode>();
		currFileNode = null;
		config = c;
		if (config.setOfLangSpecs == null || config.setOfLangSpecs.isEmpty())
			config.setOfLangSpecs = setOfLangSpec;
		fireToProjectChangeListeners();
	}

	/**
	 * Initialisier den ASTManager neu.
	 */
	public void reinit() {
		Map<String, LanguageSpecification> setOfLangSpec = config.setOfLangSpecs;
		config = new ProjectConfig();
		config.setOfLangSpecs = setOfLangSpec;

		setOfFileNodes = new TreeSet<FileNode>();
		setOfClassNodes = new TreeSet<ASTClassNode>();
		setOfTopLevelClassNodes = new TreeSet<ASTClassNode>();
		setOfTopLevelFunctionNodes = new TreeSet<ASTFunctionNode>();
		setOfFunctionNodes = new TreeSet<ASTFunctionNode>();
		listOfNodes = new ArrayList<ASTNode>();
		config.currInstrumentationScheme = new IScheme("");
		currFileNode = null;
		fireToProjectChangeListeners();
	}

	/**
	 * Speichert das Projekt in einer XML-Datei.
	 * 
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean saveProject() {

		Logger.logln("\t Saving project " + config.name);

		Element root = new Element("Project");
		Element name = new Element("Name");
		name.setText(config.getName());
		root.addContent(name);
		Element lang = new Element("Language");
		lang.setText(config.language);
		root.addContent(lang);
		Element prefix = new Element("Prefix");
		prefix.setText(config.prefix);
		root.addContent(prefix);
		Element backupExtension = new Element("BackupExtension");
		backupExtension.setText(config.backupExtension);
		root.addContent(backupExtension);
		Element projectDir = new Element("ProjectDir");
		projectDir.setText(config.projectDir);
		root.addContent(projectDir);
		Element execDir = new Element("ExecDir");
		execDir.setText(config.execDir);
		root.addContent(execDir);
		if (config.antLocation != null && new File(config.antLocation).exists()) {
			Element antLocation = new Element("AntLocation");
			antLocation.setText(config.antLocation);
			root.addContent(antLocation);
		}
		if (config.antBuildFile != null && new File(config.antBuildFile).exists()) {
			Element antBuildFile = new Element("AntBuildFile");
			antBuildFile.setText(config.antBuildFile);
			root.addContent(antBuildFile);
		}
		if (config.runScript != null && new File(config.runScript).exists()) {
			Element runScript = new Element("RunScript");
			runScript.setText(config.runScript);
			root.addContent(runScript);
		}
		Element sources = new Element("SourceFiles");
		root.addContent(sources);
		for (String s : config.setOfFileNames) {
			Element file = new Element("File");
			file.setText(s);
			sources.addContent(file);
		}
		Element ischemes = new Element("ISchemes");
		root.addContent(ischemes);

		for (IScheme is : config.getListOfISchemes()) {
			if (is instanceof GlobalInstrumentationScheme)
				continue;
			Element ischeme = new Element("IScheme");
			ischemes.addContent(ischeme);

			Element isName = new Element("Name");
			isName.setText(is.getName());
			ischeme.addContent(isName);
			Element isDesc = new Element("Description");
			isDesc.setText(is.getDescription());
			ischeme.addContent(isDesc);

			Element[] lvl = new Element[4];
			lvl[0] = new Element("Level0");
			lvl[1] = new Element("Level1");
			lvl[2] = new Element("Level2");
			lvl[3] = new Element("Level3");

			for (String s : is.getLevelMap().keySet()) {
				Element structur = new Element("Structure");
				structur.setText(s);
				lvl[is.getLevelMap().get(s)].addContent(structur);
			}

			if (lvl[0].getContentSize() > 0)
				ischeme.addContent(lvl[0]);
			if (lvl[1].getContentSize() > 0)
				ischeme.addContent(lvl[1]);
			if (lvl[2].getContentSize() > 0)
				ischeme.addContent(lvl[2]);
			if (lvl[3].getContentSize() > 0)
				ischeme.addContent(lvl[3]);

		}

		// org.jdom.Document doc = new org.jdom.Document(root, new DocType("Project", "project.dtd"));

		XMLOutputter outp = new XMLOutputter();
		outp.setFormat(Format.getPrettyFormat());
		try {
			outp.output(root, new FileOutputStream(new File(config.name + ".project")));
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
			return false;
		}
		return true;
	}

	/**
	 * ffnet das Projekt anhand einer XML-Datei ohne ProgressMonitor (SOTA-ATM).
	 * 
	 * @param file
	 *            die XML-Projektdatei
	 * @param recover
	 *            ob die originalen Quelldateien wiederhergestellt werden sollen
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean openProject(String file, boolean recover) {

		if (file == null || !(new File(file)).exists()) {
			Logger.logln("No such file");
			return false;
		}

		Logger.log("\t Opening Project " + file + " ... ");

		reinit();

		if (openProjectConfiguration(file)) {
			Logger.logln("OK.");
		} else {
			Logger.logln("Failed.");
			return false;
		}
		if (recover)
			restoreProject();
		parseProject();
		// backupProject();
		fireToGuiUpdateListeners();

		return true;
	}

	/**
	 * Liest die Konfiguration des Projektes ein.
	 * 
	 * @param file
	 *            die XML-Projektdatei
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean openProjectConfiguration(String file) {
		SAXBuilder builder = new SAXBuilder();
		org.jdom.Document doc = null;
		try {
			doc = builder.build(new File(file));

		} catch (JDOMException e) {
			e.printStackTrace();
			return false;
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}

		Element root = doc.getRootElement();
		config.name = root.getChildText("Name");
		config.language = root.getChildText("Language");
		config.prefix = root.getChildText("Prefix");
		config.backupExtension = root.getChildText("BackupExtension");
		config.projectDir = root.getChildText("ProjectDir");
		config.execDir = root.getChildText("ExecDir");
		if (root.getChild("AntLocation") != null)
			config.antLocation = root.getChildText("AntLocation");
		if (root.getChild("AntBuildFile") != null)
			config.antBuildFile = root.getChildText("AntBuildFile");
		if (root.getChild("RunScript") != null)
			config.runScript = root.getChildText("RunScript");
		Element sourceFiles = root.getChild("SourceFiles");

		for (Object o : sourceFiles.getChildren()) {
			if (!(o instanceof Element))
				continue;
			Element e = (Element) o;
			config.setOfFileNames.add(e.getText());
		}

		TreeSet<IScheme> setOfISchemes = config.getListOfISchemes();
		for (Object o : root.getChild("ISchemes").getChildren()) {
			if (!(o instanceof Element))
				continue;
			Element e = (Element) o;
			IScheme is = new IScheme(e.getChildText("Name"));
			if (e.getChild("Description") != null)
				is.setDescription(e.getChildText("Description"));
			if (e.getChild("Level0") != null) {
				for (Object o2 : e.getChild("Level0").getChildren()) {
					if (!(o2 instanceof Element))
						continue;
					is.addInstrumentationLevel(((Element) o2).getText(), 0);
				}
			}
			if (e.getChild("Level1") != null) {
				for (Object o2 : e.getChild("Level1").getChildren()) {
					if (!(o2 instanceof Element))
						continue;
					is.addInstrumentationLevel(((Element) o2).getText(), 1);
				}
			}
			if (e.getChild("Level2") != null) {
				for (Object o2 : e.getChild("Level2").getChildren()) {
					if (!(o2 instanceof Element))
						continue;
					is.addInstrumentationLevel(((Element) o2).getText(), 2);
				}
			}
			if (e.getChild("Level3") != null) {
				for (Object o2 : e.getChild("Level3").getChildren()) {
					if (!(o2 instanceof Element))
						continue;
					is.addInstrumentationLevel(((Element) o2).getText(), 3);
				}
			}

			setOfISchemes.add(is);

		}

		config.projectStatus = ProjectConfig.CONFIGURED;

		// definiert die globalen Instrumentationsschemata
		setOfISchemes.add(new GlobalInstrumentationScheme("Level 1 instrumentation", 1));
		setOfISchemes.add(new GlobalInstrumentationScheme("Level 2 instrumentation", 2));
		setOfISchemes.add(new GlobalInstrumentationScheme("Level 3 instrumentation", 3));
		return true;
	}

	/**
	 * ffnet ein Projekt mit einem ProgressMonitor (GUI).
	 * 
	 * @param file
	 *            die Projektdatei
	 * @param workbenchWindow
	 *            das zugehrige Window
	 * 
	 * @param recover
	 *            ob die originalen Quelldateien wiederhergestellt werden sollen
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean openProject(String file, IWorkbenchWindow workbenchWindow, boolean recover) {

		if (file == null || !(new File(file)).exists()) {
			Logger.logln("No such file");
			return false;
		}

		Logger.log("\t Opening Project " + file + " ... ");

		reinit();

		if (openProjectConfiguration(file)) {
			Logger.logln("OK.");
		} else {
			Logger.logln("Failed.");
			return false;
		}

		if (recover)
			restoreProject();
		parseProject(workbenchWindow);
		// backupProject();
		fireToGuiUpdateListeners();

		return true;
	}

	/**
	 * Liest die Sprachspezifikation der untersttzten Programmiersprachen fr SOTA ein.
	 * 
	 * @return true, wenn erfolgreich, sonst false
	 */
	@SuppressWarnings("unchecked")
	public boolean readLanguageSpecs() {
		File file = new File("language.spec");
		boolean specFound = false;

		Logger.logln("\t Checking LanguageSpecifiations:");

		// Parsen der XML-Datei
		SAXBuilder builder = new SAXBuilder();
		org.jdom.Document doc = null;
		try {
			doc = builder.build(file);

		} catch (JDOMException e) {
			Logger.logln("\t Failed to parse " + file.getName() + "!");
			Logger.logln(e.getMessage());
			LanguageSpecification spec = new JavaLanguageSpecification();
			Logger.log("\t Using Builtin-Specification for Java ... ");
			if(spec.checkSpecification()) {
				Logger.logln("OK.");
				config.setOfLangSpecs.put(spec.name, spec);
				return true;
			}
			else {
				Logger.logln("Failed.");
				return false;
			}
		} catch (IOException e) {
			Logger.logln("\t Failed to read " + file.getName() + "!");
			Logger.logln(e.getMessage());
			LanguageSpecification spec = new JavaLanguageSpecification();
			Logger.log("\t Using Builtin-Specification for Java ... ");
			if(spec.checkSpecification()) {
				Logger.logln("OK.");
				config.setOfLangSpecs.put(spec.name, spec);
				return true;
			}
			else {
				Logger.logln("Failed.");
				return false;
			}
		}

		// Anlegen der LanguageSpecification
		Element root = doc.getRootElement();

		for (Element ls : (List<Element>) root.getChildren("LanguageSpecification")) {

			LanguageSpecification spec = new LanguageSpecification();
			spec.name = ls.getChildText("Name");

			Logger.log("\t\t " + spec.name + " ... ");

			for (Element e : (List<Element>) ls.getChild("Extensions").getChildren())
				spec.extensions.add(e.getText());

			spec.astBuilder = ls.getChildText("ASTBuilder");
			if (ls.getChild("LogReader") != null)
				spec.logReader = ls.getChildText("LogReader");
			if (ls.getChild("CFGBuilder") != null)
				spec.cfgBuilder = ls.getChildText("CFGBuilder");
			if (ls.getChild("PartitionScanner") != null)
				spec.partitionScanner = ls.getChildText("PartitionScanner");
			if (ls.getChild("SourceViewerConfiguration") != null)
				spec.sourceViewerConfiguration = ls.getChildText("SourceViewerConfiguration");
			if (ls.getChild("MetricsManager") != null)
				spec.metricManager = ls.getChildText("MetricsManager");
			if (ls.getChild("LogReader") != null)
				spec.logReader = ls.getChildText("LogReader");
			if (ls.getChild("TestLogManager") != null)
				spec.testLogManager = ls.getChildText("TestLogManager");

			if (spec.checkSpecification()) {
				Logger.logln("OK.");
				config.setOfLangSpecs.put(spec.name, spec);
				specFound = true;
			} else {
				Logger.logln("Failed.");
			}

		}
		return specFound;
	}

	/**
	 * Startet die Kompilation des Projektes anhand eines Ant-Buildfiles, ohne Progressmonitor (vgl.
	 * ProjectBuilder-Klasse).
	 * 
	 * @param antBuildFile
	 *            das Ant-BuildFile
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean compileProject(String antBuildFile) {

		if (antBuildFile == null)
			antBuildFile = config.antBuildFile;

		Logger.logln("\t Building project ...");
		String antLocation = config.antLocation;
		if (antLocation == null || !(new File(antLocation).isFile())) {
			Logger.logln("Error: location of ant needed to compile!");
			return false;
		}
		File scriptFile = null;
		try {
			scriptFile = new File(antLocation);
			if (!scriptFile.exists()) {
				Logger.logln("\t Ant not found at: " + antLocation);
				config.exceptionMessage = "Failed to build project.\n\n" + "\t Ant not found at: " + antLocation;
				return false;
			}

			ProcessBuilder processBuilder = new ProcessBuilder(scriptFile.getAbsolutePath(), "clean", "build", "-f",
					antBuildFile);
			processBuilder.directory(scriptFile.getParentFile());
			processBuilder.redirectErrorStream(true);
			Process process = processBuilder.start();

			Scanner scanner = new Scanner(process.getInputStream());
			while (scanner.hasNextLine()) {
				Logger.logln(scanner.nextLine());
			}
			scanner.close();
			Logger.logln(process.waitFor());

		} catch (IOException e1) {
			Logger.logln("ioexception: " + e1.toString());
			if (Application.hasInstance())
				Application.getInstance().getConfiguration().exceptionMessage = "Failed to build project with:\n\n"
						+ scriptFile.getAbsolutePath() + " clean build -f " + antBuildFile + ".\n\n" + e1.getMessage();
			return false;
		} catch (InterruptedException e) {
			Logger.logln("interrupted exception: " + e.toString());
			return false;
		}
		return true;
	}

	/**
	 * Startet das Projekt anhand eines Startskripts, ohne Progressmonitor (vgl. ProjectRunner-Klasse).
	 * 
	 * @param runScript
	 *            das Startskript
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean runProject(String runScript) {
		try {
			if (runScript == null || !(new File(runScript).isFile())) {
				Logger.logln("Error: runscript " + runScript + " can't be found!");
				return false;
			}

			String arguments[] = { runScript, "1" };

			Logger.logln("\t Running project ...");
			ProcessBuilder processBuilder = new ProcessBuilder(arguments);
			processBuilder.directory(new File(runScript).getParentFile());
			Process process = processBuilder.start();
			Scanner scanner = new Scanner(process.getInputStream());
			while (scanner.hasNextLine()) {
				Logger.logln(scanner.nextLine());
			}
			scanner.close();
			Logger.logln(process.waitFor());

		} catch (IOException e1) {
			Logger.logln("ioexception: " + e1.toString());
			Application.getInstance().getConfiguration().exceptionMessage = "Failed to run project with:\n\n"
					+ Application.getInstance().getConfiguration().runScript + "\n\n" + e1.getMessage();
			return false;
		} catch (InterruptedException e) {
			Logger.logln("interrupted exception: " + e.toString());
			return false;
		}
		return true;
	}

	/** ************************************* */
	/* Projektfunktionen (aus GUI/ATM) */
	/** ************************************* */

	/**
	 * Fgt eine Datei in das Projekt hinzu.
	 * 
	 * @param s
	 *            die Datei samt Pfad
	 * @return die Projektkonfiguration
	 */
	public ProjectConfig addFile(String s) {
		if (config.projectStatus >= ProjectConfig.INSTRUMENTED)
			return null;
		else if (config.projectStatus == ProjectConfig.PARSED)
			reinit(config);

		config.projectStatus = ProjectConfig.CONFIGURED;
		config.setOfFileNames.add(s);
		Logger.logln("\t Adding file " + s + ".");
		// fireChangeToListeners();
		return config;
	}

	/**
	 * Gibt die aktuelle Konfiguration zurck.
	 * 
	 * @return die Konfiguration
	 */
	public ProjectConfig getConfiguration() {
		return this.config;
	}

	/**
	 * Startet das Parsen des Projektes mit einem ProgressMonitor.
	 * 
	 * @param window
	 *            IWorkbenchWindow fr den ProgressMonitor oder null, wenn ohne GUI gestartet (SOTA-ATM)
	 * @return die ProjektKonfiguration
	 */
	public boolean parseProject(IWorkbenchWindow window) {

		ProgressMonitorDialog pmd;

		if (config.projectStatus == ProjectConfig.EMPTY)
			return false;
		try {
			flag[0] = false;

			// ohne Progressmonitor (SOTA-ATM)
			if (window == null) {
				int nrFilesParsed = 0;
				int i = 0, m = config.setOfFileNames.size();

				for (String s : config.setOfFileNames) {

					boolean success = false;

					Logger.logln("\t Parsing " + config.setOfFileNames.size() + " files:");

					Logger.log("\t\t " + s + " ... ");

					try {
						Class<IASTBuilder> astBuilder = LangClassProvider.getASTBuilder(config);

						success = astBuilder.newInstance().buildAST(this, s);

					} catch (InstantiationException e) {
						e.printStackTrace();
					} catch (IllegalAccessException e) {
						e.printStackTrace();
					}

					if (success)
						Logger.logln("OK.");
					else {
						Logger.logln("Failed.");
						config.setOfSkippedFiles.add(s);
					}
				}
				Logger.logln("\t " + nrFilesParsed + " TestLog" + (nrFilesParsed != 1 ? "s" : "") + " read.");

				// mit Progressmonitor (GUI)
			} else {

				pmd = new ProgressMonitorDialog(window.getShell()) {
					@Override
					protected void cancelPressed() {
						flag[0] = true;
						super.cancelPressed();
					}
				};
				pmd.run(true, true, new FileParser(config.setOfFileNames, this));
				if (flag[0]) {
					config.message = "Parsing canceled.";
					Application.getInstance().getASTManager().reinit();
					Logger.logln("\n\t Parsing canceled.");
					return false;
				} else
					config.errorMessage = "Parsing successful." + getFileNodes().size() + " Files parsed.";
			}

		} catch (Exception e) {
			if (window == null) {
				Logger.logln(e.getMessage() + "\n" + e.toString());
			} else if (e instanceof InvocationTargetException) {
				MessageDialog.openError(window.getShell(), "Parseerror", e.getCause().getMessage());
				Logger.logln(e.getCause().toString());
				config.errorMessage = getFileNodes().size() + "Error while parsing project.";
			} else {
				MessageDialog.openError(window.getShell(), "Parseerror", e.getMessage());
				Logger.logln(e.toString());
				Logger.logln(e.getStackTrace().toString());
				config.errorMessage = getFileNodes().size() + "Error while parsing project.";

			}
			return false;
		}

		// Behandlung fehlerhaft geparster Dateien
		for (String s : config.setOfSkippedFiles) {
			if (config.setOfFileNames.contains(s))
				config.setOfFileNames.remove(s);
		}

		config.projectStatus = ProjectConfig.PARSED;
		if (!config.getListOfTestLogs().isEmpty())
			config.projectStatus = ProjectConfig.LOGREAD;

		if (config.setOfSkippedFiles.isEmpty())
			Logger.logln("\t Project parsed.");
		else {
			System.out.println("\t Project parsed."
					+ (config.setOfSkippedFiles.isEmpty() ? "" : " Skipped " + config.setOfSkippedFiles.size()
							+ " files."));
			String fileList = "";
			for (String s : config.setOfSkippedFiles) {
				fileList += "\n " + s;
			}
			MessageDialog.openInformation(window.getShell(), "ParseExceptions", "Skipped "
					+ config.setOfSkippedFiles.size() + " Files because of ParseExceptions." + fileList);
			config.setOfSkippedFiles.clear();
		}

		// statische Analyse
		try {
			LangClassProvider.getMetricManager(config).newInstance().computeMetrics(this);
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}

		// Update
		checkSourceStatus();
		fireToProjectChangeListeners();

		fireToGuiUpdateListeners();

		return true;
	}

	/**
	 * Fhrt das Parsen des Projektes als Thread mit Progressmonitor durch.
	 * 
	 */
	public class FileParser implements IRunnableWithProgress {
		TreeSet<String> set;

		ASTManager ast;

		/**
		 * Der Konstruktor.
		 * 
		 * @param set
		 *            die Menge an Dateien
		 * @param ast
		 *            der ASTManager
		 */
		private FileParser(TreeSet<String> set, ASTManager ast) {
			this.set = set;
			this.ast = ast;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.operation.IRunnableWithProgress#run(org.eclipse.core.runtime.IProgressMonitor)
		 */
		public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {

			int nrFilesParsed = 0;
			int i = 0, m = set.size();

			Logger.logln("\t Parsing " + set.size() + " files:");
			monitor.beginTask("Reading Files", set.size());

			for (String s : set) {
				boolean success = false;
				if (flag[0])
					return;
				try {

					Logger.log("\t\t " + s + " ... ");

					Class<IASTBuilder> astBuilder = LangClassProvider.getASTBuilder(config);

					success = astBuilder.newInstance().buildAST(ast, s);
				} catch (InstantiationException e) {
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					e.printStackTrace();
				}

				if (success) {
					nrFilesParsed++;
					Logger.logln("OK.");
				} else {
					Logger.logln("Failed.");
					config.setOfSkippedFiles.add(s);
				}

				monitor.worked(1);
				monitor.subTask(s + " (" + ++i + "/" + m + ")");

			}
			monitor.done();
			// Logger.logln("\t " + nrFilesParsed + " Files" + (nrFilesParsed != 1 ? "s" : "") + " parsed.");
		}

	}

	/**
	 * Parst das Projekt ohne Progressmonitor
	 * 
	 * @return true, wenn erfolgreich,sonst false
	 */
	public boolean parseProject() {
		return parseProject(null);
	}

	/**
	 * @param fn
	 */
	public void removeFileNode(FileNode fn) {

		Logger.logln("\t Removing " + fn.getFullName() + " from Project.");

		for (ASTClassNode cn : fn.getClassNodes()) {
			if (setOfTopLevelClassNodes.contains(cn))
				setOfTopLevelClassNodes.remove(cn);
			if (setOfClassNodes.contains(cn))
				setOfClassNodes.remove(cn);
		}

		for (ASTFunctionNode fcn : fn.getFunctionNodes()) {
			if (setOfTopLevelFunctionNodes.contains(fcn))
				setOfTopLevelFunctionNodes.remove(fcn);
			if (setOfFunctionNodes.contains(fcn))
				setOfFunctionNodes.remove(fcn);
		}

		if (setOfFileNodes.contains(fn)) {
			setOfFileNodes.remove(fn);

		}
	}

	/**
	 * Instrumentiert das Projekt nach aktueller Konfiguration.
	 * 
	 * @return die Konfiguration
	 */
	public boolean instrumentProject() {
		if (!backupProject()) {
			Logger.logln("\t Instrumentation failed! Project is already instrumented!");
			return false;
		}
		Logger.logln("\t Instrumenting " + setOfFileNodes.size() + " Files ...");
		String s;
		for (FileNode fn : setOfFileNodes) {
			Logger.log("\t\t " + fn.getFullName() + " ... ");
			if (fn.isInstrumented() || fn.isDirty()) {
				Logger.logln("is already instrumented. Skipping file.");
				continue;
			}

			// erstellt ASCContent fr File
			fn.createASCContentForScope(fn, config.getCurrentIScheme());

			// File schreiben
			try {
				BufferedWriter out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(fn.getFullName())));
				for (ASTNode n = fn.getStartNode(); n != fn.getEndNode(); n = n.getNextNode()) {
					s = n.getASCContent();
					out.write(s, 0, s.length());
				}
				out.close();
			} catch (IOException e) {
				Logger.logln("IOException.");
				continue;
			}
			fn.setDirty(true);
			fn.clearASCContent();
			Logger.logln("OK.");
		}
		fireToProjectChangeListeners();
		return true;
	}

	/**
	 * Liest Logfiles mit einem ProgressMonitor ein und verarbeitet diese (GUI).
	 * 
	 * @param set
	 *            die Menge der Logdateien
	 * @param window
	 *            das aktuelle Window
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean readLog(TreeSet<String> set, IWorkbenchWindow window) {
		if (config.projectStatus < ProjectConfig.PARSED)
			return false;

		ProgressMonitorDialog pmd;
		flag[0] = false;

		// Einlesen der Logs
		try {
			pmd = new ProgressMonitorDialog(window.getShell()) {

				@Override
				protected void cancelPressed() {
					flag[0] = true;
					super.cancelPressed();
				}
			};
			pmd.run(true, true, new ASTLogReader(set, this));
		} catch (Exception e) {
			MessageDialog.openError(window.getShell(), "Error", e.getMessage());
			return false;
		}

		// Verarbeitung der Logdaten
		try {
			LangClassProvider.getTestLogManager(config).newInstance().completeTestLogs(this);
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}

		// Update
		Application.getInstance().updateStatusLine(null, true);
		fireToGuiUpdateListeners();
		return true;
	}

	/**
	 * Liest Logfiles mit einem ProgressMonitor ein und verarbeitet diese (SOTA-ATM).
	 * 
	 * @param set
	 *            die Menge der Logdateien
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean readLog(TreeSet<String> set) {
		if (config.projectStatus < ProjectConfig.PARSED)
			return false;
		int nrTestsRead = 0;

		// Einlesen der Logdaten
		for (String s : set) {

			TestLog testLog = null;

			try {
				Class<LogReader> logReader;
				logReader = (Class<LogReader>) Class.forName(LangClassProvider.getLogReader(config));

				testLog = logReader.newInstance().readTestLogs(s, this);

				if (testLog != null) {
					config.addTestLog(testLog);
					nrTestsRead++;
				} else
					config.addBadTestFile(s, new File(s));
			} catch (ClassNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (InstantiationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}
		Logger.logln("\t " + nrTestsRead + " TestLog" + (nrTestsRead != 1 ? "s" : "") + " read.");

		// Verarbeitung der Logdaten
		try {
			LangClassProvider.getTestLogManager(config).newInstance().completeTestLogs(this);
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}

		// Update
		fireToGuiUpdateListeners();
		return true;
	}

	/**
	 * Klasse zum Einlesen von Logdateien als Thread mit ProgressMonitor.
	 * 
	 */
	public class ASTLogReader implements IRunnableWithProgress {
		TreeSet<String> set;

		ASTManager ast;

		/**
		 * Der Konstruktor.
		 * 
		 * @param set
		 *            die Menge an Logdateien
		 * @param ast
		 *            der ASTManager
		 */
		private ASTLogReader(TreeSet<String> set, ASTManager ast) {
			this.set = set;
			this.ast = ast;
		}

		/*
		 * (non-Javadoc)
		 * 
		 * @see org.eclipse.jface.operation.IRunnableWithProgress#run(org.eclipse.core.runtime.IProgressMonitor)
		 */
		@SuppressWarnings("unchecked")
		public void run(IProgressMonitor monitor) throws InvocationTargetException, InterruptedException {

			int nrTestsRead = 0;
			int i = 0, m = set.size();
			monitor.beginTask("Reading Logfiles", set.size());

			for (String s : set) {
				if (flag[0])
					return;
				TestLog testLog = null;

				try {
					Class<LogReader> logReader;
					logReader = (Class<LogReader>) Class.forName(LangClassProvider.getLogReader(config));

					ast.getConfiguration().tempString = "";
					testLog = logReader.newInstance().readTestLogs(s, ast);

					if (testLog != null) {
						config.addTestLog(testLog);
						nrTestsRead++;
					} else {
						config.addBadTestFile(s, new File(s));
					}
				} catch (ClassNotFoundException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (InstantiationException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				} catch (IllegalAccessException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
				monitor.worked(1);
				monitor.subTask(s + " (" + ++i + "/" + m + ")");

			}
			monitor.done();
			Logger.logln("\t " + nrTestsRead + " TestLog" + (nrTestsRead != 1 ? "s" : "") + " read.");
		}

	}

	/**
	 * Appliziert einen Testfall auf den AST und ermglicht dessen Auswertung.
	 * 
	 * @param t
	 *            der Testfall
	 */
	public void applyTestCase(TestCase t) {
		TestLogManager.applyTestCase(this, t);
	}

	/**
	 * Berechnet die berdeckungsmae fr die applizierten Testflle.
	 */
	public void computeCoverage() {
		try {
			config.setMetrics(LangClassProvider.getMetricManager(config).newInstance().computeCoverage(this));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Setzt alle applizierten Testdaten zurck.
	 */
	public void resetHits() {
		TestLogManager.resetHits(this);
	}

	/**
	 * Fgt ein IScheme zu dem Projekt hinzu.
	 * 
	 * @param scheme
	 *            das IScheme
	 */
	public void addInstrumentationScheme(IScheme scheme) {
		config.addIScheme(scheme);
		saveProject();
		fireToGuiUpdateListeners();
	}

	/**
	 * Konstruiert den Kontrollflussgraphen fr eine Funktion.
	 * 
	 * @param fn
	 *            der Funktionsknoten
	 * @param gcm
	 *            der Graphmanager der CFGView
	 * @return der konstruierte DirectedGraph
	 */
	public DirectedGraph buildCFG(ASTFunctionNode fn, GraphManager gcm) {
		try {
			return LangClassProvider.getCFGBuilder(config).newInstance().buildCFG(fn, gcm);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return new DirectedGraph();
	}

	/**
	 * Legt fest, ob die berdeckung in den Views farblich hervorgehoben werden soll.
	 * 
	 * @param b
	 *            true, wenn ja, sonst false
	 */
	public void showCoverage(boolean b) {
		config.showCoverage(b);
		fireToGuiUpdateListeners();
	}

	/**
	 * Liefert die Information, ob die berdeckung farblich angezeigt werden soll.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean showCoverage() {
		return config.showCoverage();
	}

	/** ************************************* */
	/* allgemeine Getters und Setters */
	/** ************************************* */

	/**
	 * Gibt das TreeSet aller Dateiknoten zurck.
	 * 
	 * @return die FileNodes
	 */
	public TreeSet<FileNode> getFileNodes() {
		return setOfFileNodes;
	}

	/**
	 * Gibt TreeSet aller Klassenknoten zurck.
	 * 
	 * @return die ASTClassNodes
	 */
	public TreeSet<ASTClassNode> getClassNodes() {
		return setOfClassNodes;
	}

	/**
	 * Gibt TreeSet aller Top-Level-Klassenknoten zurck.
	 * 
	 * @return die Top-Level-ASTClassNodes
	 */
	public TreeSet<ASTClassNode> getTopLevelClassNodes() {
		return setOfTopLevelClassNodes;
	}

	/**
	 * Gibt TreeSet aller Funktionsknoten zurck.
	 * 
	 * @return die ASTFunctionNodes
	 */
	public TreeSet<ASTFunctionNode> getFunctionNodes() {
		return setOfFunctionNodes;
	}

	/**
	 * Gibt TreeSet aller Top-Level-Funktionsknoten zurck
	 * 
	 * @return die Top-Level-ASTFunctionNodes
	 */
	public TreeSet<ASTFunctionNode> getTopLevelFunctionNodes() {
		return setOfTopLevelFunctionNodes;
	}

	/**
	 * Gibt die Metrik-Daten des Projektes zurck.
	 * 
	 * @return die Metriken
	 */
	public Metrics getProjectMetrics() {
		return config.getMetrics();
	}

	/**
	 * Gibt Knoten nach ID zurck.
	 * 
	 * @param id
	 *            die ID
	 * @return der Knoten
	 */
	public ASTNode getNodeByID(int id) {
		if (id >= listOfNodes.size()) {
			Logger.logln("Error@ASTManager.getNodeByID() - " + id + " not in List");
			return null;
		}
		return listOfNodes.get(id);
	}

	/**
	 * Gibt alle Knoten in einer Liste zurck.
	 * 
	 * @return die vollstndige Knotenliste
	 */
	public List<ASTNode> getAllNodes() {
		return listOfNodes;
	}

	/**
	 * Gibt den letzten Knoten zurck.
	 * 
	 * @return der letzte Knoten
	 */
	public ASTNode getLastNode() {
		return currFileNode.getLastNode();
	}

	/**
	 * Gibt den akutellen Projektstatus zurck.
	 * 
	 * @return der Status
	 */
	public int getProjectStatus() {
		return config.projectStatus;
	}

	/**
	 * Gibt den Status der Quelldateien zurck.
	 * 
	 * @return der Status
	 */
	public int getSourceStatus() {
		return config.sourceStatus;
	}

	/**
	 * Gibt den Status der geparsten Quelldateien zurck.
	 * 
	 * @return der Status
	 */
	public int getParsedSourceStatus() {
		return config.parsedSourceStatus;
	}

	/** *************************************************** */
	/* Schnittstellen zum ASTBuilder fr die AST-Erstellung */
	/** *************************************************** */

	/**
	 * Fgt einen neuen Knoten in den AST ein
	 * 
	 * @param n
	 *            der ASTNode
	 */
	public void append(ASTNode n) {
		currFileNode.append(n);
		listOfNodes.add(n.getID(), n);

		if (n instanceof ASTClassNode) {
			if (VERBOSE) {
				Logger.logln("ASTManager.addClassNode: " + ((ASTClassNode) n).getName());
			}

			setOfClassNodes.add(((ASTClassNode) n));
			if (((ASTClassNode) n).isTopLevel())
				setOfTopLevelClassNodes.add(((ASTClassNode) n));
			currFileNode.addClassNode((ASTClassNode) n);
		} else if (n instanceof ASTFunctionNode) {
			if (VERBOSE) {
				Logger.logln("ASTManager.addFunctionNode: " + ((ASTFunctionNode) n).getName());
			}
			setOfFunctionNodes.add(((ASTFunctionNode) n));
			if (((ASTFunctionNode) n).isTopLevel() || ((ASTFunctionNode) n).isFirstLevelFunction())
				setOfTopLevelFunctionNodes.add(((ASTFunctionNode) n));

			currFileNode.addFunctionNode((ASTFunctionNode) n);
		}
	}

	/**
	 * Fgt einen neuen Fileknoten in den AST ein.
	 * 
	 * @param fn
	 *            der FileNode
	 */
	public void addFileNode(FileNode fn) {
		if (VERBOSE) {
			Logger.logln("ASTManager.addFileNode: " + fn.getName());
		}
		setOfFileNodes.add(fn);
		currFileNode = fn;
	}

	/** ************************************* */
	/* Backup + Restore-Funktionen */
	/** ************************************* */

	/**
	 * Sichert die Quelldateien des Projektes.
	 * 
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean backupProject() {
		boolean success = true;
		Logger.logln("\t Backing up " + config.setOfFileNames.size() + " Files:");
		for (String file : config.setOfFileNames) {
			success = backupFile(file) & success;
		}
		if (!success)
			Logger.logln("\t No full backup! Some files were instrumented!");
		return success;
	}

	/**
	 * Stellt die originalen Quelldateien des Projekts wieder her.
	 * 
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean restoreProject() {
		boolean b = true;
		Logger.logln("\t Restoring " + setOfFileNodes.size() + " Files:");
		for (String file : config.setOfFileNames)
			b = restoreFile(file) & b;
		checkSourceStatus();
		return b;
	}

	/**
	 * Sichert eine Quelldatei.
	 * 
	 * @param fileName
	 *            die Datei
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean backupFile(String fileName) {

		Logger.log("\t\t " + fileName + " ... ");

		if (isInstrumented(fileName)) {
			Logger.logln("is instrumented! No backup.");
			return false;
		}
		try {
			FileInputStream in = new FileInputStream(fileName);
			FileOutputStream out = new FileOutputStream(fileName + "." + config.backupExtension);
			/*
			 * Logger.log("\t Backing up: " + fileName + " -> " + fileName + "." + config.backupExtension);
			 */

			byte buf[] = new byte[4096];
			int len;
			while ((len = in.read(buf)) > 0)
				out.write(buf, 0, len);
			out.close();
			in.close();
		} catch (IOException e) {
			Logger.logln("IOException");
			Logger.logln(e.toString());
			return false;
		}

		Logger.logln("OK.");

		return true;
	}

	/**
	 * Stellt eine originale Quelldatei wieder her.
	 * 
	 * @param fileName
	 *            die Datei
	 * @return true, wenn erfolgreich, sonst false
	 */
	public boolean restoreFile(String fileName) {
		try {
			Logger.log("\t\t " + fileName + "." + config.backupExtension + " -> " + fileName + " ... ");

			FileInputStream in = new FileInputStream(fileName + "." + config.backupExtension);
			FileOutputStream out = new FileOutputStream(fileName);

			byte buf[] = new byte[4096];
			int len;
			while ((len = in.read(buf)) > 0)
				out.write(buf, 0, len);
			out.close();
			in.close();
		} catch (IOException e) {
			Logger.logln("Failed.");
			e.printStackTrace();
			return false;
		}
		Logger.logln("OK.");
		return true;
	}

	/**
	 * Teste, ob diese Datei schon instrumentiert ist.
	 * 
	 * @param file
	 *            die Datei
	 * @return true, wenn instrumentiert, sonst false
	 */
	public static boolean isInstrumented(String file) {
		try {
			BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
			for (int i = 0; i < 3; i++)
				if (in.readLine().startsWith(" /* ASC")) {
					in.close();
					return true;
				}
			in.close();
		} catch (IOException e) {
			return false;
		}
		return false;
	}

	/**
	 * Testet den Quellcodestatus des Projektes.
	 * 
	 * @return true, wenn mind. eine Datei instrumentiert ist, sonst false
	 */
	public boolean checkSourceStatus() {

		boolean projectIsDirty = false;
		boolean hasParsedDirty = false;
		for (FileNode fn : setOfFileNodes) {
			String file = fn.getFullName();
			boolean fileIsDirty = false;
			try {
				BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(file)));
				for (int i = 0; i < 3; i++) {
					String s = in.readLine();

					if (s.startsWith(" /* ASC")) {
						projectIsDirty = true;
						fileIsDirty = true;
					}
				}
				fn.setDirty(fileIsDirty);

				if (fn.isParsedDirty())
					hasParsedDirty = true;

				in.close();
			} catch (IOException e) {
				Logger.logln("IOException@ASTManager.checkStatus");
				projectIsDirty = true;
				fn.setDirty(true);
			}

		}
		if (projectIsDirty) {
			config.sourceStatus = ProjectConfig.DIRTY;
			if (hasParsedDirty)
				config.parsedSourceStatus = ProjectConfig.DIRTY;
			else
				config.parsedSourceStatus = ProjectConfig.CLEAN;
		} else {
			config.sourceStatus = ProjectConfig.CLEAN;
			config.parsedSourceStatus = ProjectConfig.CLEAN;
		}

		if (Application.hasInstance())
			Application.getInstance().updateStatusLine(null, false);

		return projectIsDirty;

	}

	/** ************************************* */
	/* Listener-Verwaltung */
	/** ************************************* */

	/**
	 * Fgt einen ProjectChangeListener hinzu.
	 * 
	 * @param listener
	 *            der Listener
	 */
	synchronized public void addProjectChangeListener(ProjectChangeListener listener) {
		if (!changeListeners.contains(listener))
			changeListeners.add(listener);
	}

	/**
	 * Entfernt einen ProjectChangeListener.
	 * 
	 * @param listener
	 *            der Listener
	 */
	synchronized public void removeProjectChangeListener(ProjectChangeListener listener) {
		changeListeners.remove(listener);
	}

	/**
	 * Ruft fr alle ProjectChangeListener die Funktion projectChanged(this) auf.
	 */
	synchronized public void fireToProjectChangeListeners() {
		for (ProjectChangeListener l : changeListeners) {
			l.projectChanged(this);
		}
		fireToGuiUpdateListeners();
	}

	/**
	 * Fgt einen GuiUpdateListener hinzu.
	 * 
	 * @param listener
	 *            der Listener
	 */
	synchronized public void addGuiUpdateListener(GuiUpdateListener listener) {
		if (!guiUpdateListeners.contains(listener))
			guiUpdateListeners.add(listener);
	}

	/**
	 * Entfernt einen GuiUpdateListener.
	 * 
	 * @param listener
	 *            der Listener
	 */
	synchronized public void removeGuiUpdateListener(GuiUpdateListener listener) {
		guiUpdateListeners.remove(listener);
	}

	/**
	 * Ruft an allen GuiUpdateListener die Funktion guiUpdate(this) auf.
	 */
	synchronized public void fireToGuiUpdateListeners() {
		for (GuiUpdateListener l : guiUpdateListeners) {
			l.guiUpdate(this);
		}
	}

	/**
	 * Fgt eine ProjectTreeView hinzu.
	 * 
	 * @param view
	 *            die View
	 */
	synchronized public void addProjectTreeView(IProjectTreeView view) {
		if (!projectTreeViews.contains(view))
			projectTreeViews.add(view);
	}

	/**
	 * Entfernt eine ProjectTreeView.
	 * 
	 * @param view
	 *            die View
	 */
	synchronized public void removeProjectTreeView(IProjectTreeView view) {
		projectTreeViews.remove(view);
	}

	/**
	 * Whlt an allen ProjectTreeViwes ein Objekt aus und aktualisiert die Darstellung.
	 * 
	 * @param o
	 *            das Objekt
	 */
	synchronized public void select(Object o) {
		for (IProjectTreeView a : projectTreeViews) {
			a.select(o);
		}
		fireToGuiUpdateListeners();
	}

	/** ************************************* */
	/* Print-Funktionen (kommandozeile) */
	/** ************************************* */

	/**
	 * Ausgabe aller Knoten des AST mit Informationen.
	 */
	public void printNodes() {
		for (FileNode fn : setOfFileNodes)
			fn.printNodes();
	}

	/**
	 * Ausgabe der aller Funktionen im AST.
	 * 
	 */
	public void printFunctions() {
		Logger.logln("\n Parsed Classes and Functions:");
		for (FileNode fn : setOfFileNodes) {
			Logger.logln("\t File: " + fn.getName());
			fn.printFunctions(2);
		}
	}

	/**
	 * Ausgabe der Kontrollflussgraphen des AST.
	 * 
	 */
	public void printCFG() {
		for (FileNode fn : setOfFileNodes)
			fn.printCFG();
	}

	/**
	 * Ausgabe der Bedingungen.
	 */
	public void printConditions() {
		for (FileNode fn : setOfFileNodes)
			fn.printConditions();
	}

	/**
	 * Ausgabe der geparsten Dateien
	 * 
	 */
	public void print() {
		for (FileNode fn : setOfFileNodes)
			fn.print();
	}
}
