package sota.ast.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Stack;

import sota.Logger;
import sota.app.Application;
import sota.ast.nodes.ASTBranchNode;
import sota.ast.nodes.ASTCaseNode;
import sota.ast.nodes.ASTCatchNode;
import sota.ast.nodes.ASTCompoundNode;
import sota.ast.nodes.ASTDoNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIfNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.ASTJumpNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.ASTStatementNode;
import sota.ast.nodes.ASTSwitchNode;
import sota.ast.nodes.ASTTryNode;

/**
 * 
 * Klasse mit statischen Methoden zur Berechnung der Boundary-Interior-Pfadberdeckung (BI).
 * 
 */
public class BIPathManager {

	protected static boolean INFO = false;
	protected static boolean VERBOSE = false;
	protected static boolean VERBOSE2 = false;
	protected static boolean VERBOSE3 = false;

	protected static ASTFunctionNode currFunctionNode = null;

	/**
	 * Kapselt die Informationen zu einem Pfad.
	 * 
	 */
	protected static class BIPathInfo {
		// Anzahl der Pfade
		public long nrPaths = 1;

		// Abbildung einer Knoten-ID eines Sprungzielknotens auf die Menge der zu ihm fhrenden BI-Pfade ber break,
		// return, throw
		public HashMap<Integer, Long> closedPaths = new HashMap<Integer, Long>();

		// Abbildung einer Knoten-ID eines Sprungzielknotens auf die Menge der zu ihm fhrenden BI-Pfade ber continue
		public HashMap<Integer, Long> continuedPaths = new HashMap<Integer, Long>();

		public String stringValue() {
			return "\n Path (" + nrPaths + ")\n\tclosedPaths: " + closedPaths + "\n\tcontinuedPaths: " + continuedPaths
					+ "\n";
		}
	}

	/**
	 * Berechnet die Anzahl der mglichen Pfade fr den Boundary-Interior-Pfadtest.
	 * 
	 * @param fn
	 *            der ASTFunctionNode der entsprechenden Funktion
	 */
	public static void computeNrBIPaths(ASTFunctionNode fn) {
		currFunctionNode = fn;
		// rekursive Bestimmung
		BIPathInfo pi = computeBIPaths(fn.getEntryNode(), fn.getEndNode());

		// Verarbeitung der noch offenen Pfade
		if (pi.closedPaths.containsKey(fn.getEndNode().getID())) {
			pi.nrPaths += pi.closedPaths.remove(fn.getEndNode().getID());
		}
		if (INFO)
			System.out.println(fn.getIDString() + "\n" + pi.nrPaths);
		if (!pi.closedPaths.isEmpty())
			System.out.println("closedPath not empty!\n" + pi.closedPaths);
		if (!pi.continuedPaths.isEmpty())
			System.out.println("continuedPaths not empty!\n" + pi.continuedPaths);
		fn.getMetrics().nrBIPaths = pi.nrPaths;
	}

	/**
	 * Berechnet PfadInfo fr die BIPfade eines Quellcodeabschnitt
	 * 
	 * @param start
	 *            Beginn des Abschnittes
	 * @param end
	 *            Ende des Abschnittes
	 * @return die berechnete PfadInfo
	 */
	protected static BIPathInfo computeBIPaths(ASTNode start, ASTNode end) {
		if (VERBOSE)
			System.out.println("----- computeBIPaths " + start.getID() + " -> " + end.getID() + " -----");
		BIPathInfo pi = new BIPathInfo();
		ASTNode endNode = null;

		// Verfolgt alle mglihen Pfade, erstellt bei Verzweigungen rekursiv die Pfadinfos und verknpft diese geeigenet
		ASTNode n = start;
		while (n != end) {
			// IF: die parallelen True- und False-Pfade werden an den aktuellen Pfad angehngt
			if (n instanceof ASTIfNode) {
				endNode = ((ASTIfNode) n).getEndNode();
				BIPathInfo truePaths = computeBIPaths(((ASTIfNode) n).getTrueNode(), ((ASTIfNode) n).getElseNode());
				BIPathInfo falsePaths = computeBIPaths(((ASTIfNode) n).getFalseNode(), endNode);
				pi = joinSequentialPaths(pi, joinParallelPaths(truePaths, falsePaths), endNode);
				if (endNode == null)
					System.out.println(n.getNodeString());
				n = endNode;
				// normale Schleife: die Pfade der Schleife werden als IterationsPfade verarbeitet
			} else if (n instanceof ASTIterationNode && !((ASTIterationNode) n).hasDoNode()) {
				endNode = ((ASTIterationNode) n).getEndNode();
				BIPathInfo loopPaths = computeBIPaths(((ASTIterationNode) n).getLoopNode(), endNode);
				pi = joinIterationPath(pi, loopPaths, (ASTIterationNode) n);
				if (endNode == null)
					System.out.println(n.getNodeString());

				n = endNode;
				// Do-While-Schleife: die Pfade der Schleife werden als IterationsPfade verarbeitet
			} else if (n instanceof ASTDoNode) {
				endNode = ((ASTDoNode) n).getEndNode();
				BIPathInfo loopPaths = computeBIPaths(((ASTDoNode) n).getLoopNode(), endNode);
				pi = joinIterationPath(pi, loopPaths, ((ASTDoNode) n).getIterationNode());
				if (endNode == null)
					System.out.println(n.getNodeString());

				n = endNode;
				// Switch: alle Cases werden als parallele Pfade verarbeitet
			} else if (n instanceof ASTSwitchNode) {
				endNode = ((ASTSwitchNode) n).getEndNode();
				BIPathInfo summedPaths = null;
				BIPathInfo tempPaths;
				for (ASTCaseNode cn : ((ASTSwitchNode) n).getListOfCases()) {
					tempPaths = computeBIPaths(cn, cn.getSwitchNode().getEndNode());
					if (summedPaths == null)
						summedPaths = tempPaths;
					else
						summedPaths = joinParallelPaths(summedPaths, tempPaths);
				}
				if (!((ASTSwitchNode) n).hasDefault())
					summedPaths.nrPaths++;
				pi = joinSequentialPaths(pi, summedPaths, endNode);
				if (endNode == null)
					System.out.println(n.getNodeString());

				n = endNode;

				// Ausnahmebehandlung: Catches werden als parellele Pfade behandelt
			} else if (n instanceof ASTTryNode) {
				ASTTryNode tn = (ASTTryNode) n;
				endNode = tn.getEndNode();
				// die Pfade des Try-Blocks
				BIPathInfo tempPaths = computeBIPaths(n.getNextNode(), tn.getEndOfTryBlock());
				// die Pfade der Exception-Auswertung (ohne Finally) werden als
				// parallel betrachtet - untere Schranke fr BI-Pfadtest
				for (ASTCatchNode cn : tn.getListOfCatches()) {
					if (!cn.isFinally()) {
						tempPaths = joinParallelPaths(tempPaths, computeBIPaths(cn.getNextNode(), cn
								.getEndOfCatchBlock()));
					}
				}
				if (tn.hasFinally()) {
					tempPaths = joinExceptionPaths(tempPaths, computeBIPaths(tn.getFinallyNode().getNextNode(), tn
							.getFinallyNode().getEndOfCatchBlock()), endNode);
				}
				pi = joinSequentialPaths(pi, tempPaths, tn.getEndNode());
				if (tn.getEndNode() == null)
					System.out.println(n.getNodeString());

				n = tn.getEndNode();

				// Sprnge fhren zur ffnung von closed- bzw. continued Paths
			} else if (n instanceof ASTJumpNode && !n.isVirtual()) {
				ASTJumpNode jn = (ASTJumpNode) n;
				ASTNode tn = ((ASTJumpNode) n).getTargetNode();
				if (tn.getID() >= end.getID() || tn.getID() <= start.getID()) {
					// continues werden in continuedPaths gespeichert - treten
					// nur in Schleifen auf
					if (jn.isContinue()) {
						if (pi.continuedPaths.containsKey(tn.getID()))
							pi.continuedPaths.put(tn.getID(), pi.continuedPaths.remove(tn.getID()) + pi.nrPaths);
						else
							pi.continuedPaths.put(tn.getID(), pi.nrPaths);

						// throws werden immer! als return betrachtet, d.h.
						// untere Schranke der BIPfade
					} else if (jn.isThrow()) {
						if (pi.closedPaths.containsKey(currFunctionNode.getEndNode().getID()))
							pi.closedPaths.put(currFunctionNode.getEndNode().getID(), pi.closedPaths
									.remove(currFunctionNode.getEndNode().getID())
									+ pi.nrPaths);
						else
							pi.closedPaths.put(currFunctionNode.getEndNode().getID(), pi.nrPaths);
						// break und return werden zu ihrem Zielknoten verbunden
					} else {
						if (pi.closedPaths.containsKey(tn.getID()))
							pi.closedPaths.put(tn.getID(), pi.closedPaths.remove(tn.getID()) + pi.nrPaths);
						else
							pi.closedPaths.put(tn.getID(), pi.nrPaths);
					}
					pi.nrPaths = 0;
					return pi;
				}
				// 2. JumpNode beendet nur einen CompoundBlock oder springt zu
				// einem Label im Suchbereich
				if (((ASTJumpNode) n).getTargetNode() == null)
					System.out.println(n.getNodeString());

				n = ((ASTJumpNode) n).getTargetNode();

			} else if (n instanceof ASTCompoundNode && !((ASTCompoundNode) n).isVirtual()
					&& ((ASTCompoundNode) n).isStartNode() && ((ASTCompoundNode) n).getEndNode().getID() < end.getID()) {
				endNode = ((ASTCompoundNode) n).getEndNode();
				pi = joinSequentialPaths(pi, computeBIPaths(n.getNextNode(), endNode), endNode);
				if (endNode == null)
					System.out.println(n.getNodeString());
				n = endNode;
				// Behandlung von geteilten Statements (inkl. innerer Klasendefinition)
			} else if (n instanceof ASTStatementNode && ((ASTStatementNode) n).isSepareted()) {
				if (((ASTStatementNode) n).getEndOfStatement() == null)
					System.out.println(n.getNodeString());

				n = ((ASTStatementNode) n).getEndOfStatement();
			} else
				n = n.getNextNode();
		}
		return pi;
	}

	/**
	 * Vereinigt die Pfadinfos zweier paralleler Pfade (true-false bei If, case bei Switch), d.h. addiert Pfade.
	 * 
	 * @param first
	 *            ein Pfad
	 * @param second
	 *            ein paralleler Pfad
	 * @return die parallele Vereinigung der Pfade
	 */
	protected static BIPathInfo joinParallelPaths(BIPathInfo first, BIPathInfo second) {
		if (VERBOSE)
			System.out.println("----- joinParallelPaths -----" + first.stringValue() + " + " + second.stringValue());

		// alle weiterhin geschlossenen Pfade vom ersten auf zweiten Pfadverlauf
		// bertragen
		for (Integer i : second.closedPaths.keySet()) {
			if (first.closedPaths.containsKey(i))
				first.closedPaths.put(i, first.closedPaths.remove(i) + second.closedPaths.get(i));
			else
				first.closedPaths.put(i, second.closedPaths.get(i));
		}

		// alle continue-Pfade von ersten auf zweiten Pfadverlauf bertragen
		for (Integer i : second.continuedPaths.keySet()) {
			if (first.continuedPaths.containsKey(i))
				first.continuedPaths.put(i, first.continuedPaths.remove(i) + second.continuedPaths.get(i));
			else
				first.continuedPaths.put(i, second.continuedPaths.get(i));
		}

		// offene Pfade bertragen
		first.nrPaths += second.nrPaths;
		if (VERBOSE)
			System.out.println(" = " + first.stringValue());

		return first;
	}

	/**
	 * Vereingt die Pfadinfos zweier aufeinanderfolgender Pfadverlufe an einem Endknoten. Offene Pfade zum Endknoten
	 * werden dort wieder eingefgt.
	 * 
	 * @param head
	 *            PfadInfo des Ursprungspfades
	 * @param tail
	 *            PfadInfo des Teilpfades
	 * @param endNode
	 *            der vereinigende Knoten des verzweigenden Konstrukts
	 * @return die in Reihe vereinigte PfadInfo
	 */
	protected static BIPathInfo joinSequentialPaths(BIPathInfo head, BIPathInfo tail, ASTNode endNode) {
		if (VERBOSE)
			System.out.println("----- joinSequentialPaths -----" + head.stringValue() + " -> " + tail.stringValue());

		// alle abgeschlossenen Pfade im Tail, die auf den endNode verweisen,
		// wieder zu den offenen Pfaden hinzufgen
		if (!tail.closedPaths.isEmpty() && tail.closedPaths.keySet().contains(endNode.getID()))
			tail.nrPaths += tail.closedPaths.remove(endNode.getID());

		// alle weiterhin geschlossenen Pfade vom Tail auf Head bertragen,
		// dabei mit den
		// offenen Pfaden des Heads multiplizieren!
		for (Integer i : tail.closedPaths.keySet()) {
			if (head.closedPaths.containsKey(i))
				head.closedPaths.put(i, head.closedPaths.remove(i) + head.nrPaths * tail.closedPaths.get(i));
			else
				head.closedPaths.put(i, head.nrPaths * tail.closedPaths.get(i));
		}

		// alle continue-Pfade von Tail auf Head weitergeben, dabei mit den
		// offenen Pfaden des Heads multiplizieren!
		for (Integer i : tail.continuedPaths.keySet()) {
			if (head.continuedPaths.containsKey(i))
				head.continuedPaths.put(i, head.continuedPaths.remove(i) + head.nrPaths * tail.continuedPaths.get(i));
			else
				head.continuedPaths.put(i, head.nrPaths * tail.continuedPaths.get(i));
		}

		// offene Pfade vom Tail zum Head hinzufgen, d.h. multiplizieren
		head.nrPaths *= tail.nrPaths;

		if (VERBOSE)
			System.out.println(" =  " + head.stringValue());

		// alles Head und Tail vereinigt, Head weitergeben
		return head;
	}

	/**
	 * Berechnet die PfadInfo der Schleife.
	 * 
	 * @param head
	 *            die PfadInfo des Verlaufes bis zum Eingang der Iteration
	 * @param loop
	 *            die PfadInfo eines Loopdurchlaufes
	 * @param iterationNode
	 *            der entsprechende ASTIterationNode
	 * @return die vereinigte PfadInfo
	 */
	protected static BIPathInfo joinIterationPath(BIPathInfo head, BIPathInfo loop, ASTIterationNode iterationNode) {
		if (VERBOSE)
			System.out
					.println("----- joinIterationPaths (" + iterationNode.getID() + " -> "
							+ iterationNode.getEndNode().getID() + ") ------" + head.stringValue() + " + "
							+ loop.stringValue());

		ASTNode endNode = iterationNode.getEndNode();
		long nrContinuePathsToIterationNode = 0;
		long nrClosedPathsToEndNode = 0;

		// wenn abgeschlossenen Pfade im Loop auf den endNode verweisen, dann
		// diese entfernen und merken
		if (!loop.closedPaths.isEmpty() && loop.closedPaths.keySet().contains(endNode.getID()))
			nrClosedPathsToEndNode = loop.closedPaths.remove(endNode.getID());

		// wenn continue-Pfade im Loop auf den iterationNode verweisen, dann
		// entfernen
		// und diese merken
		if (!loop.continuedPaths.isEmpty() && loop.continuedPaths.keySet().contains(iterationNode.getID()))
			nrContinuePathsToIterationNode = loop.continuedPaths.remove(iterationNode.getID());

		// alle weiterhin geschlossenen Pfade vom Loop auf Head bertragen,
		// dabei ein- und zweifachen Schleifendurchlauf beachten!
		for (Integer i : loop.closedPaths.keySet()) {
			if (head.closedPaths.containsKey(i))
				head.closedPaths.put(i, head.closedPaths.remove(i) + head.nrPaths * loop.closedPaths.get(i)
						* (1 + loop.nrPaths + nrContinuePathsToIterationNode));
			else
				head.closedPaths.put(i, head.nrPaths * loop.closedPaths.get(i)
						* (1 + loop.nrPaths + nrContinuePathsToIterationNode));
		}

		// alle weiteren continue-Pfade von Tail auf Head weitergeben,
		// dabei ein- und zweifachen Schleifendurchlauf beachten!
		for (Integer i : loop.continuedPaths.keySet()) {
			if (head.continuedPaths.containsKey(i))
				head.continuedPaths.put(i, head.continuedPaths.remove(i) + head.nrPaths * loop.continuedPaths.get(i)
						* (1 + loop.nrPaths + nrContinuePathsToIterationNode));
			else
				head.continuedPaths.put(i, head.nrPaths * loop.continuedPaths.get(i)
						* (1 + loop.nrPaths + nrContinuePathsToIterationNode));
		}

		// Pfade fr Boundary-Interior berechnen
		// 1. wenn keine Do-While-Schleife: dann der eine Pfad, der Schleife
		// umgeht
		// 2. fr einzelnen Durchlauf: LoopPfade + alle die zum Schleifenende
		// (break) springen + alle die zum Schleifenkopf springen(und dann
		// abbrechen ->
		// 1xcontinue und dann skip)
		// 3. fr genau zwei Durchlufe:
		// alle Pfade, die nicht Schleife abbrechen, mal alle Pfade, die
		// Schleife fhren
		// (LoopPfade + die zum Schleifenkopf springen)
		// mal
		// LoopPfade + die zum Schleifenkopf oder zum Schleifenende
		// springen)

		head.nrPaths *= (iterationNode.hasDoNode() ? 0 : 1)
				+ (loop.nrPaths + nrContinuePathsToIterationNode + nrClosedPathsToEndNode)
				+ ((loop.nrPaths + nrContinuePathsToIterationNode) * (loop.nrPaths + nrContinuePathsToIterationNode + nrClosedPathsToEndNode));

		// alles Head und Tail vereinigt, Head weitergeben
		if (VERBOSE)
			System.out.println(" = " + head.stringValue());
		return head;
	}

	/**
	 * Vereingt die Pfadinfos der vereinigten Pfade von Try + allen Catches mit den folgenden finally-Pfaden. Ruft
	 * joinSequentialPaths() mit gleichen Parametern auf, aber sorgt vorher dafr, dass alle Sprungpfade aus dem
	 * Try-Konstrukt auch durch den finallyNode laufen. Wichtig: Sprnge im finally berschreiben alle Sprnge der
	 * Try-Catchblcke
	 * 
	 * @param head
	 *            PfadInfo des Ursprungspfades (Try-Catch)
	 * @param tail
	 *            PfadInfo des Finally-Pfades
	 * @param endNode
	 *            der vereinigende Knoten des Try-Konstrukts
	 * @return die vereinigte PfadInfo
	 */
	protected static BIPathInfo joinExceptionPaths(BIPathInfo head, BIPathInfo tail, ASTNode endNode) {
		if (VERBOSE)
			System.out.println("----- joinExceptionPaths -----" + head.stringValue() + " -> " + tail.stringValue());

		// alle Pfade, auch die breaks und returns laufen durch finally durch,
		// d.h. alle mssen mit Pfadanzahl im finallyNode multipliziert werden
		for (Integer i : head.closedPaths.keySet())
			head.closedPaths.put(i, head.closedPaths.remove(i) * tail.nrPaths);

		for (Integer i : head.continuedPaths.keySet())
			head.continuedPaths.put(i, head.continuedPaths.remove(i) * tail.nrPaths);

		return joinSequentialPaths(head, tail, endNode);

	}

	/**
	 * Erstellt fr alle aktuellen TestLogs die BiPfade.
	 * 
	 * @param ast
	 *            der aktuelle ASTManager
	 */
	public static void createCoveredBIPathsSet(ASTManager ast) {
		for (TestLog t : ast.getConfiguration().getListOfTestLogs())
			for (TestCase tc : t.getListOfTestCases())
				for (ArrayList<Integer> l : tc.getListOfPaths())
					createCoveredBIPathsSet(ast, tc, l);

	}

	/**
	 * Erstellt fr eine Pfadliste die BIPfade, d.h. berechnet ein Long-Zahl, welche eindeutig den BI-Pfad
	 * reprsentiert.
	 * 
	 * @param ast
	 *            der ASTManager
	 * @param t
	 *            das entsprechende TestLog
	 * @param list
	 *            die Pfadliste aus dem TestLog
	 */
	protected static void createCoveredBIPathsSet(ASTManager ast, TestCase t, ArrayList<Integer> list) {
		Stack<ASTIterationNode> stackOfIterationNodes = new Stack<ASTIterationNode>();
		Stack<Integer> stackOfLoops = new Stack<Integer>();

		if (list.size() < 2)
			return;

		ASTFunctionNode functionNode = ast.getNodeByID(list.get(1)).getFunctionNode();
		long biPath = 0;
		int nrDecisions = 0;

		ASTNode lastDecision = functionNode.getEntryNode();

		if (VERBOSE2)
			System.out.println("\n Convert to BIPaths");

		if (VERBOSE2) {
			System.out.println(list);
			System.out.println(pathToString(list));
		}

		S: for (int index = 2; index < list.size(); index++) {

			ASTNode indexNode = ast.getNodeByID(list.get(index));

			// nichtverzweigende Knoten uninteressant (Level3)
			if (indexNode.getControlingBranchNode() == null)
				continue;

			// innerhalb eines 2+x.ten Schleifendurchlaufes -> uninteressant
			while (!stackOfIterationNodes.empty() && stackOfLoops.peek() > 2) {
				// wenn Iteration verlassen,dann vom Stack nehmen
				if (!stackOfIterationNodes.peek().isWithinBranchConstruct(indexNode.getID())) {
					stackOfIterationNodes.pop();
					stackOfLoops.pop();
					// wenn in Iteration, dann die Entscheidung berspringen, da uninteressant
				} else
					continue S;
			}

			ASTBranchNode branchNode = indexNode.getControlingBranchNode();

			// suche Entscheidung ohne geloggten Knoten (Switch ohne case, try ohne catch, Iteration ohne Loop
			// zwischen letzter und nchster Entscheidung:
			if (lastDecision != null) {
				ASTNode n = lastDecision;
				while (n != branchNode && n != null) {
					if (n instanceof ASTJumpNode)
						n = ((ASTJumpNode) n).getTargetNode();
					else if (n instanceof ASTStatementNode && ((ASTStatementNode) n).isSepareted())
						n = ((ASTStatementNode) n).getEndOfStatement();
					/*
					 * else if(n instanceof ASTTryNode) { biPath = addDecisionToBIPath(biPath, ++nrDecisions, 0); int
					 * nrBranches = ((ASTTryNode)n).getListOfCatches().size() + (((ASTTryNode)n).hasFinally() ? 0 : 1);
					 * // Entscheidungsbreite aufaddieren nrDecisions += (int) Math.ceil(Math.log(nrBranches) /
					 * Math.log(2)) - 1; }
					 */else if (n instanceof ASTCatchNode && !((ASTCatchNode) n).isFinally()) {
						if (((ASTCatchNode) n).getTryNode().hasFinally())
							n = ((ASTCatchNode) n).getTryNode().getFinallyNode();
						else
							n = ((ASTCatchNode) n).getTryNode().getEndNode();
					} else if (n instanceof ASTSwitchNode) {
						int nrBranches = ((ASTSwitchNode) n).getListOfCases().size()
								+ (((ASTSwitchNode) n).hasDefault() ? 0 : 1);

						if (VERBOSE2)
							Logger.logln("Switch: " + 0 + " / " + nrBranches);
						// decision ist 0 wenn Switch bersprungen, sonst Case-zahl
						biPath = addDecisionToBIPath(biPath, ++nrDecisions, 0);
						// Entscheidungsbreite aufaddieren
						nrDecisions += (int) Math.ceil(Math.log(nrBranches) / Math.log(2)) - 1;
						n = ((ASTSwitchNode) n).getEndNode();
					} else if (n instanceof ASTIterationNode) {
						if (VERBOSE2)
							Logger.logln("Iteration: " + 0);
						biPath = addDecisionToBIPath(biPath, ++nrDecisions, 0);
						n = ((ASTIterationNode) n).getEndNode();
					} else {
						n = n.getNextNode();
					}
				}
			}

			lastDecision = branchNode;

			// binre Entscheidung bei If
			if (branchNode instanceof ASTIfNode) {
				ASTIfNode in = (ASTIfNode) branchNode;

				if (indexNode == in.getTrueNode()) {
					if (VERBOSE2)
						Logger.logln("If-True " + indexNode.getID());
					biPath = addDecisionToBIPath(biPath, ++nrDecisions, 0);
					continue;
				} else if (indexNode == in.getFalseNode()) {
					if (VERBOSE2)
						Logger.logln("If-False " + indexNode.getID());
					biPath = addDecisionToBIPath(biPath, ++nrDecisions, 1);
					continue;
				}
			}

			// breite Entscheidung bei Switch
			if (branchNode instanceof ASTSwitchNode) {
				ASTSwitchNode sn = (ASTSwitchNode) branchNode;

				// Anzahl mglicher Verzweigungen
				int nrBranches = sn.getListOfCases().size() + (sn.hasDefault() ? 0 : 1);
				int decision = 0;
				int nrCases = 0;
				// nchste ID muss ein Case sein, oder ende von Switch
				for (ASTCaseNode cn : sn.getListOfCases()) {
					nrCases++;
					if (cn == indexNode) {
						// CaseNode gefunden
						decision = nrCases;
						break;
					}

				}
				if (VERBOSE2)
					Logger.logln("Switch: " + decision + " / " + nrBranches);
				// decision ist 0 wenn Switch bersprungen, sonst Case-zahl
				biPath = addDecisionToBIPath(biPath, ++nrDecisions, decision);
				// Entscheidungsbreite aufaddieren
				nrDecisions += (int) Math.ceil(Math.log(nrBranches) / Math.log(2)) - 1;
				continue;

				// breite Entscheidung bei Try
			}

			if (branchNode instanceof ASTTryNode) {
				ASTTryNode tn = (ASTTryNode) branchNode;
				ArrayList<ASTCatchNode> listOfCatches = tn.getListOfCatches();

				// Anzahl mglicher Verzweigungen
				int nrBranches = listOfCatches.size() + (tn.hasFinally() ? 0 : 1);
				int decision = 0;
				int nrCatches = 0;

				for (ASTCatchNode cn : listOfCatches) {
					nrCatches++;
					if (cn == indexNode) {
						// Durchlauf = 0 als Entscheidung
						if (!cn.isFinally())
							decision = nrCatches;
					}
				}

				if (VERBOSE2)
					Logger.logln("Try: " + decision + " / " + nrBranches);
				biPath = addDecisionToBIPath(biPath, ++nrDecisions, decision);
				// Entscheidungsbreite aufaddieren
				nrDecisions += (int) Math.ceil(Math.log(nrBranches) / Math.log(2)) - 1;
				continue;
			}
			// falls LoopNode, dann auch Eintritt in die Schleife
			if (branchNode instanceof ASTIterationNode) {
				ASTIterationNode in = (ASTIterationNode) branchNode;

				// wenn schon auf Stack, dann Wiederholung
				if (stackOfIterationNodes.contains(in)) {
					// falls zweite Wiederholung, dann noch Pfad merken
					if (stackOfLoops.peek() < 2) {
						stackOfLoops.push(stackOfLoops.pop() + 1);
						biPath = addDecisionToBIPath(biPath, ++nrDecisions, 1);
					}
					// ab 2. Wiederholung Pfad ignorieren

					// wenn nich auf Stack, dann neuer Eintritt
				} else {
					stackOfIterationNodes.push(in);
					stackOfLoops.push(new Integer(1));
					biPath = addDecisionToBIPath(biPath, ++nrDecisions, 1);
				}
				continue;
			}
		}
		// printBIPath(functionNode.getID(), biPath);
		t.addCoveredBIPath(functionNode.getID(), biPath);
		return;

	}

	/**
	 * Fgt eine Entscheidung an den BIPfad an.
	 * 
	 * @param biPath
	 *            der aktuelle BIPfad
	 * @param nrDecisions
	 *            wie breit die Entscheidung ist (Switch, Try)
	 * @param i
	 *            die Entscheidung binr kodiert , 0/1 bei einfacher, sonst lnger
	 * @return der BIPfad mit angefgter Entscheidung
	 */
	protected static long addDecisionToBIPath(long biPath, int nrDecisions, int i) {
		long bi = biPath | (i << (nrDecisions - 1));
		if (VERBOSE3) {
			printBIPath(0, biPath);
			System.out.println("adding: " + i);
			printBIPath(0, bi);
		}
		return bi;
	}

	/**
	 * Ausgabe eines Pfades in Binrform auf Kommandozeile.
	 * 
	 * @param f
	 *            die zugehrige Funktion
	 * @param i
	 *            der Pfad
	 */
	protected static void printBIPath(int f, long i) {
		System.out.print("Function " + f + ":");
		int z = 1;
		int c = 0;
		for (int j = 0; j < 64; j++) {
			if ((c % 8) == 0)
				System.out.print(" ");
			System.out.print(((i & z) == 0) ? "0" : "1");
			z = z << 1;
			c++;
		}
		System.out.println();
	}

	/**
	 * Erzeugt aus einer Liste aus Knoten-IDs einen informativen String ber den Pfad.
	 * 
	 * @param l
	 *            die Liste der IDs
	 * @return der erstellte String
	 */
	protected static String pathToString(ArrayList<Integer> l) {
		List<ASTNode> al = Application.getInstance().getASTManager().getAllNodes();
		HashSet<ASTNode> trues = new HashSet<ASTNode>();
		HashSet<ASTNode> falses = new HashSet<ASTNode>();
		HashSet<ASTNode> loops = new HashSet<ASTNode>();
		HashSet<ASTNode> iexits = new HashSet<ASTNode>();
		String s = "[ Level: " + l.get(0) + " >";

		for (int i = 1; i < l.size(); i++) {
			ASTNode n = al.get(l.get(i));
			if (n instanceof ASTIterationNode) {
				s += i + " It (" + n.getID() + "), ";
				loops.add(((ASTIterationNode) n).getLoopNode());
				iexits.addAll(((ASTIterationNode) n).getSetOfExitNodes());
			} else if (n instanceof ASTIfNode) {
				s += i + " If (" + n.getID() + "), ";
				trues.add(((ASTIfNode) n).getTrueNode());
				falses.add(((ASTIfNode) n).getFalseNode());
			} else if (n instanceof ASTSwitchNode) {
				s += i + " Sw (" + n.getID() + "), ";
			} else if (n instanceof ASTCaseNode) {
				s += i + " Cs (" + n.getID() + "), ";
			} else if (n instanceof ASTCatchNode) {
				s += i + " Ct (" + n.getID() + "), ";
			} else if (n instanceof ASTTryNode) {
				s += i + " Ty (" + n.getID() + "), ";
			} else if (n instanceof ASTFunctionNode) {
				s += i + " Fu (" + n.getID() + "), ";
			} else if (trues.contains(n)) {
				s += i + " tN (" + n.getID() + "), ";
			} else if (falses.contains(n)) {
				s += i + " fN (" + n.getID() + "), ";
			} else if (iexits.contains(n)) {
				s += i + " iX (" + n.getID() + "), ";
			} else if (loops.contains(n)) {
				s += i + " iL (" + n.getID() + "), ";
			} else {
				s += i + " (" + n.getID() + "), ";
			}
			if (i % 13 == 0)
				s += "\n";
		}
		return s + "]";
	}

}
