package sota.ast.util;

import java.util.HashSet;

import sota.ast.nodes.ASTCaseNode;
import sota.ast.nodes.ASTCatchNode;
import sota.ast.nodes.ASTClassNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIfNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.ASTStatementNode;
import sota.ast.nodes.ASTSwitchNode;
import sota.ast.nodes.ASTTryNode;
import sota.ast.nodes.FileNode;
import sota.ast.nodes.IConditionNode;

/**
 * Berechnung der berdeckungsmae anhand applizieter Testlogs.
 */

public class CoverageManager {

	/**
	 * Fhrt berdeckungsberechnung anhand der aktuell applizierten HitLists fr alle ScopeNodes des AST und das Projekt
	 * als Ganzes durch.
	 * 
	 * @param ast
	 *            Der aktuelle ASTManager
	 * @return die Metriken des Projekts
	 */
	public static Metrics computeCoverage(ASTManager ast) {

		for (ASTNode n : ast.getAllNodes()) {

			// berechne die Bedingungsberdeckung fr alle ConditionNodes im AST
			if (n instanceof IConditionNode) {
				((IConditionNode) n).computeConditionCoverage();
			}
		}

		// berechne berdeckungsmae fr FunctionNodes
		for (ASTFunctionNode fn : ast.getFunctionNodes())
			computeCoverageOfFunction(fn);

		// rechne hoch fr ClassNodes
		for (ASTClassNode cn : ast.getClassNodes()) {
			Metrics cm = new Metrics();

			for (ASTFunctionNode fn : cn.getFunctionNodes())
				cm.sumUp(fn.getMetrics());

			cm.nrClasses = cn.getClassNodes().size();
			cm.nrFunctions = cn.getFunctionNodes().size();
			cm.nrLines = cn.getEndNode().getLineNr() - cn.getLineNr();

			cm.computeCoverageMetrics();

			cn.setMetrics(cm);
		}

		// rechne hoch fr FileNodes
		for (FileNode fn : ast.getFileNodes()) {
			Metrics fm = new Metrics();

			for (ASTClassNode cn : fn.getClassNodes())
				fm.sumUp(cn.getMetrics());

			fm.nrClasses = fn.getClassNodes().size();
			fm.nrFunctions = fn.getFunctionNodes().size();
			fm.nrLines = fn.getEndNode().getLineNr();

			fm.computeCoverageMetrics();

			fn.setMetrics(fm);
		}

		// rechne hoch fr Projekt
		Metrics projectMetric = new Metrics();
		for (FileNode fn : ast.getFileNodes()) {
			projectMetric.sumUp(fn.getMetrics());
			projectMetric.nrLines += fn.getMetrics().nrLines;
		}

		projectMetric.nrInnerClasses = 0;
		for (ASTClassNode cn : ast.getClassNodes())
			if (!cn.isTopLevel())
				projectMetric.nrInnerClasses++;

		projectMetric.nrFiles = ast.getFileNodes().size();
		projectMetric.nrClasses = ast.getClassNodes().size();
		projectMetric.nrFunctions = ast.getFunctionNodes().size();

		projectMetric.computeCoverageMetrics();

		ast.getConfiguration().setMetrics(projectMetric);

		return projectMetric;
	}

	/**
	 * Berechnet die berdeckungsmae fr eine Funktion.
	 * 
	 * @param fn
	 *            der ASTFunctionNode, fr welchen die berdeckungmae berechnet werden sollen
	 * @return die berdeckungsmae der Funktion als Metric-Objekt
	 */
	protected static Metrics computeCoverageOfFunction(ASTFunctionNode fn) {
		Metrics m = new Metrics();

		// if(fn.getID() == 5 )
		// System.out.println(fn.getNodeString());

		// FEEC-berdeckung
		m.nrFunctions++;
		if (fn.getNrHits() > 0)
			m.nrEnteredFunctions++;

		for (ASTStatementNode exit : fn.getSetOfExitNodes()) {
			m.nrFunctionExits++;
			if (exit.getNrHits() > 0)
				m.nrReachedExits++;
		}

		// durchlaufe die Funktion bis zum Ende und sammel berdeckungsdaten
		ASTNode n = fn.getStartNode();
		while (n != fn.getEndNode()) {

			// falls innere Klasse, dann berspringen
			if (n instanceof ASTClassNode) {
				m.nrInnerClasses++;
				n = ((ASTClassNode) n).getEndNode();
			}

			// falls aktuelle Node ein n.virtuelle Statement ist, sammel
			// berdeckungsdaten
			if (n instanceof ASTStatementNode && !n.isVirtual()) {

				// Statements -> fr C0 berdeckung
				m.nrStatements++;
				if (((ASTStatementNode) n).getNrHits() > 0)
					m.nrCoveredStatements++;

				// Branches -> fr C1 berdeckung (
				if (n instanceof ASTIfNode) {
					m.nrBranches = m.nrBranches + 2;
					if (((ASTIfNode) n).getTrueNode().getNrHits() > 0)
						m.nrCoveredBranches++;
					if (((ASTIfNode) n).getFalseNode().getNrHits() > 0)
						m.nrCoveredBranches++;
				}

				if (n instanceof ASTSwitchNode) {
					for (ASTCaseNode cn : ((ASTSwitchNode) n).getListOfCases()) {
						m.nrBranches++;
						if (cn.getNrSelects() > 0)
							m.nrCoveredBranches++;
					}
				}

				if (n instanceof ASTTryNode) {
					m.nrBranches++;
					if (((ASTTryNode) n).getNrHitsWithoutException() > 0) {
						m.nrCoveredBranches++;
					}
					for (ASTCatchNode cn : ((ASTTryNode) n).getListOfCatches()) {
						if (!cn.isFinally()) {
							m.nrBranches++;
							if (cn.getNrHits() > 0)
								m.nrCoveredBranches++;
						}
					}
				}

				if (n instanceof ASTIterationNode) {
					ASTIterationNode in = (ASTIterationNode) n;
					if (in.hasDoNode()) {
						m.nrBranches = m.nrBranches + 2;
						if (in.getNrMultipleLoops() > 0)
							m.nrCoveredBranches++;
						if (in.getEndNode().getNrHits() > 0)
							m.nrCoveredBranches++;
					} else {
						m.nrBranches = m.nrBranches + 1 + (in.getEndNode().isReachable() ? 1 : 0);
						if (in.getNrLoops() > 0)
							m.nrCoveredBranches++;
						if (in.getEndNode().getNrHits() > 0)
							m.nrCoveredBranches++;
					}
				}

			}

			// sammelt fr Condition die Auswertunsdaten
			if (n instanceof IConditionNode && ((IConditionNode) n).getCondition() != null) {
				Metrics temp = ((IConditionNode) n).getMetrics();

				m.nrConditionNodes++;

				m.nrAtoms += temp.nrAtoms;
				m.nrAtomEvaluations += temp.nrAtomEvaluations;
				m.nrConditions += temp.nrConditions;
				m.nrConditionEvaluations += temp.nrConditionEvaluations;
				m.nrMcdcPairs += temp.nrMcdcPairs;
				m.nrAtomCombinations += temp.nrAtomCombinations;
				m.nrAtomCombinationEvaluations += temp.nrAtomCombinationEvaluations;
			}
			n = n.getNextNode();
		}

		// merke lokale, statische Mae
		m.nrInnerClasses = fn.getClassNodes().size();
		m.nrLines = fn.getEndNode().getLineNr() - fn.getLineNr();

		// bertrage vorher berechnete, statische Mae
		m.cyclomaticComplexity = fn.getMetrics().cyclomaticComplexity;
		m.essentialComplexity = fn.getMetrics().essentialComplexity;

		// Pfadinfos bertragen
		m.nrBIPaths = fn.getMetrics().nrBIPaths;
		m.nrModBIPaths = fn.getMetrics().nrModBIPaths;
		m.mapOfModBIPaths.putAll(fn.getMetrics().mapOfModBIPaths);
		m.mapOfCoveredModBIPaths.putAll(fn.getMetrics().mapOfCoveredModBIPaths);

		m.nrCoveredBIPaths = fn.getMetrics().setOfCoveredBIPaths.size();
		m.nrCoveredModBIPaths = 0;
		for (HashSet<Long> set : fn.getMetrics().mapOfCoveredModBIPaths.values())
			m.nrCoveredModBIPaths += set.size();

		// Berechne die Metriken aus den gesammelten Daten
		m.computeCoverageMetrics();

		// speicher Metrik im ScopeNode
		fn.setMetrics(m);

		return m;
	}

}
