package sota.ast.util;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Stack;

import sota.Logger;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.IConditionNode;

/**
 * Liest eine Logdatei ein und erstellt aus den gelesenen Daten ein TestCase mit Pfadinformationen.
 * 
 */
public class LogReader {

	protected static boolean VERBOSE = false;
	protected static FileInputStream input;
	protected static int intValue;
	protected static int pathID = -1;
	protected static boolean endOfFile = false;
	protected static boolean endOfLog = false;
	protected static Stack<ASTIterationNode> stackOfIterationNodes;

	/**
	 * Liest eine Logdatei ein und erstellt einen TestLog-Datensatz daraus.
	 * 
	 * @param logfile
	 *            die Logdatei
	 * @param ast
	 *            der ASTManager
	 * @return das erstellte Testlog.
	 */
	public TestLog readTestLogs(String logfile, ASTManager ast) {
		endOfFile = false;

		try {
			input = new FileInputStream(logfile);
		} catch (FileNotFoundException e) {
			Logger.logln("\t LogFile " + logfile + " couldn't be found.");
			return null;
		}
		Logger.logln("\t Reading LogFile " + logfile + " ...");
		TestLog t = readLog(input, logfile, ast);
		try {
			input.close();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		return t;
	}

	/**
	 * Liest die Logdaten aus einem Stream und schreibt PfadListe in einzelne TestCases.
	 * 
	 * @param input
	 *            der InputStream
	 * @param log
	 *            die Logdatei
	 * @param ast
	 *            der ASTManger
	 * @return das erstellte Testlog
	 */
	protected static TestLog readLog(FileInputStream input, String log, ASTManager ast) {
		TestLog testLog = null;
		TestCase testCase = null;

		ArrayList<String> list = new ArrayList<String>();

		String projectName, description, iScheme;
		int currId = 0;
		boolean skipIt = false;
		do {

			ArrayList<ArrayList<Integer>> pathList = new ArrayList<ArrayList<Integer>>();

			// Header einlesen
			try {
				intValue = read4Bytes();
				if (intValue > 100)
					return testLog;
				byte b[] = new byte[intValue];
				input.read(b);
				projectName = new String(b);
				Logger.logln("\t\tProjectName = " + projectName);

				skipIt = false;

				if (!projectName.equals(ast.getConfiguration().getName())) {
					if (ast.getConfiguration().tempString.equals(""))
						ast.getConfiguration().tempString = projectName;
					Logger.logln("\t Testcase " + log + " (" + currId + ")" + " belongs to project " + projectName
							+ ". Skipping ...");
					skipIt = true;
					list.add(projectName);
				}

				intValue = read4Bytes();
				b = new byte[intValue];
				input.read(b);
				description = new String(b);
				Logger.logln("\t\tDescription = " + description);

				intValue = read4Bytes();
				b = new byte[intValue];
				input.read(b);
				iScheme = new String(b);
				Logger.logln("\t\tiScheme = " + iScheme);

				if (!skipIt) {
					if (testLog == null)
						testLog = new TestLog(log, projectName, log.substring(log.lastIndexOf("\\") + 1, log
								.lastIndexOf(".")), description, iScheme);
					testCase = new TestCase(testLog.getListOfTestCases().size() + 1, log, projectName, log.substring(
							log.lastIndexOf("\\") + 1, log.lastIndexOf(".")), description, iScheme);
				}

			} catch (IOException e) {
				Logger.logln("LogReader: IOError");
				return testLog;
			}

			// Neuinitialisieren des Iterationstacks
			stackOfIterationNodes = new Stack<ASTIterationNode>();

			// Auslesen der Logdaten
			do {
				try {
					intValue = read4Bytes();
				} catch (IOException e) {
					Logger.logln("LogReader: IOError");
					return testLog;
				}

				// falls EOF erreicht, loglesen abbrechen
				if (endOfFile) {
					if (VERBOSE)
						Logger.logln("EOF");

					break;
				}
				if (intValue == 0) {
					if (VERBOSE)
						Logger.logln("0 -> break");
					break;
				}

				if (!skipIt) {
					if (isPathId(intValue)) {

						if (VERBOSE)
							Logger.logln(intToString(intValue) + " PathID: " + clearHeader(intValue));
						pathID = clearHeader(intValue);

						// falls neuer Path -> Pathlist initialisieren
						while (pathList.size() <= pathID)
							pathList.add(new ArrayList<Integer>());
					} else if (isNodeId(intValue)) {
						if (VERBOSE)
							Logger.logln(intToString(intValue) + " NodeID: " + clearHeader(intValue));
						currId = clearHeader(intValue);

						// aktuellen Knoten im aktuellen Pfad anfgen
						pathList.get(pathID).add(currId);

						// wenn Iterationskonstrukte verlassen wurden, dann
						// diese vom Stack nehmen
						while (stackOfIterationNodes.size() > 0) {
							if (!stackOfIterationNodes.peek().isWithinBranchConstruct(currId))
								stackOfIterationNodes.pop();
							else
								break;
						}
					} else {
						if (isBoolean(intValue)) {
							if (VERBOSE)
								Logger.logln(intToString(intValue) + " " + getNrBooleans(intValue) + " Boolean: "
										+ intToBooleanString(intValue) + "\tCurrID: " + currId);
							intValue = clearHeader(intValue);

							// Boolean-String in truthVectorMap merken
							if (ast.getNodeByID(currId) instanceof IConditionNode) {
								testCase.addTruthVector(currId, intToBooleanString(intValue));
								// System.out.println("TV added");
								// entferne ggf. die ConditionNodes, um
								// Essential- bzw. FullPathlists zu erstellen
								IConditionNode node = (IConditionNode) ast.getNodeByID(currId);

								// kein IConditionNode kommt bei Level 1 und 2 in den PathLists vor
								// alle IConditionNodes bleiben in Level 3 enthalten, auer:
								// IterationNode kommt bei Level 3 nur einmal am Eingang vor

								if (pathList.get(pathID).get(0) != 3) {
									if (VERBOSE)
										Logger.logln("removing conditionNode: "
												+ pathList.get(pathID).get(pathList.get(pathID).size() - 1));
									pathList.get(pathID).remove(pathList.get(pathID).size() - 1);
								} else if (node instanceof ASTIterationNode) {
									if (stackOfIterationNodes.contains(node))
										pathList.get(pathID).remove(pathList.get(pathID).size() - 1);
									else
										stackOfIterationNodes.push((ASTIterationNode) node);
								}
							} else {
								System.out.println("\tCan't find owner of Booleans! (" + currId
										+ "). Skipping TestCase.");
								list.add(projectName);
								skipIt = true;
							}
						}
						// momentan undefiniert
						else {
							Logger.logln("LogReader: Header undefined!");
							return testLog;
						}
					}
				}
			} while (!endOfFile && !endOfLog);
			endOfLog = false;
			if (!skipIt) {
				Logger.logln("\t\t" + (pathList.size() - 1) + " Paths read.");
				testCase.addListOfPaths(pathList);
				testLog.addTestCase(testCase);
			}
		} while (!endOfFile);

		if (testLog != null)
			testLog.addBadTestCases(list);
		return testLog;
	}

	/**
	 * Wandelt Bytearray in int um.
	 * 
	 * @param b
	 *            das byte-Array
	 * @return der int-Wert
	 */
	protected static int bytes2int(byte[] b) {
		int a = 0;
		for (int i = 3; i >= 0; i--) {
			if (b[i] >= 0)
				a |= b[i];
			else
				a |= 256 + b[i];
			if (i != 0)
				a <<= 8;
		}
		return a;
	}

	/**
	 * Liest vier Byte aus dem Stream.
	 * 
	 * @return der ermittelte Int-Wert
	 * @throws IOException
	 */
	protected static int read4Bytes() throws IOException {
		// vier Byte lesen
		int i;
		byte b[] = new byte[4];
		for (int c = 0; c < 4; c++) {
			i = input.read();

			if (i == -1) {
				endOfFile = true;
				return 0;
			}
			b[c] = (byte) i;
		}
		return bytes2int(b);
	}

	/**
	 * Wandelt Int-Wert in ein binren 0-1-Darstellung als String um.
	 * 
	 * @param value
	 *            der Int-Wert
	 * @return der binre 0-1-String
	 */
	protected static String intToString(int value) {
		String s = "";
		for (int i = 0; i < 32; i++) {
			if ((value & 1) == 1)
				s = "1" + s;
			else
				s = "0" + s;
			value >>>= 1;
		}
		return s;
	}

	/**
	 * Wandelt Int-Kodierung von booleans in eine Liste von Boolean-Werten in 0-1-Darstellung als String um.
	 * 
	 * @param booleans
	 *            der Int-Wert mit Header
	 * @return der binre 0-1-String
	 */
	protected static String intToBooleanString(int booleans) {
		String s = "";
		int nrBooleans = getNrBooleans(booleans);
		for (int i = 0; i < nrBooleans; i++) {
			if ((booleans & 1) == 1)
				s += "1";
			else
				s += "0";
			booleans >>>= 1;
		}
		return s;
	}

	/**
	 * Wandelt Int-Kodierung von booleans in einen boolean-Array um.
	 * 
	 * @param booleans
	 *            der Int-Wert mit Header
	 * @return der boolean-Array
	 */
	protected static boolean[] intToBooleanArray(int booleans) {
		int nrBooleans = getNrBooleans(booleans);
		boolean b[] = new boolean[nrBooleans];
		for (int i = 0; i < nrBooleans; i++) {
			if ((booleans & 1) == 1)
				b[i] = true;
			else
				b[i] = false;
			booleans >>>= 1;
		}
		return b;
	}

	/**
	 * Ermittelt die Anzahl an boolean Werten in einer Int-Kodierung aus dessen Header.
	 * 
	 * @param i
	 *            der Int-Wert
	 * @return die Anzahl an boolean-Werten
	 */
	protected static int getNrBooleans(int i) {
		return (clearHeader(i) >>> 24);
	}

	/**
	 * Lscht den Header eines Ints.
	 * 
	 * @param i
	 *            das Int
	 * @return das Int mit gelschtem Header
	 */
	protected static int clearHeader(int i) {
		return i &= 0x3FFFFFFF;
	}

	/**
	 * Entscheidet, ob das Int eine Path-ID kodiert.
	 * 
	 * @param value
	 *            das Int
	 * @return true, wenn Path-ID, false sonst
	 */
	protected static boolean isPathId(int value) {
		return ((value & 0xC0000000) == 0);
	}

	/**
	 * Entscheidet, ob das Int eine Knoten-ID kodiert.
	 * 
	 * @param value
	 *            das Int
	 * @return true, wenn Knoten-ID, false sonst
	 */
	protected static boolean isNodeId(int value) {
		return ((value & 0xC0000000) == 0x40000000);
	}

	/**
	 * Entscheidet, ob das Int einen Boolean-Array kodiert.
	 * 
	 * @param value
	 *            das Int
	 * @return true, wenn Boolean-Array, false sonst
	 */
	protected static boolean isBoolean(int value) {
		return ((value & 0xC0000000) == 0x80000000);
	}
}