package sota.ast.util;

import sota.Logger;
import sota.ast.nodes.ASTCaseNode;
import sota.ast.nodes.ASTCatchNode;
import sota.ast.nodes.ASTClassNode;
import sota.ast.nodes.ASTCompoundNode;
import sota.ast.nodes.ASTDoNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIfNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.ASTJumpNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.ASTStatementNode;
import sota.ast.nodes.ASTSwitchNode;
import sota.ast.nodes.ASTTryNode;

/**
 * Berechnet die statischen und dynamischen Metriken fr das Projekt.
 */
public class MetricManager {

	/**
	 * Berechnet die statischen Metriken fr das Projekt.
	 * 
	 * @param ast
	 *            der ASTManager
	 */
	public static void computeMetrics(ASTManager ast) {
		for (ASTFunctionNode fn : ast.getFunctionNodes()) {
			computeCyclomaticComplexity(fn);
			computeEssentialComplexity(fn);
			computeNrBIPaths(fn);
			computeNrModBIPaths(fn);
		}
		computeCoverage(ast);
	}

	/**
	 * Berechnet die dynamischen (berdeckungs-) Metriken fr das Projekt.
	 * 
	 */
	public static Metrics computeCoverage(ASTManager ast) {
		return CoverageManager.computeCoverage(ast);
	}

	/**
	 * Berechnet zyklomatische Komplexitt fr Funktionen.
	 * 
	 * @param fn
	 *            der FunctionNode
	 */
	protected static void computeCyclomaticComplexity(ASTFunctionNode fn) {
		int nrBranches = 0;

		// durchlaufe die Funktion bis zum Ende
		ASTNode n = fn.getStartNode();
		while (n != fn.getEndNode()) {

			if (n instanceof ASTClassNode) {
				n = ((ASTClassNode) n).getEndNode();
			} else {
				// sammel Daten fr cyclomatische Komplexitt
				if (n instanceof ASTStatementNode && !n.isVirtual()) {
					if (n instanceof ASTIfNode)
						nrBranches++;
					if (n instanceof ASTIterationNode)
						nrBranches++;
					if (n instanceof ASTSwitchNode)
						nrBranches += ((ASTSwitchNode) n).getNrOfCases() - (((ASTSwitchNode) n).hasDefault() ? 1 : 0);
					if (n instanceof ASTTryNode)
						nrBranches += ((ASTTryNode) n).getListOfCatches().size()
								- (((ASTTryNode) n).hasFinally() ? 1 : 0);
				}
				n = n.getNextNode();
			}
		}
		fn.getMetrics().cyclomaticComplexity = nrBranches + 1;
	}

	/**
	 * Berechnet essentielle Komplexitt fr Funktionen.
	 * 
	 * @param fn
	 *            der FunctionNode
	 */
	protected static void computeEssentialComplexity(ASTFunctionNode fn) {
		// System.out.println(fn.getName());
		int temp = computeEssentialComplexity(fn.getStartNode(), fn.getEndNode());

		fn.getMetrics().essentialComplexity = temp < 0 ? 1 : temp + 1;
		// System.out.println(fn.getName() + "X");
	}

	/**
	 * Berechnet die essentielle Komplexitt fr Codeabschnitte.
	 * 
	 * @param startNode
	 *            der Startknoten
	 * @param endNode
	 *            der Endknoten
	 * @return 0 falls Codeabschnitte nur reduzierbare CFG-Elemente enthlt, d.h. keine Jumps<br>
	 *         -1 falls Codeabschnitt mit Jump endet aber sonst keine nichtreduzierbaren Konstrukte enthlt<br>
	 *         m>0 falls Codeabschnitt nicht reduzierbar ist, m enthlt essentielle Komplexitt
	 */
	protected static int computeEssentialComplexity(ASTNode startNode, ASTNode endNode) {
		// System.out.println(startNode.getID() + " -> " + endNode.getID());
		int ec = 0;
		ASTNode n = startNode;

		while (n.getID() < endNode.getID()) {
			// System.out.println(n.getID() + " < " + endNode.getID());

			if (n instanceof ASTIfNode) {
				int tv, fv;
				tv = computeEssentialComplexity(((ASTIfNode) n).getTrueNode(), ((ASTIfNode) n).getElseNode());
				fv = computeEssentialComplexity(((ASTIfNode) n).getFalseNode(), ((ASTIfNode) n).getEndNode());
				// falls eins von beiden Zweigen nicht reduzierbar,
				// dann das if selber nicht reduzierbar
				if (tv != 0 || fv != 0) {
					// Komplexitten aufsummieren und If dazuzhlen
					ec += (tv > 0 ? tv : 0) + (fv > 0 ? fv : 0) + 1;
				}
				n = ((ASTIfNode) n).getEndNode();
			} else if (n instanceof ASTIterationNode && !((ASTIterationNode) n).hasDoNode()) {
				int temp = computeEssentialComplexity(n.getNextNode(), ((ASTIterationNode) n).getEndNode());
				// falls Iteration nicht reduzierbar
				if (temp != 0)
					ec += (temp > 0 ? temp : 0) + 1;
				n = ((ASTIterationNode) n).getEndNode();
			} else if (n instanceof ASTDoNode) {
				int temp = computeEssentialComplexity(n.getNextNode(), ((ASTDoNode) n).getEndNode());
				// falls DoIteration nicht reduzierbar
				if (temp != 0)
					ec += (temp > 0 ? temp : 0) + 1;
				n = ((ASTDoNode) n).getEndNode();
			} else if (n instanceof ASTSwitchNode) {
				int temp = 0;
				int switchTemp = 0;
				for (ASTCaseNode cn : ((ASTSwitchNode) n).getListOfCases()) {
					temp = computeEssentialComplexity(cn, cn.getEndOfCaseBlock());
					if (temp > 0)
						switchTemp += temp;
				}
				// falls mind. ein Case nich reduzierbar
				// dann Switch mit raufrechnen
				if (switchTemp > 0)
					ec += switchTemp + ((ASTSwitchNode) n).getNrOfCases() - (((ASTSwitchNode) n).hasDefault() ? 1 : 0);

				// falls alle Cases reduzierbar,
				// aber ein Case verlsst Switchconstrukt (z.b.: return)
				// dann Switch mit raufrechnen
				else {
					if (((ASTSwitchNode) n).getSetOfExitNodes().size() > 1) {
						ec += ((ASTSwitchNode) n).getNrOfCases() - (((ASTSwitchNode) n).hasDefault() ? 1 : 0);
					}
				}
				n = ((ASTSwitchNode) n).getEndNode();

			} else if (n instanceof ASTTryNode) {
				int temp = 0;
				int tryTemp = 0;

				// try-Block
				temp = computeEssentialComplexity(((ASTTryNode) n).getEntryOfTryBlock(), ((ASTTryNode) n)
						.getEndOfTryBlock());
				if (temp > 0)
					tryTemp += temp;

				// catches
				for (ASTCatchNode cn : ((ASTTryNode) n).getListOfCatches()) {
					temp = computeEssentialComplexity(cn, cn.getEndOfCatchBlock());
					if (temp > 0)
						tryTemp += temp;
				}
				// falls mind. ein Catch nich reduzierbar
				// dann Try mit raufrechnen
				if (tryTemp > 0)
					ec += tryTemp + ((ASTTryNode) n).getListOfCatches().size()
							- (((ASTTryNode) n).hasFinally() ? 1 : 0);

				// falls alle Catches reduzierbar,
				// aber ein Catch verlsst Tryconstrukt (z.b.: return)
				// dann Try mit raufrechnen
				else {
					if (((ASTTryNode) n).getSetOfExitNodes().size() > 1) {
						ec += ((ASTTryNode) n).getListOfCatches().size() - (((ASTTryNode) n).hasFinally() ? 1 : 0);
					}
				}
				n = ((ASTTryNode) n).getEndNode();

			} else if (n instanceof ASTCompoundNode && ((ASTCompoundNode) n).isStartNode()) {
				int temp = computeEssentialComplexity(n.getNextNode(), ((ASTCompoundNode) n).getEndNode());
				// falls Compound nicht reduzierbar
				if (temp > 0)
					ec += temp;
				n = ((ASTCompoundNode) n).getEndNode();

			} else if (n instanceof ASTJumpNode && !n.isVirtual()) {
				// gib Komplexitt oder -1 zurck
				return ec != 0 ? ec : -1;
			} else {
				n = n.getNextNode();
			}

		}
		return ec;
	}

	/**
	 * Berechnt die Anzahl an Boundary-Interior-Pfade durch eine Funktion.
	 * 
	 * @param fn
	 *            der FunctionNode
	 */
	protected static void computeNrBIPaths(ASTFunctionNode fn) {
		try {
			BIPathManager.computeNrBIPaths(fn);
		} catch (Exception e) {
			Logger.logln("Exception@MetricManager.computeNrBIPaths (" + fn.getFullyQualifiedName(true, true) + "):\n"
					+ e.toString());
			Logger.logln(e.getStackTrace());
		}
	}

	/**
	 * Berechnt die Anzahl an Modifizierten Boundary-Interior-Pfade durch eine Funktion.
	 * 
	 * @param fn
	 *            der FunctionNode
	 */
	protected static void computeNrModBIPaths(ASTFunctionNode fn) {
		try {
			ModBIPathManager.computeNrModBIPaths(fn);
		} catch (Exception e) {
			Logger.logln("Exception@MetricManager.computeNrModBIPaths (" + fn.getFullyQualifiedName(true, true) + ")\n"
					+ e.toString());
			Logger.logln(e.getStackTrace());
		}
	}
}
