package sota.ast.util;

import java.util.HashMap;
import java.util.HashSet;

/**
 * Kapselt alle Informationen zu statischen und berdeckungsmaen von Projektstrukturen.
 * 
 */
public class Metrics {

	// Daten fr berdeckungsberechnung

	public int nrFunctions = 0;
	public int nrEnteredFunctions = 0;

	public int nrFunctionExits = 0;
	public int nrReachedExits = 0;

	public int nrStatements = 0;
	public int nrCoveredStatements = 0;

	public int nrBranches = 0;
	public int nrCoveredBranches = 0;

	// Bedingungsberdeckung

	public int nrAtoms = 0;
	public int nrAtomEvaluations = 0;

	public int nrConditions = 0;
	public int nrConditionEvaluations = 0;

	public int nrMcdcPairs = 0;

	public int nrAtomCombinations = 0;
	public int nrAtomCombinationEvaluations = 0;

	// Pfadberdeckung

	public long nrModBIPaths = 0;
	public long nrCoveredModBIPaths = 0;

	public long nrBIPaths = 0;
	public long nrCoveredBIPaths = 0;

	// Zuordnung: (Funktion | Schleife) -> Anzahl der ModBI-Pfade
	public HashMap<Integer, Long> mapOfModBIPaths = new HashMap<Integer, Long>();

	// Zuordnung: (Funktion | Schleife) -> Menge aller berdeckter ModBI-Pfade
	public HashMap<Integer, HashSet<Long>> mapOfCoveredModBIPaths = new HashMap<Integer, HashSet<Long>>();

	// Menge aller berdeckten BI-Pfade
	public HashSet<Long> setOfCoveredBIPaths = new HashSet<Long>();

	// Komplexittsmae

	public int nrLines = 0;
	public int cyclomaticComplexity = 0;
	public int essentialComplexity = 0;
	public int nrConditionNodes = 0;

	// Komplexittsmae, nur Projekt (fr Report)

	public int nrFiles = 0;
	public int nrClasses = 0;
	public int nrInnerClasses = 0;

	// berdeckungsmae

	public float feec = 0;
	public float c0 = 0;
	public float c1 = 0;
	public float c2 = 0;
	public float c3 = 0;
	public float mbi = 0;
	public float bi = 0;
	public float mmdc = 0;
	public float mcdc = 0;

	/**
	 * Summiert zwei Metriken auf.
	 * 
	 * @param m
	 *            die aufzusummierenden Metrik
	 */
	public void sumUp(Metrics m) {
		nrFunctions += m.nrFunctions;
		nrEnteredFunctions += m.nrEnteredFunctions;
		nrFunctionExits += m.nrFunctionExits;
		nrReachedExits += m.nrReachedExits;

		nrStatements += m.nrStatements;
		nrCoveredStatements += m.nrCoveredStatements;

		nrBranches += m.nrBranches;
		nrCoveredBranches += m.nrCoveredBranches;

		nrModBIPaths += m.nrModBIPaths;
		mapOfModBIPaths.putAll(m.mapOfModBIPaths);
		nrCoveredModBIPaths += m.nrCoveredModBIPaths;

		nrBIPaths += m.nrBIPaths;
		nrCoveredBIPaths += m.nrCoveredBIPaths;

		nrAtoms += m.nrAtoms;
		nrAtomEvaluations += m.nrAtomEvaluations;

		nrConditions += m.nrConditions;
		nrConditionEvaluations += m.nrConditionEvaluations;

		nrMcdcPairs += m.nrMcdcPairs;

		nrAtomCombinations += m.nrAtomCombinations;
		nrAtomCombinationEvaluations += m.nrAtomCombinationEvaluations;

		nrInnerClasses += m.nrInnerClasses;
		nrConditionNodes += m.nrConditionNodes;

		if (cyclomaticComplexity < m.cyclomaticComplexity)
			cyclomaticComplexity = m.cyclomaticComplexity;
		if (essentialComplexity < m.essentialComplexity)
			essentialComplexity = m.essentialComplexity;
	}

	/**
	 * Berechnet die berdeckungsmae dieser Metrik.
	 */
	public void computeCoverageMetrics() {

		if (nrFunctions > 0)
			feec = 100.0f * (nrEnteredFunctions + nrReachedExits) / (nrFunctions + nrFunctionExits);

		if (nrStatements > 0) {
			c0 = 100.0f * nrCoveredStatements / nrStatements;
		}

		if (nrBranches > 0) {
			c1 = 100.0f * nrCoveredBranches / nrBranches;
		}

		if (nrAtoms > 0) {
			c2 = 100.0f * nrAtomEvaluations / (nrAtoms * 2);
			mmdc = 100.0f * nrConditionEvaluations / (nrConditions * 2);
			mcdc = 100.0f * nrMcdcPairs / nrAtoms;
			c3 = 100.f * nrAtomCombinationEvaluations / nrAtomCombinations;
		}

		if (nrModBIPaths > 0) {
			if (nrCoveredModBIPaths > nrModBIPaths)
				nrCoveredModBIPaths = nrModBIPaths;
			mbi = 100.0f * nrCoveredModBIPaths / nrModBIPaths;
			// Kappen, da nrMBIPaths nur Minimum ist (-> Exceptionpaths)?
			if (mbi > 100)
				mbi = 100;
		}

		if (nrBIPaths > 0) {
			if (nrCoveredBIPaths > nrBIPaths)
				nrCoveredBIPaths = nrBIPaths;
			bi = 100.0f * nrCoveredBIPaths / nrBIPaths;
			// Kappen, da nrBIPaths nur Minimum ist (-> Exceptionpaths)
			if (bi > 100)
				bi = 100;
		}

	}

	/**
	 * Liefert die Anzahl an mglichen ModBIPfaden, die durch Schleife oder den Hauptpfad der Funktion laufen.
	 * 
	 * @param id
	 *            die ID eines ASTIterationNodes oder eines ASTFunctionNodes
	 * @return die Anzahl an mglichen ModBIPfaden
	 */
	public long getNrModBIPaths(int id) {
		return mapOfModBIPaths.get(id);
	}

	/**
	 * Liefert die Anzahl an berdeckten ModBIPfaden, die durch Schleife oder den Hauptpfad der Funktion laufen.
	 * 
	 * @param id
	 *            die ID eines ASTIterationNodes oder eines ASTFunctionNodes
	 * @return die Anzahl an berdeckten ModBIPfaden
	 */
	public long getNrCoveredModBIPaths(int id) {
		if (mapOfCoveredModBIPaths.get(id) == null)
			return 0;
		return mapOfCoveredModBIPaths.get(id).size();
	}

	/**
	 * Setzt alle berdeckungsdaten wieder auf null.
	 */
	public void reset() {
		setOfCoveredBIPaths.clear();
		mapOfCoveredModBIPaths.clear();
		nrEnteredFunctions = 0;
		nrReachedExits = 0;
		nrCoveredStatements = 0;
		nrCoveredBranches = 0;
		nrAtomEvaluations = 0;
		nrConditionEvaluations = 0;
		nrMcdcPairs = 0;
		nrAtomCombinationEvaluations = 0;
		nrCoveredModBIPaths = 0;
		nrCoveredBIPaths = 0;
		computeCoverageMetrics();
	}
}
