package sota.ast.util;

import java.io.File;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.TreeSet;

import sota.Logger;
import sota.gui.graph.ZoomManager;
import sota.lang.LanguageSpecification;

/**
 * Container fr Projektdaten auf AST- und GUI-Ebene.
 */
public class ProjectConfig {

	// Projekt- und Quellenstatus
	public static final int EMPTY = 0;
	public static final int CONFIGURED = 1;
	public static final int PARSED = 2;
	public static final int INSTRUMENTED = 3;
	public static final int LOGREAD = 4;
	public static final int CLEAN = 0;
	public static final int DIRTY = 1;

	public int projectStatus = EMPTY;
	public int sourceStatus = CLEAN;
	public int parsedSourceStatus = CLEAN;

	// gelesene Sprachspezifikationen
	public Map<String, LanguageSpecification> setOfLangSpecs;

	// aktuelle Projektdaten
	public String name = "";
	public String language = "";
	public String prefix = "asc";
	public String backupExtension = "backup";
	public String projectDir = "";
	public String execDir = "";
	public String srcDir = "";
	public TreeSet<IScheme> setOfISchemes;
	public IScheme currInstrumentationScheme;
	public TreeSet<String> setOfFileNames;

	public String antLocation = "";
	public String antBuildFile = "";
	public String runScript = "";

	public Metrics metrics;

	public TreeSet<String> setOfSkippedFiles = null;

	// Daten fr Start-Test
	public boolean newTest = false;
	public String testName = "";
	public String testDesc = "";
	public String runArgs = "";

	public String iSchemeName = "";

	// Daten fr Logreading
	private HashMap<String, TestLog> mapOfTestLogs;
	private HashMap<String, File> mapOfBadTestFiles;
	private HashMap<String, String> mapOfBadTestFileProjects;
	private TreeSet<String> setOfLogFiles;

	// Darstellungsdaten
	public boolean showCoverage = true;
	public boolean rerunProject = false;
	public boolean buildProject = false;
	public boolean runProject = false;

	public TreeSet<TestCase> listOfSelectedTestCases = null;
	public ArrayList<String> listOfSelectedBadTestFiles = null;
	public TreeSet<TestLog> listOfSelectedTestLogs = null;

	private boolean pinSourceView = true;
	private boolean showPathNumbers = false;
	private ZoomManager zoomManager = new ZoomManager();
	private Object scopeNodeSelection = null;
	public String exceptionMessage = null;
	public String errorMessage = null;
	public String message = null;
	public boolean buildSuccessFlag = false;
	public String tempString = "";

	/**
	 * Der Konstruktor zur Initialisierung der Daten.
	 */
	public ProjectConfig() {
		projectStatus = EMPTY;
		setOfISchemes = new TreeSet<IScheme>();
		mapOfTestLogs = new HashMap<String, TestLog>();
		mapOfBadTestFiles = new HashMap<String, File>();
		mapOfBadTestFileProjects = new HashMap<String, String>();
		setOfLogFiles = new TreeSet<String>();
		setOfFileNames = new TreeSet<String>();
		setOfSkippedFiles = new TreeSet<String>();
		metrics = new Metrics();
		zoomManager = new ZoomManager();
		setOfLangSpecs = new HashMap<String, LanguageSpecification>();
	}

	/**
	 * Der Konstruktor zur Initalisierung eines Projekts.
	 * 
	 * @param name
	 *            der Name des Projekts
	 * @param language
	 *            die verwendete Sprache
	 * @param projectDir
	 *            das Projektverzeichnis
	 */
	public ProjectConfig(String name, String language, String projectDir) {
		this(name, language, projectDir, projectDir, projectDir);
	}

	/**
	 * Der Konstruktor zur Initalisierung eines Projekts.
	 * 
	 * @param name
	 *            der Name des Projekts
	 * @param language
	 *            die verwendete Sprache
	 * @param projectDir
	 *            das Projektverzeichnis
	 * @param execDir
	 *            das Ausfhrungsverzeicnis
	 */
	public ProjectConfig(String name, String language, String projectDir, String execDir) {
		this(name, language, projectDir, execDir, projectDir);
	}

	/**
	 * Der Konstruktor zur Initalisierung eines Projekts.
	 * 
	 * @param name
	 *            der Name des Projekts
	 * @param language
	 *            die verwendete Sprache
	 * @param projectDir
	 *            das Projektverzeichnis
	 * @param execDir
	 *            das Ausfhrungsverzeicnis
	 * @param srcDir
	 *            das Quellcodeverzeichnis
	 */
	public ProjectConfig(String name, String language, String projectDir, String execDir, String srcDir) {
		this.name = name;
		this.language = language;
		this.projectDir = projectDir;
		this.execDir = execDir;
		this.srcDir = srcDir;
		projectStatus = EMPTY;
		setOfISchemes = new TreeSet<IScheme>();
		mapOfTestLogs = new HashMap<String, TestLog>();
		mapOfBadTestFiles = new HashMap<String, File>();
		mapOfBadTestFileProjects = new HashMap<String, String>();
		setOfLogFiles = new TreeSet<String>();
		setOfFileNames = getSources(srcDir);
		setOfSkippedFiles = new TreeSet<String>();
		metrics = new Metrics();
		zoomManager = new ZoomManager();
		setOfLangSpecs = new HashMap<String, LanguageSpecification>();
		projectStatus = CONFIGURED;
	}

	/**
	 * Bestimmt, ob die Konfiguration vollstndig ist.
	 * 
	 * @return true, wenn vollstndig, sonst false
	 */
	public boolean isComplete() {
		return (name != "" && language != "" && backupExtension != "" && prefix != "" && (new File(execDir)).exists() && (new File(
				projectDir)).exists());
	}

	/**
	 * Bestimmt, ob die erste Wizardseite der Projekterstellung vollstndig ist.
	 * 
	 * @return true, wenn vollstndig, sonst false
	 */
	public boolean firstPageComplete() {
		return (name != "" && language != "" && backupExtension != "" && prefix != ""
				&& (new File(projectDir)).exists() && /*
													 * (new File( srcDir )).exists () &&
													 */(new File(execDir)).exists());
	}

	/**
	 * Bestimmt, ob die zweite Wizardseite der Projekterstellung vollstndig ist.
	 * 
	 * @return true, wenn vollstndig, sonst false
	 */
	public boolean secondPageComplete() {
		return (projectDir != "");
	}

	/**
	 * Setzt die Menge der Quellcodedateien des Projektes.
	 * 
	 * @param setOfFileNames
	 *            die Menge der Dateien
	 * 
	 */
	public void setSetOfFileNames(TreeSet<String> setOfFileNames) {
		this.setOfFileNames = setOfFileNames;
		this.projectStatus = CONFIGURED;
	}

	/**
	 * Liefert die Menge der ISchemes.
	 * 
	 * @return die ISchemes
	 */
	public TreeSet<IScheme> getListOfISchemes() {
		return setOfISchemes;
	}

	/**
	 * Setzt die Liste der ISchemes.
	 * 
	 * @param listOfISchemes
	 *            die Liste der ISchemes
	 */
	public void setListOfISchemes(TreeSet<IScheme> listOfISchemes) {
		this.setOfISchemes = listOfISchemes;
	}

	/**
	 * Fgt ein IScheme hinzu.
	 * 
	 * @param scheme
	 *            das IScheme
	 */
	public void addIScheme(IScheme scheme) {
		setOfISchemes.add(scheme);
	}

	/**
	 * Bestimmt das aktuell verwendete IScheme.
	 * 
	 * @param scheme
	 *            das IScheme
	 */
	public void setCurrentInstrumentationScheme(IScheme scheme) {
		currInstrumentationScheme = scheme;
	}

	/**
	 * Bestimmt das aktuell verwendete IScheme.
	 * 
	 * @param s
	 *            der IScheme-Name
	 * @return true, wenn existent, sonst falls
	 */
	public boolean setCurrentInstrumentationScheme(String s) {
		if (s == null || s.length() == 0) {
			currInstrumentationScheme = null;
			return true;
		}
		for (IScheme scheme : setOfISchemes)
			if (s.equals(scheme.getName())) {
				currInstrumentationScheme = scheme;
				return true;
			}
		return false;
	}

	/**
	 * Gibt ein IScheme zurck.
	 * 
	 * @param s
	 *            der IScheme-Name
	 * @return das IScheme oder null
	 */
	public IScheme getIScheme(String s) {
		if (s == null || s.length() == 0)
			return null;

		for (IScheme scheme : setOfISchemes)
			if (s.equals(scheme.getName()))
				return scheme;

		return null;
	}

	/**
	 * Liefert das aktuelle IScheme.
	 * 
	 * @return das IScheme
	 */
	public IScheme getCurrentIScheme() {
		return currInstrumentationScheme;
	}

	/**
	 * Liefert den Projektnamen.
	 * 
	 * @return der Name
	 */
	public String getName() {
		return name;
	}

	/**
	 * Legt die Daten fr den Test fest.
	 * 
	 * @param test
	 *            der Testname
	 * @param desc
	 *            die Beschreibung
	 * @param scheme
	 *            das Ischeme
	 * @param rerun
	 *            ob letzte Testkonfiguration nochmal verwendet werden soll
	 * @param build
	 *            ob das Projekt kompiliert werden soll
	 * @param run
	 *            ob das Projekt gestartet werden soll
	 */
	public void setNewTest(String test, String desc, String scheme, boolean rerun, boolean build, boolean run, String args) {
		testName = test;
		testDesc = desc;
		iSchemeName = scheme;
		newTest = true;
		rerunProject = rerun;
		buildProject = build;
		runProject = run;
		runArgs = args;
	}

	/**
	 * Liefert die Meng an TestLogs.
	 * 
	 * @return die TestLogs
	 */
	public TreeSet<TestLog> getListOfTestLogs() {
		return new TreeSet<TestLog>(mapOfTestLogs.values());
	}

	/**
	 * Liefert die Menge an invaliden Testdateien
	 * 
	 * @return die Dateinamen
	 */
	public ArrayList<String> getListOfBadTestFiles() {
		return new ArrayList<String>(mapOfBadTestFiles.keySet());
	}

	/**
	 * Liefert die Abbildung einer invaliden Logdatei auf ihr (zu diesem inkompatiblen) Projektnamen.
	 * 
	 * @return die Map
	 */
	public HashMap<String, String> getMapOfBadTestFileProjects() {
		return mapOfBadTestFileProjects;
	}

	/**
	 * Fgt eine Testdatei hinzu.
	 * 
	 * @param file
	 *            die Datei
	 */
	public void addTestFile(String file) {
		setOfLogFiles.add(file);
	}

	/**
	 * Liefert die Liste an Testdateien.
	 * 
	 * @return die Testdateien
	 */
	public TreeSet<String> getListOfTestFiles() {
		return setOfLogFiles;
	}

	/**
	 * Gibt an,ob die berdeckung farblich hervorgehoben werden soll.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean showCoverage() {
		return (showCoverage && mapOfTestLogs != null && mapOfTestLogs.size() > 0);
	}

	/**
	 * Legt fest, ob die berdeckung farblich hervorgehoben werden soll.
	 * 
	 * @param showCoverage
	 *            true, wenn ja, sonst false
	 */
	public void showCoverage(boolean showCoverage) {
		this.showCoverage = showCoverage;
	}

	/**
	 * ndert die Entscheidung, ob die berdeckung farblich hervorgehoben werden soll.
	 */
	public void invertCoverage() {
		showCoverage = !showCoverage;
	}

	/**
	 * @param name
	 * @param file
	 */
	public void addBadTestFile(String name, File file) {
		mapOfBadTestFiles.put(name, file);
		if (tempString != "")
			mapOfBadTestFileProjects.put(name, tempString);
	}

	/**
	 * Fgt eine invalides Testdatei hinzu.
	 * 
	 * @param f
	 *            die Datei
	 */
	public void removeBadTestFile(String f) {
		if (this.mapOfBadTestFiles.containsKey(f))
			this.mapOfBadTestFiles.remove(f);

		if (this.mapOfBadTestFileProjects.containsKey(f))
			this.mapOfBadTestFileProjects.remove(f);
	}

	/**
	 * Fgt ein TestLog hinzu.
	 * 
	 * @param testLog
	 *            das TestLog
	 */
	public void addTestLog(TestLog testLog) {
		if (testLog == null)
			return;
		if (this.mapOfTestLogs.containsKey(testLog.getName()))
			this.mapOfTestLogs.remove(testLog.getName());
		this.mapOfTestLogs.put(testLog.getName(), testLog);
		testLog.computeInvokationMap();
		for (TestCase testCase : testLog.getListOfTestCases())
			testCase.computeInvokationMap();

	}

	/**
	 * Entfernt ein Testlog.
	 * 
	 * @param t
	 *            das TestLog
	 */
	public void removeTestLog(TestLog t) {
		if (this.mapOfTestLogs.containsKey(t.getName()))
			this.mapOfTestLogs.remove(t.getName());
	}

	/**
	 * Liefert die Projektmetrik.
	 * 
	 * @return die Metrik des Projekts
	 */
	public Metrics getMetrics() {
		return metrics;
	}

	/**
	 * Setzt die Projektmetrik
	 * 
	 * @param metrics
	 *            die neue Metrik des Projekts
	 */
	public void setMetrics(Metrics metrics) {
		this.metrics = metrics;
	}

	/**
	 * Invertiert die Entscheidung, ob der SourceViewer in der CFGView gepinnt werden soll.
	 */
	public void invertPinSourceView() {
		pinSourceView = !pinSourceView;
	}

	/**
	 * Liefert die Entscheidung, ob der SourceViewer in der CFGView gepinnt werden soll.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean getPinSourceView() {
		return pinSourceView;
	}

	/**
	 * Legt die Entscheidung fest, ob der SourceViewer in der CFGView gepinnt werden soll.
	 * 
	 * @param checked
	 *            true, wenn ja, sonst false
	 */
	public void setPinSourceView(boolean checked) {
		pinSourceView = checked;

	}

	/**
	 * Legt fest, ob die Anzahl der Kantendurchlufe in der CFGView angezeigt werden soll.
	 * 
	 * @param checked
	 *            true fr ja, aonst false
	 */
	public void showPathNumbers(boolean checked) {
		showPathNumbers = checked;
	}

	/**
	 * Liefert die Entscheidung, ob die Anzahl der Kantendurchlufe in der CFGView angezeigt werden soll.
	 * 
	 * @return true, falls ja, sonst false
	 */
	public boolean showPathNumbers() {
		return showPathNumbers;
	}

	/**
	 * Liefert den aktuellen ZoomManager.
	 * 
	 * @return der ZoomManager
	 */
	public ZoomManager getZoomManager() {
		return zoomManager;
	}

	/**
	 * Legt den ausgewhlen ScopeNode fest.
	 * 
	 * @param scopeNodeSelection
	 *            die Auswahl des ScopeNodes
	 */
	public void setScopeNodeSelection(Object scopeNodeSelection) {
		this.scopeNodeSelection = scopeNodeSelection;
	}

	/**
	 * Liefert die Auswahl des ScopeNodes.
	 * 
	 * @return der ausgewhlte ScopeNode
	 */
	public Object getScopeNodeSelection() {
		return scopeNodeSelection;
	}

	/**
	 * Liefert die Sprachspezifikation fr eine Sprache.
	 * 
	 * @param s
	 *            die Sprache
	 * @return die Sprachspezifikation
	 */
	public LanguageSpecification getLangSpecification(String s) {
		return setOfLangSpecs.get(s);
	}

	/**
	 * Liefert die Menge der Sprachspezifikationen.
	 * 
	 * @return die Sprachspezifikationen
	 */
	public Collection<LanguageSpecification> getLangSpecs() {
		return setOfLangSpecs.values();
	}

	/**
	 * Liefert rekursiv die Menge aller Quelldateien fr ein Verzeichnis.
	 * 
	 * @param s
	 *            das Verzeichnis
	 * @return die Menge der Quelldateien
	 */
	public static TreeSet<String> getSources(String s) {

		TreeSet<String> set = new TreeSet<String>();

		File path = new File(s);
		if (path.isFile()) {
			set.add(path.getAbsolutePath());
			return set;
		}

		if (path.isDirectory()) {
			for (File f : path.listFiles()) {
				if (f.isFile() && f.getName().endsWith(".java"))
					set.add(f.getAbsolutePath());
				else if (f.isDirectory())
					set.addAll(getSources(f.getAbsolutePath()));
			}
		}
		return set;
	}

	/**
	 * Ausgabefunktion der Daten fr die Kommandozeile.
	 */
	public void print() {
		Logger.logln("Name: " + name);
		Logger.logln("Language: " + language);
		Logger.logln("newTest: " + (newTest ? "X" : "-"));
		Logger.logln("buildProject: " + (buildProject ? "X" : "-"));
		Logger.logln("runProject: " + (runProject ? "X" : "-"));
		Logger.logln("rerunProject: " + (rerunProject ? "X" : "-"));
		Logger.logln("ProjectStatus: " + projectStatus);
		Logger.logln("SourceStatus: " + sourceStatus);
		Logger.logln("ProjectDir: " + projectDir);
		Logger.logln("ExecDir: " + execDir);
		Logger.logln("BackUpExtension: " + backupExtension);
		Logger.logln("ASCPrefix: " + prefix);
		Logger.logln("InstrumentationSchemes:");
		for (IScheme scheme : setOfISchemes)
			Logger.logln("\t" + scheme.getName());

		Logger.log("Files:");
		for (String file : setOfFileNames)
			Logger.logln("\t" + file);

		for (LanguageSpecification s : setOfLangSpecs.values()) {
			Logger.logln(s.name);
			Logger.logln(s.astBuilder);
		}
	}

}