package sota.ast.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.TreeSet;

import sota.Logger;

/**
 * Enthlt alle Informationen zu einem Testfall. Das umfasst die Pfadinformationen (Pfadlisten), die aus der Logdatei
 * gelesen wurden, und die daraus aufbereiteten Information (Hitlisten, Maps, etc.) fr die schnelle Applikation der
 * Daten auf den AST.
 * 
 */
public class TestCase implements ITestContainer, Comparable<TestCase> {

	private String projectName = "";
	private String name = "";
	private String description = "";
	private String iSchemeName = "";
	private String fileName = "";

	private int number;

	private boolean isCompleted = false;

	// Liste von Pfadlisten fr alle Testdurchlufe durch LogReader erstellt
	private ArrayList<ArrayList<Integer>> listOfPaths;

	// aus Pfadlisten durch BIPathManager bzw. ModBIPathManager erstellt
	private HashMap<Integer, HashSet<Long>> mapOfCoveredBIPaths;
	private HashMap<Integer, HashSet<Long>> mapOfCoveredModBIPaths;

	// durch TestLogManager erstellte berdeckungslisten

	// Hitmap aller Anweisungsknoten : id -> nrHits
	private HashMap<Integer, Integer> hitMap;
	// IterationNode.id -> Anzahl der Schleifenauslassungen
	private HashMap<Integer, Integer> iterationSkipsMap;
	// IterationNode.id -> Anzahl der einmaligen Schleifenausfhrungen
	private HashMap<Integer, Integer> iterationSingleLoopsMap;
	// IterationNode.id -> Anzahl der mehrmaligen Schleifenausfhrungen
	private HashMap<Integer, Integer> iterationDoubleLoopsMap;
	// CaseNode.id -> Anzahl der Auswahlen durch Switch
	private HashMap<Integer, Integer> selectionMap;
	// TryNode.id -> Anzahl der Auswertung ohne Ausnahme
	private HashMap<Integer, Integer> tryWithoutCatchMap;
	// IConditionNode -> Wahrheitsvektoren der atomaren Auswertung
	private HashMap<Integer, TreeSet<String>> tvMap;
	// FunctionNode.id -> Anzahl der Funktionsaufrufe nach Level 1, 2, 3
	HashMap<Integer, int[]> invokationMap = new HashMap<Integer, int[]>();

	/**
	 * Konstruktor, der den Testfall erstellt.
	 * 
	 * @param number
	 *            die Nummer des Testfalls im Testlog
	 * @param fileName
	 *            der Dateiname
	 * @param projectName
	 *            der Projektname
	 * @param testName
	 *            der Testname
	 * @param description
	 *            die Beschreibung
	 * @param iScheme
	 *            das IScheme
	 */
	public TestCase(int number, String fileName, String projectName, String testName, String description, String iScheme) {
		this.number = number;
		this.fileName = fileName;
		this.projectName = projectName;
		this.description = description;
		listOfPaths = new ArrayList<ArrayList<Integer>>();
		mapOfCoveredBIPaths = new HashMap<Integer, HashSet<Long>>();
		mapOfCoveredModBIPaths = new HashMap<Integer, HashSet<Long>>();
		tvMap = new HashMap<Integer, TreeSet<String>>();
		hitMap = new HashMap<Integer, Integer>();
		tryWithoutCatchMap = new HashMap<Integer, Integer>();
		iterationSkipsMap = new HashMap<Integer, Integer>();
		iterationSingleLoopsMap = new HashMap<Integer, Integer>();
		iterationDoubleLoopsMap = new HashMap<Integer, Integer>();
		selectionMap = new HashMap<Integer, Integer>();
		name = testName;
		this.iSchemeName = iScheme;
	}

	/**
	 * Liefert den Dateinamen.
	 * 
	 * @return der Dateiname
	 */
	public String getFileName() {
		return fileName;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.util.ITestContainer#getDescription()
	 */
	public String getDescription() {
		return description;
	}

	/**
	 * Liefert die Beschreibung
	 * 
	 * @param description
	 *            die Beschreibung
	 */
	public void setDescription(String description) {
		this.description = description;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.util.ITestContainer#getISchemeName()
	 */
	public String getISchemeName() {
		return iSchemeName;
	}

	/**
	 * Legt das IScheme fest.
	 * 
	 * @param iScheme
	 *            das IScheme
	 */
	public void setISchemeName(String iScheme) {
		this.iSchemeName = iScheme;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.util.ITestContainer#getName()
	 */
	public String getName() {
		return name;
	}

	/**
	 * Liefert den Projektnamen
	 * 
	 * @return der Projektname
	 */
	public String getProjectName() {
		return projectName;
	}

	/**
	 * fgt einen berdeckten BIPfad hinzu
	 * 
	 * @param functionID
	 *            die Funktion
	 * @param biPath
	 *            der Pfad
	 */
	public void addCoveredBIPath(int functionID, long biPath) {
		if (mapOfCoveredBIPaths.keySet().contains(functionID))
			mapOfCoveredBIPaths.get(functionID).add(biPath);
		else {
			HashSet<Long> set = new HashSet<Long>();
			set.add(biPath);
			mapOfCoveredBIPaths.put(functionID, set);
		}
	}

	/**
	 * Liefert die Tabelle der berdeckten BIPfade.
	 * 
	 * @return die Tabelle
	 */
	public HashMap<Integer, HashSet<Long>> getMapOfCoveredBIPaths() {
		return mapOfCoveredBIPaths;
	}

	/**
	 * Fgt einen ModBI-Pfad fr einen Knoten hinzu (Iteration oder Funkion)
	 * 
	 * @param id
	 *            die ID des Knotens
	 * @param modBIPath
	 *            der Pfad
	 */
	public void addCoveredModBIPath(int id, long modBIPath) {
		if (mapOfCoveredModBIPaths.keySet().contains(id))
			mapOfCoveredModBIPaths.get(id).add(modBIPath);
		else {
			HashSet<Long> set = new HashSet<Long>();
			set.add(modBIPath);
			mapOfCoveredModBIPaths.put(id, set);
		}
	}

	/**
	 * Liefert die Tabelle der berdeckten ModBIPfade.
	 * 
	 * @return die Tabelle
	 */
	public HashMap<Integer, HashSet<Long>> getMapOfCoveredModBIPaths() {
		return mapOfCoveredModBIPaths;
	}

	/**
	 * Fgt eine Pfadliste hinzu.
	 * 
	 * @param pathList
	 *            die Pfadliste
	 */
	public void addListOfPaths(ArrayList<ArrayList<Integer>> pathList) {
		for (ArrayList<Integer> l : pathList) {
			if (l.size() >= 2)
				listOfPaths.add(l);
		}
	}

	/**
	 * Lieferte die Pfadliste.
	 * 
	 * @return die Pfadliste
	 * 
	 */
	public ArrayList<ArrayList<Integer>> getListOfPaths() {
		return listOfPaths;
	}

	/**
	 * Legt die Tabelle der KnotenTreffer fest.
	 * 
	 * @param hitMap
	 *            die Tabelle
	 */
	public void setHitMap(HashMap<Integer, Integer> hitMap) {
		this.hitMap = hitMap;
	}

	/**
	 * Fgt eine neue Tabelle mit Knotentreffer hinzu.
	 * 
	 * @param newHitMap
	 *            die neue Tabelle
	 */
	public void addHitMap(HashMap<Integer, Integer> newHitMap) {
		// Logger.log("adding Hits");
		for (Integer id : newHitMap.keySet())
			if (hitMap.containsKey(id))
				hitMap.put(id, this.hitMap.remove(id) + newHitMap.get(id));
			else
				hitMap.put(id, newHitMap.get(id));
		// Logger.log("finished: adding Hits");
	}

	/**
	 * Liefert die Tabelle mit Knotentreffern.
	 * 
	 * @return die Tabelle
	 */
	public HashMap<Integer, Integer> getHitMap() {
		return hitMap;
	}

	/**
	 * Ausgabefunktion fr Kommandozeile.
	 */
	public void print() {
		Logger.log("Projectname: " + projectName + ", Testname: " + name + ", IScheme : " + iSchemeName + "\n");
	}

	/**
	 * Ausgabefunktion der Treffertabelle.
	 */
	public void printHitMap() {
		Logger.log("HitMap:");
		for (Integer i : hitMap.keySet()) {
			System.out.print(i + " " + hitMap.get(i) + ", ");
		}
		Logger.log("");
		Logger.log("SelectionMap:");
		for (Integer i : selectionMap.keySet()) {
			System.out.print(i + " " + selectionMap.get(i) + ", ");
		}
		Logger.log("");
	}

	/**
	 * Ausgabefunktion fr Pfade.
	 * 
	 * @param l
	 *            die Pfadliste
	 */
	public static void printPath(ArrayList<Integer> l) {
		Logger.log("Path:");
		for (Integer i : l) {
			Logger.log(i + " ");
		}
		Logger.logln("");
	}

	/**
	 * Liefert die Tabelle der Schleifenauslassungen.
	 * 
	 * @return die Tabelle
	 */
	public HashMap<Integer, Integer> getIterationSkipsMap() {
		return iterationSkipsMap;
	}

	/**
	 * Fgt eine Tabelle mit Schleifenauslassungen hinzu.
	 * 
	 * @param map
	 *            die Tabelle
	 */
	public void addIterationSkipsMap(HashMap<Integer, Integer> map) {
		// Logger.log("adding Hits");
		for (Integer id : map.keySet())
			if (iterationSkipsMap.containsKey(id)) {
				iterationSkipsMap.put(id, this.hitMap.remove(id) + map.get(id));
			} else
				iterationSkipsMap.put(id, map.get(id));
		// Logger.log("finished: adding Hits");

	}

	/**
	 * Liefert die Tabelle der einmaligen Schleifendurchlufen.
	 * 
	 * @return die Tabelle
	 */
	public HashMap<Integer, Integer> getIterationSingleLoopsMap() {
		return iterationSingleLoopsMap;
	}

	/**
	 * Fgt eine Tabelle mit einmaligen Schleifendurchlufen hinzu.
	 * 
	 * @param iterationLoopsMap
	 *            die Tabelle
	 */
	public void addIterationLoopsMap(HashMap<Integer, Integer> iterationLoopsMap) {
		this.iterationSingleLoopsMap = iterationLoopsMap;
	}

	/**
	 * Liefert die Tabelle der mehrmaligen Schleifendurchlufen.
	 * 
	 * @return die Tabelle
	 */
	public HashMap<Integer, Integer> getIterationDoubleLoopsMap() {
		return iterationDoubleLoopsMap;
	}

	/**
	 * Fgt eine Tabelle mit mehrmaligen Schleifendurchlufen hinzu.
	 * 
	 * @param iterationDoubleLoopsMap
	 *            die Tabelle
	 */
	public void setIterationDoubleLoopsMap(HashMap<Integer, Integer> iterationDoubleLoopsMap) {
		this.iterationDoubleLoopsMap = iterationDoubleLoopsMap;
	}

	/**
	 * Fgt den Wahrheitsvektor einer bedingungsauswertung hinzu.
	 * 
	 * @param id
	 *            der Bedingungsknoten
	 * @param tv
	 *            der Wahrheitsvektor in Stringdarstellung
	 */
	public void addTruthVector(int id, String tv) {
		if (!tvMap.keySet().contains(id))
			tvMap.put(id, new TreeSet<String>());
		tvMap.get(id).add(tv);
	}

	/**
	 * Liefert die Tabelle der Wahrheitsvektoren.
	 * 
	 * @return die Tabelle
	 */
	public HashMap<Integer, TreeSet<String>> getTvMap() {
		return tvMap;
	}

	/**
	 * Liefert die Tabelle der Selektionsberdeckung (fr Switch).
	 * 
	 * @return die Tabelle.
	 */
	public HashMap<Integer, Integer> getSelectionMap() {
		return selectionMap;
	}

	/**
	 * Liefert die Anzahl der Selektionsberdeckungen fr einen Case.
	 * 
	 * @param id
	 *            die ID des CaseNodes
	 * @return die Anzahl der Auswertungen
	 */
	public int getSelection(int id) {
		if (selectionMap.containsKey(id))
			return selectionMap.get(id);
		return 0;
	}

	/**
	 * Fgt eine Selektionsauswertung hinzu.
	 * 
	 * @param id
	 *            die ID des CaseNodes
	 * @param nrSelects
	 *            die Anzahl der Auswertungen
	 */
	public void addSelection(int id, int nrSelects) {
		if (selectionMap.keySet().contains(id))
			selectionMap.put(id, selectionMap.remove(id) + nrSelects);
		else
			selectionMap.put(id, nrSelects);
	}

	/**
	 * Liefert die Tabelle mit den Ausnahmeauswertungen ohne Ausnahmevorkommen.
	 * 
	 * @return die Tabelle
	 */
	public HashMap<Integer, Integer> getTryWithoutExceptionMap() {
		return tryWithoutCatchMap;
	}

	/**
	 * Fgt eine Ausnahmeauswertungen ohne Ausnahmevorkommen hinzu.
	 * 
	 * @param id
	 *            die ID des TryNodes
	 * @param nr
	 *            die Anzahl der Auswertungen ohne Ausnahmevorkommen
	 */
	public void addTryWithoutException(int id, int nr) {
		if (tryWithoutCatchMap.keySet().contains(id))
			tryWithoutCatchMap.put(id, tryWithoutCatchMap.remove(id) + nr);
		else
			tryWithoutCatchMap.put(id, nr);
	}

	/**
	 * Fgt Schleifenauslassungen hinzu.
	 * 
	 * @param id
	 *            die ID des IterationsNodes
	 * @param nrSkips
	 *            die Anzahl
	 */
	public void addIterationSkip(int id, int nrSkips) {
		if (iterationSkipsMap.keySet().contains(id))
			iterationSkipsMap.put(id, iterationSkipsMap.remove(id) + nrSkips);
		else
			iterationSkipsMap.put(id, nrSkips);

	}

	/**
	 * Fgt einmalige Schleifenausfhrungen hinzu.
	 * 
	 * @param id
	 *            die ID des IterationsNodes
	 * @param nrSingleLoops
	 *            die Anzahl
	 */
	public void addIterationSingleLoop(int id, int nrSingleLoops) {
		if (iterationSingleLoopsMap.keySet().contains(id))
			iterationSingleLoopsMap.put(id, iterationSingleLoopsMap.remove(id) + nrSingleLoops);
		else
			iterationSingleLoopsMap.put(id, nrSingleLoops);

	}

	/**
	 * Fgt eine mehrmalige Schleifenausfhrung hinzu.
	 * 
	 * @param id
	 *            die ID des IterationsNodes
	 * @param nrDoubleLoops
	 *            die Anzahl
	 */
	public void addIterationDoubleLoop(int id, int nrDoubleLoops) {
		if (iterationDoubleLoopsMap.keySet().contains(id))
			iterationDoubleLoopsMap.put(id, iterationDoubleLoopsMap.remove(id) + nrDoubleLoops);
		else
			iterationDoubleLoopsMap.put(id, nrDoubleLoops);

	}

	/**
	 * Berechnet aus den Pfadlisten die Anzahl der Funktionsaufrfe nach Level in dem Test.
	 */
	public void computeInvokationMap() {
		invokationMap = new HashMap<Integer, int[]>();
		for (ArrayList<Integer> l : listOfPaths)
			if (l.size() > 1) {
				if (invokationMap.containsKey(l.get(1))) {
					int[] i = invokationMap.remove(l.get(1));
					i[0]++;
					i[l.get(0)]++;
					invokationMap.put(l.get(1), i);
				} else {
					int[] i = new int[4];
					i[0] = 1;
					i[l.get(0)] = 1;
					invokationMap.put(l.get(1), i);
				}
			}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.util.ITestContainer#getInvokationMap()
	 */
	public HashMap<Integer, int[]> getInvokationMap() {
		return invokationMap;
	}

	/**
	 * Liefert die Stringdarstellung eines Pfades.
	 * 
	 * @param list
	 *            der Pfad
	 * @return die Stringdarstellung
	 */
	public static String getPathString(ArrayList<Integer> list) {
		String s = "Path: ";
		for (Integer i : list) {
			s += i + " ";
		}
		return s + "\n";
	}

	/**
	 * Liefert die Nummer innerhalb des TestLogs.
	 * 
	 * @return die Nummer
	 */
	public int getNumber() {
		return number;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	@Override
	public int compareTo(TestCase arg0) {

		int c = name.compareTo(arg0.getName());
		if (c != 0)
			return c;
		return number - arg0.getNumber();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.ast.util.ITestContainer#getNrOfPaths()
	 */
	@Override
	public int getNrOfPaths() {
		return listOfPaths.size();
	}

	/**
	 * Setzt das TestCase auf vervollstndigt.
	 * 
	 * @param b
	 *            true, wenn vollstndig, sonst false
	 */
	public void setCompleted(boolean b) {
		isCompleted = b;
	}

	/**
	 * Liefert die Aussage, ob das TestCase vollstndig ist.
	 * 
	 * @return true, wenn ja, sonst false
	 */
	public boolean isCompleted() {
		return isCompleted;
	}

	/**
	 * Liefert die Stringdarstellung von BI-Pfaden.
	 * 
	 * @param hashSet
	 *            die Menge an codierten Pfaden
	 * @return der String
	 */
	public static String getPathString(HashSet<Long> hashSet) {

		String result = "";
		// System.out.println("\npath: " + hashSet);

		for (Long p : hashSet) {
			int z = 1;
			int c = 0;
			String s = "";
			for (int j = 0; j < 64; j++) {
				if ((c % 8) == 0)
					s += " ";
				s += (p & z) == 0 ? "0" : "1";
				z = z << 1;
				c++;

			}
			// System.out.print(s);
			s = s.substring(0, s.lastIndexOf('1') + 1);
			// System.out.println( " -> " + s);*/
			if (s.length() == 0)
				s = "0";
			result += s + ", ";
			// System.out.println("result: " + result);
		}
		return result;
	}

	/**
	 * Summiert die berdeckungsdaten zweier Testflle auf.
	 * 
	 * @param tc
	 *            der zweite Testfall
	 */
	public void sumUp(TestCase tc) {
		for (Integer i : tc.getHitMap().keySet()) {
			if (hitMap.containsKey(i))
				hitMap.put(i, hitMap.remove(i) + tc.getHitMap().get(i));
			else
				hitMap.put(i, tc.getHitMap().get(i));
		}

		for (Integer i : tc.getTvMap().keySet()) {
			if (tvMap.containsKey(i))
				tvMap.get(i).addAll(tc.getTvMap().get(i));
			else
				tvMap.put(i, tc.getTvMap().get(i));
		}

		for (Integer i : tc.getMapOfCoveredModBIPaths().keySet()) {
			if (!mapOfCoveredModBIPaths.containsKey(i))
				mapOfCoveredModBIPaths.put(i, new HashSet<Long>());
			mapOfCoveredModBIPaths.get(i).addAll(tc.getMapOfCoveredModBIPaths().get(i));
		}

		for (Integer i : tc.getMapOfCoveredBIPaths().keySet()) {
			if (!mapOfCoveredBIPaths.containsKey(i))
				mapOfCoveredBIPaths.put(i, new HashSet<Long>());
			mapOfCoveredBIPaths.get(i).addAll(tc.getMapOfCoveredBIPaths().get(i));
		}

		for (Integer i : tc.getSelectionMap().keySet()) {
			if (selectionMap.containsKey(i))
				selectionMap.put(i, selectionMap.remove(i) + tc.getSelectionMap().get(i));
			else
				selectionMap.put(i, tc.getSelectionMap().get(i));
		}

		for (Integer i : tc.getIterationSkipsMap().keySet()) {
			if (iterationSkipsMap.containsKey(i))
				iterationSkipsMap.put(i, iterationSkipsMap.remove(i) + tc.getIterationSkipsMap().get(i));
			else
				iterationSkipsMap.put(i, tc.getIterationSkipsMap().get(i));
		}

		for (Integer i : tc.getIterationSingleLoopsMap().keySet()) {
			if (iterationSingleLoopsMap.containsKey(i))
				iterationSingleLoopsMap.put(i, iterationSingleLoopsMap.remove(i)
						+ tc.getIterationSingleLoopsMap().get(i));
			else
				iterationSingleLoopsMap.put(i, tc.getIterationSingleLoopsMap().get(i));
		}

		for (Integer i : tc.getIterationDoubleLoopsMap().keySet()) {
			if (iterationDoubleLoopsMap.containsKey(i))
				iterationDoubleLoopsMap.put(i, iterationDoubleLoopsMap.remove(i)
						+ tc.getIterationDoubleLoopsMap().get(i));
			else
				iterationDoubleLoopsMap.put(i, tc.getIterationDoubleLoopsMap().get(i));
		}
	}
}
