package sota.ast.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Stack;

import sota.Logger;
import sota.ast.nodes.ASTBranchNode;
import sota.ast.nodes.ASTCaseNode;
import sota.ast.nodes.ASTCatchNode;
import sota.ast.nodes.ASTDoNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.ASTJumpNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.ASTStatementNode;
import sota.ast.nodes.ASTSwitchNode;
import sota.ast.nodes.ASTTryNode;
import sota.ast.nodes.FileNode;
import sota.ast.nodes.IConditionNode;
import sota.ast.nodes.IStatementNode;

/**
 * Klasse mit statischen Methoden zur Berechnung und Verarbeitung von HitsLists. Ausgangspunkt sind die durch den
 * LogReader gelesenen TestCases. Die dort gespeicherten pathLists werden hier aufbereitet zu hitlists, die eine
 * schnelle Applikation auf den AST ermglichen.
 */
public class TestLogManager {

	protected static Stack<ASTIterationNode> stackOfIterationNodes;
	protected static Stack<Integer> stackOfLoopCounters;

	protected static final boolean VERBOSE = false;

	/**
	 * Wandelt die Pfadlisten der eingelesenen TestLogs in HitLists um, d.h. macht fr jedes TestLog: 1. durchlaufe alle
	 * Pfade und erhhe Hits, berechne Verzweigungsinfos 2. erstelle aus diesen Infos die HitList fr das TestLog und
	 * merke diese 3. lsche Hits und fahre bei nchstem TestLog fort
	 * 
	 * @param ast
	 *            der aktuelle ASTManager
	 */
	public static void completeTestLogs(ASTManager ast) {

		// fr alle Testflle
		for (TestLog testLog : ast.getConfiguration().getListOfTestLogs()) {

			Inner: for (TestCase testCase : testLog.getListOfTestCases()) {

				if (testCase.isCompleted())
					continue Inner;

				// Verteilung lschen
				resetHits(ast);

				// appliziere Pfadlisten auf AST und bestimme Verzweigungen
				for (ArrayList<Integer> l : testCase.getListOfPaths()) {
					if (l.size() >= 2) {
						// je nach Instrumentierungslevel Pfade auftragen
						if (l.get(0) != 3)
							applyEssentialTestLogPath(ast, l);
						else
							applyFullTestLogPath(ast, l);
					}
				}
				// berdeckungs- und Verzweigungsinformationen der Knoten
				// einsammeln
				collectHitList(ast, testCase);
				testCase.setCompleted(true);

			}
		}

		// Verteilung lschen
		resetHits(ast);

		BIPathManager.createCoveredBIPathsSet(ast);
		ModBIPathManager.createCoveredModBIPathsSet(ast);

		if (ast.getConfiguration().getListOfTestLogs().size() > 0)
			ast.getConfiguration().projectStatus = ProjectConfig.LOGREAD;

	}

	/**
	 * Berechnet die HitLists und Verzweigungen fr Funktionspfade nach Instrumentierung Level 3. Alle
	 * Anweisungensknoten sind im LogPath enthalten, diese mssen lediglich aufaddiert und dann lediglich die
	 * Verzweigungsinfo berechnet werden fr Switch, Try, Iteration. Zu merken: aktuelles Iterationskonstrukt fr Anzahl
	 * der Loops
	 * 
	 * @param ast
	 *            der aktuelle ASTManager
	 * @param list
	 *            die zu applizierende Pfadliste
	 */
	protected static void applyFullTestLogPath(ASTManager ast, ArrayList<Integer> list) {
		stackOfIterationNodes = new Stack<ASTIterationNode>();
		stackOfLoopCounters = new Stack<Integer>();

		// n hlt den Knoten beim Durchlauf durch die Funktion
		ASTNode n = ast.getNodeByID(list.get(1));

		// finde Funktion und erhhe Zhler
		ASTFunctionNode functionNode = n.getFunctionNode();
		functionNode.incNrHits();

		if (VERBOSE)
			Logger.logln("FullPath: " + TestCase.getPathString(list));

		// durchlaufe PathList
		for (int i = 1; i < list.size(); i++) {
			n = ast.getNodeByID(list.get(i));

			// appliziere Hits fr alle Statements
			if (n instanceof ASTStatementNode) {
				// System.out.println(n.getID() + " incs");
				((ASTStatementNode) n).incNrHits();
			}
			// fr Switch entscheiden, zu welchem Case gesprungen wird
			if (n instanceof ASTSwitchNode) {
				ASTNode temp = ast.getNodeByID(list.get(i + 1));
				if (temp instanceof ASTCaseNode) {
					((ASTCaseNode) temp).incNrSelects();
				}
				// IterationNode markiert Eintritt in Iterationkonstrukt, es knnen beliebig viele
				// Schleifendurchgnge folgen (siehe unten). Knoten und Loopzhler kommmen auf Stack.
			} else if (n instanceof ASTIterationNode) {
				ASTIterationNode in = ((ASTIterationNode) n);
				if (!stackOfIterationNodes.contains(in)) {
					stackOfLoopCounters.push(0);
					stackOfIterationNodes.push(in);
				}
			}
			// fr Try entscheide, ob eine Ausnahme geworfen wurde oder nicht
			if (n instanceof ASTTryNode) {
				ASTTryNode tn = (ASTTryNode) n;
				int j = i;
				T: while (true) {
					if (j >= list.size()) {
						tn.incNrHitsWithoutException();
						break T;
					}
					// falls es finally ist, dann erhhe Zhler und brich ab
					if (tn.hasFinally() && tn.getFinallyNode().getID() == list.get(j)) {
						tn.incNrHitsWithoutException();
						break T;
					}
					// falls endNode vorgekommt, dann erhhe Zhler und
					// brich ab
					if (tn.getEndNode().getID() <= list.get(j)) {
						tn.incNrHitsWithoutException();
						break T;
					}
					// falls es ein catch ist, dann brich ab
					for (ASTCatchNode cn : tn.getListOfCatches()) {
						if (cn.getID() == list.get(j)) {
							break T;
						}
					}
					j++;
				}
				// fr Iterationkonstrukte: entscheide den Loop-Verlauf
			} else
				while (!stackOfIterationNodes.isEmpty()) {
					// wenn LoopNode gefunden, dann erhhe LoopCounter
					if (stackOfIterationNodes.peek().getLoopNode() == n) {
						int c = stackOfLoopCounters.pop();
						stackOfLoopCounters.push(c + 1);
						break;
						// false ExitNode gefunden oder Iteration verlassen, dann setze Loop-Verlauf
						// und wiederhole Schleife
					} else if (stackOfIterationNodes.peek().getSetOfExitNodes().contains(n)
							|| !stackOfIterationNodes.peek().isWithinBranchConstruct(n.getID())) {
						int c = stackOfLoopCounters.pop();
						ASTIterationNode in = stackOfIterationNodes.pop();
						if (c == 0)
							in.incNrSkips();
						else if (c == 1)
							in.incNrSingleLoops();
						else
							in.incNrMultipleLoops();
						// wenn nix von beidem, dann Schleife abbrechen
					} else
						break;

				}
		}
	}

	/**
	 * Berechnet die HitLists und Verzweigungen fr Funktionspfade nach Instrumentierung Level 1 und 2. In EP stehen
	 * nur: case, catch, true/false und loop. Beginnend mit Funktionseingang propagiere Hits solange, bis der BranchNode
	 * des nchsten Pfadknotens erreicht wird, dann folge der Entscheidung bis zum nchsten. Beim Propagieren zu
	 * beachten: - wenn switch getroffen wird, dann wurde dieses bersprungen, - wenn try getroffen wurde, dann wurde
	 * dieses ohne catch ausgefhrt, - wenn iteration kam, dann wurde geskippt. Auf Stack bis Ende des Konstrukts zu
	 * halten: Try,Iteration, Switch
	 * 
	 * @param ast
	 *            der aktuelle ASTManager
	 * @param list
	 *            die zu applizierende Pfadliste
	 */
	protected static void applyEssentialTestLogPath(ASTManager ast, ArrayList<Integer> list) {

		stackOfIterationNodes = new Stack<ASTIterationNode>();
		stackOfLoopCounters = new Stack<Integer>();

		// n hlt den Knoten beim Durchlauf durch die Funktion
		ASTNode n = ast.getNodeByID(list.get(1));

		// finde Funktion und erhhe Zhler
		ASTFunctionNode functionNode = n.getFunctionNode();
		functionNode.incNrHits();

		if (VERBOSE)
			Logger.logln("EssentialPath: " + TestCase.getPathString(list));

		if (VERBOSE)
			System.out.println(TestCase.getPathString(list));

		// durchlaufe die EssentialPathList und propagiere Hits

		// in der essentiellen Liste tauchen nur auf:
		// - functionNode.getEntryNode().getID()
		// - ifNode.getTrue/FalseNode().getID()
		// - iterationNode.getLoopNode().getID()
		// - switchNode.getListOfCases()*.getID()
		// - tryNode.getListOfCatches()*.getID()
		for (int index = 2; index <= list.size(); index++) {
			if (VERBOSE)
				System.out.println("index: " + index);
			// der nchste geloggte Knoten
			ASTNode temp = null;
			if (index < list.size())
				temp = ast.getNodeByID(list.get(index));
			ASTStatementNode indexNode;
			if (temp instanceof ASTStatementNode)
				indexNode = (ASTStatementNode) temp;
			else
				continue;

			// der zugehrige BranchNode
			ASTBranchNode branchNode = indexNode.getControlingBranchNode();

			if (VERBOSE && branchNode != null && indexNode != null)
				System.out.println("indexNode: " + indexNode.getID() + " branchNode: " + branchNode.getID());

			if (branchNode == null) {
				if (functionNode.getSetOfExitNodes().contains(ast.getNodeByID(list.get(index)))) {
					if (VERBOSE)
						Logger.logln("found function-exit: " + list.get(index) + " (" + index + ")");
				} else {
					Logger.logln("Error@TestLogManager: no controlingBranchNode for LogNode " + list.get(index)
							+ " (index: " + index + ")");
				}
				continue;
			}

			n = propagateHits(n, branchNode);

			if (VERBOSE)
				System.out.println("stopped propagating at: " + n == null ? "null" : n.getID());

			// functionexit
			if (n == null)
				return;
			// Hits vom Branch und Index-Node erhhen
			if (!stackOfIterationNodes.contains(branchNode))
				// Iteration nur einmal erhhen, nicht bei jedem Schleifendurchgang
				branchNode.incNrHits();
			indexNode.incNrHits();
			// wenn SwitchNode-Entscheidung, dann Select des Case erhhen
			if (branchNode instanceof ASTSwitchNode) {
				if (indexNode instanceof ASTCaseNode) {
					((ASTCaseNode) indexNode).incNrSelects();
				}
				// wenn Iteration
			} else if (branchNode instanceof ASTIterationNode) {
				// falls im Stack, dann wiederholte Iteration, d.h. Zhler erhhen
				if (stackOfIterationNodes.contains(branchNode)) {
					stackOfLoopCounters.push(stackOfLoopCounters.pop() + 1);
					// sonst mit loopzhler auf Stack packen
				} else {
					stackOfIterationNodes.push((ASTIterationNode) branchNode);
					stackOfLoopCounters.push(new Integer(1));
				}
			} else if (branchNode instanceof ASTTryNode) {
				// what to do?
			} else
				while (!stackOfIterationNodes.empty()) {
					if (!stackOfIterationNodes.peek().isWithinBranchConstruct(n.getID())) {
						ASTIterationNode in = stackOfIterationNodes.pop();
						int c = stackOfLoopCounters.pop();
						if (c == 1)
							in.incNrSingleLoops();
						else if (c > 1)
							in.incNrMultipleLoops();
						else
							in.incNrSkips();
					} else
						break;
				}
			n = indexNode.getNextNode();
		}

		n = propagateHits(n, null);
		if (n == n.getFunctionNode().getEndNode())
			((ASTStatementNode) n).incNrHits();
	}

	/**
	 * Propagiert die Hits ber ein Codesegment.
	 * 
	 * @param n
	 *            der Anfangsknoten
	 * @param endNode
	 *            der Endknoten
	 * @return der Endknoten
	 */
	protected static ASTNode propagateHits(ASTNode n, ASTNode endNode) {
		// durchlaufe alle Knoten und propagiere Hits bis BranchNode gefunden wird
		S: while (n != endNode && n != null && n != n.getFunctionNode().getEndNode()) {

			if (VERBOSE)
				System.out.println("propagating " + n.getID() + (endNode != null ? " (" + endNode.getID() + ")" : ""));

			// Catches werden bersprungen, sonst wren sie geloggt
			if (n instanceof ASTCatchNode && !((ASTCatchNode) n).isFinally()) {
				n = ((ASTCatchNode) n).getTryNode().getEndOfStatement();
				continue;
			}

			// propagiere Hits
			if (n instanceof ASTStatementNode) {
				if (!(n instanceof ASTIterationNode) || !stackOfIterationNodes.contains(n))
					((ASTStatementNode) n).incNrHits();
				// berspring innere Klassen
				if (((ASTStatementNode) n).isSepareted()) {
					n = ((ASTStatementNode) n).getEndOfStatement();
					continue S;
				}
			}
			// folge Sprngen
			if (n instanceof ASTJumpNode) {
				n = ((ASTJumpNode) n).getTargetNode();
				continue S;
			}
			// TryNode ohne Ausnahmewurf (sonst wrde Schleife abbrechen)
			// erhhe nrHitsWithoutException
			if (n instanceof ASTTryNode) {
				((ASTTryNode) n).incNrHitsWithoutException();
				// Iteration mit Skip, sonst kme LoopNode
			} else if (n instanceof ASTIterationNode) {
				if (!stackOfIterationNodes.contains(n))
					((ASTIterationNode) n).incNrSkips();
				n = ((ASTIterationNode) n).getEndNode();
				continue;
				// falls DoNode, dann Suche abbrechen, da als nchstes Loop kommen muss
			} else if (n instanceof ASTDoNode)
				break S;
			else
				while (!stackOfIterationNodes.empty()) {
					if (!stackOfIterationNodes.peek().isWithinBranchConstruct(n.getID())) {
						ASTIterationNode in = stackOfIterationNodes.pop();
						int c = stackOfLoopCounters.pop();
						if (c == 1)
							in.incNrSingleLoops();
						else if (c > 1)
							in.incNrMultipleLoops();
						else
							in.incNrSkips();
					} else
						break;

				}
			n = n.getNextNode();
		}
		return n;
	}

	/**
	 * Luft ber den AST und erstellt fr das TestLog die HitLists.
	 * 
	 * @param ast
	 *            der aktuelle ASTManager
	 * @param testCase
	 *            das TestLog, fr welches die Hitlists erstellt werden
	 */
	protected static void collectHitList(ASTManager ast, TestCase testCase) {
		HashMap<Integer, Integer> hitMap = new HashMap<Integer, Integer>();

		// fr alle Funktionen
		for (ASTFunctionNode fn : ast.getFunctionNodes()) {
			ASTNode n = fn;
			while (n != fn.getEndNode()) {

				// merke Hits fr Statements
				if (n instanceof IStatementNode && ((IStatementNode) n).getNrHits() > 0) {
					hitMap.put(n.getID(), ((IStatementNode) n).getNrHits());
				}

				// merke Selections fr CaseNodes
				if (n instanceof ASTCaseNode && ((ASTCaseNode) n).getNrSelects() > 0) {
					testCase.addSelection(n.getID(), ((ASTCaseNode) n).getNrSelects());
				}

				// merke BI-Informationen fr IterationNode
				if (n instanceof ASTIterationNode) {
					// merke Skips
					if (((ASTIterationNode) n).getNrSkips() > 0)
						testCase.addIterationSkip(n.getID(), ((ASTIterationNode) n).getNrSkips());
					// merke SingleLoops
					if (((ASTIterationNode) n).getNrSingleLoops() > 0)
						testCase.addIterationSingleLoop(n.getID(), ((ASTIterationNode) n).getNrSingleLoops());
					// merke DoubleLoops
					if (((ASTIterationNode) n).getNrMultipleLoops() > 0)
						testCase.addIterationDoubleLoop(n.getID(), ((ASTIterationNode) n).getNrMultipleLoops());
				}

				// merke bei Try-Catch-Blcken, wie oft keine Exception
				if (n instanceof ASTTryNode) {
					if (((ASTTryNode) n).getNrHitsWithoutException() > 0)
						testCase.addTryWithoutException(n.getID(), ((ASTTryNode) n).getNrHitsWithoutException());
				}

				n = n.getNextNode();
			}
			if (n instanceof IStatementNode && ((IStatementNode) n).getNrHits() > 0) {
				hitMap.put(n.getID(), ((IStatementNode) n).getNrHits());
				((IStatementNode) n).setNrHits(0);
			}
		}
		testCase.addHitMap(hitMap);
	}

	/**
	 * Appliziert die vorher berechneten HitLists (incl. Selections und Truthvectors) auf den AST.
	 * 
	 * @param ast
	 *            der aktuelle ASTManager
	 * @param testCase
	 *            das TestLog mit den zu applizierenden HitLists
	 */
	public static void applyTestCase(ASTManager ast, TestCase testCase) {

		// appliziere BIPfade auf entsprechende Funktionen
		for (Integer id : testCase.getMapOfCoveredBIPaths().keySet()) {
			// System.out.println(id + "\n" + ast.getNodeByID(id).toString());
			if (ast.getNodeByID(id) instanceof ASTFunctionNode) {
				((ASTFunctionNode) ast.getNodeByID(id)).getMetrics().setOfCoveredBIPaths.addAll(testCase
						.getMapOfCoveredBIPaths().get(id));
				// System.out.println(id + " - "
				// +test.getMapOfBIPaths().get(id));
			}
		}

		// appliziere ModBIPfade auf entsprechende Funktionen/Iterationen,
		// d.h. aktualisiere die Metrics der entsprechenden Funktion
		for (Integer id : testCase.getMapOfCoveredModBIPaths().keySet()) {
			ASTFunctionNode fn = ast.getNodeByID(id).getFunctionNode();

			if (!fn.getMetrics().mapOfCoveredModBIPaths.containsKey(id))
				fn.getMetrics().mapOfCoveredModBIPaths.put(id, new HashSet<Long>());
			fn.getMetrics().mapOfCoveredModBIPaths.get(id).addAll(testCase.getMapOfCoveredModBIPaths().get(id));
		}

		// appliziert HitLists
		for (Integer id : testCase.getHitMap().keySet())
			if (ast.getNodeByID(id) instanceof IStatementNode)
				((IStatementNode) ast.getNodeByID(id)).incNrHits(testCase.getHitMap().get(id));

		// appliziert TruthVectors
		for (Integer id : testCase.getTvMap().keySet()) {
			((IConditionNode) ast.getNodeByID(id)).fillConditions(testCase.getTvMap().get(id));
		}

		// appliziert Selections
		for (Integer id : testCase.getSelectionMap().keySet()) {
			if (ast.getNodeByID(id) instanceof ASTCaseNode) {
				// Logger.log(test.getSelection(id));
				((ASTCaseNode) ast.getNodeByID(id)).incNrSelects(testCase.getSelection(id));
			}
		}

		// appliziert IterationSkips
		for (Integer id : testCase.getIterationSkipsMap().keySet())
			if (ast.getNodeByID(id) instanceof ASTIterationNode)
				((ASTIterationNode) ast.getNodeByID(id)).incNrSkips(testCase.getIterationSkipsMap().get(id));

		// appliziert IterationSingleLoops
		for (Integer id : testCase.getIterationSingleLoopsMap().keySet())
			if (ast.getNodeByID(id) instanceof ASTIterationNode)
				((ASTIterationNode) ast.getNodeByID(id))
						.incNrSingleLoops(testCase.getIterationSingleLoopsMap().get(id));

		// appliziert IterationDoubleLoops
		for (Integer id : testCase.getIterationDoubleLoopsMap().keySet())
			if (ast.getNodeByID(id) instanceof ASTIterationNode)
				((ASTIterationNode) ast.getNodeByID(id)).incNrMultipleLoops(testCase.getIterationDoubleLoopsMap().get(
						id));

		// appliziert TrysWithoutExceptions
		for (Integer id : testCase.getTryWithoutExceptionMap().keySet())
			if (ast.getNodeByID(id) instanceof ASTTryNode)
				((ASTTryNode) ast.getNodeByID(id)).setNrHitsWithoutException(testCase.getTryWithoutExceptionMap().get(
						id));

	}

	/**
	 * Setzt alle berdeckungsinformationen wieder auf Null.
	 * 
	 * @param ast
	 *            der aktuelle ASTManager
	 */
	public static void resetHits(ASTManager ast) {
		for (FileNode fn : ast.getFileNodes())
			fn.resetHits();
	}

}
