package sota.gui.graph;

import java.util.ArrayList;

import org.eclipse.draw2d.geometry.Insets;
import org.eclipse.draw2d.graph.DirectedGraph;
import org.eclipse.draw2d.graph.Edge;
import org.eclipse.draw2d.graph.EdgeList;
import org.eclipse.draw2d.graph.Node;
import org.eclipse.draw2d.graph.NodeList;

import sota.Logger;
import sota.app.Application;
import sota.ast.nodes.ASTCaseNode;
import sota.ast.nodes.ASTCatchNode;
import sota.ast.nodes.ASTClassNode;
import sota.ast.nodes.ASTDoNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIfNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.ASTJumpNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.ASTSwitchNode;
import sota.ast.nodes.ASTTryNode;

/**
 * Kapselt die Funktionalitt, Kontrollflussgraphen aus (Java-)Funktionen zu konstruieren, sowie auf diesen Graphen
 * Operationen, wie z.B. Frben, Tooltip-Text erstellen, auszufhren.
 */
public class CFGBuilder {

	private static ArrayList<GraphNode> listOfJumpNodes;

	private static EdgeList edges;

	private static NodeList nodes;

	private static ZoomManager zoomManager;

	/**
	 * Liefert den DirectedGraph fr eine Funktion.
	 * 
	 * @param fn
	 *            Der ASTFunctionNode, dessen Funktion modelliert werden soll.
	 * @param gcm
	 *            Der verwendete GraphManager, der die Verknpfung GraphNode<->Node<->Widget und
	 *            GraphEdge<->Edge<->Widget speichert.
	 * @return Der konstruierte DirectedGraph
	 */
	public static DirectedGraph buildCFG(ASTFunctionNode fn, GraphManager gcm) {
		listOfJumpNodes = new ArrayList<GraphNode>();
		nodes = new NodeList();
		edges = new EdgeList();
		gcm.reset();

		zoomManager = Application.getInstance().getConfiguration().getZoomManager();

		// erster Knoten wird aus dem ASTFunctionNode generiert
		GraphNode root = new GraphNode(fn);
		// rekursiver Konstruktion der restlichen Knoten beginnend vom
		// Nachfolger bis zum letzten Knoten in der Funktion
		GraphNode son = buildCFG(root, fn.getNextNode(), fn.getEndNode());
		GraphNode end = new GraphNode(fn.getEndNode(), "function-end");

		// fhrt ein direkter Pfad zum Endknoten, wird er hier angehngt
		if (son != null)
			son.addChild(end);

		// der GraphManager wird mit den Knoten gefllt
		gcm.putGraph(root);

		// falls kein direkter Pfad zum Endknoten fhrt, wird er separat in den
		// Graph gegeben und danach verknpft
		if (son == null)
			gcm.putGraph(end);

		// alle Sprungziele werden verknpft
		for (GraphNode n : listOfJumpNodes) {
			// Logger.logln(n.astNode.getNodeString());
			ASTJumpNode jn = (ASTJumpNode) n.astNode;

			// hole dir GraphNode des Ziel-ASTNodes
			GraphNode target = gcm.astNode2graphNode(jn.getTargetNode());
			if (target != null) {
				// fge entsprechende Kante ein
				GraphEdge e = n.addChild(target);
				if (jn.isThrow())
					e.isExceptionEdge = true;

				// der ASTNode des Sprungziels ist nicht als GraphNode
				// gespeichert!
			} else
				Logger.logln("CFGBuilder: no GraphNode for " + jn.getTargetNode().getID()
						+ " as a target for jumpnode " + jn.getID());
		}

		// konstruiert restliche Knoten und Kanten fr DirectedGraph
		// (draw2d-nodes, edges)
		buildFigureGraph(root, 1);

		// der GraphManager wird vervollstndigt
		gcm.updateNodeAndEdgeReferences();

		// der DirectedGraph wird zurckgegeben
		DirectedGraph graph = new DirectedGraph();
		graph.edges = edges;
		graph.nodes = nodes;
		return graph;
	}

	/**
	 * Rekursive Funktion zum Erstellen des Kontrollflussgraphen.
	 * 
	 * @param father
	 *            Der GraphNode, an denen dieser Teilgraph angehngt wird.
	 * @param startNode
	 *            Beginn des Quellcodeabschnittes, der den Subgraphen bildet.
	 * @param endNode
	 *            Ende des Quellcodeabschnittes, der den Subgraphen bildet.
	 * @return Der letzte GraphNode, an welchem der Kontrollfluss beim Verlassen des Subgraphen weitergeht. Null, wenn
	 *         der Subgraph mit einem Sprungknoten endet. Die Sprungkante wird spter eingefgt.
	 */

	protected static GraphNode buildCFG(GraphNode father, ASTNode startNode, ASTNode endNode) {
		ASTNode n = startNode;
		GraphNode son = null, firstSon, secondSon, grandSon = null;
		// System.out.println(n.getID() + " -> " + endNode.getID());

		// solange das Ende der Funktion nicht erreicht wurde...
		do {
			// suche den nchsten kontrollflussrelevanten Knoten in der
			// AST-Knotenliste
			while (!(n instanceof ASTClassNode || n instanceof ASTIfNode || n instanceof ASTIterationNode
					|| n instanceof ASTDoNode || n instanceof ASTSwitchNode || n instanceof ASTJumpNode
					|| n instanceof ASTTryNode || n.getID() >= endNode.getID()))
				n = n.getNextNode();

			// falls eine Klasse beginnt, springe zum Ende der Klassendefinition
			if (n instanceof ASTClassNode) {
				n = ((ASTClassNode) n).getEndNode();
			}

			// falls nicht-virtuelle Sprungknoten auftreten, breche
			// KFG-Berechnung ab und merke Sprungknoten
			else if (n instanceof ASTJumpNode && !n.isVirtual()) {
				// Logger.logln(n.getNodeString());
				father.addChild(son = new GraphNode(n));
				listOfJumpNodes.add(son);
				return null;
			}

			// fr If-Knoten...
			else if (n instanceof ASTIfNode) {

				// ... konstruiere Eingang ...
				father.addChild(son = new GraphNode(n));
				father = son;

				// ... Wahr-Knoten und rekursiv den KFG des Wahr-Blocks ...
				father.addChild(firstSon = new GraphNode(((ASTIfNode) n).getTrueNode(), "true"));
				firstSon = buildCFG(firstSon, ((ASTIfNode) n).getTrueNode(), ((ASTIfNode) n).getElseNode());

				// ... falls n.virtueller Else-Knoten existiert: Falsch-Knoten
				// und rekursiv den KFG des Falsch-Blocks ...
				if (!((ASTIfNode) n).getElseNode().isVirtual()) {
					father.addChild(secondSon = new GraphNode(((ASTIfNode) n).getFalseNode(), "false"));
					secondSon = buildCFG(secondSon, ((ASTIfNode) n).getFalseNode(), ((ASTIfNode) n).getEndNode());
				}
				// sonst nur Kante vom If zum Ende
				else {
					secondSon = father;
				}

				// ... den If-Ausgangsknoten, falls er im KFG erreichbar ist ...
				if (firstSon != null || secondSon != null) {
					grandSon = new GraphNode(((ASTIfNode) n).getEndNode(), "if-end");
					if (firstSon != null)
						firstSon.addChild(grandSon);
					if (secondSon != null)
						secondSon.addChild(grandSon);

					// ... setze KFG-Untersuchung am Ende des If-Blocks fort
					n = ((ASTIfNode) n).getEndNode();
					father = grandSon;
				}
				// ansonsten breche KFG-Untersuchung hier ab
				else
					return null;
			}

			// fr for- und while-Knoten ...
			else if (n instanceof ASTIterationNode && !((ASTIterationNode) n).hasDoNode()) {

				// ... konstruiere Eingang ...
				father.addChild(son = new GraphNode(n));
				father = son;

				// ...Loop-Knoten und Subgraph in der Schleife
				father.addChild(firstSon = new GraphNode(((ASTIterationNode) n).getLoopNode(), "iter-body"));
				if ((son = buildCFG(firstSon, ((ASTIterationNode) n).getLoopNode(), ((ASTIterationNode) n).getEndNode())) != null)
					firstSon = son;

				// ... die rcklufige Kante zur nchsten Iteration
				firstSon.addLoopChild(father);

				// ... und den Iterationsausgang, falls er erreicht werden kann (-> while(true) ohne break, nur return)
				if (((ASTIterationNode) n).getSetOfExitNodes().contains(((ASTIterationNode) n).getEndNode())) {
					father.addChild(son = new GraphNode(((ASTIterationNode) n).getEndNode(), "iter-end"));
					n = ((ASTIterationNode) n).getEndNode();
					father = son;
				} else
					return null;
			}

			// fr Do-while-Knoten
			else if (n instanceof ASTDoNode) {

				// ... konstruiere Eingang ...
				father.addChild(son = new GraphNode(n));
				father = son;

				// ... und Loop-Knoten samt Schleifen-subgraph
				father.addChild(firstSon = new GraphNode(((ASTDoNode) n).getLoopNode(), "iter-body"));
				if ((son = buildCFG(firstSon, ((ASTDoNode) n).getLoopNode(), ((ASTDoNode) n).getEndNode())) != null)
					firstSon = son;

				// ... den Iterations-Knoten
				firstSon.addChild(son = new GraphNode(((ASTDoNode) n).getIterationNode()));

				// ... die rcklufige Kante zur nchsten Iteration beginned am
				// Do-Knoten
				son.addLoopChild(father);
				firstSon = son;

				// ... und den Iterationsausgang
				firstSon.addChild(son = new GraphNode(((ASTDoNode) n).getEndNode(), "do-end"));

				n = ((ASTDoNode) n).getEndNode();
				father = son;
			}

			// fr Switch-case-Knoten
			else if (n instanceof ASTSwitchNode) {

				// der switch-Knonten selbst
				father.addChild(son = new GraphNode(n));
				father = son;

				// enthlt immer den Ausgang vom letzen Case,
				// wenn dort Kontrollfluss nicht weiterfhrt (bei break), dann
				// null
				GraphNode lastCase = null;

				firstSon = null;
				for (ASTCaseNode cn : ((ASTSwitchNode) n).getListOfCases()) {

					// erstelle Case und hnge an Switch an
					father.addChild(firstSon = new GraphNode(cn));

					// eventuell erstelle Verbindung zu vorherigem Case-Ausgang
					if (lastCase != null) {
						if (cn.isDirectReachable())
							lastCase.addChild(firstSon);
					}

					// falls kein leeres Case, dann Case-Block-Pfad erstellen
					if (!(cn.getNextNode() instanceof ASTCaseNode)) {
						firstSon = buildCFG(firstSon, cn.getNextNode(), cn.getEndOfCaseBlock());
					}
					lastCase = firstSon;
				}

				// falls der Kontrollfluss aus Switch-ausgang herausfhrt
				if (((ASTSwitchNode) n).getEndNode().isReachable()) {

					// endKnoten erstellen
					son = new GraphNode(((ASTSwitchNode) n).getEndNode(), "switch-end");

					// falls Kontrollfluss aus letztem case herausfhrt,
					// dort einhngen
					if (firstSon != null)
						firstSon.addChild(son);

					// falls kein Default, dann im Switch einhngen
					if (!((ASTSwitchNode) n).hasDefault())
						father.addChild(son);

					// wenn kein direkter Weg zum Endknoten, dann an JumpNode
					// anhngen,
					// damit der GraphNode nicht verloren geht
					if (!((ASTSwitchNode) n).getEndNode().isDirectReachable()) {
						for (GraphNode gn : listOfJumpNodes)
							if (gn.astNode instanceof ASTJumpNode)
								if (((ASTJumpNode) gn.astNode).getTargetNode() == ((ASTSwitchNode) n).getEndNode())
									gn.addChild(son);
					}

					// am Ende des Switch-konstruktes weitermachen
					n = ((ASTSwitchNode) n).getEndNode();
					father = son;

				} else {
					return null;
				}
			}

			// fr Try-Konstrukte
			else if (n instanceof ASTTryNode) {
				ASTTryNode tn = (ASTTryNode) n;
				ArrayList<GraphNode> listOfEdgesToEndNode = new ArrayList<GraphNode>();
				GraphNode finallyNode = null, finallyEndNode = null;

				// Try-Node
				father.addChild(son = new GraphNode(n));
				father = son;

				// Try-Block
				father.addChild(son = new GraphNode(tn.getEntryOfTryBlock(), "try-block"));
				son = buildCFG(son, tn.getEntryOfTryBlock(), tn.getEndOfTryBlock());

				// fr alle CatchNodes ...
				for (ASTCatchNode cn : tn.getListOfCatches()) {

					// falls Finally
					if (cn.isFinally()) {
						// Node erstellen und an Try-Node hngen
						father.addChild(finallyNode = new GraphNode(cn)).isExceptionEdge = true;
						// finally-Block konstruieren
						finallyEndNode = buildCFG(finallyNode, cn.getNextNode(), cn.getEndOfCatchBlock());
					}
					// sonst bei Catch-Nodes
					else {
						// Node erstellen und an Try-Node hngen
						father.addChild(firstSon = new GraphNode(cn)).isExceptionEdge = true;

						// Catch-Block konstruieren
						firstSon = buildCFG(firstSon, cn.getNextNode(), cn.getEndOfCatchBlock());
						// wenn Catch-Block direkten KFG-Ausgang hat, merken fr
						// Anbindung an End-Node
						if (firstSon != null)
							listOfEdgesToEndNode.add(firstSon);
					}
				}

				GraphNode tryEndNode = null;

				// gibt es Bedarf fr ein Try-End-Knoten?
				// d.h. finally (wenn da) muss fortfhren und Try-Block oder ein
				// Catch-Block mssen Ende erreichen
				if ((!tn.hasFinally() || finallyEndNode != null) && (son != null || listOfEdgesToEndNode.size() > 0)) {

					tryEndNode = new GraphNode(tn.getEndNode(), "try-end");

					// Kante aus Finally zum TryEndNode
					if (finallyEndNode != null) {
						GraphEdge fe = finallyEndNode.addChild(tryEndNode);
						if (son == null && listOfEdgesToEndNode.size() == 0)
							fe.isExceptionEdge = true;
					}
				}

				// Kanten vom Try-Block zum End-Knoten oder FinallyNode ziehen
				if (son != null) {
					if (finallyNode != null)
						son.addChild(finallyEndNode);
					else
						son.addChild(tryEndNode);// .isExceptionEdge(true);
				}

				if (finallyNode != null) {
					// Kanten von fortfhrenden Catches zum Finally-Knoten
					// ziehen
					for (GraphNode gn : listOfEdgesToEndNode) {
						gn.addChild(finallyNode);// .isExceptionEdge(true);
					}

				} else {
					// Kanten von fortfhrenden Catches zum TryEnd-Knoten
					// ziehen
					for (GraphNode gn : listOfEdgesToEndNode) {
						gn.addChild(tryEndNode);// .isExceptionEdge(true);
					}
				}

				// KFG am Ende des Try-Catch-Blockes weiterkonstruieren
				n = tn.getEndNode();
				father = tryEndNode;
			}

			n = n.getNextNode();
		} while (n.getID() <= endNode.getID());
		return father;
	}

	/**
	 * Konstruiert rekursive Nodes und Edges fr den DirectedGraph aus der GraphNode-Struktur.
	 * 
	 * @param rootGraphNode
	 *            Der aktuelle GraphNode, zu welchem Nodes und Edges konstruiert werden sollen.
	 * @param constraint
	 *            Offset zu Wichtung der geordneten Node-Platzierung.
	 */
	@SuppressWarnings( { "unchecked", "deprecation" })
	protected static void buildFigureGraph(GraphNode rootGraphNode, int constraint) {
		Node rootNode, childNode;
		Edge edge;

		// wenn schon betrachtet -> Abbruch
		if (rootGraphNode.isPainted)
			return;
		rootGraphNode.isPainted = true;

		// bestimme bzw. konstruiere Ausgangknoten (rootNode) fr aktuellen
		// GraphNode
		if (rootGraphNode.node == null) {
			nodes.add(rootNode = new Node(rootGraphNode.toString()));
			rootGraphNode.node = rootNode;
			rootNode.setRowConstraint(constraint);
		} else {
			rootNode = rootGraphNode.node;
		}

		// System.out.println(rootNode.width + " " + rootNode.height);
		rootNode.height = zoomManager.getNodeHeight();
		rootNode.width = zoomManager.getNodeWidth();
		rootNode.setPadding(new Insets(zoomManager.getPadding(), zoomManager.getPadding(), zoomManager.getPadding(),
				zoomManager.getPadding()));

		// Breite des Node fr ASTFunctionNode und endNode festlegen
		if (rootGraphNode.astNode instanceof ASTFunctionNode) {
			rootNode.width = rootGraphNode.toString().length() < 12 ? 2 * zoomManager.getNodeWidth()
					: (int) (8 * rootGraphNode.toString().length() * (1.f - (1.f * zoomManager.getZoomLevel())
							/ (zoomManager.getMaxZoomLevel() + 1)));
			// rootNode.setPadding(new Insets(30, 50, zoomManager.getPadding(),zoomManager.getPadding()));
		} else if (rootGraphNode.buttonString.equals("function-end"))
			rootNode.width = 2 * zoomManager.getNodeWidth();

		// alle ausgehenden GraphEdges durchlaufen und entsprechende
		// Kanten+Knoten konstruieren
		for (GraphEdge e : rootGraphNode.outgoingEdges) {
			GraphNode n = e.targetNode;
			if (n.node == null) {
				nodes.add(childNode = new Node(n.toString()));
				n.node = childNode;
				childNode.setRowConstraint(constraint + 1);
			} else {
				childNode = n.node;
			}
			// inverted Edges sind rcklufige Kanten, die Zirkel in den Graphen
			// einfhren und daher umgangen werden mssen
			if (e.isInverted) {
				edges.add(edge = new Edge(childNode, rootNode));
				edge.isFeedback = true;
				// edges.add(edge = new Edge(rootNode, childNode));
			} else
				edges.add(edge = new Edge(rootNode, childNode));
			e.edge = edge;

			// rekursive fr Kinder aufrufen
			buildFigureGraph(n, constraint + 1);
		}
	}

}