/*******************************************************************************
 * Copyright (c) 2005 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *     IBM Corporation - initial API and implementation
 *     
 *     
 * Ursprungsauthor: Daniel Lee
 *******************************************************************************/
package sota.gui.graph;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import org.eclipse.draw2d.AbsoluteBendpoint;
import org.eclipse.draw2d.ActionEvent;
import org.eclipse.draw2d.ActionListener;
import org.eclipse.draw2d.BendpointConnectionRouter;
import org.eclipse.draw2d.Button;
import org.eclipse.draw2d.ChopboxAnchor;
import org.eclipse.draw2d.ColorConstants;
import org.eclipse.draw2d.Figure;
import org.eclipse.draw2d.FigureCanvas;
import org.eclipse.draw2d.IFigure;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.LineBorder;
import org.eclipse.draw2d.MidpointLocator;
import org.eclipse.draw2d.MouseEvent;
import org.eclipse.draw2d.MouseListener;
import org.eclipse.draw2d.Panel;
import org.eclipse.draw2d.PolygonDecoration;
import org.eclipse.draw2d.PolylineConnection;
import org.eclipse.draw2d.XYLayout;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.draw2d.graph.DirectedGraph;
import org.eclipse.draw2d.graph.DirectedGraphLayout;
import org.eclipse.draw2d.graph.Edge;
import org.eclipse.draw2d.graph.Node;
import org.eclipse.draw2d.graph.NodeList;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.PlatformUI;

import sota.Logger;
import sota.app.Application;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.Condition;
import sota.ast.nodes.IConditionNode;
import sota.ast.util.ColorManager;
import sota.ast.util.Metrics;
import sota.gui.preferences.PreferenceConstants;

/**
 * Klasse, die einen DirectedGraph zeichnet.
 * 
 */
public class CFGPlotter {

	/** Contents of the demo */
	protected IFigure contents;

	private FigureCanvas fc;

	private DirectedGraph graph = new DirectedGraph();

	private static GraphManager gcm;

	private static ZoomManager zoomManager;

	static HashMap<Edge, PolylineConnection> edgeMap = new HashMap<Edge, PolylineConnection>();

	/**
	 * Der Konstruktor.
	 * 
	 * @param composite
	 *            das Composite zum Zeichnen
	 * @param gcm
	 *            der GraphManager mit den Daten des Graphes
	 */
	public CFGPlotter(Composite composite, GraphManager gcm) {

		CFGPlotter.gcm = gcm;
		zoomManager = Application.getInstance().getConfiguration().getZoomManager();

		composite.setLayoutData(new GridData(GridData.FILL_VERTICAL));
		composite.setLayout(new GridLayout());

		fc = new FigureCanvas(composite);
		fc.setContents(contents = getContents());
		fc.getViewport().setContentsTracksHeight(true);
		fc.getViewport().setContentsTracksWidth(true);
		fc.setLayoutData(new GridData(GridData.FILL_BOTH));
	}

	/**
	 * Erstellt und liefert die Figure fr den Graph.
	 * 
	 * @return die erstelle Figure
	 */
	protected IFigure getContents() {
		new DirectedGraphLayout().visit(graph);
		IFigure content = buildFigureFromGraph(graph);
		return content;
	}

	/**
	 * Setzt den Plotter auf einen neuen Graphen an und aktualisiert die Darstellung.
	 * 
	 * @param graph
	 *            der neue Graph
	 */
	public void setContents(DirectedGraph graph) {
		this.graph = graph;
		edgeMap = new HashMap<Edge, PolylineConnection>();
		new DirectedGraphLayout().visit(graph);
		fc.setContents(buildFigureFromGraph(graph));
		zoomManager = Application.getInstance().getConfiguration().getZoomManager();
	}

	/**
	 * Beginnt das Zeichnen des Graphes.
	 * 
	 * @param graph
	 *            der DirectedGraph
	 * @return die erstellte Figure
	 */
	protected static Figure buildFigureFromGraph(DirectedGraph graph) {
		Figure contents = new Panel();

		contents.setBackgroundColor(ColorConstants.white);
		contents.setLayoutManager(new XYLayout());

		for (int i = 0; i < graph.nodes.size(); i++) {
			Node node = graph.nodes.getNode(i);
			buildNodeFigure(contents, node);
		}

		for (int i = 0; i < graph.edges.size(); i++) {
			Edge edge = graph.edges.getEdge(i);
			buildEdgeFigure(contents, edge);
		}

		return contents;
	}

	static class TopOrBottomAnchor extends ChopboxAnchor {
		public TopOrBottomAnchor(IFigure owner) {
			super(owner);
		}

		public Point getLocation(Point reference) {
			Point p;
			p = getOwner().getBounds().getCenter();
			getOwner().translateToAbsolute(p);
			if (reference.y < p.y)
				p = getOwner().getBounds().getTop();
			else
				p = getOwner().getBounds().getBottom();
			getOwner().translateToAbsolute(p);
			return p;
		}
	}

	static class LeftOrRightAnchor extends ChopboxAnchor {
		public LeftOrRightAnchor(IFigure owner) {
			super(owner);
		}

		public Point getLocation(Point reference) {
			Point p;
			p = getOwner().getBounds().getCenter();
			getOwner().translateToAbsolute(p);
			if (reference.x < p.x)
				p = getOwner().getBounds().getLeft();
			else
				p = getOwner().getBounds().getRight();
			getOwner().translateToAbsolute(p);
			return p;
		}
	}

	/**
	 * Konstruiert eine Figure aus einer Edge, inklusive der Frbungen.
	 * 
	 * @param contents
	 *            die Parent-Figure
	 * @param edge
	 *            die Edge
	 */
	protected static void buildEdgeFigure(Figure contents, Edge edge) {
		GraphEdge graphEdge = gcm.edge2GraphEdge(edge);
		PolylineConnection conn = connection(edge);
		graphEdge.connection = conn;
		graphEdge.updateStyle();

		PolygonDecoration dec = new PolygonDecoration();
		dec.setScale(zoomManager.getArrowXScale(), zoomManager.getArrowYScale());
		conn.setTargetDecoration(dec);
		// System.out.println(dec.getSize());
		// System.out.println(dec.getSize());
		conn.setPoints(edge.getPoints());

		graphEdge.hitLabel = new Label("");
		graphEdge.hitLabel.setForegroundColor(Application.getColor(ColorManager.BLACK));
		conn.add(graphEdge.hitLabel, new MidpointLocator(conn, 0));

		contents.add(conn);

		edgeMap.put(edge, conn);
	}

	/**
	 * Konstruiert eine Figure aus einem Node, inklusive der Labels und Frbungen.
	 * 
	 * @param contents
	 *            die Parent-Figure
	 * @param node
	 *            der Node
	 */
	protected static void buildNodeFigure(Figure contents, Node node) {

		final GraphNode graphNode = gcm.node2GraphNode(node);
		if (graphNode == null)
			Logger.logln("ASTGraph: couldn't retrieve GraphNode for " + node.toString() + " (" + node.hashCode()
					+ ") in node2graphNode");

		final Button button;

		if (zoomManager.showButtonText()) {
			button = new Button(node.data.toString());
			/*
			 * if(zoomManager.getZoomLevel()>0) button.setFont(zoomManager.getFont());
			 */
		} else
			button = new Button("");

		// button.setFont(zoomManager.getFontSize())
		button.setBackgroundColor(ColorConstants.lightGray);
		button.setOpaque(true);
		button.setBorder(new LineBorder());

		button.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent arg0) {

				if (graphNode.astNode.getFunctionNode() == null)
					return;
				gcm.refreshSourceViewer(graphNode, graphNode.astNode);
			}

		});

		button.addMouseListener(new MouseListener() {

			public void mouseDoubleClicked(MouseEvent arg0) {
				NodeInfoDialog nfd = new NodeInfoDialog(Display.getCurrent().getActiveShell(), graphNode);
				nfd.open();
			}

			public void mousePressed(MouseEvent arg0) {
			}

			public void mouseReleased(MouseEvent arg0) {
			}

		});

		if (node.incoming.isEmpty() || graphNode.astNode.getFunctionNode().getEndNode() == graphNode.astNode)
			button.setBorder(new LineBorder(2));
		node.data = button;

		graphNode.button = button;

		contents.add(button, new Rectangle(node.x, node.y, node.width, node.height));

		if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_CONDITIONCOVERAGE_LABELS))
			if ((graphNode.astNode instanceof IConditionNode && ((IConditionNode) graphNode.astNode)
					.hasNonTrivialCondition())
					|| (graphNode.astNode instanceof ASTFunctionNode && ((ASTFunctionNode) graphNode.astNode)
							.getMetrics().nrAtoms > 0)) {
				graphNode.labelC2 = new Label();
				graphNode.labelC2.setBackgroundColor(Application
						.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
				graphNode.labelC2.setOpaque(true);
				graphNode.labelC2.setBorder(new LineBorder());
				graphNode.labelC2.setVisible(false);
				contents.add(graphNode.labelC2, new Rectangle(node.x + node.width + 1, node.y
						+ zoomManager.getCondYOffset(), zoomManager.getCondHeight(), zoomManager.getCondHeight()));

				graphNode.labelMMDC = new Label();
				graphNode.labelMMDC.setBackgroundColor(Application
						.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
				graphNode.labelMMDC.setOpaque(true);
				graphNode.labelMMDC.setBorder(new LineBorder());
				graphNode.labelMMDC.setVisible(false);
				contents.add(graphNode.labelMMDC, new Rectangle(node.x + node.width + 1, node.y
						+ zoomManager.getCondYOffset() + zoomManager.getCondYDist(), zoomManager.getCondHeight(),
						zoomManager.getCondHeight()));

				graphNode.labelMCDC = new Label();
				graphNode.labelMCDC.setBackgroundColor(Application
						.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
				graphNode.labelMCDC.setOpaque(true);
				graphNode.labelMCDC.setBorder(new LineBorder());
				graphNode.labelMCDC.setVisible(false);
				contents.add(graphNode.labelMCDC, new Rectangle(node.x + node.width + 1, node.y
						+ zoomManager.getCondYOffset() + 2 * zoomManager.getCondYDist(), zoomManager.getCondHeight(),
						zoomManager.getCondHeight()));

				graphNode.labelC3 = new Label();
				graphNode.labelC3.setBackgroundColor(Application
						.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
				graphNode.labelC3.setOpaque(true);
				graphNode.labelC3.setBorder(new LineBorder());
				graphNode.labelC3.setVisible(false);
				contents.add(graphNode.labelC3, new Rectangle(node.x + node.width + 1, node.y
						+ zoomManager.getCondYOffset() + 3 * zoomManager.getCondYDist(), zoomManager.getCondHeight(),
						zoomManager.getCondHeight()));
			}

		if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_PATHCOVERAGE_LABELS)) {
			int indent = 0;
			if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_CONDITIONCOVERAGE_LABELS))
				indent = zoomManager.getCondHeight() + 1;
			if (graphNode.astNode instanceof ASTIterationNode) {
				Condition cond = (((ASTIterationNode) graphNode.astNode).getCondition());
				if (cond == null || cond.getMetrics().nrAtoms == 0)
					indent = 0;
				graphNode.labelMBI = new Label();
				graphNode.labelMBI.setBackgroundColor(Application
						.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
				graphNode.labelMBI.setOpaque(true);
				graphNode.labelMBI.setBorder(new LineBorder());
				graphNode.labelMBI.setVisible(false);
				contents.add(graphNode.labelMBI, new Rectangle(node.x + node.width + 1 + indent, node.y
						+ zoomManager.getCondYOffset(), zoomManager.getCondHeight(), zoomManager.getCondHeight()));
			}
			if (graphNode.astNode instanceof ASTFunctionNode) {
				Metrics m = ((ASTFunctionNode) graphNode.astNode).getMetrics();
				if (m.nrAtoms == 0)
					indent = 0;

				graphNode.labelMBI = new Label();
				graphNode.labelMBI.setBackgroundColor(Application
						.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
				graphNode.labelMBI.setOpaque(true);
				graphNode.labelMBI.setBorder(new LineBorder());
				graphNode.labelMBI.setVisible(false);
				contents.add(graphNode.labelMBI, new Rectangle(node.x + node.width + 1 + indent, node.y
						+ zoomManager.getCondYOffset(), zoomManager.getCondHeight(), zoomManager.getCondHeight()));
				graphNode.labelFMBI = new Label();
				graphNode.labelFMBI.setBackgroundColor(Application
						.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
				graphNode.labelFMBI.setOpaque(true);
				graphNode.labelFMBI.setBorder(new LineBorder());
				graphNode.labelFMBI.setVisible(false);
				contents.add(graphNode.labelFMBI, new Rectangle(node.x + node.width + 1 + indent, node.y
						+ zoomManager.getCondYOffset() + zoomManager.getCondYDist(), zoomManager.getCondHeight(),
						zoomManager.getCondHeight()));
				graphNode.labelBI = new Label();
				graphNode.labelBI.setBackgroundColor(Application
						.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
				graphNode.labelBI.setOpaque(true);
				graphNode.labelBI.setBorder(new LineBorder());
				graphNode.labelBI.setVisible(false);
				contents.add(graphNode.labelBI, new Rectangle(node.x + node.width + 1 + indent, node.y
						+ zoomManager.getCondYOffset() + 2 * zoomManager.getCondYDist(), zoomManager.getCondHeight(),
						zoomManager.getCondHeight()));
			}
		}

		if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_BASICOVERAGE_LABELS)
				&& graphNode.astNode instanceof ASTFunctionNode) {
			int indent = 0;
			if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_CONDITIONCOVERAGE_LABELS)
					&& ((ASTFunctionNode) graphNode.astNode).getMetrics().nrAtoms > 0)
				indent = zoomManager.getCondHeight() + 1;
			if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_PATHCOVERAGE_LABELS))
				indent += zoomManager.getCondHeight() + 1;

			graphNode.labelFEEC = new Label();
			graphNode.labelFEEC.setBackgroundColor(Application
					.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
			graphNode.labelFEEC.setOpaque(true);
			graphNode.labelFEEC.setBorder(new LineBorder());
			graphNode.labelFEEC.setVisible(false);
			contents.add(graphNode.labelFEEC, new Rectangle(node.x + node.width + 1 + indent, node.y
					+ zoomManager.getCondYOffset(), zoomManager.getCondHeight(), zoomManager.getCondHeight()));
			graphNode.labelC0 = new Label();
			graphNode.labelC0.setBackgroundColor(Application
					.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
			graphNode.labelC0.setOpaque(true);
			graphNode.labelC0.setBorder(new LineBorder());
			graphNode.labelC0.setVisible(false);
			contents.add(graphNode.labelC0, new Rectangle(node.x + node.width + 1 + indent, node.y
					+ zoomManager.getCondYOffset() + zoomManager.getCondYDist(), zoomManager.getCondHeight(),
					zoomManager.getCondHeight()));
			graphNode.labelC1 = new Label();
			graphNode.labelC1.setBackgroundColor(Application
					.getColorFromPreferenceStore(PreferenceConstants.CFG_DEF_COV_LABEL));
			graphNode.labelC1.setOpaque(true);
			graphNode.labelC1.setBorder(new LineBorder());
			graphNode.labelC1.setVisible(false);
			contents.add(graphNode.labelC1, new Rectangle(node.x + node.width + 1 + indent, node.y
					+ zoomManager.getCondYOffset() + 2 * zoomManager.getCondYDist(), zoomManager.getCondHeight(),
					zoomManager.getCondHeight()));

		}

		graphNode.setToolTip();
	}

	/**
	 * Konstruiert die Verbindung einer Edge.
	 * 
	 * @param e
	 *            die Edge
	 * @return the connection die erstellte Verbindung
	 */
	protected static PolylineConnection connection(Edge e) {
		PolylineConnection conn = new PolylineConnection();
		conn.setConnectionRouter(new BendpointConnectionRouter());
		List<AbsoluteBendpoint> bends = new ArrayList<AbsoluteBendpoint>();
		NodeList nodes = e.vNodes;
		if (nodes != null) {
			for (int i = 0; i < nodes.size(); i++) {
				Node n = nodes.getNode(i);
				int x = n.x;
				int y = n.y;
				bends.add(new AbsoluteBendpoint(x, y));
				bends.add(new AbsoluteBendpoint(x, y + n.height));
			}
		}
		conn.setRoutingConstraint(bends);
		return conn;
	}
}