package sota.gui.graph;

import java.util.HashMap;

import org.eclipse.draw2d.graph.Edge;
import org.eclipse.draw2d.graph.Node;
import org.eclipse.jface.preference.PreferenceConverter;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocumentPartitioner;
import org.eclipse.jface.text.rules.FastPartitioner;
import org.eclipse.jface.text.source.SourceViewer;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.ui.PlatformUI;

import sota.Logger;
import sota.app.Application;
import sota.ast.nodes.ASTCaseNode;
import sota.ast.nodes.ASTCatchNode;
import sota.ast.nodes.ASTCompoundNode;
import sota.ast.nodes.ASTDoNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIfNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.ASTJumpNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.ASTStatementNode;
import sota.ast.nodes.ASTSwitchNode;
import sota.ast.nodes.ASTTryNode;
import sota.ast.nodes.IConditionNode;
import sota.ast.nodes.IStatementNode;
import sota.ast.util.ASTManager;
import sota.ast.util.ProjectConfig;
import sota.gui.GuiUpdateListener;
import sota.gui.preferences.PreferenceConstants;
import sota.lang.LangClassProvider;

/**
 * Diese Klasse verwaltet die Datenstrukturen des CFG, ermglicht ihre Abbildung aufeinander und die Akualisierung der
 * Graph-Informationen.
 * 
 */
public class GraphManager implements GuiUpdateListener {

	private HashMap<ASTNode, GraphNode> ast2graphNode = new HashMap<ASTNode, GraphNode>();
	private HashMap<Node, GraphNode> node2graphNode = new HashMap<Node, GraphNode>();
	private HashMap<Edge, GraphEdge> edge2graphEdge = new HashMap<Edge, GraphEdge>();
	private SourceViewer sourceViewer;
	private IDocumentPartitioner partitioner;
	private Document fileDocument;
	private int topLineNr = -1;
	private boolean firstNodeSelection = true;
	private ASTFunctionNode currentFunction = null;

	/**
	 * Der Konstruktor.
	 */
	public GraphManager() {
		Application.getInstance().getASTManager().addGuiUpdateListener(this);
	}

	/**
	 * Lscht die Darstellung im SourceViwer.
	 */
	public void clearSourceViewer() {
		fileDocument.set("");
		sourceViewer.refresh();
	}

	/**
	 * Aktualisiert den SourceViewer, indem die Zeile des ausgewhlten Knoten hervorgehoben wird.
	 * 
	 * @param gn
	 *            der GraphNode
	 * @param n
	 *            der zugehrige ASTNode *
	 */
	public void refreshSourceViewer(GraphNode gn, ASTNode n) {
		// System.out.println("refresh A");
		Document x = Application.getFileDocument();
		if (x == null)
			return;

		int oldTopIndex = sourceViewer.getTopIndex();
		sourceViewer.getTextWidget().setRedraw(false);
		fileDocument.set(x.get());
		// System.out.println("a");
		if (n != null) {
			int lineNr = n.getLineNr();

			if (gn != null)
				if (gn.buttonString.equals("true") || gn.buttonString.equals("iter-body")) {

					ASTNode temp = n;
					while (temp.isVirtual() || (temp instanceof ASTCompoundNode)) {
						temp = temp.getNextNode();
						lineNr = temp.getLineNr();
						if (temp instanceof ASTCompoundNode && !((ASTCompoundNode) temp).isStartNode()) {
							lineNr = n.getLineNr() + 1;
							break;
						}
					}

				}
			// System.out.println("b");
			try {
				int startLineOffset = fileDocument.getLineOffset(n.getFunctionNode().getLineNr() - 1);
				int length = fileDocument.getLineOffset(n.getFunctionNode().getEndNode().getLineNr()) - startLineOffset;
				sourceViewer.setVisibleRegion(startLineOffset, length);
			} catch (BadLocationException e) {
				e.printStackTrace();
			}
			// System.out.println("c");
			if (!Application.getInstance().getConfiguration().getPinSourceView()
					|| currentFunction != n.getFunctionNode() || firstNodeSelection) {
				int center = ((sourceViewer.getBottomIndex() - sourceViewer.getTopIndex()) / 2) + 1;
				topLineNr = lineNr - center;
				sourceViewer.setTopIndex(topLineNr);
				if (currentFunction != n.getFunctionNode()) {
					currentFunction = n.getFunctionNode();
					firstNodeSelection = true;
				} else
					firstNodeSelection = false;
			} else
				sourceViewer.setTopIndex(oldTopIndex);

			// System.out.println("d");
			sourceViewer.getTextWidget().setLineBackground(lineNr - n.getFunctionNode().getLineNr(), 1,
					Application.getColorFromPreferenceStore(PreferenceConstants.CFG_CURR_LINE_COLOR));
		}

		sourceViewer.getTextWidget().setRedraw(true);
		// System.out.println("refresh X");
	}

	/**
	 * Initialisiert den SourceViewer.
	 */
	public void initSourceViewer() {
		// System.out.println("Init SV A");
		ProjectConfig config = Application.getInstance().getConfiguration();
		fileDocument.set("");
		topLineNr = -1;
		firstNodeSelection = true;
		try {
			// System.out.println("SV: " + sourceViewer + " p: "+ partitioner + " fd: " + fileDocument);
			sourceViewer.unconfigure();
			sourceViewer.configure(LangClassProvider.getSourceViewerConfiguration(config).newInstance());

			if (partitioner != null)
				partitioner.disconnect();
			partitioner = new FastPartitioner(LangClassProvider.getPartitionScanner(config).newInstance(),
					LangClassProvider.getPartitionScannerTypes(config));

			fileDocument.setDocumentPartitioner(LangClassProvider.getPartitioning(config), partitioner);

			partitioner.connect(fileDocument);

		} catch (InstantiationException e) {
			Logger.logln(e.toString());
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			Logger.logln(e.toString());
			e.printStackTrace();
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.lang.java.ast.util.GuiUpdateListener#guiUpdate(sota.ast.util.ASTManager)
	 */
	public void guiUpdate(ASTManager ast) {

		// falls TestLogs vorhanden sind, dann frbe Graph nach berdeckung
		if (ast.getConfiguration().getListOfTestLogs().size() > 0 && ast.showCoverage()
				&& ast.getConfiguration().parsedSourceStatus == ProjectConfig.CLEAN) {

			// createCoverageText();

			// frbe erstmal alle Kanten rot (n. berdeckt)
			for (GraphEdge e : edge2graphEdge.values())
				e.showWithTraversals(0);

			// frbe Knoten nach berdeckung und ausgehende
			// Kanten, wenn berdeckt
			for (GraphNode n : ast2graphNode.values()) {
				n.setToolTip();
				applyCoverageColor(n);
			}
		}

		// ansonsten frbe Graph monochrom
		else {
			// frbe Knoten grau
			for (GraphNode n : node2graphNode.values()) {
				n.setToolTip();
				n.applyColor(Application.getColorFromPreferenceStore(PreferenceConstants.CFG_NODE_DEF_COLOR));
				ASTNode astNode = n.astNode;
				if ((astNode instanceof IConditionNode && ((IConditionNode) astNode).hasNonTrivialCondition())
						|| (n.astNode instanceof ASTFunctionNode && ((ASTFunctionNode) n.astNode).getMetrics().nrAtoms > 0)) {
					n.labelC2.setVisible(false);
					n.labelMMDC.setVisible(false);
					n.labelMCDC.setVisible(false);
					n.labelC3.setVisible(false);
				}
				/*
				 * if (n.astNode instanceof ASTFunctionNode) n.labelFEEC.setVisible(false);
				 */

			}

			// frbe Kanten schwarz
			for (GraphEdge e : edge2graphEdge.values())
				e.showDefault();
		}
	}

	/**
	 * Ermittelt zu einem Node den GraphNode.
	 * 
	 * @param node
	 *            der Node
	 * @return der GraphNode
	 */
	public GraphNode node2GraphNode(Node node) {
		return node2graphNode.get(node);
	}

	/**
	 * Ermittelt zu einer Edge die GraphEdge.
	 * 
	 * @param edge
	 *            die Edge
	 * @return die GraphEdge
	 */
	public GraphEdge edge2GraphEdge(Edge edge) {
		return edge2graphEdge.get(edge);
	}

	/**
	 * Ermittelt zu einem ASTNode den GraphNode.
	 * 
	 * @param n
	 *            der ASTNode
	 * @return der GraphNode
	 */
	public GraphNode astNode2graphNode(ASTNode n) {
		if (!ast2graphNode.containsKey(n))
			return null;
		return ast2graphNode.get(n);
	}

	/**
	 * Fgt einen GraphNode in den Manager hinzu.
	 * 
	 * @param graphNode
	 *            der GraphNode
	 */
	public void putGraph(GraphNode graphNode) {
		if (ast2graphNode.keySet().contains(graphNode.astNode)) {
			// System.out.println("Graphmanager: Graphnode already in
			// ast2graphNode");
			return;
		}
		// System.out.println("put: " + root.astNode.getID());
		ast2graphNode.put(graphNode.astNode, graphNode);
		for (GraphEdge e : graphNode.outgoingEdges) {
			putGraph(e.targetNode);
		}
	}

	/**
	 * Fhrt ein Update der Kanten- und Knotenreferenzen durch.
	 */
	public void updateNodeAndEdgeReferences() {
		for (GraphNode n : ast2graphNode.values()) {
			/*
			 * System.out.println("put in node2grapheNode: " + n.astNode.getID() + " - " + n.toString() + " - " +
			 * n.node);
			 */
			node2graphNode.put(n.node, n);
			for (GraphEdge e : n.outgoingEdges)
				edge2graphEdge.put(e.edge, e);
		}
	}

	/**
	 * Lscht alle Knoten und Kanten.
	 */
	public void reset() {
		ast2graphNode.clear();
		node2graphNode.clear();
		edge2graphEdge.clear();
	}

	/**
	 * Setzt den SourceViwer.
	 * 
	 * @param sourceViewer
	 *            der SourceViewer
	 */
	public void setSourceViewer(SourceViewer sourceViewer) {
		this.sourceViewer = sourceViewer;
		fileDocument = new Document();
		sourceViewer.setEditable(false);
		sourceViewer.getTextWidget().setIndent(10);
		sourceViewer.setDocument(fileDocument);
		// refreshSourceViewer(
	}

	/**
	 * Ermittelt zu einem ASTNode den String fr die Knotenbeschriftung.
	 * 
	 * @param graphNode
	 *            der GraphNode
	 * @return der entsprechende String
	 */
	public static String astNodeToString(GraphNode graphNode) {
		if (graphNode.astNode instanceof ASTFunctionNode)
			return ((ASTFunctionNode) graphNode.astNode).getName();
		if (graphNode.astNode instanceof ASTSwitchNode)
			return "switch";
		if (graphNode.astNode instanceof ASTDoNode)
			return "do";
		if (graphNode.astNode instanceof ASTIterationNode)
			return "iteration";
		if (graphNode.astNode instanceof ASTCaseNode) {
			if (((ASTCaseNode) graphNode.astNode).isDefault())
				return "default";
			return "case";
		}
		if (graphNode.astNode instanceof ASTIfNode)
			return "if";
		if (graphNode.astNode instanceof ASTTryNode)
			return "try";
		if (graphNode.astNode instanceof ASTCatchNode) {
			if (((ASTCatchNode) graphNode.astNode).isFinally())
				return "finally";
			return "catch";
		}
		if (graphNode.astNode instanceof ASTJumpNode) {
			if (((ASTJumpNode) graphNode.astNode).isReturn())
				return "return";
			if (((ASTJumpNode) graphNode.astNode).isBreak())
				return "break";
			if (((ASTJumpNode) graphNode.astNode).isContinue())
				return "continue";
			if (((ASTJumpNode) graphNode.astNode).isThrow())
				return "throw";
		}
		return null;
	}

	/**
	 * Bestimmt, ob dieser GraphNode berdeckt ist oder nicht
	 * 
	 * @param graphNode
	 *            Der zu testende GraphNode.
	 * @return Wahr, wenn der GraphNode berdeckt wurde, sonst falsch.
	 */
	public static boolean isCovered(GraphNode graphNode) {
		if (graphNode.button == null)
			return false;
		if (graphNode.astNode instanceof IStatementNode)
			return ((IStatementNode) graphNode.astNode).getNrHits() > 0;
		if (graphNode.astNode instanceof ASTDoNode)
			return ((ASTDoNode) graphNode.astNode).getIterationNode().getNrHits() > 0;
		// + ((ASTDoNode) graphNode.astNode).getIterationNode()
		// .getLoopNode().getNrHits() > 0;
		return false;
	}

	/**
	 * Bestimmt die Anzahl der Hits fr einen GraphNode.
	 * 
	 * @param graphNode
	 *            der GraphNode
	 * @return die Anzahl der berdeckungen
	 */
	public static int getNrHits(GraphNode graphNode) {
		if (graphNode.button == null)
			return 0;
		if (graphNode.astNode instanceof IStatementNode)
			return ((IStatementNode) graphNode.astNode).getNrHits();
		if (graphNode.astNode instanceof ASTDoNode)
			return ((ASTDoNode) graphNode.astNode).getLoopNode().getNrHits();
		// + ((ASTDoNode) graphNode.astNode).getIterationNode()
		// .getLoopNode().getNrHits() > 0;
		return 0;
	}

	/**
	 * Frbt den GraphNode und die ausgehenden GraphEdges je nach Frbung des zugehrigen ASTNodes
	 * 
	 * @param graphNode
	 *            Der zu frbende GraphNode.
	 */
	public static void applyCoverageColor(GraphNode graphNode) {
		ASTNode astNode = graphNode.astNode;

		RGB COVEREDNODE = PreferenceConverter.getColor(PlatformUI.getPreferenceStore(),
				PreferenceConstants.CFG_NODE_HIT_COLOR);
		RGB NOTCOVEREDNODE = PreferenceConverter.getColor(PlatformUI.getPreferenceStore(),
				PreferenceConstants.CFG_NODE_MISS_COLOR);
		RGB BRANCHINGNODE = PreferenceConverter.getColor(PlatformUI.getPreferenceStore(),
				PreferenceConstants.CFG_BRANCHINIG_NODE_COLOR);
		boolean showBranchingNodeColor = PlatformUI.getPreferenceStore().getBoolean(
				PreferenceConstants.CFG_SHOW_BRANCHING_NODE_COLOR);

		if (graphNode.button == null) {
			Logger.logln("no button at " + astNode.getID());
			return;
		}

		graphNode.setCoverageLabels();

		if (astNode.getFunctionNode().getEndNode().getID() == astNode.getID()) {
			int c = 0;
			for (ASTStatementNode n : astNode.getFunctionNode().getSetOfExitNodes())
				c += n.getNrHits();
			if (c > 0)
				graphNode.applyColor(COVEREDNODE);
			else
				graphNode.applyColor(NOTCOVEREDNODE);

		} else if (isCovered(graphNode)) {

			graphNode.applyColor(COVEREDNODE);

			switch (graphNode.outgoingEdges.size()) {
			case 0:
				return;
			case 1:
				GraphEdge ge = graphNode.outgoingEdges.iterator().next();
				ge.showWithTraversals(getNrHits(graphNode));
				return;
			default:
				if (astNode instanceof ASTIfNode) {
					ASTIfNode ifn = (ASTIfNode) astNode;

					if (showBranchingNodeColor
							&& (ifn.getTrueNode().getNrHits() == 0 || ifn.getFalseNode().getNrHits() == 0))
						graphNode.applyColor(BRANCHINGNODE);

					for (GraphEdge e : graphNode.outgoingEdges) {
						if (e.targetNode.astNode == ifn.getTrueNode()) {
							e.showWithTraversals(ifn.getTrueNode().getNrHits());

						} else {
							e.showWithTraversals(ifn.getFalseNode().getNrHits());
						}
					}
					return;

				} else if (astNode instanceof ASTIterationNode) {
					ASTIterationNode in = (ASTIterationNode) astNode;

					if (in.hasDoNode()) {
						if (showBranchingNodeColor && (in.getNrMultipleLoops() == 0))
							graphNode.applyColor(BRANCHINGNODE);
					} else if (showBranchingNodeColor && in.getLoopNode().getNrHits() == 0)
						graphNode.applyColor(BRANCHINGNODE);

					for (GraphEdge e : graphNode.outgoingEdges) {
						if (e.targetNode.astNode == in.getDoNode()) {
							e.showWithTraversals(in.getLoopNode().getNrHits() - in.getNrHits());
						} else if (e.targetNode.astNode == in.getLoopNode()) {
							e.showWithTraversals(in.getNrLoops());
						} else if (e.targetNode.astNode == in.getEndNode()) {
							e.showWithTraversals(in.getEndNode().getNrHits());
						}
					}

					return;

				} else if (astNode instanceof ASTSwitchNode) {
					boolean allCasesSelected = true;

					for (GraphEdge e : graphNode.outgoingEdges) {
						if (e.targetNode.astNode instanceof ASTCaseNode) {
							if (((ASTCaseNode) e.targetNode.astNode).getNrSelects() == 0)
								allCasesSelected = false;
							e.showWithTraversals(((ASTCaseNode) e.targetNode.astNode).getNrSelects());
						}
					}

					if (showBranchingNodeColor && !allCasesSelected)
						graphNode.applyColor(BRANCHINGNODE);

					return;
				} else if (astNode instanceof ASTTryNode) {
					boolean allCatchesVisited = true;

					for (GraphEdge e : graphNode.outgoingEdges) {
						if (e.targetNode.astNode instanceof ASTCatchNode) {
							if (((ASTCatchNode) e.targetNode.astNode).getNrHits() == 0)
								allCatchesVisited = false;
							e.showWithTraversals(((ASTCatchNode) e.targetNode.astNode).getNrHits());
						} else if (e.targetNode.astNode instanceof ASTStatementNode)
							e.showWithTraversals(((ASTStatementNode) e.targetNode.astNode).getNrHits());

					}

					if (showBranchingNodeColor && !allCatchesVisited)
						graphNode.applyColor(BRANCHINGNODE);

					return;
				}
			}
		} else {
			graphNode.applyColor(NOTCOVEREDNODE);
		}

	}

	/**
	 * Ordnet jedem berdeckungsgrad eine Farbe zu.
	 * 
	 * @param p
	 *            der berdeckungsgrad
	 * @return die Farbe
	 */
	public static Color percentage2Color(float p) {
		return Application.getInstance().getColorManager().percentage2Color(p);
	}

}
