package sota.gui.graph;

import java.util.HashSet;

import org.eclipse.draw2d.Button;
import org.eclipse.draw2d.Label;
import org.eclipse.draw2d.graph.Node;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.RGB;
import org.eclipse.ui.PlatformUI;

import sota.app.Application;
import sota.ast.nodes.ASTCaseNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIfNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.ASTStatementNode;
import sota.ast.nodes.ASTSwitchNode;
import sota.ast.nodes.ASTTryNode;
import sota.ast.nodes.IConditionNode;
import sota.ast.nodes.IStatementNode;
import sota.ast.util.Metrics;
import sota.gui.preferences.PreferenceConstants;

/**
 * Knoten des Kontrollflussesgraphen, der sowohl entsprechenden ASTNode als auch den Node des DirectedGraph, welcher in
 * Draw2d den Graphen reprsentiert und selber auf den Button verweist, hlt.
 * 
 */
public class GraphNode {
	HashSet<GraphEdge> incomingEdges;

	HashSet<GraphEdge> outgoingEdges;

	public ASTNode astNode;

	// Knoten des DirectedGraph
	Node node;
	Button button;

	Label labelFEEC;
	Label labelC0;
	Label labelC1;

	Label labelMBI;
	Label labelFMBI;
	Label labelBI;

	Label labelC2;
	Label labelMMDC;
	Label labelMCDC;
	Label labelC3;

	boolean isPainted;

	String buttonString;

	boolean isJumpNode;

	public String firstMCDC = "", secondMCDC = "";

	/**
	 * Der Konstruktor.
	 * 
	 * @param n
	 *            der entsprechende ASTNode
	 */
	public GraphNode(ASTNode n) {
		incomingEdges = new HashSet<GraphEdge>();
		outgoingEdges = new HashSet<GraphEdge>();
		astNode = n;
		node = null;
		isPainted = false;
		buttonString = n.getID() + "";

	}

	/**
	 * Der Konstruktor fr einen beschrifteten Knoten.
	 * 
	 * @param n
	 *            der Knoten
	 * @param s
	 *            die vorgegebene Beschriftung
	 */
	public GraphNode(ASTNode n, String s) {
		this(n);
		buttonString = s;
	}

	/**
	 * Appliziert eine Farbe auf den Knoten.
	 * 
	 * @param rgb
	 *            der RGB-Wert der Farbe
	 */
	public void applyColor(RGB rgb) {
		if (button == null)
			return;
		button.setBackgroundColor(Application.getInstance().getColorManager().getColor(rgb));
	}

	/**
	 * Appliziert eine Farb eauf den Knoten.
	 * 
	 * @param color
	 *            die Farbe
	 */
	public void applyColor(Color color) {
		if (button == null)
			return;
		button.setBackgroundColor(color);
	}

	/**
	 * Fgt dem Knoten ein Kind hinzu.
	 * 
	 * @param child
	 *            das Kind
	 * @return die verbindende Kante
	 */
	public GraphEdge addChild(GraphNode child) {
		for (GraphEdge ge : outgoingEdges)
			if (ge.targetNode == child)
				return ge;
		GraphEdge edge = new GraphEdge(this, child);
		child.incomingEdges.add(edge);
		this.outgoingEdges.add(edge);
		return edge;
	}

	/**
	 * Fgt eine Schleife zu einem Knoten hinzu.
	 * 
	 * @param child
	 *            das Kind
	 */
	public void addLoopChild(GraphNode child) {
		GraphEdge edge = new GraphEdge(this, child);
		child.incomingEdges.add(edge);
		this.outgoingEdges.add(edge);
		edge.isInverted = true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		String s = GraphManager.astNodeToString(this);
		if (s != null)
			return s;

		return buttonString;
	}

	/**
	 * Bestimmt den informativen ToolTip-Text, abhngig von der berdeckung.
	 */
	public void setToolTip() {

		String s = " " + toString() + "\n";

		if (button == null)
			return;

		if (astNode.getFunctionNode().getEndNode().getID() == astNode.getID()) {
			int c = 0;
			for (ASTStatementNode n : astNode.getFunctionNode().getSetOfExitNodes())
				c += n.getNrHits();
			s += " Function Exit \n nrHits: " + c + " \n\n ID: " + astNode.getID() + " \tLine: " + astNode.getLineNr()
					+ " ";

		} else if (astNode instanceof ASTIterationNode) {

			s += " nrHits: " + ((ASTIterationNode) astNode).getNrHits() + " \n nrSkips: "
					+ ((ASTIterationNode) astNode).getNrSkips() + " \n nrSingleLoops: "
					+ ((ASTIterationNode) astNode).getNrSingleLoops() + " \n nrMultipleLoops: "
					+ ((ASTIterationNode) astNode).getNrMultipleLoops() + " \n nrLoops: "
					+ ((ASTIterationNode) astNode).getNrLoops() + " \n\n ID: " + astNode.getID() + " \tLine: "
					+ astNode.getLineNr() + " ";

		} else if (astNode instanceof ASTCaseNode) {

			s += " nrHits: " + ((ASTCaseNode) astNode).getNrHits() + " \n nrSelects: "
					+ ((ASTCaseNode) astNode).getNrSelects() + " \n\n ID: " + astNode.getID() + " \tLine: "
					+ astNode.getLineNr() + " ";

		} else if (astNode instanceof ASTSwitchNode) {

			s += " nrHits: " + ((ASTSwitchNode) astNode).getNrHits() + " \n Selections: "
					+ ((ASTSwitchNode) astNode).getSelectionString();
			for (ASTCaseNode cn : ((ASTSwitchNode) astNode).getListOfCases())
				if (cn.isDefault())
					s += " \n default: " + cn.getNrSelects();
				else
					s += " \n case " + cn.getConstant() + ": " + cn.getNrSelects();
			s += " \n\n ID: " + astNode.getID() + " \tLine: " + astNode.getLineNr() + " ";

		} else if (astNode instanceof ASTIfNode) {

			s += " nrHits: " + ((IStatementNode) astNode).getNrHits() + " \n True: "
					+ ((ASTIfNode) astNode).getTrueNode().getNrHits() + " \n False: "
					+ ((ASTIfNode) astNode).getFalseNode().getNrHits() + " \n\n ID: " + astNode.getID() + " \tLine: "
					+ astNode.getLineNr() + " ";

		} else if (astNode instanceof ASTTryNode) {

			s += " nrHits: " + ((ASTTryNode) astNode).getNrHits() + " \n w/o Exception: "
					+ ((ASTTryNode) astNode).getNrHitsWithoutException() + " \n\n ID: " + astNode.getID() + " \tLine: "
					+ astNode.getLineNr() + " ";

		} else if (astNode instanceof IStatementNode) {

			s += " nrHits: " + ((IStatementNode) astNode).getNrHits() + " \n\n ID: " + astNode.getID() + " \tLine: "
					+ astNode.getLineNr() + " ";

		} else {
			s += " ID: " + astNode.getID() + " \tLine: " + astNode.getLineNr() + " ";
		}

		button.setToolTip(new Label(s));

		if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_CONDITIONCOVERAGE_LABELS))
			if ((astNode instanceof IConditionNode && ((IConditionNode) astNode).hasNonTrivialCondition())
					|| (astNode instanceof ASTFunctionNode && ((ASTFunctionNode) astNode).getMetrics().nrAtoms > 0)) {
				Metrics m;
				if (astNode instanceof ASTFunctionNode)
					m = ((ASTFunctionNode) astNode).getMetrics();
				else
					m = ((IConditionNode) astNode).getMetrics();

				labelC2.setToolTip(new Label("  C2: " + m.c2 + "%  "));
				labelMMDC.setToolTip(new Label("  MMDC: " + m.mmdc + "%  "));
				labelMCDC.setToolTip(new Label("  MCDC: " + m.mcdc + "%  "));
				labelC3.setToolTip(new Label("  C3: " + m.c3 + "%  "));
			}

		if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_PATHCOVERAGE_LABELS)) {
			if (astNode instanceof ASTIterationNode) {
				labelMBI.setToolTip(new Label("  MBI: " + ((ASTIterationNode) astNode).getMetrics().mbi + "% "));
			}
			if (astNode instanceof ASTFunctionNode) {
				Metrics m = ((ASTFunctionNode) astNode).getMetrics();
				labelMBI.setToolTip(new Label("  MBI: " + m.mbi + "% "));
				labelFMBI.setToolTip(new Label("  FMBI: " + " " + 100.0 * m.getNrCoveredModBIPaths(astNode.getID())
						/ m.getNrModBIPaths(astNode.getID()) + "%"));

				if (m.nrModBIPaths == m.getNrModBIPaths(astNode.getID()))
					labelMBI.setToolTip(new Label("  MBI: ---"));
				else
					labelMBI.setToolTip(new Label("  MBI: " + 100.0
							* (m.nrCoveredModBIPaths - m.getNrCoveredModBIPaths(astNode.getID()))
							/ (m.nrModBIPaths - m.getNrModBIPaths(astNode.getID())) + "%"));

				labelBI.setToolTip(new Label("  BI: " + m.bi + "% "));
			}

		}
		if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_BASICOVERAGE_LABELS))
			if (astNode instanceof ASTFunctionNode) {
				Metrics m = ((ASTFunctionNode) astNode).getMetrics();
				labelFEEC.setToolTip(new Label("  FEEC: " + m.feec + "% "));
				labelC0.setToolTip(new Label("  C0: " + m.c0 + "% "));
				labelC1.setToolTip(new Label("  C1: " + m.c1 + "% "));
			}

	}

	/**
	 * Aktualisiert die berdeckungslabel.
	 */
	public void setCoverageLabels() {

		if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_CONDITIONCOVERAGE_LABELS))
			if ((astNode instanceof IConditionNode && ((IConditionNode) astNode).hasNonTrivialCondition())
					|| (astNode instanceof ASTFunctionNode && ((ASTFunctionNode) astNode).getMetrics().nrAtoms > 0)) {
				Metrics m;
				if (astNode instanceof ASTFunctionNode)
					m = ((ASTFunctionNode) astNode).getMetrics();
				else
					m = ((IConditionNode) astNode).getCondition().getMetrics();

				labelC2.setBackgroundColor(percentage2Color(m.c2));
				labelC2.setVisible(true);
				labelMMDC.setBackgroundColor(percentage2Color(m.mmdc));
				labelMMDC.setVisible(true);
				labelMCDC.setBackgroundColor(percentage2Color(m.mcdc));
				labelMCDC.setVisible(true);
				labelC3.setBackgroundColor(percentage2Color(m.c3));
				labelC3.setVisible(true);
			}
		if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_PATHCOVERAGE_LABELS)) {
			if (astNode instanceof ASTIterationNode) {
				labelMBI.setBackgroundColor(percentage2Color(((ASTIterationNode) astNode).getMetrics().mbi));
				labelMBI.setVisible(true);
			}
			if (astNode instanceof ASTFunctionNode) {
				labelMBI.setBackgroundColor(percentage2Color(((ASTFunctionNode) astNode).getMetrics().mbi));
				labelMBI.setVisible(true);

				labelFMBI.setBackgroundColor(percentage2Color(((ASTFunctionNode) astNode).getMetrics().mbi));
				labelFMBI.setVisible(true);

				labelBI.setBackgroundColor(percentage2Color(((ASTFunctionNode) astNode).getMetrics().bi));
				labelBI.setVisible(true);

			}

		}
		if (PlatformUI.getPreferenceStore().getBoolean(PreferenceConstants.CFG_SHOW_BASICOVERAGE_LABELS))
			if (astNode instanceof ASTFunctionNode) {
				labelFEEC.setBackgroundColor(percentage2Color(((ASTFunctionNode) astNode).getMetrics().feec));
				labelFEEC.setVisible(true);
				labelC0.setBackgroundColor(percentage2Color(((ASTFunctionNode) astNode).getMetrics().c0));
				labelC0.setVisible(true);
				labelC1.setBackgroundColor(percentage2Color(((ASTFunctionNode) astNode).getMetrics().c1));
				labelC1.setVisible(((ASTFunctionNode) astNode).getMetrics().nrBranches > 0);
			}

	}

	/**
	 * Ordnet jedem berdeckungsgrad eine Farbe zu.
	 * 
	 * @param p
	 *            der berdeckungsgrad
	 * @return die Farbe
	 */
	private Color percentage2Color(float p) {
		return GraphManager.percentage2Color(p);
	}

}
