package sota.gui.provider;

import java.util.Formatter;

import org.eclipse.jface.viewers.ILabelProviderListener;
import org.eclipse.jface.viewers.ITableColorProvider;
import org.eclipse.jface.viewers.ITableLabelProvider;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Image;

import sota.app.Application;
import sota.ast.nodes.ASTClassNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.FileNode;
import sota.ast.nodes.IScopeNode;
import sota.ast.util.Metrics;
import sota.ast.util.ProjectConfig;
import sota.gui.views.CoverageView;

/**
 * Der LabelProvider fr die Klasse CoverageView.
 * 
 */
public class CoverageViewLabelProvider implements ITableLabelProvider, ITableColorProvider {

	private static String EMPTY = "  ---  ";

	private CoverageView coverageView;

	/**
	 * Der Konstruktor
	 * 
	 * @param coverageView
	 *            die CoverageView-Instanz
	 */
	public CoverageViewLabelProvider(CoverageView coverageView) {
		super();
		this.coverageView = coverageView;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.viewers.ITableLabelProvider#getColumnImage(java.lang.Object, int)
	 */
	public Image getColumnImage(Object element, int columnIndex) {
		if (columnIndex != 0)
			return null;
		if (element instanceof ASTFunctionNode) {
			if (((ASTFunctionNode) element).isPrivate())
				return Application.getInstance().getImageRegistry().get("privatemethod");
			if (((ASTFunctionNode) element).isProtected())
				return Application.getInstance().getImageRegistry().get("protectedmethod");
			return Application.getInstance().getImageRegistry().get("method");

		}

		if (element instanceof ASTClassNode) {
			if (((ASTClassNode) element).isTopLevel())
				return Application.getInstance().getImageRegistry().get("class");
			return Application.getInstance().getImageRegistry().get("innerclass");
		}

		if (element instanceof FileNode)
			return Application.getInstance().getImageRegistry().get("jfile");
		return null;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.viewers.ITableLabelProvider#getColumnText(java.lang.Object, int)
	 */
	public String getColumnText(Object element, int columnIndex) {
		StringBuilder sb = new StringBuilder();
		Formatter formatter = new Formatter(sb);

		boolean showFlat = coverageView.showFlat();
		boolean showPackages = coverageView.showPackageNames();
		boolean showFiles = coverageView.showFiles();
		boolean percentInfo = coverageView.showPercentage();

		Metrics m;
		// Zeilenname
		if (columnIndex == 0) {
			if (element instanceof String)
				return "Project " + (String) element;

			if (element instanceof IScopeNode) {
				if (showFlat) {
					return ((IScopeNode) element).getFullyQualifiedName(showPackages, showFiles);
				} else {
					if (showPackages) {
						if (element instanceof FileNode)
							return ((IScopeNode) element).getFullyQualifiedName(showPackages, showFiles);
						if (!showFiles && element instanceof ASTClassNode && ((ASTClassNode) element).isTopLevel())
							return ((IScopeNode) element).getFullyQualifiedName(showPackages, showFiles);
					}
					return ((IScopeNode) element).getName();

				}
			}
		}

		if (element instanceof String)
			m = Application.getInstance().getASTManager().getProjectMetrics();
		else if (element instanceof IScopeNode)
			m = ((IScopeNode) element).getMetrics();
		else
			return EMPTY;

		switch (columnIndex) {
		case 1:
			if (m.nrFunctions == 0 && !(element instanceof ASTFunctionNode))
				return EMPTY;
			formatter.format("%3.2f", m.feec);
			if (percentInfo)
				return sb.toString() + "%";
			return (m.nrEnteredFunctions + m.nrReachedExits) + "/" + (m.nrFunctions + m.nrFunctionExits);
		case 2:
			if (m.nrStatements == 0)
				return EMPTY;
			formatter.format("%3.2f", m.c0);
			if (percentInfo)
				return sb.toString() + "%";
			return m.nrCoveredStatements + "/" + m.nrStatements;
		case 3:
			if (m.nrBranches == 0)
				return EMPTY;
			formatter.format("%3.2f", m.c1);
			if (percentInfo)
				return sb.toString() + "%";
			return m.nrCoveredBranches + "/" + m.nrBranches;
		case 4:
			if (m.nrAtoms == 0)
				return EMPTY;
			formatter.format("%3.2f", m.c2);
			if (percentInfo)
				return sb.toString() + "%";
			return m.nrAtomEvaluations + "/" + 2 * m.nrAtoms;
		case 5:
			if (m.nrAtoms == 0)
				return EMPTY;
			formatter.format("%3.2f", m.mmdc);
			if (percentInfo)
				return sb.toString() + "%";
			return m.nrConditionEvaluations + "/" + 2 * m.nrConditions;
		case 6:
			if (m.nrAtoms == 0)
				return EMPTY;
			formatter.format("%3.2f", m.mcdc);
			if (percentInfo)
				return sb.toString() + "%";
			return m.nrMcdcPairs + "/" + m.nrAtoms;
		case 7:
			if (m.nrAtoms == 0)
				return EMPTY;
			formatter.format("%3.2f", m.c3);
			if (percentInfo)
				return sb.toString() + "%";
			return m.nrAtomCombinationEvaluations + "/" + m.nrAtomCombinations;
		case 8:
			if (m.nrModBIPaths == 0)
				return EMPTY;
			formatter.format("%3.2f", m.mbi);
			if (percentInfo)
				return sb.toString() + "%";
			return m.nrCoveredModBIPaths + "/" + m.nrModBIPaths;
		case 9:
			if (m.nrBIPaths == 0)
				return EMPTY;
			formatter.format("%3.2f", m.bi);
			if (percentInfo)
				return sb.toString() + "%";
			return m.nrCoveredBIPaths + "/" + m.nrBIPaths;
		default:
			return EMPTY;
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.viewers.IBaseLabelProvider#addListener(org.eclipse.jface.viewers.ILabelProviderListener)
	 */
	public void addListener(ILabelProviderListener listener) {
		// TODO Auto-generated method stub

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.viewers.IBaseLabelProvider#dispose()
	 */
	public void dispose() {
		// TODO Auto-generated method stub

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.viewers.IBaseLabelProvider#isLabelProperty(java.lang.Object, java.lang.String)
	 */
	public boolean isLabelProperty(Object element, String property) {
		// TODO Auto-generated method stub
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.jface.viewers.IBaseLabelProvider#removeListener(org.eclipse.jface.viewers.ILabelProviderListener)
	 */
	public void removeListener(ILabelProviderListener listener) {
		// TODO Auto-generated method stub

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.viewers.ITableColorProvider#getBackground(java.lang.Object, int)
	 */
	public Color getBackground(Object element, int columnIndex) {
		ProjectConfig config = Application.getInstance().getConfiguration();
		if (!config.showCoverage() || config.parsedSourceStatus == ProjectConfig.DIRTY)
			return null;
		Metrics m;
		if (element instanceof String)
			m = Application.getInstance().getASTManager().getProjectMetrics();
		else if (element instanceof IScopeNode)
			m = ((IScopeNode) element).getMetrics();
		else
			return null;

		switch (columnIndex) {
		case 1:
			if (m.nrFunctions == 0 && !(element instanceof ASTFunctionNode))
				return null;
			return percentage2Color(m.feec);
		case 2:
			if (m.nrStatements == 0)
				return null;
			return percentage2Color(m.c0);
		case 3:
			if (m.nrBranches == 0)
				return null;
			return percentage2Color(m.c1);
		case 4:
			if (m.nrAtoms == 0)
				return null;
			return percentage2Color(m.c2);
		case 5:
			if (m.nrAtoms == 0)
				return null;
			return percentage2Color(m.mmdc);
		case 6:
			if (m.nrAtoms == 0)
				return null;
			return percentage2Color(m.mcdc);
		case 7:
			if (m.nrAtoms == 0)
				return null;
			return percentage2Color(m.c3);
		case 8:
			if (m.nrModBIPaths == 0)
				return null;
			return percentage2Color(m.mbi);
		case 9:
			if (m.nrBIPaths == 0)
				return null;
			return percentage2Color(m.bi);
		default:
			return null;
		}
	}

	/**
	 * Ordnet einem berdeckungsgrad eine Farbe zu.
	 * 
	 * @param p
	 *            der berdeckungsgrad
	 * @return die Farbe
	 */
	private Color percentage2Color(float p) {
		return Application.getInstance().getColorManager().percentage2Color(p);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.viewers.ITableColorProvider#getForeground(java.lang.Object, int)
	 */
	public Color getForeground(Object element, int columnIndex) {
		return null;
	}
}
