package sota.gui.views;

import org.eclipse.draw2d.graph.DirectedGraph;
import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.text.source.SourceViewer;
import org.eclipse.jface.text.source.VerticalRuler;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.SashForm;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.ViewPart;

import sota.app.Application;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.util.ASTManager;
import sota.ast.util.ProjectConfig;
import sota.gui.GuiUpdateListener;
import sota.gui.graph.CFGPlotter;
import sota.gui.graph.GraphManager;

/**
 * Stellt die Ansicht des Kontrollflussgraphen dar.
 * 
 */
public class CFGView extends ViewPart implements GuiUpdateListener, ISelectionListener, IPropertyChangeListener {

	public static final String ID = "sota.cfgview";

	private CFGPlotter cfgPlotter;
	private GraphManager gcm;
	private ASTFunctionNode functionNode = null;
	private final boolean PIN_SOURCE = true;
	private final boolean SHOW_NUMBERS = true;

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createPartControl(Composite parent) {

		createActions();
		SashForm sash = new SashForm(parent, SWT.VERTICAL);

		gcm = new GraphManager();
		cfgPlotter = new CFGPlotter(sash, gcm);

		SourceViewer sourceViewer = new SourceViewer(sash, new VerticalRuler(10), SWT.V_SCROLL | SWT.H_SCROLL
				| SWT.BORDER | SWT.READ_ONLY);

		sourceViewer.getTextWidget().setWordWrap(false);
		gcm.setSourceViewer(sourceViewer);
		PlatformUI.getPreferenceStore().addPropertyChangeListener(this);

		sash.setWeights(new int[] { 5, 1 });

		getViewSite().getPage().addSelectionListener(this);
		Application.getInstance().getASTManager().addGuiUpdateListener(this);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#setFocus()
	 */
	public void setFocus() {
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.ISelectionListener#selectionChanged(org.eclipse.ui.IWorkbenchPart,
	 * org.eclipse.jface.viewers.ISelection)
	 */
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		// //System.out.println("Selection A");
		if (part instanceof ProjectView) {
			if (selection instanceof IStructuredSelection) {
				Object o = ((IStructuredSelection) selection).getFirstElement();
				if (o instanceof ASTFunctionNode) {
					functionNode = (ASTFunctionNode) o;
					gcm.refreshSourceViewer(null, functionNode);
					DirectedGraph graph = Application.getInstance().getASTManager().buildCFG(functionNode, gcm);
					cfgPlotter.setContents(graph);
					gcm.guiUpdate(Application.getInstance().getASTManager());
				} else {
					gcm.initSourceViewer();
					cfgPlotter.setContents(new DirectedGraph());
				}
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.lang.java.ast.util.GuiUpdateListener#guiUpdate(sota.ast.util.ASTManager)
	 */
	public void guiUpdate(ASTManager ast) {
		// //System.out.println("update A");
		if (ast.getConfiguration().getScopeNodeSelection() != functionNode) {
			// System.out.println("not functionNode");
			if (ast.getConfiguration().getScopeNodeSelection() instanceof ASTFunctionNode) {
				// System.out.println("is Function");
				functionNode = (ASTFunctionNode) ast.getConfiguration().getScopeNodeSelection();
				gcm.refreshSourceViewer(null, functionNode);
				DirectedGraph graph = Application.getInstance().getASTManager().buildCFG(functionNode, gcm);
				cfgPlotter.setContents(graph);
				gcm.guiUpdate(Application.getInstance().getASTManager());
			} else if (ast.getConfiguration().projectStatus >= ProjectConfig.PARSED) {
				// System.out.println("is parsed");
				gcm.initSourceViewer();
				cfgPlotter.setContents(new DirectedGraph());
			} else {
				// System.out.println("not parsed");
				gcm.clearSourceViewer();
				cfgPlotter.setContents(new DirectedGraph());
			}
		}
		// System.out.println("update X");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.util.IPropertyChangeListener#propertyChange(org.eclipse.jface.util.PropertyChangeEvent)
	 */
	public void propertyChange(PropertyChangeEvent event) {
		// System.out.println("PropertyChange A");
		if (event.getProperty().startsWith("CFG")) {
			if (Application.getInstance().getConfiguration().getScopeNodeSelection() instanceof ASTFunctionNode) {
				functionNode = (ASTFunctionNode) Application.getInstance().getConfiguration().getScopeNodeSelection();
				gcm.refreshSourceViewer(null, functionNode);
				DirectedGraph graph = Application.getInstance().getASTManager().buildCFG(functionNode, gcm);
				cfgPlotter.setContents(graph);
				gcm.guiUpdate(Application.getInstance().getASTManager());
			} else {
				gcm.initSourceViewer();
				cfgPlotter.setContents(new DirectedGraph());
			}
		}
		// System.out.println("PropertyChange X");
	}

	/**
	 * Erstellt die Aktionen fr diese Ansicht.
	 */
	private void createActions() {
		Action pinSourceViewAction = new Action("Pin SourceView", IAction.AS_CHECK_BOX) {
			public void run() {
				Application.getInstance().getConfiguration().setPinSourceView(isChecked());
				Application.getInstance().getASTManager().fireToGuiUpdateListeners();
			}

		};

		Action showPathNumbersAction = new Action("Show Number of Paths", IAction.AS_CHECK_BOX) {
			public void run() {
				// if((isChecked()&&!Application.getInstance().getConfiguration().showPathNumbers())||
				// (!isChecked()&&Application.getInstance().getConfiguration().showPathNumbers()))
				Application.getInstance().getConfiguration().showPathNumbers(isChecked());
				// //System.out.println(Application.getInstance().getConfiguration().showPathNumbers());
				Application.getInstance().getASTManager().fireToGuiUpdateListeners();
			}

		};

		Action zoomInAction = new Action("Zoom In", IAction.AS_PUSH_BUTTON) {
			public void run() {
				if (functionNode != null && Application.getInstance().getConfiguration().getZoomManager().zoomIn()) {
					DirectedGraph graph = Application.getInstance().getASTManager().buildCFG(functionNode, gcm);
					cfgPlotter.setContents(graph);
					gcm.guiUpdate(Application.getInstance().getASTManager());
				}
			}

		};

		Action zoomOutAction = new Action("Zoom Out", IAction.AS_PUSH_BUTTON) {
			public void run() {
				if (functionNode != null && Application.getInstance().getConfiguration().getZoomManager().zoomOut()) {
					DirectedGraph graph = Application.getInstance().getASTManager().buildCFG(functionNode, gcm);
					cfgPlotter.setContents(graph);
					gcm.guiUpdate(Application.getInstance().getASTManager());
				}
			}

		};

		pinSourceViewAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("pin"));
		pinSourceViewAction.setToolTipText("Pin SourceView");
		pinSourceViewAction.setChecked(PIN_SOURCE);
		Application.getInstance().getConfiguration().setPinSourceView(PIN_SOURCE);

		showPathNumbersAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor(
				"shownumbers"));
		showPathNumbersAction.setToolTipText("Show Number of Paths");
		showPathNumbersAction.setChecked(SHOW_NUMBERS);
		Application.getInstance().getConfiguration().showPathNumbers(SHOW_NUMBERS);

		zoomInAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("zoomin"));
		zoomInAction.setToolTipText("Zoom In");

		zoomOutAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("zoomout"));
		zoomOutAction.setToolTipText("Zoom Out");

		IToolBarManager tmgr = getViewSite().getActionBars().getToolBarManager();
		tmgr.add(showPathNumbersAction);
		tmgr.add(pinSourceViewAction);
		tmgr.add(zoomInAction);
		tmgr.add(zoomOutAction);
	}
}