package sota.gui.views;

import java.util.Comparator;

import org.eclipse.jface.action.Action;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.jface.viewers.DoubleClickEvent;
import org.eclipse.jface.viewers.IDoubleClickListener;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.jface.viewers.TreeViewer;
import org.eclipse.swt.SWT;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.part.ViewPart;

import sota.app.Application;
import sota.ast.nodes.IScopeNode;
import sota.ast.util.ASTManager;
import sota.gui.GuiUpdateListener;
import sota.gui.ProjectChangeListener;
import sota.gui.provider.IProjectTreeView;
import sota.gui.provider.ProjectContentProvider;
import sota.gui.provider.ProjectViewLabelProvider;

/**
 * Reprsentiert die Ansicht fr die Projektstrukturen.
 * 
 */
public class ProjectView extends ViewPart implements GuiUpdateListener, ProjectChangeListener, ISelectionListener,
		IProjectTreeView {
	public static final String ID = "sota.projectview";

	private Action expandAllAction, collapseAllAction, sortAction, showHierarchicalAction, showFlatAction;
	private TreeViewer tv;
	private boolean showFlat = false;
	private boolean showPackages = false;
	private boolean showFiles = true;
	private boolean showClasses = true;;
	private boolean showInnerClasses = true;
	private boolean reverseSort = false;

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createPartControl(Composite parent) {

		createActions();
		tv = new TreeViewer(parent, SWT.MULTI | SWT.H_SCROLL | SWT.V_SCROLL | SWT.BORDER);
		tv.setContentProvider(new ProjectContentProvider(this));
		tv.setLabelProvider(new ProjectViewLabelProvider(this));
		tv.setInput(Application.getInstance().getASTManager());
		Application.getInstance().getASTManager().addProjectChangeListener(this);
		Application.getInstance().getASTManager().addGuiUpdateListener(this);
		Application.getInstance().getASTManager().addProjectTreeView(this);
		getSite().setSelectionProvider(tv);
		getViewSite().getPage().addSelectionListener(this);

		tv.addDoubleClickListener(new IDoubleClickListener() {
			public void doubleClick(DoubleClickEvent event) {
				Application.getInstance().getASTManager().select(
						((IStructuredSelection) event.getSelection()).getFirstElement());
			}
		});

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.lang.java.ast.util.ProjectChangeListener#projectChanged(sota.ast.util.ASTManager)
	 */
	public void projectChanged(ASTManager ast) {
		tv.setInput(ast);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.lang.java.ast.util.GuiUpdateListener#guiUpdate(sota.ast.util.ASTManager)
	 */
	public void guiUpdate(ASTManager ast) {
		tv.refresh();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#setFocus()
	 */
	public void setFocus() {
		tv.getControl().setFocus();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.ISelectionListener#selectionChanged(org.eclipse.ui.IWorkbenchPart,
	 * org.eclipse.jface.viewers.ISelection)
	 */
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		if (part == this) {
			if (selection instanceof IStructuredSelection) {
				Application.getInstance().getConfiguration().setScopeNodeSelection(
						((IStructuredSelection) selection).getFirstElement());
			}
		} else if (part instanceof ISchemeView)
			if (selection instanceof IStructuredSelection) {
				tv.refresh();
			}
	}

	/**
	 * Erstellt die Aktionen fr diese Ansicht.
	 */
	private void createActions() {
		expandAllAction = new Action("Expand All") {
			public void run() {
				tv.getTree().setRedraw(false);
				tv.expandAll();
				tv.getTree().setRedraw(true);
			}
		};

		collapseAllAction = new Action("Collapse All") {
			public void run() {
				tv.getTree().setRedraw(false);
				tv.collapseAll();
				tv.getTree().setRedraw(true);
			}
		};

		sortAction = new Action("Sort") {
			public void run() {
				reverseSort = !reverseSort;
				tv.setInput(Application.getInstance().getASTManager());
			}
		};

		Action showPackageAction = new Action("Show PackageNames", IAction.AS_CHECK_BOX) {
			public void run() {
				tv.getTree().setRedraw(false);
				showPackages = !showPackages;
				tv.refresh();
				tv.getTree().setRedraw(true);
			}
		};

		Action showFilesAction = new Action("Show Files", IAction.AS_CHECK_BOX) {
			public void run() {
				tv.getTree().setRedraw(false);
				showFiles = !showFiles;
				tv.refresh();
				tv.getTree().setRedraw(true);
			}
		};

		Action showClassesAction = new Action("Show Classes", IAction.AS_CHECK_BOX) {
			public void run() {
				tv.getTree().setRedraw(false);
				showClasses = !showClasses;
				tv.refresh();
				tv.getTree().setRedraw(true);
			}
		};

		Action showInnerClassesAction = new Action("Show Inner Classes", IAction.AS_CHECK_BOX) {
			public void run() {
				tv.getTree().setRedraw(false);
				showInnerClasses = !showInnerClasses;
				tv.refresh();
				tv.getTree().setRedraw(true);
			}
		};

		showFlatAction = new Action("Flat Presentation", IAction.AS_RADIO_BUTTON) {
			public void run() {
				showFlat(true);
			}
		};

		showHierarchicalAction = new Action("Hierarchical Presentation", IAction.AS_RADIO_BUTTON) {
			public void run() {
				showFlat(false);
			}
		};

		showFlatAction.setChecked(showFlat);
		showHierarchicalAction.setChecked(!showFlat);

		showPackageAction.setChecked(showPackages);
		showFilesAction.setChecked(showFiles);
		showClassesAction.setChecked(showClasses);
		showInnerClassesAction.setChecked(showInnerClasses);

		sortAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("sort"));
		expandAllAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("expandall"));
		collapseAllAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("collapseall"));
		showPackageAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("package"));
		showFilesAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("jfile"));
		showClassesAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("class"));
		showInnerClassesAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor(
				"innerclass"));
		showFlatAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor("flat"));
		showHierarchicalAction.setImageDescriptor(Application.getInstance().getImageRegistry().getDescriptor(
				"hierarchical"));

		IToolBarManager tmgr = getViewSite().getActionBars().getToolBarManager();

		IMenuManager mm = getViewSite().getActionBars().getMenuManager();

		mm.add(showHierarchicalAction);
		mm.add(showFlatAction);
		mm.add(new Separator());
		mm.add(showPackageAction);
		mm.add(showFilesAction);
		mm.add(showClassesAction);
		mm.add(showInnerClassesAction);

		tmgr.add(sortAction);
		tmgr.add(expandAllAction);
		tmgr.add(collapseAllAction);

	}

	/**
	 * Legt fest, ob die Strukturen hierarchisch oder flach dargestellt werden sollen.
	 * 
	 * @param b
	 *            true, fr flache Darstellung, sonst false
	 */
	private void showFlat(boolean b) {
		tv.getTree().setRedraw(false);
		showFlat = b;
		showFlatAction.setChecked(b);
		showHierarchicalAction.setChecked(!b);
		tv.refresh();
		tv.getTree().setRedraw(true);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.gui.provider.IProjectTreeView#showProjectLine()
	 */
	public boolean showProjectLine() {
		return false;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.gui.provider.IProjectTreeView#showPackageNames()
	 */
	public boolean showPackageNames() {
		return showPackages;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.gui.provider.IProjectTreeView#showFiles()
	 */
	public boolean showFiles() {
		return showFiles;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.gui.provider.IProjectTreeView#showClasses()
	 */
	public boolean showClasses() {
		return showClasses;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.gui.provider.IProjectTreeView#showInnerClasses()
	 */
	public boolean showInnerClasses() {
		return showInnerClasses;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.gui.provider.IProjectTreeView#showFlat()
	 */
	public boolean showFlat() {
		return showFlat;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.gui.provider.IProjectTreeView#getComparator()
	 */
	@SuppressWarnings("unchecked")
	public Comparator getComparator() {
		return new Comparator() {

			@Override
			public int compare(Object firstObject, Object secondObject) {
				if (!(firstObject instanceof IScopeNode && secondObject instanceof IScopeNode))
					return 0;

				IScopeNode first = (IScopeNode) firstObject, second = (IScopeNode) secondObject;
				int factor = 1;
				if (reverseSort)
					factor = -1;

				return factor
						* first.getFullyQualifiedName(showPackageNames(), showFiles()).compareTo(
								second.getFullyQualifiedName(showPackageNames(), showFiles()));

			}
		};
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.gui.provider.IProjectTreeView#select(java.lang.Object)
	 */
	@Override
	public void select(Object o) {
		tv.setSelection(new StructuredSelection(o), true);
		// getViewSite().getSelectionProvider().setSelection(tv.getSelection());
	}
}