package sota.gui.views;

import java.util.ArrayList;

import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.Document;
import org.eclipse.jface.text.IDocumentPartitioner;
import org.eclipse.jface.text.rules.FastPartitioner;
import org.eclipse.jface.text.source.SourceViewer;
import org.eclipse.jface.text.source.VerticalRuler;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.ISelectionListener;
import org.eclipse.ui.IWorkbenchPart;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.part.ViewPart;

import sota.app.Application;
import sota.ast.nodes.ASTDoNode;
import sota.ast.nodes.ASTIfNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.FileNode;
import sota.ast.nodes.IScopeNode;
import sota.ast.nodes.IStatementNode;
import sota.ast.util.ASTManager;
import sota.ast.util.ColorManager;
import sota.ast.util.ProjectConfig;
import sota.gui.GuiUpdateListener;
import sota.gui.ProjectChangeListener;
import sota.gui.preferences.PreferenceConstants;
import sota.lang.LangClassProvider;

/**
 * Klasse fr die Darstellung des Quellcodes in einer Ansicht.
 * 
 */
public class SourceView extends ViewPart implements ISelectionListener, ProjectChangeListener, GuiUpdateListener,
		IPropertyChangeListener {

	public static String ID = "sota.sourceview";

	private Document fileDocument = new Document();
	private SourceViewer sourceViewer;
	private ArrayList<Integer> coveredLinesList;
	private ArrayList<Integer> notCoveredLinesList;
	private Display display;
	private Object currFileSelection = null;
	private IDocumentPartitioner partitioner = null;

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#createPartControl(org.eclipse.swt.widgets.Composite)
	 */
	public void createPartControl(Composite parent) {
		display = PlatformUI.getWorkbench().getDisplay();

		sourceViewer = new SourceViewer(parent, new VerticalRuler(10), SWT.V_SCROLL | SWT.H_SCROLL | SWT.BORDER
				| SWT.READ_ONLY);
		sourceViewer.getTextWidget().setWordWrap(false);
		sourceViewer.getTextWidget().setFont(new Font(display, new FontData("Courier New", 9, 0)));

		sourceViewer.setDocument(fileDocument);
		sourceViewer.getTextWidget().setIndent(10);

		PlatformUI.getPreferenceStore().addPropertyChangeListener(this);
		getViewSite().getPage().addSelectionListener(this);
		Application.getInstance().getASTManager().addProjectChangeListener(this);
		Application.getInstance().getASTManager().addGuiUpdateListener(this);
	}

	/**
	 * Initialisiert den Partitioner neu, wenn das Projekt gendert wurde.
	 */
	public void initializePartioner() {
		ProjectConfig config = Application.getInstance().getConfiguration();
		sourceViewer.unconfigure();
		try {
			sourceViewer.configure(LangClassProvider.getSourceViewerConfiguration(config).newInstance());

			if (partitioner != null)
				partitioner.disconnect();

			partitioner = new FastPartitioner(LangClassProvider.getPartitionScanner(config).newInstance(),
					LangClassProvider.getPartitionScannerTypes(config));

			partitioner.connect(fileDocument);
			fileDocument.setDocumentPartitioner(LangClassProvider.getPartitioning(config), partitioner);

		} catch (InstantiationException e) {
			// e.printStackTrace();
		} catch (IllegalAccessException e) {
			// e.printStackTrace();
		}

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.part.WorkbenchPart#setFocus()
	 */
	public void setFocus() {
		sourceViewer.getTextWidget().getShell().setFocus();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.ui.ISelectionListener#selectionChanged(org.eclipse.ui.IWorkbenchPart,
	 * org.eclipse.jface.viewers.ISelection)
	 */
	public void selectionChanged(IWorkbenchPart part, ISelection selection) {
		if (selection instanceof IStructuredSelection) {
			if (part instanceof ProjectView) {
				currFileSelection = ((IStructuredSelection) selection).getFirstElement();
				showSource();
			}
		}
	}

	/**
	 * Zeigt den Quellcode erneut an.
	 */
	public void showSource() {
		int startLineOffset, length;

		if (currFileSelection == null) {
			sourceViewer.setDocument(new Document(""));
			sourceViewer.refresh();
			return;
		}
		sourceViewer.getTextWidget().setRedraw(false);
		sourceViewer.resetVisibleRegion();

		// baue FileDocument neu auf
		if (currFileSelection instanceof IScopeNode) {
			IScopeNode sn = (IScopeNode) currFileSelection;
			buildFileDocument(sn.getFileNode());
			sourceViewer.setDocument(fileDocument);

			if (sn instanceof FileNode) {
				applyLineCoverage(sn.getFirstLineNr() - 1, sn.getLastLineNr());
				sourceViewer.setSelectedRange(0, 0);
				// zeige nur den Code der gewhlten Struktur
			} else {
				try {
					startLineOffset = fileDocument.getLineOffset(sn.getFirstLineNr() - 1);

					length = fileDocument.getLineOffset(sn.getLastLineNr()) - startLineOffset;
					sourceViewer.setVisibleRegion(startLineOffset, length);
					applyLineCoverage(sn.getFirstLineNr() - 1, sn.getLastLineNr());
				} catch (BadLocationException e) {
					e.printStackTrace();
				}
			}
		}
		sourceViewer.getTextWidget().setRedraw(true);
		sourceViewer.getTextWidget().redraw();
		Application.setFileDocument(fileDocument);
	}

	/**
	 * Appliziert Farben die Zeilen des Quellcodes, je nach berdeckung.
	 * 
	 * @param startLine
	 *            der Beginn des zu verarbeitenden Segmentes
	 * @param endLine
	 *            dessen Ende
	 */
	private void applyLineCoverage(int startLine, int endLine) {
		if (Application.getInstance().getASTManager().getSourceStatus() != ProjectConfig.DIRTY
				&& Application.getInstance().getConfiguration().showCoverage()) {
			for (Integer i : coveredLinesList) {
				if (i > startLine && i < endLine) {

					sourceViewer.getTextWidget().setLineBackground(i - 1 - startLine, 1,
							Application.getColorFromPreferenceStore(PreferenceConstants.SOURCE_LINE_HIT_COLOR));
				}
			}

			for (Integer i : notCoveredLinesList) {
				if (i > startLine && i < endLine) {

					sourceViewer.getTextWidget().setLineBackground(i - 1 - startLine, 1,
							Application.getColorFromPreferenceStore(PreferenceConstants.SOURCE_LINE_MISS_COLOR));
				}
			}
		} else
			sourceViewer.getTextWidget().setLineBackground(0, sourceViewer.getTextWidget().getLineCount(),
					Application.getColor(ColorManager.WHITE));
		sourceViewer.getTextWidget().redraw();
	}

	// erzwinge Neuerstellen des SourceTextes
	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.lang.java.ast.util.ProjectChangeListener#projectChanged(sota.ast.util.ASTManager)
	 */
	public void projectChanged(ASTManager ast) {
		currFileSelection = null;
		if (ast.getConfiguration().projectStatus >= ProjectConfig.PARSED)
			initializePartioner();
		else
			showSource();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.lang.java.ast.util.GuiUpdateListener#guiUpdate(sota.ast.util.ASTManager)
	 */
	public void guiUpdate(ASTManager ast) {
		showSource();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.util.IPropertyChangeListener#propertyChange(org.eclipse.jface.util.PropertyChangeEvent)
	 */
	public void propertyChange(PropertyChangeEvent evt) {
		if (evt.getProperty().startsWith("source")) {
			sourceViewer.unconfigure();
			try {
				sourceViewer.configure(LangClassProvider.getSourceViewerConfiguration(
						Application.getInstance().getConfiguration()).newInstance());
			} catch (InstantiationException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			}
			showSource();
		}
	}

	/**
	 * Erstellt ein Document aus einem FileNode.
	 * 
	 * @param fn
	 *            der FileNode
	 */
	public void buildFileDocument(FileNode fn) {
		FileNode fileNode = fn;
		final ASTNode sn = fileNode.getStartNode();
		final ASTNode en = fileNode.getEndNode();

		final StringBuffer s = new StringBuffer();
		coveredLinesList = new ArrayList<Integer>();
		notCoveredLinesList = new ArrayList<Integer>();

		ASTNode n = sn;
		while (n != en) {

			s.append(n.getContent());

			if (n instanceof IStatementNode && !n.isVirtual()) {
				if (((IStatementNode) n).getNrHits() > 0)
					coveredLinesList.add(n.getLineNr());
				else
					notCoveredLinesList.add(n.getLineNr());
			} else if (n instanceof ASTDoNode) {
				if (((ASTDoNode) n).getIterationNode().getNrHits() > 0)
					coveredLinesList.add(n.getLineNr());
				else
					notCoveredLinesList.add(n.getLineNr());
			}
			if (n instanceof ASTIfNode && !((ASTIfNode) n).getElseNode().isVirtual()) {
				if (((ASTIfNode) n).getFalseNode().getNrHits() > 0)
					coveredLinesList.add(((ASTIfNode) n).getElseNode().getLineNr());
				else
					notCoveredLinesList.add(((ASTIfNode) n).getElseNode().getLineNr());
			}

			n = n.getNextNode();
		}

		s.append(n.getContent() + "\n\n");

		fileDocument.set(s.toString());

	}

	/**
	 * Liefert die Liste der berdeckten Zeilen.
	 * 
	 * @return die Liste der berdeckten Zeilen
	 */
	public ArrayList<Integer> getCoveredLinesList() {
		return coveredLinesList;
	}

	/**
	 * Liefert die Liste der nicht berdeckten Zeilen.
	 * 
	 * @return die Liste der nicht berdeckten Zeilen
	 */
	public ArrayList<Integer> getNotCoveredLinesList() {
		return notCoveredLinesList;
	}

	/**
	 * Liefert das aktuelle Document
	 * 
	 * @return das Document
	 */
	public Document getFileDocument() {
		return fileDocument;
	}

}
