package sota.lang.java.ast.util;

import java.util.Vector;

import sota.ast.nodes.Condition;

/**
 * bernimmt die Formattierung der Instrumentationsanweisungen nach einem einheilichen Schema.
 * 
 */
public class JavaASCFormatter {

	// die Definition der zu verwendenden Elemente der Instrumentierung

	/** Linebreak-Zeichen */
	private static final String NL = "\n";

	/** Tabulator-Zeichen */
	private static final String TAB = "\t";

	/** der Beginn eines Kommentars */
	private static final String STARTCOMMENT = " /* ";

	/** das Ende eines Kommentars */
	private static final String ENDCOMMENT = " */ ";

	/** der Prfix fr eine Kommentarzeile */
	private static final String PREFIX = "ASC";

	/** der Name der Logfunktion */
	private static final String LOGFUNCTION = "ASCLogger.log";

	/** der Name der Funktion, die den Zeitstempel liefert */
	private static final String TSFUNCTION = "ASCLogger.getTimeStamp()";

	/** der Name der Variablen, die den Zeitstempel liefert */
	private static final String PATHIDVAR = "asc_path_id";

	/** der Prfix der boolschen Variablen */
	private static final String BOOLVARHEAD = "asc_boolean_";

	/** die erste (und dritte) Introzeile */
	private static final String INTRO0 = " /*************************************************************/\n";

	/** die zweite Introzeile */
	private static final String INTRO1 = "Instrumented source code created by SOTA v1.0";

	/**
	 * Erstellt das Intro der Datei.
	 * 
	 * @return der erstellte String
	 */
	public static String insertIntro() {
		return INTRO0 + comment(PREFIX + "      " + INTRO1 + "     ") + NL + INTRO0;

	}

	/**
	 * Fgt den Logger-Import in die Datei ein.
	 * 
	 * @return der erstellte String
	 */
	public static String insertASCLoggerImport() {
		return NL + comment(PREFIX) + TAB + "import " + "sota.ASCLogger" + ";" + TAB + comment("Logger") + NL;
	}

	/**
	 * Deklariert am Funktionsanfang die ntigen Variablen.
	 * 
	 * @param nrPrimaryConditions
	 *            die Anzahl an primren Bedingungen
	 * @param nrAtoms
	 *            die Anzahl an Atomen
	 * @return der erstellte String
	 */
	public static String declarationOfBooleans(int nrPrimaryConditions, Vector<Integer> nrAtoms) {
		String s = NL + comment(PREFIX + TAB + "BEGIN DECLARATION OF BOOLEANS");
		for (int i = 0; i < nrPrimaryConditions; i++)
			for (int j = 0; j < nrAtoms.get(i); j++)
				s += NL + comment(PREFIX) + TAB + "boolean " + BOOLVARHEAD + i + "_" + j + " = true;";
		s += NL + comment(PREFIX + TAB + "END DECLARATION OF BOOLEANS");
		return s + NL;
	}

	/**
	 * Liefert einen String mit analysierten Logkommandos fr einen ConditionNode.
	 * 
	 * @param id
	 *            die zu loggende ID
	 * @param suffix
	 *            der erluternde Kommentar, z.b. die ID
	 * @param cond
	 *            die Bedingung
	 * @return der erstellte String
	 */
	public static String logCondition(int id, String suffix, Condition cond) {
		if (cond == null)
			return NL + comment(PREFIX + TAB + "SKIPPING FOR-EACH-CONDITION");

		String s = "";
		String[] _s;

		s += NL + comment(PREFIX + TAB + "BEGIN LOGGING CONDITION");

		_s = logAtoms(cond);
		s += _s[0];

		s += NL + comment(PREFIX + TAB + "LOGGING OF CONDITIONS OWNER");

		s += logId(id, suffix);
		s += NL + comment(PREFIX + TAB + "LOGGING OF ATOMS");
		s += _s[1];

		// erste Boolvariable ist bei komplexen Bedingungen auch
		// Auswertungsvariable
		if (!cond.isAtom()) {
			s += NL + comment(PREFIX + TAB + "FINAL CONDITION EVALUATION");
			s += NL + comment(PREFIX) + TAB + getBooleanName(cond, 0) + " = " + conditionsToASCString(cond) + ";";
		}
		s += NL + comment(PREFIX + TAB + "END LOGGING CONDITION") + NL;
		return s;
	}

	/**
	 * Liefert rekursiv analysierte Logkommandos - s[0] liefert die Variablenauswertung, s[1] liefert die Logstrings -
	 * dazwischen muss fr Schleifenupdate der iteratingNode geloggt werden!
	 * 
	 * @param cond
	 *            die Bedingung
	 * @return der erstellte String
	 */
	private static String[] logAtoms(Condition cond) {
		String[] s = { "", "" };
		String[] _s;

		if (cond == null)
			return s;

		if (cond.isEmpty()) {
			String VARNAME = getBooleanName(cond);
			s[0] += NL + comment(PREFIX) + TAB + VARNAME + " = true" + ";";
			s[1] += NL + comment(PREFIX + TAB + "(No need to log true-Condition)");
			return s;
		}

		if (cond.isAtom()) {
			String VARNAME = getBooleanName(cond);
			s[0] += NL + comment(PREFIX) + TAB + VARNAME + " = " + cond.getContent() + ";";
			s[1] += NL + comment(PREFIX) + TAB + LOGFUNCTION + "(" + VARNAME + ");";
			return s;
		}

		if (cond.isOperator()) {
			// unrer Operator -> nur rechts Auswerten
			if (cond.getLeftCondition() == null)
				return logAtoms(cond.getRightCondition());

			if (!cond.hasShortCircuit()) {
				_s = logAtoms(cond.getLeftCondition());
				s[0] += _s[0];
				s[1] += _s[1];
				_s = logAtoms(cond.getRightCondition());
				s[0] += _s[0];
				s[1] += _s[1];
				return s;
			}

			_s = logAtoms(cond.getLeftCondition());
			s[0] += _s[0];
			s[1] += _s[1];

			if (cond.getShortCircuitValue())
				s[0] += NL + comment(PREFIX) + TAB + "if( ! " + conditionsToASCString(cond.getLeftCondition()) + ") {";
			else
				s[0] += NL + comment(PREFIX) + TAB + "if( " + conditionsToASCString(cond.getLeftCondition()) + ") {";
			_s = logAtoms(cond.getRightCondition());
			s[0] += _s[0] + "}";
			s[1] += _s[1];
		}
		return s;
	}

	/**
	 * Liefert den For-Init-String.
	 * 
	 * @param s
	 *            der Initialisierungsausdruck
	 * @return der erstellte String
	 */
	public static String insertForInit(String s) {
		return NL + comment(PREFIX) + TAB + s + ";" + TAB + comment("ForInit");
	}

	/**
	 * Liefert den For-Update-String.
	 * 
	 * @param s
	 *            die Update-Anweisung
	 * @return der erstellte String
	 */
	public static String insertForUpdate(String s) {
		return NL + comment(PREFIX) + TAB + s + ";" + TAB + comment("ForUpdate");
	}

	/**
	 * Liefert den String fr die Auswertung von If-Knoten.
	 * 
	 * @param prefix
	 *            der Prefix der Auswertung - "if("
	 * @param cond
	 *            die Bedingung
	 * @param postfix
	 *            der Postfix der Auswertung - ")"
	 * @param suffix
	 *            der Kommentar fr diese Quellcodezeile
	 * @return der erstellte String
	 */
	public static String insertIfStatement(String prefix, Condition cond, String postfix, String suffix) {
		String s = NL + comment(PREFIX) + TAB + prefix;
		s += getBooleanName(cond, 0);
		s += postfix + TAB + comment(suffix);
		return s;
	}

	/**
	 * Liefert den String fr den Variablennamen zur Auswertung einer Bedingung.
	 * 
	 * @param c
	 *            die Bedingung
	 * @param a
	 *            der Index
	 * @return der erstellte String
	 */
	public static String getBooleanName(Condition c, int a) {
		return BOOLVARHEAD + c.getConditionID() + "_" + a;
	}

	/**
	 * Liefert den String fr den Variablennamen zur Auswertung einer Bedingung.
	 * 
	 * @param c
	 *            die Bedingung
	 * @return der erstellte String
	 */
	public static String getBooleanName(Condition c) {
		return BOOLVARHEAD + c.getConditionID() + "_" + c.getAtomID();
	}

	/**
	 * Liefert die String-Darstellung einer Bedingung.
	 * 
	 * @param c
	 *            die Bedingung
	 * @return der erstellte String
	 */
	public static String conditionsToASCString(Condition c) {
		if (c.isEmpty())
			return "true";
		if (c.isAtom())
			return getBooleanName(c);
		if (c.isOperator()) {
			if (c.getLeftCondition() != null)
				return "(" + conditionsToASCString(c.getLeftCondition()) + " " + c.getContent() + " "
						+ conditionsToASCString(c.getRightCondition()) + ")";
			return c.getContent() + conditionsToASCString(c.getRightCondition());
		}
		return "";
	}

	/**
	 * Liefert den Log-String fr eine ID.
	 * 
	 * @param id
	 *            die zu loggende ID
	 * @param suffix
	 *            ein Kommentar als Suffix
	 * @return der erstellte String
	 */
	public static String logId(int id, String suffix) {
		return NL + comment(PREFIX) + TAB + LOGFUNCTION + "(" + id + "," + PATHIDVAR + ");" + TAB + comment(suffix)
				+ NL;
	}

	/**
	 * Liefert den String, der eine Quellcodezeile einfgt.
	 * 
	 * @param s
	 *            der einzufgende String
	 * @return der erstellte String
	 */
	public static String insert(String s) {
		return NL + comment(PREFIX) + TAB + s + NL;
	}

	/**
	 * Liefert den String, der eine Quellcodezeile mit Kommentar einfgt.
	 * 
	 * @param s
	 *            der einzufgende String
	 * @param suffix
	 *            der Kommentar
	 * @return der erstellte String
	 */
	public static String insert(String s, String suffix) {
		return NL + comment(PREFIX) + TAB + s + TAB + comment(suffix) + NL;
	}

	/**
	 * Liefert den String zur Defintion der Pfadvariablen (Zeitstempel).
	 * 
	 * @return der erstellte String
	 */
	public static String insertPIDInit() {
		return NL + comment(PREFIX) + TAB + "int " + PATHIDVAR + " = " + TSFUNCTION + ";" + NL;
	}

	/**
	 * Liefert den String, der einen Kommentar in den Quellcode einfgt
	 * 
	 * @param comment
	 *            der Kommentar
	 * @return der erstellte String
	 */
	public static String insertComment(String comment) {
		return NL + comment(PREFIX + TAB + comment) + NL;
	}

	/**
	 * Liefert den String, der einen Block beginnt.
	 * 
	 * @return der erstellte String
	 */
	public static String startCompound() {
		return NL + comment(PREFIX) + TAB + "{";
	}

	/**
	 * Liefert den String, der einen Block beendet.
	 * 
	 * @return der erstellte String
	 */
	public static String endCompound() {
		return NL + comment(PREFIX) + TAB + "}";
	}

	/**
	 * Liefert den String, der einen Kommentar einfasst.
	 * 
	 * @param s
	 *            der Kommentar
	 * @return der erstellte String
	 */
	private static String comment(String s) {
		return STARTCOMMENT + s + ENDCOMMENT;
	}

}
