package sota.lang.java.ast.util;

import java.io.FileInputStream;
import java.util.HashMap;
import java.util.Stack;

import sota.Logger;
import sota.ast.nodes.ASTClassNode;
import sota.ast.nodes.ASTCompoundNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIfNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.ASTLabelNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.ASTStatementNode;
import sota.ast.nodes.ASTSwitchNode;
import sota.ast.nodes.ASTTryNode;
import sota.ast.nodes.FileNode;
import sota.ast.nodes.IScopeNode;
import sota.ast.util.ASTManager;
import sota.ast.util.ProjectConfig;
import sota.lang.IASTBuilder;
import sota.lang.java.ast.nodes.JavaASTCaseNode;
import sota.lang.java.ast.nodes.JavaASTCatchNode;
import sota.lang.java.ast.nodes.JavaASTClassNode;
import sota.lang.java.ast.nodes.JavaASTCompoundNode;
import sota.lang.java.ast.nodes.JavaASTDoNode;
import sota.lang.java.ast.nodes.JavaASTFunctionNode;
import sota.lang.java.ast.nodes.JavaASTIfNode;
import sota.lang.java.ast.nodes.JavaASTIterationNode;
import sota.lang.java.ast.nodes.JavaASTJumpNode;
import sota.lang.java.ast.nodes.JavaASTLabelNode;
import sota.lang.java.ast.nodes.JavaASTStatementNode;
import sota.lang.java.ast.nodes.JavaASTSwitchNode;
import sota.lang.java.ast.nodes.JavaASTTryNode;
import sota.lang.java.ast.nodes.JavaCondition;
import sota.lang.java.ast.nodes.JavaFileNode;
import sota.lang.java.ast.parser.JavaParser;
import sota.lang.java.ast.parser.ParseException;
import sota.lang.java.ast.parser.TokenMgrError;

/**
 * Erstellt fr die bergebenen Dateien den AST und verknpft die Knoten geeignet.
 * 
 */
public class JavaASTBuilder implements IASTBuilder {

	private FileNode currFileNode;
	private Stack<ASTNode> stackOfNodes;
	private JavaSymbolTable st;
	private JavaJumpManager jm;
	private HashMap<String, IScopeNode> scopeMap; // Zuordnung von Scope
	private FileInputStream input;
	private ProjectConfig config;
	private ASTManager ast;
	private static TermData currTerm = new TermData();
	private static Stack<TermData> currStack;
	private int nrAnonymousClasses;
	private int nrEnteredStatements = 0;

	/**
	 * Konstruktor des JavaASTBuilders
	 */
	public JavaASTBuilder() {
		nrAnonymousClasses = 0;
		st = new JavaSymbolTable();
		jm = new JavaJumpManager();
		scopeMap = new HashMap<String, IScopeNode>();
		stackOfNodes = new Stack<ASTNode>();
		currStack = new Stack<TermData>();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see sota.lang.IASTBuilder#buildAST(sota.ast.util.ASTManager, java.lang.String)
	 */
	public boolean buildAST(ASTManager manager, String fileName) {

		ast = manager;
		config = ast.getConfiguration();

		if (!config.language.equals("Java"))
			return false;

		st = new JavaSymbolTable();

		try {
			input = new FileInputStream(fileName);
		} catch (java.io.FileNotFoundException e) {
			Logger.log("couldn't be found!");
			return false;
		}

		currFileNode = new JavaFileNode(fileName, ast.getAllNodes().size());

		scopeMap.put(currFileNode.getFullName(), currFileNode);
		// scopeNodeStack.push(fn);
		ast.addFileNode(currFileNode);

		jm = new JavaJumpManager();
		jm.addScopeNode(currFileNode);
		try {
			JavaParser.parse(input, this);
		} catch (TokenMgrError e) {
			Logger.logln(e.getLocalizedMessage());
			ast.removeFileNode(currFileNode);
			return false;
		} catch (ParseException e) {
			Logger.logln(e.getLocalizedMessage());
			ast.removeFileNode(currFileNode);
			return false;
		}

		jm.linkNodes(ast);
		jm.removeScopeNode();

		// Logger.logln("x: " +
		// currFileNode.getFunctionNodes().size());

		for (ASTFunctionNode fn : currFileNode.getFunctionNodes())
			fn.completeFunction(ast);

		return true;
	}

	/**
	 * Liefert den letzten erstellten Knoten.
	 * 
	 * @return der letzte erstellte Knoten
	 */
	public ASTNode getLastNode() {
		return ast.getLastNode();
	}

	/**
	 * Setzt die Datei auf instrumentiert.
	 */
	public void isInstrumented() {
		Logger.logln("\t\t " + currFileNode.getFullName() + " is instrumented.");
		currFileNode.isInstrumented(true);
	}

	/**
	 * Fgt einen virutellen (leeren) ASTNode an.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @return der erstellte ASTNode
	 */
	public ASTNode appendNode(int ln) {
		ASTNode n = new ASTNode(ln);
		if (!withinStatement()) {
			n.setFunctionNode(jm.getCurrFunctionNode());

			ast.append(n);
		}
		return n;
	}

	/**
	 * Fgt einen nicht-virtuellen ASTNode ein.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt des Strings
	 * @return der erstellte ASTNode
	 */
	public ASTNode appendNode(int ln, String s) {
		ASTNode n = new ASTNode(ln, s);
		if (!withinStatement()) {
			n.setFunctionNode(jm.getCurrFunctionNode());
			ast.append(n);
		}
		return n;
	}

	/**
	 * Fgt einen ASTCatchNode ein.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param e
	 *            der Case-Ausdruck
	 * @return der ASTCatchNode
	 */
	public JavaASTCatchNode appendCatch(int ln, String s, String e) {
		JavaASTCatchNode n = new JavaASTCatchNode(ln, s, e);
		if (!withinStatement()) {
			n.setFunctionNode(jm.getCurrFunctionNode());
			ast.append(n);
		}
		return n;
	}

	/**
	 * Fgt einen ASTCatchNode fr ein Finally ein.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @return der ASTCatchNode
	 */
	public JavaASTCatchNode appendFinally(int ln, String s) {
		JavaASTCatchNode n = new JavaASTCatchNode(ln, s);
		if (!withinStatement()) {
			n.setFunctionNode(jm.getCurrFunctionNode());
			ast.append(n);
		}
		return n;
	}

	/**
	 * Fgt den Beginn eines Blocks ein.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @return der ASTCompoundNode
	 */
	public JavaASTCompoundNode beginCompound(int ln, String s) {
		JavaASTCompoundNode cn = new JavaASTCompoundNode(ln, s);
		if (!withinStatement()) {
			cn.setFunctionNode(jm.getCurrFunctionNode());
			stackOfNodes.push(cn);

			ast.append(cn);
		}
		return cn;
	}

	/**
	 * Fgt das Ende eines Blocks ein und verknpft mit zugehrigen ASTCompoundNode.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @return der ASTCompoundNode
	 */
	public JavaASTCompoundNode endCompound(int ln, String s) {
		JavaASTCompoundNode en = new JavaASTCompoundNode(ln, s);
		if (!withinStatement()) {
			ast.append(en);
			en.setFunctionNode(jm.getCurrFunctionNode());

			if (stackOfNodes.peek() instanceof ASTCompoundNode)
				((ASTCompoundNode) stackOfNodes.pop()).setEndNode(en);
			else
				Logger.logln("missing start compound");
		}
		return en;
	}

	/**
	 * Fgt eine virtuelle Anweisung ein.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @return der ASTStatementNode
	 */
	public JavaASTStatementNode appendStatement(int ln) {
		JavaASTStatementNode sn = new JavaASTStatementNode(ln);
		if (!withinStatement()) {
			ast.append(sn);
			sn.setFunctionNode(jm.getCurrFunctionNode());
		}
		return sn;
	}

	/**
	 * Fgt eine nicht-virtuelle Anweisung ein.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @return der ASTStatementNode
	 */
	public JavaASTStatementNode appendStatement(int ln, String s) {
		JavaASTStatementNode sn = new JavaASTStatementNode(ln, s);
		if (!withinStatement()) {
			ast.append(sn);
			sn.setFunctionNode(jm.getCurrFunctionNode());
		}

		return sn;
	}

	/**
	 * Fgt einen Sprung ein.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param l
	 *            das Label
	 * @return der ASTJumpNode
	 */
	public JavaASTJumpNode appendJump(int ln, String s, String l) {
		JavaASTJumpNode n = new JavaASTJumpNode(ln, s, l);
		if (!withinStatement()) {
			n.setFunctionNode(jm.getCurrFunctionNode());
			ast.append(n);
		}
		return n;
	}

	/**
	 * Fgt ein If ein.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param cond
	 *            die Bedingung
	 * @return der ASTIfNode
	 */
	public JavaASTIfNode beginIf(int ln, String s, JavaCondition cond) {

		ASTNode entryNode = ast.getLastNode();
		JavaASTIfNode n = new JavaASTIfNode(ln, s, cond);
		if (!withinStatement()) {
			if (entryNode instanceof ASTLabelNode) {
				while (entryNode.getPrevNode() instanceof ASTLabelNode)
					entryNode = entryNode.getPrevNode();
			} else
				entryNode = appendNode(ln);

			n.setEntryNode(entryNode);
			n.setFunctionNode(jm.getCurrFunctionNode());
			jm.add(n);
			ast.append(n);
		}
		return n;
	}

	/**
	 * Beendet eine If-Struktur
	 * 
	 * @param ln
	 *            die Zeilennummer
	 */
	public void endIf(int ln) {
		if (!withinStatement()) {
			ASTStatementNode endNode = appendStatement(ln);
			ASTNode n = jm.remove();
			if (n instanceof ASTIfNode)
				((ASTIfNode) n).setEndNode(endNode);
			else
				Logger.logln("error@astbuilder - no if on stack");
		}
	}

	/**
	 * Beginnt ein Package.
	 * 
	 * @param packageName
	 *            der Packagename
	 */
	public void openPackage(String packageName) {
		if (!withinStatement()) {
			st.openPackage(packageName);
			currFileNode.setPackageName(packageName);
			addImportedPackage(packageName);
		}
	}

	/**
	 * Fgt importierte Packages hinzu.
	 * 
	 * @param packageName
	 *            der PackageName
	 */
	public void addImportedPackage(String packageName) {
		st.addImportedPackage(packageName);
	}

	/**
	 * Fgt importierte Klasse hinzu.
	 * 
	 * @param className
	 *            die Klasse
	 */
	public void addImportedClass(String className) {
		st.addImportedClass(className);
	}

	/**
	 * Fgt einen Typ hinzu.
	 * 
	 * @param typeName
	 *            der Name des Typs
	 */
	public void addTypeName(String typeName) {
		st.putType(typeName);
	}

	/**
	 * ffnet einen neuen Scope.
	 * 
	 * @param s
	 *            der Scopename
	 */
	public void openScope(String s) {
		st.openScope(s);
	}

	/**
	 * ffnet einen anonymen Scope.
	 */
	public void openScope() {
		st.openScope();
	}

	/**
	 * Schliet einen Scope.
	 */
	public void closeScope() {
		st.closeScope();
	}

	/**
	 * Beginnt ein Switch.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Ausdruck im Switch
	 * @return der ASTSwitchNode
	 */
	public JavaASTSwitchNode beginSwitch(int ln, String s) {

		ASTNode entryNode = ast.getLastNode();
		JavaASTSwitchNode n = new JavaASTSwitchNode(ln, s);
		if (!withinStatement()) {
			if (entryNode instanceof ASTLabelNode) {
				while (entryNode.getPrevNode() instanceof ASTLabelNode)
					entryNode = entryNode.getPrevNode();
			} else
				entryNode = appendNode(ln);

			n.setEntryNode(entryNode);
			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());
			jm.add(n);
		}
		return n;
	}

	/**
	 * Beendet Switch-Block
	 * 
	 * @param ln
	 *            die Zeilennummer
	 */
	public void endSwitch(int ln) {
		if (!withinStatement()) {
			ASTStatementNode endNode = appendStatement(ln);
			ASTNode n = jm.remove();
			if (n instanceof ASTSwitchNode)
				((ASTSwitchNode) n).setEndNode(endNode);
			else
				Logger.logln("error@astbuilder - no switch on stack");
		}
	}

	/**
	 * Fgt ein Default-Case an.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @return der ASTCaseNode
	 */
	public JavaASTCaseNode appendCase(int ln, String s) {
		JavaASTCaseNode n = new JavaASTCaseNode(ln, s);
		if (!withinStatement()) {
			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());

			jm.getSwitchNode().addCaseNode(n);
		}
		return n;
	}

	/**
	 * Fgt ein normales Case an.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param c
	 *            der Case-Ausdruck
	 * @return der ASTCaseNode
	 */
	public JavaASTCaseNode appendCase(int ln, String s, String c) {
		JavaASTCaseNode n = new JavaASTCaseNode(ln, s, c);
		if (!withinStatement()) {
			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());

			jm.getSwitchNode().addCaseNode(n);
		}
		return n;
	}

	/**
	 * Beginnt eine While-Schleife.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param cond
	 *            die Bedingung
	 * @return der ASTIterationNode
	 */
	public JavaASTIterationNode beginWhileIteration(int ln, String s, JavaCondition cond) {
		ASTNode entryNode = ast.getLastNode();
		JavaASTIterationNode n = new JavaASTIterationNode(ln, s, cond);
		if (!withinStatement()) {
			if (entryNode instanceof ASTLabelNode) {
				while (entryNode.getPrevNode() instanceof ASTLabelNode)
					entryNode = entryNode.getPrevNode();
			} else
				entryNode = appendNode(ln);

			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());

			jm.add(n);
		}
		return n;
	}

	/**
	 * Beginnt eine For-Each-Schleife.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @return der ASTIterationNode
	 */
	public JavaASTIterationNode beginForEachIteration(int ln, String s) {

		ASTNode entryNode = ast.getLastNode();
		JavaASTIterationNode n = new JavaASTIterationNode(ln, s);
		if (!withinStatement()) {
			if (entryNode instanceof ASTLabelNode) {
				while (entryNode.getPrevNode() instanceof ASTLabelNode)
					entryNode = entryNode.getPrevNode();
			} else
				entryNode = appendNode(ln);

			ast.append(n);
			// jm.getCurrFunctionNode().addConditionNode(n);
			n.setFunctionNode(jm.getCurrFunctionNode());

			jm.add(n);
		}
		return n;
	}

	/**
	 * Beginnt eine vollstndige For-Schleife.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param s1
	 *            der For-Init-Ausdruck
	 * @param cond
	 *            die Bedingung
	 * @param s2
	 *            der For-Update-Ausdruck
	 * @return der ASTIterationNode
	 */
	public JavaASTIterationNode beginForIteration(int ln, String s, String s1, JavaCondition cond, String s2) {

		ASTNode entryNode = ast.getLastNode();
		if (cond == null)
			cond = new JavaCondition();
		JavaASTIterationNode n = new JavaASTIterationNode(ln, s, s1, cond, s2);
		if (!withinStatement()) {
			if (entryNode instanceof ASTLabelNode) {
				while (entryNode.getPrevNode() instanceof ASTLabelNode)
					entryNode = entryNode.getPrevNode();
			} else
				entryNode = appendNode(ln);

			// jm.getCurrFunctionNode().addConditionNode(n);
			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());

			jm.add(n);
		}
		return n;
	}

	/**
	 * Begint Do-While-Iteration.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @return der ASTDoNode
	 */
	public JavaASTDoNode beginDoIteration(int ln, String s) {

		ASTNode entryNode = ast.getLastNode();
		JavaASTDoNode n = new JavaASTDoNode(ln, s);
		if (!withinStatement()) {
			if (entryNode instanceof ASTLabelNode) {
				while (entryNode.getPrevNode() instanceof ASTLabelNode)
					entryNode = entryNode.getPrevNode();
			} else
				entryNode = appendNode(ln);

			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());

			jm.add(n);
		}
		return n;
	}

	/**
	 * Beendet Do-While-Iteration.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param cond
	 *            die Bedingung
	 * @param endLn
	 *            die Zeilennummer des Endes
	 * @return der ASTIterationNode
	 */
	public JavaASTIterationNode endDoIteration(int ln, String s, JavaCondition cond, int endLn) {
		JavaASTIterationNode n = new JavaASTIterationNode(ln, s, cond);

		if (!withinStatement()) {
			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());
			n.setEndNode(appendStatement(endLn));
			jm.remove();
		}
		return n;
	}

	/**
	 * Beendet eine Iteration.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 */
	public void endIteration(int ln) {
		if (!withinStatement()) {
			ASTNode n = jm.remove();
			((ASTIterationNode) n).setEndNode(appendStatement(ln));
		}
	}

	/**
	 * Fgt ein Label hinzu.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param l
	 *            das Label
	 */
	public void appendLabel(int ln, String s, String l) {
		if (!withinStatement()) {
			JavaASTLabelNode n = new JavaASTLabelNode(ln, s, l);
			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());
			jm.add(n);
		}
	}

	/**
	 * Fgt eine Throw-Anweisung hinzu.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param l
	 *            das Geworfene
	 */
	public void appendThrow(int ln, String s, String l) {
		if (!withinStatement()) {

			JavaASTJumpNode n = new JavaASTJumpNode(ln, s, l);

			n.isThrow(true);
			n.addTryNodes(jm.getStackOfTryNodes());
			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());
			// Throw geht auf Funktionsende, wenn kein Try offen ist, oder Throw erst im Catch-Block vorkommt
			if (jm.getStackOfTryNodes().size() == 0
					|| (jm.getTryNode().getEndOfTryBlock() != null && jm.getTryNode().getEndOfTryBlock().getID() < n
							.getID())) {
				n.setTarget(jm.getCurrFunctionNode());
			} else
				n.setTarget(jm.getTryNode());
			jm.add(n);
		}
	}

	/**
	 * Fgt eine Break-Anweisung hinzu.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param l
	 *            das Label
	 */
	public void appendBreak(int ln, String s, String l) {
		if (!withinStatement()) {
			JavaASTJumpNode n = new JavaASTJumpNode(ln, s, l);
			n.isBreak(true);
			n.addTryNodes(jm.getStackOfTryNodes());
			n.setFunctionNode(jm.getCurrFunctionNode());
			jm.add(n);
			ast.append(n);
		}
	}

	/**
	 * Fgt eine Continue-Anweisung hinzu
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param l
	 *            das Label
	 */
	public void appendContinue(int ln, String s, String l) {
		if (!withinStatement()) {
			JavaASTJumpNode n = new JavaASTJumpNode(ln, s, l);
			n.isContinue(true);
			n.addTryNodes(jm.getStackOfTryNodes());
			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());
			jm.add(n);
		}
	}

	/**
	 * Fgt eine Return-Anweisung hinzu.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @param l
	 *            das Label
	 */
	public void appendReturn(int ln, String s, String l) {
		if (!withinStatement()) {
			JavaASTJumpNode n = new JavaASTJumpNode(ln, s, l);
			n.isReturn(true);
			ast.append(n);
			n.setFunctionNode(jm.getCurrFunctionNode());
			jm.add(n);
		}
	}

	/**
	 * Beginnt Ausnahmebehandlugn.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param s
	 *            der Inhalt
	 * @return der ASTTryNode
	 */
	public JavaASTTryNode beginTry(int ln, String s) {
		ASTNode entryNode = ast.getLastNode();
		JavaASTTryNode n = new JavaASTTryNode(ln, s);
		if (!withinStatement()) {
			ast.append(n);
			n.setEntryNode(entryNode);
			n.setFunctionNode(jm.getCurrFunctionNode());

			jm.add(n);
		}
		return n;
	}

	/**
	 * Fgt die einen virtuellen Knoten fr den Eintritt in den Try-Blocks ein.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 */
	public void setEntryOfTryBlock(int ln) {
		if (!withinStatement()) {
			JavaASTStatementNode entryNode = new JavaASTStatementNode(ln);
			ast.append(entryNode);
			jm.getTryNode().setEntryOfTryBlock(entryNode);
			entryNode.setFunctionNode(jm.getCurrFunctionNode());
		}
	}

	/**
	 * Bestimmt das Ende des Tryblocks.
	 */
	public void setEndOfTryBlock() {
		if (!withinStatement()) {
			jm.getTryNode().setEndOfTryBlock(ast.getLastNode());
		}
	}

	/**
	 * Beendet die Ausnahmebehandlung.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 */
	public void endTry(int ln) {
		if (!withinStatement()) {
			ASTStatementNode endNode = appendStatement(ln);
			ASTNode n = jm.remove();
			if (n instanceof ASTTryNode)
				((ASTTryNode) n).setEndNode(endNode);
			else
				Logger.logln("error@astbuilder - no try on stack");
		}
	}

	/**
	 * Beginnt eine neue Klasse.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param content
	 *            der Inhalt
	 * @param name
	 *            der Name
	 * @param extension
	 *            die Superklasse
	 * @return der ASTClassNode
	 */
	public ASTClassNode beginClass(int ln, String content, String name, String extension) {
		String absScope;

		if (name == null)
			name = new String("<" + content + "><" + ++nrAnonymousClasses + ">");
		ASTClassNode cn = new JavaASTClassNode(ln, content, st.getCurrScope(), name, jm.getScopeNode());
		// System.out.println(st.getCurrScope() + " - " + cn.getName() + " " + cn.isTopLevel());
		if (!withinStatement()) {
			ast.append(cn);
			cn.setFunctionNode(jm.getCurrFunctionNode());

			jm.addScopeNode(cn);
			if (extension != null)
				st.putExtension(name, extension);

			absScope = currFileNode.getFullName();
			if (!cn.getScope().equals("")) {
				absScope += "::" + cn.getScope();
			}

			// Anonyme Klassen/Structs auch so im Scope einfgen
			if (cn.getName().equals("")) {
				scopeMap.put(absScope + "::<anonymous>", cn);
			} else {
				scopeMap.put(absScope + "::" + cn.getName(), cn);

			}

			if (scopeMap.containsKey(absScope)) {
				scopeMap.get(absScope).addClassNode(cn);
			} else {
				Logger.logln("Error: ASTBuilder.addClass() - Scope " + absScope + " not found.");
				/*
				 * for (String s : scopeMap.keySet()) System.out.println("\t" + s);
				 */
			}

			openScope(name);
		}
		return cn;
	}

	/**
	 * Beendet eine Klasse.
	 */
	public void endClass() {
		if (!withinStatement()) {
			jm.removeScopeNode().setEndNode(ast.getLastNode());
			closeScope();
		}
	}

	/**
	 * Beginnt eine Funktion.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 * @param content
	 *            der Inhalt
	 * @param modifiers
	 *            die Modifiers der Definition.
	 * @param name
	 *            der Name der Funktion
	 * @param arguments
	 *            die Argumente der Funktion
	 * @return der ASTFunctionNode
	 */
	public ASTFunctionNode beginFunction(int ln, String content, String modifiers, String name, String arguments) {

		String absScope;
		ASTFunctionNode fn = new JavaASTFunctionNode(ln, content, st.getCurrScope(), // scope
				modifiers, name, // name
				arguments, // arguments
				jm.getScopeNode());
		if (!withinStatement()) {
			ast.append(fn);

			absScope = currFileNode.getFullName();
			if (!fn.getScope().equals("")) {
				absScope += "::" + fn.getScope();

			}

			scopeMap.put(absScope + "::" + fn.getName(), fn);

			if (scopeMap.containsKey(absScope)) {
				scopeMap.get(absScope).addFunctionNode(fn);
			} else {
				Logger.logln("Error: ASTBuilder.addFunction() - Scope " + absScope + "not found.");
				for (String s : scopeMap.keySet())
					Logger.logln("\t" + s);
			}
			jm.addFunction(fn);

			openScope(fn.getName());
			clearTerm();
		}
		return fn;
	}

	/**
	 * Fgt virtuellen Knoten fr den Funktionseingang hinzu.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 */
	public void addFunctionEntryNode(int ln) {
		if (!withinStatement()) {
			JavaASTStatementNode entryNode = new JavaASTStatementNode(ln);
			ast.append(entryNode);
			entryNode.setFunctionNode(jm.getCurrFunctionNode());
			jm.getCurrFunctionNode().setEntryNode(entryNode);
		}
	}

	/**
	 * Beendet eine Funktion.
	 * 
	 * @param ln
	 *            die Zeilennummer
	 */
	public void endFunction(int ln) {
		if (!withinStatement()) {
			JavaASTStatementNode endNode = new JavaASTStatementNode(ln);
			ast.append(endNode);
			endNode.setFunctionNode(jm.getCurrFunctionNode());
			jm.getCurrFunctionNode().setEndNode(endNode);
			jm.removeFunction();
			st.closeScope();
		}
	}

	/**
	 * Zeigt an, dass eine Anweisung begonnen wurde.
	 */
	public void enterStatement() {
		nrEnteredStatements++;

	}

	/**
	 * Zeigt das Beenden einer Anweisung an.
	 */
	public void leaveStatement() {
		nrEnteredStatements--;

	}

	/**
	 * Entscheidet, ob man sich innerhalb einer Anweisung befindet.
	 * 
	 * @return wahr, wenn innerhalb einer Anweisung, sonst false
	 */
	private boolean withinStatement() {
		return nrEnteredStatements != 0;
	}

	/*************************************************************************/
	/* Typbestimmung fr Ausdrcke. (Momentan nicht genutzt) */
	/*************************************************************************/

	/**
	 * Container fr die Sicherung von Typdaten einer Variablen.
	 * 
	 */
	private static class TermData {
		String type;

		// String id, suffix;

		public TermData() {
			type = "";
		}

		public void empty() {
			type = "";
		}

		@SuppressWarnings("unused")
		public void print() {
			Logger.logln("currTerm: " + type);
		}
	}

	/**
	 * Speichert TermData auf Stack.
	 */
	public void storeTermData() {
		currStack.push(currTerm);
		currTerm = new TermData();

	}

	/**
	 * Holt TermData vom Stack.
	 */
	public void restoreTermData() {
		currTerm = currStack.pop();
	}

	/**
	 * Beginnt eine Typdeklaration.
	 * 
	 * @param t
	 *            der Typ
	 */
	public void beginDeclaration(String t) {
		// currTerm.type = checkType(currTerm.type);
		currTerm.type = t;
		if (currTerm.type == null)
			currTerm.type = "void";
	}

	/**
	 * Beendet eine Typdeklaration.
	 */
	public void endDeclaration() {
		currTerm.empty();
	}

	/**
	 * Legt den Typ fr den bergebenen Term fest.
	 * 
	 * @param s
	 *            die Variable
	 */
	public void setType(String s) {
		// currTerm.type = checkType(s);
		currTerm.type = s;
		if (currTerm.type == null)
			currTerm.type = "void";
	}

	/**
	 * Liefert den Typ des aktuellen Terms.
	 * 
	 * @return der Typ
	 */
	public String getType() {
		return currTerm.type;
	}

	/**
	 * Lscht den aktuellen Term.
	 */
	public void clearTerm() {
		currTerm.empty();
	}

	/**
	 * Erweitert den aktuellen Typ eines Terms.
	 * 
	 * @param s
	 *            die Erweiterung (z.b.: [])
	 */
	public void appendToType(String s) {
		currTerm.type += s;
	}

	/**
	 * Sucht vollqualifizierten Typ zu einem Typ. (ungenutzt)
	 * 
	 * @param s
	 *            der zu untersuchende Typ
	 * @return der vollqualifizierte Typ oder null
	 */
	public String checkType(String s) {
		return st.checkType(s);
	}

	/**
	 * Sucht den Typ fr einen String. (ungenutzt)
	 * 
	 * @param s
	 *            der Ausdruck
	 * @return der Typ oder null
	 */
	public String lookUp(String s) {
		return st.lookUp(s);
	}

}