package sota.lang.java.ast.util;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Stack;

import sota.Logger;
import sota.ast.nodes.ASTBranchNode;
import sota.ast.nodes.ASTCompoundNode;
import sota.ast.nodes.ASTDoNode;
import sota.ast.nodes.ASTFunctionNode;
import sota.ast.nodes.ASTIfNode;
import sota.ast.nodes.ASTIterationNode;
import sota.ast.nodes.ASTJumpNode;
import sota.ast.nodes.ASTLabelNode;
import sota.ast.nodes.ASTNode;
import sota.ast.nodes.ASTSwitchNode;
import sota.ast.nodes.ASTTryNode;
import sota.ast.nodes.IScopeNode;
import sota.ast.util.ASTManager;

/**
 * Verwaltet die Sprungziele whrend des Parseprozesses.
 * */
class JavaJumpManager {

	// alle BranchNodes und DoNodes
	private Stack<ASTNode> stackOfNodes = new Stack<ASTNode>(); // geschachtelte Statements

	private Stack<ASTFunctionNode> stackOfFunctions = new Stack<ASTFunctionNode>();

	private Stack<ASTTryNode> stackOfTryNodes = new Stack<ASTTryNode>();

	private Stack<IScopeNode> stackOfScopeNodes = new Stack<IScopeNode>();

	private Stack<ASTJumpNode> stackOfJumpNodes = new Stack<ASTJumpNode>(); // alle Jumps mit Label

	private HashMap<ASTFunctionNode, ArrayList<ASTLabelNode>> mapOfLabelNodes = new HashMap<ASTFunctionNode, ArrayList<ASTLabelNode>>();

	// alle Jumps ohne Label
	private HashMap<ASTJumpNode, ASTNode> tableOfBreaks = new HashMap<ASTJumpNode, ASTNode>();

	private HashMap<ASTJumpNode, ASTNode> tableOfContinues = new HashMap<ASTJumpNode, ASTNode>();

	private HashMap<ASTJumpNode, ASTNode> tableOfReturns = new HashMap<ASTJumpNode, ASTNode>();

	private HashMap<ASTJumpNode, ASTNode> tableOfThrows = new HashMap<ASTJumpNode, ASTNode>();

	public static HashSet<ASTFunctionNode> setOfFunctionNodes = new HashSet<ASTFunctionNode>();

	/**
	 * Fgt relevanten Knoten hinzu. (Iteration, Switch, Try, If, Do, Jump, Label)
	 * 
	 * @param n
	 *            der ASTNode
	 */
	public void add(ASTNode n) { // ASTIteration, ASTSwitch, ASTTry, ASTIfNode, ASTDoNode
		if (n instanceof ASTJumpNode) {
			ASTJumpNode jn = (ASTJumpNode) n;
			if (jn.isBreak()) {
				if (jn.getLabel() != "")
					stackOfJumpNodes.push(jn);
				else
					// break ohne Label bricht Iteration oder Switch ab, kein Try oder If
					tableOfBreaks.put(jn, getIterationOrSwitchNode());
			} else if (jn.isContinue()) {
				if (jn.getLabel() != "")
					stackOfJumpNodes.push(jn);
				else
					tableOfContinues.put(jn, getIterationNode());
			} else if (jn.isThrow()) {
				tableOfThrows.put(jn, stackOfFunctions.peek());
			} else if (jn.isReturn()) {
				tableOfReturns.put(jn, stackOfFunctions.peek());
			}

		} else if (n instanceof ASTLabelNode) {
			if (!mapOfLabelNodes.containsKey(n.getFunctionNode()))
				mapOfLabelNodes.put(n.getFunctionNode(), new ArrayList<ASTLabelNode>());
			mapOfLabelNodes.get(n.getFunctionNode()).add(((ASTLabelNode) n));
		}

		// ASTIteration, ASTSwitch, ASTTry, ASTIfNode, ASTDoNode
		else {
			stackOfNodes.push(n);
			if (n instanceof ASTTryNode)
				stackOfTryNodes.push((ASTTryNode) n);
		}
	}

	/**
	 * Entfernt einen Knoten vom Stack.
	 * 
	 * @return der ASTNode
	 */
	public ASTNode remove() {
		ASTNode n = stackOfNodes.pop();
		if (n instanceof ASTTryNode)
			stackOfTryNodes.pop();
		return n;
	}

	/**
	 * Liefert den aktuellen Try-Node
	 * 
	 * @return der ASTTrynode
	 */
	public ASTTryNode getTryNode() {
		return stackOfTryNodes.peek();
	}

	/**
	 * Liefert den Stack mit Try-Nodes
	 * 
	 * @return der Stack
	 */
	public Stack<ASTTryNode> getStackOfTryNodes() {
		return stackOfTryNodes;

	}

	/**
	 * Fgt eine Funktion hinzu.
	 * 
	 * @param fn
	 *            der ASTFunctionNode
	 */
	public void addFunction(ASTFunctionNode fn) {
		stackOfFunctions.push(fn);
		stackOfScopeNodes.push(fn);
	}

	/**
	 * Entfernt eine Funktion.
	 */
	public void removeFunction() {
		stackOfFunctions.pop();
		stackOfScopeNodes.pop();
	}

	/**
	 * Verlinkt alle Knoten des JumpManagers.
	 * 
	 * @param ast
	 *            der ASTManager
	 */
	public void linkNodes(ASTManager ast) {
		ASTJumpNode jn;
		ASTFunctionNode fn;
		ASTLabelNode ln, temp;
		ASTNode dn;
		ArrayList<ASTLabelNode> listOfLabels;

		// Jump to Label
		while (!stackOfJumpNodes.empty()) {
			jn = stackOfJumpNodes.pop();
			ln = null;

			listOfLabels = mapOfLabelNodes.get(jn.getFunctionNode());

			search: for (int i = 0; i < listOfLabels.size(); i++) {
				temp = listOfLabels.get(i);
				if (temp.getID() > jn.getID())
					break search;
				if (temp.getLabel().equals(jn.getLabel()))
					ln = temp;
			}

			if (ln == null)
				Logger.logln("couldn't find label: " + jn.getFunctionNode().getIDString() + jn.getLabel());
			ln.addJumpNode(jn);
			if (jn.isGoto()) {
				jn.setTarget(ln);
			} else if (jn.isContinue()) {
				ln.addJumpNode(jn);
				while (ln.getNextNode() instanceof ASTLabelNode) {
					ln = (ASTLabelNode) ln.getNextNode();
				}
				if (ln.getNextNode() instanceof ASTIterationNode) {
					jn.setTarget(ln.getNextNode());
				} else if (ln.getNextNode() instanceof ASTDoNode) {
					jn.setTarget(((ASTDoNode) ln.getNextNode()).getIterationNode());
				}
			} else if (jn.isBreak()) {
				while (ln.getNextNode() instanceof ASTLabelNode) {
					ln = (ASTLabelNode) ln.getNextNode();
				}
				if (ln.getNextNode() instanceof ASTBranchNode) {
					jn.setTarget(((ASTBranchNode) ln.getNextNode()).getEndNode());
				} else if (ln.getNextNode() instanceof ASTDoNode) {
					jn.setTarget(((ASTDoNode) ln.getNextNode()).getEndNode());
				} else if (ln.getNextNode() instanceof ASTCompoundNode) {
					jn.setTarget(((ASTCompoundNode) ln.getNextNode()).getEndNode());
				} else {
					Logger.logln("Error@JumpManager - Break without Compound or BranchNode");
					jn.setTarget(ln);
				}
			}
		}

		// Breaks
		for (ASTNode n : tableOfBreaks.keySet()) {
			jn = (ASTJumpNode) n;
			dn = tableOfBreaks.get(jn);
			switch (dn.getType()) {
			case ASTNode.SWITCH:
				jn.setTarget(((ASTSwitchNode) dn).getEndNode());
				break;
			case ASTNode.DO:
				jn.setTarget(((ASTDoNode) dn).getEndNode());
				break;
			case ASTNode.ITERATION:
				jn.setTarget(((ASTIterationNode) dn).getEndNode());
				break;
			default:
				System.out.println("Error: JumpManager - couldn't find Jumptarget (break)!");
			}
		}

		// Continues
		for (ASTNode n : tableOfContinues.keySet()) {
			jn = (ASTJumpNode) n;
			// Logger.logln("X: " + jn.getID());
			dn = tableOfContinues.get(jn);
			switch (dn.getType()) {
			case ASTNode.DO:
				jn.setTarget(((ASTDoNode) dn).getIterationNode());
				break;
			case ASTNode.ITERATION:
				jn.setTarget(dn);
				break;
			default:
				System.out.println("Error: JumpManager - couldn't find Jumptarget (continue)!");
			}
		}

		// Returns
		for (ASTNode n : tableOfReturns.keySet()) {
			jn = (ASTJumpNode) n;
			fn = (ASTFunctionNode) tableOfReturns.get(jn);
			jn.setTarget(fn.getEndNode());
			fn.addExitNode(jn);
		}

		// Throws - Ziel wird von Funktion auf endNode umgelenkt, wenn keine Trys da sind
		for (ASTNode n : tableOfThrows.keySet()) {
			jn = (ASTJumpNode) n;
			if (jn.getTargetNode() == tableOfThrows.get(jn)) {
				jn.setTarget(((ASTFunctionNode) tableOfThrows.get(jn)).getEndNode());
				((ASTFunctionNode) tableOfThrows.get(jn)).addExitNode(jn);
			}
		}

	}

	/**
	 * Liefert aktuellen Switch.
	 * 
	 * @return der ASTSwitchNode
	 */
	public ASTSwitchNode getSwitchNode() {
		Stack<ASTNode> h = new Stack<ASTNode>();
		ASTSwitchNode sn;
		while (!stackOfNodes.empty() && !(stackOfNodes.peek() instanceof ASTSwitchNode)) {
			h.push(stackOfNodes.pop());
		}
		sn = (ASTSwitchNode) stackOfNodes.peek();
		while (!h.empty()) {
			stackOfNodes.push(h.pop());
		}
		return sn;
	}

	/**
	 * Liefert Switch- oder IterationNode vom Stack.
	 * 
	 * @return der ASTSwitchNode oder ASTIterationNode
	 */
	public ASTNode getIterationOrSwitchNode() {
		Stack<ASTNode> h = new Stack<ASTNode>();
		ASTNode dn;
		while (!stackOfNodes.empty()
				&& (stackOfNodes.peek() instanceof ASTTryNode || stackOfNodes.peek() instanceof ASTIfNode)) {
			h.push(stackOfNodes.pop());
		}
		dn = stackOfNodes.peek();
		while (!h.empty()) {
			stackOfNodes.push(h.pop());
		}
		return dn;
	}

	/**
	 * Liefert ASTIterationNode vom Stack.
	 * 
	 * @return der ASTIterationNode
	 */
	public ASTNode getIterationNode() {
		Stack<ASTNode> h = new Stack<ASTNode>();
		ASTNode dn;
		while (!stackOfNodes.empty()
				&& !((stackOfNodes.peek() instanceof ASTIterationNode) || (stackOfNodes.peek() instanceof ASTDoNode))) {

			h.push(stackOfNodes.pop());
		}
		dn = stackOfNodes.peek();
		/*
		 * if(dn instanceof ASTDoNode) dn = ((ASTDoNode)dn).getIterationNode();
		 */

		while (!h.empty()) {
			stackOfNodes.push(h.pop());
		}
		return dn;
	}

	/**
	 * Liefert aktuelle Funktion.
	 * 
	 * @return der ASTFunctionNode
	 */
	public ASTFunctionNode getCurrFunctionNode() {
		if (stackOfFunctions.size() > 0)
			return stackOfFunctions.peek();

		return null;
	}

	/**
	 * Fgt neuen ScopeNode hinzu.
	 * 
	 * @param n
	 *            der IScopeNode
	 */
	public void addScopeNode(IScopeNode n) {
		stackOfScopeNodes.push(n);
	}

	/**
	 * Liefert aktuellen Scope.
	 * 
	 * @return der IScopeNode
	 */
	public IScopeNode getScopeNode() {
		if (stackOfScopeNodes.size() > 0)
			return stackOfScopeNodes.peek();
		return null;
	}

	/**
	 * Entfernt aktuellen Scope
	 * 
	 * @return der verlassene IScopeNode
	 */
	public IScopeNode removeScopeNode() {
		return stackOfScopeNodes.pop();
	}

}